package geneticWedge.gp;

/**AbstractPopulation is the parent of all Population objects. It contains only the most basic functions and defines
additional abstract functions that must be implemented by any subclass.*/
public abstract class AbstractPopulation
{
	/**The FitnessFunction is used to select Individuals from a Population for breeding.*/
	public static enum FitnessFunction
	{
		/**Mean Square Error*/
		MSE,
		/**Root Mean Square Error*/
		RMSE,
		/**Average absolute value of error*/
		ABSOLUTE_ERROR,
		/**Number of correct binary classifications; the output must have the correct sign to give correct classification*/
		NUMBER_CORRECT_BINARY,
		/**This is the sqrt of the sum of the squares of the false negative fraction and the false positive fraction during classification.*/
		COMBINED_CLASSIFICATION_ERROR,
		/**Area under the ROC curve*/
		AUROC,
		/**Area over the REC curve*/
		AOREC,
		/**Sensitivity is the ratio TP/(TP+FN)*/
		SENSITIVITY,
		/**Specificity is the ratio TN/(TN+FP)*/
		SPECIFICITY,
		/**PPV is the positive predictive value, that is the probability that a given prediction is actually true given that it is predicted as true.*/
		PPV,
		/**NPV is the negative predictive value, that is the probability that a given prediction is actually false given that it is predicted as false.*/
		NPV,
		CLUSTERED_REGRESSION, //new 010208
		PREDEFINED //new 080208*/
	};

	/**SelectionMethod determines <em>how</em> Individuals are selected for breeding.*/
	public static enum SelectionMethod
	{
		/**Random selection*/
		RANDOM,
		/**Fitness proportional selection uses a simple roulette wheel technique to select Individuals.*/
		FITNESS_PROPORTIONAL,
		/**Rank selection selects Individuals according to their rankings within a Population*/
		RANK,
		/**Tournament selection selects the best Individual from within a randomly chosen pool (tournament) of Individuals.*/
		TOURNAMENT,
		/**Truncation Selection breeds a number of Individuals using a pool of parents and then selects the best offspring to go into the nexy generation.*/
		TRUNCATION,
		/**FUSS stands for 'fitness uniform selection scheme'. It selects from sparse regions in order to create a homogeneously distributed population */
		FUSS
	};

	//All class variables protected access so that they can be used by subclasses, in particular Population
	/**An array containing the (training) fitnesses of all Individuals in this Population*/
	protected double[] allFitnesses;
	/**The index of the Individual with the best (training) fitnesses*/
	protected int bestIndex=-1;
	/**The average (mean) training fitness of all members of the population*/
	protected double averageTrainFitness=Double.NaN;
	/**The best training fitness obtained by any member of the current population*/
	protected double bestTrainFitness=Double.NaN;
	/**The average (mean) validation fitness of all members of the population*/
	protected double averageValidationFitness=Double.NaN;
	/**The best validation fitness obtained by any member of the current population*/
	protected double bestValidationFitness=Double.NaN;
	/**The number of generations of evolution that have occurred*/
	protected int generationNumber=0;
	/**The maximum number of generations of evolution that may occur*/
	protected int maxGenerations=1000;
	/**The best training fitness obtained by any member of the current population in any generation so far*/
	protected double bestValidationFitnessSoFar=Double.NaN;
	/**The best validation fitness obtained by any member of the current population in any generation so far*/
	protected double bestTrainFitnessSoFar=Double.NaN;

	/**The starting population size. Usually the population size stays constant.
	However, if a performanceFunction is used and non-dominated Individuals are removed the population size may vary.*/
	protected int startingPopSize=0;
	/**The current population size. Usually the population size stays constant.
	However, if a performanceFunction is used and non-dominated Individuals are removed the population size may vary.*/
	protected int popSize=0;
	/**The size of the tournament used during tournament selection. The default tournament size is 4.*/
	protected int tournamentSize=4;
	/**The PopulationDescriptor that describes this Population*/
	protected PopulationDescriptor popDescriptor;
	/**The FitnessFunction used to evaluate Individuals. The default is MSE.*/
	protected FitnessFunction fitnessFunction=FitnessFunction.MSE;
	/**The number of points used to estimate the AUROC (and AOREC). Default is 100.*/
	protected int aurocSampleSize = 100;
	/**The FitnessFunction used to evaluate the AOREC. Fefault is MSE.*/
	protected FitnessFunction recFitnessFunction = FitnessFunction.MSE;



	/**Returns the fitness of the best Individual on the training data.*/
	public abstract double getBestTrainingFitness();

	/**Returns the fitness of the best Individual on the validation data.*/
	public abstract double getBestValidationFitness();

	/**Returns the average fitness of all Individuals on the training data.*/
	public abstract double getAverageTrainingFitness();

	/**Returns the average fitness of all Individuals on the validation data.*/
	public abstract double getAverageValidationFitness();


	//protected Mutation mut = new Mutation(new Mutation.MutationOperator[]{Mutation.MutationOperator.SUBTREE_REPLACEMENT});
	//protected Crossover crossover = new Crossover();

	/**Returns the number of generations of evolution that have occurred.*/
	public int getGenerationNumber()
	{
		return generationNumber;
	}

	/**Returns the initial population size. In most cases the population size is constant.
	It may vary if a performance function is used and dominated Individuals are removed from the Population.*/
	public int getInitialPopulationSize()
	{
		return startingPopSize;
	}

	/**Returns the current population size.*/
	public int getPopulationSize()
	{
		return popSize;
	}

	/**Sets the maximum number of generations for which the population may evolve.*/
	public void setMaxGenerations(int maxGenerations)
	{
		this.maxGenerations=maxGenerations;
	}

	/**Returns the maximum number of generations for which the population may evolve.*/
	public int getMaxGenerations()
	{
		return maxGenerations;
	}

	/**Sets the size of the tournament to be used in tournament selection.
	A large tournament size will lead to a quicker loss in diversity, i.e. faster convergence.
	A tournament size less than 1 or greater than or equal to the population size will not be set.*/
	public void setTournamentSize(int tournamentSize)
	{
		if(tournamentSize>0&&tournamentSize<=popSize)
			this.tournamentSize=tournamentSize;
	}

	/**Returns the tournament size to be used during tournament selection.*/
	public int getTournamentSize()
	{
		return tournamentSize;
	}

	/**Returns a Population Descriptor object, which holds information concerning the Population as a whole.*/
	public PopulationDescriptor getPopulationDescriptor()
	{
		return popDescriptor;
	}

	/**AUROC sample size is also used for REC curves and Combined Classification Error.*/
	public void setAUROCSampleSize(int sampleSize)
	{
		aurocSampleSize = sampleSize;
	}

	/**AUROC sample size is also used for REC curves and Combined Classification Error.*/
	public int getAUROCSampleSize()
	{
		return aurocSampleSize;
	}

	/**Sets the FitnessFunction to be used when constructing a REC curve.*/
	public void setRECFitnessFunction(FitnessFunction recFunction)
	{
		if(recFunction.equals(FitnessFunction.MSE)||recFunction.equals(FitnessFunction.RMSE)||recFunction.equals(FitnessFunction.ABSOLUTE_ERROR))
			recFitnessFunction = recFunction;
		else
			System.err.println("Attempt to set invalid REC FitnessFunction.");
	}

	/**Returns the FitnessFunction to be used when constructing a REC curve.*/
	public FitnessFunction getRECFitnessFunction()
	{
		return recFitnessFunction;
	}
}