package geneticWedge.gp;

import java.util.Vector;
import java.util.Random;

import geneticWedge.gp.function.*;

/**Crossover performs crossover ('sexual' recombination) between 2 Individuals, generating 1 or 2 offspring.
It holds various parameters that control the way in which crossover is performed.*/
public final class Crossover
{
	private Random rand=new Random();
	private int noOfTries=5;
	private double functionProportion=Double.NaN;
	private boolean useBrood = false;
	private int broodSize = 4;
	private int samplingFactor = 8;
	private AbstractPopulation.FitnessFunction fitnessFunction = AbstractPopulation.FitnessFunction.MSE;
	private double[][] trainingInputs = null;
	private double[] targetOutputs = null;
	private boolean useExpectedDepth = false;
	private double expectedDepth = 0.5;
	private boolean collectTransmissionInfo = false;
	private Vector<double[]> transmissionInfo=null;
	private double[] thisTransmissionInfo = new double[6];

	/**The default constructor creates a Crossover with default parameters. These are functionProportion of NaN
	(random choice of nodes) and noOfTries of 5 (before stopping the attempt to create valid offspring).
	*/
	public Crossover()
	{

	}

	/**
	* Create a Crossover object.
	* @param functionProportion The proportion of crossover points that will be chosen to be Functions. The remainder of crossover points will occur at Terminals. If set to NaN, crossover points will be seleted randomly.
	* @param noOfTries The number of times that Crossover will try to create valid offspring.
	*/
	public Crossover(double functionProportion, int noOfTries)
	{
		if(functionProportion>=0.0 && functionProportion<=1.0)
			this.functionProportion = functionProportion;
		else
			System.err.println("Attempt to set function proportion outside the allowable range [0.0,1.0]");

		if(noOfTries>0)
			this.noOfTries = noOfTries;
		else
			System.err.println("Attempt to set non-positive value of noOfTries");
	}

	/**This sets the 'expected' depth at which crossover will take place. A Poisson distribution is then created with this
	expected value in order to select the depth at which crossover takes place. If this method is not called, or it is called
	with depth outside the range 0.0-1.0 (exclusive), the crossover site will be selected randomly.*/
	public void setExpectedCrossoverDepth(double depth)
	{
		if(depth<=0.0||depth>=1.0)
		{
			useExpectedDepth = false;
			return;
		}
		useExpectedDepth = true;
		expectedDepth = depth;
	}

	/**If set to true, collectTransmissionInfo saves information concerning the relationship between parent and offspring
	during crossover operations*/
	public void collectTransmissionInfo(boolean collect)
	{
		collectTransmissionInfo = collect;
		if(collect)
			transmissionInfo=new Vector<double[]>();
	}

	/**Returns a vector containing information about the relationship between parent and offspring during crossover operations,
	provided collectTransmissionInfo has been set to true.*/
	public Vector<double[]> getTransmissionInfo()
	{
		return transmissionInfo;
	}

	private Individual[] cross(Individual ind1, Individual ind2, int crossoverPoint1, int crossoverPoint2, boolean twoOffspring, Input[] inputs, Constant[] constants)
	{
		Vector<Component> nodes1=ind1.getNodes();
		Vector<Component> nodes2=ind2.getNodes();
		thisTransmissionInfo[0] = ind1.getDepth();
		thisTransmissionInfo[1] = ind2.getDepth();
		if(crossoverPoint1>=ind1.getLength()-1||crossoverPoint2>=ind2.getLength()-1||crossoverPoint1<0||crossoverPoint2<0)
		{
			thisTransmissionInfo[2] = -1;
			thisTransmissionInfo[3] = -1;
			thisTransmissionInfo[4] = -1;
			thisTransmissionInfo[5] = -1;
			return new Individual[]{new Individual(nodes1,inputs,constants),new Individual(nodes2,inputs,constants)};
		}
		else
		{
			int end1=Utils.getEndOfFragment(nodes1,crossoverPoint1);
			int end2=Utils.getEndOfFragment(nodes2,crossoverPoint2);

			long time1 = System.currentTimeMillis();
			Vector<Component> newNodes1=new Vector<Component>();
			for(int i=0;i<=crossoverPoint1;i++)
				newNodes1.add(nodes1.get(i));
			for(int i=crossoverPoint2+1;i<=end2;i++)
				newNodes1.add(nodes2.get(i));
			for(int i=end1+1;i<nodes1.size();i++)
				newNodes1.add(nodes1.get(i));

			Individual newInd1=new Individual(newNodes1, inputs, constants);

			int[] depths1 = Utils.getFragmentDepths(nodes1,crossoverPoint1, end1);
			int[] depths2 = Utils.getFragmentDepths(nodes2,crossoverPoint2, end2);
			thisTransmissionInfo[2] = depths1[0];
			thisTransmissionInfo[3] = depths2[0];
			thisTransmissionInfo[4] = depths1[1]-depths1[0];
			thisTransmissionInfo[5] = depths2[1]-depths2[0];

			if(twoOffspring)
			{
				Vector<Component> newNodes2=new Vector<Component>();
				for(int i=0;i<=crossoverPoint2;i++)
					newNodes2.add(nodes2.get(i));
				for(int i=crossoverPoint1+1;i<=end1;i++)
					newNodes2.add(nodes1.get(i));
				for(int i=end2+1;i<nodes2.size();i++)
					newNodes2.add(nodes2.get(i));

				Individual newInd2=new Individual(newNodes2, inputs, constants);

				return new Individual[]{newInd1,newInd2};
			}
			else
				return new Individual[]{newInd1};
		}
	}

	/**This method returns an array of Individuals created by crossing 2 parents (ind1 and ind2).
	@param twoOffspring If true two offspring will be returned. If false only one offspring will be returned.
	@param maxDepth The maximum depth of the returned offspring
	@param maxLength The maximum length of the returned offspring
	*/
	public Individual[] cross(Individual ind1, Individual ind2,boolean twoOffspring, Input[] inputs, Constant[] constants, int maxLength, int maxDepth)
	{
		if(useBrood && trainingInputs!=null && targetOutputs!=null)
			return broodCross(ind1,ind2,twoOffspring,inputs,constants,maxLength,maxDepth);
		else
		{
			Individual[] offspr = simpleCross(ind1, ind2, twoOffspring, inputs, constants, maxLength, maxDepth);
			if(offspr!=null) //don't include crossovers that have failed
			{
				if(offspr[0] == null)
				{
					double temp = thisTransmissionInfo[0];
					thisTransmissionInfo[0] = thisTransmissionInfo[1];
					thisTransmissionInfo[1] = temp;
					temp = thisTransmissionInfo[2];
					thisTransmissionInfo[2] = thisTransmissionInfo[3];
					thisTransmissionInfo[3] = temp;
					temp = thisTransmissionInfo[4];
					thisTransmissionInfo[4] = thisTransmissionInfo[5];
					thisTransmissionInfo[5] = temp;
				}
				if(collectTransmissionInfo)
					transmissionInfo.add(thisTransmissionInfo.clone());
			}
			return offspr;
		}
	}

	private Individual[] broodCross(Individual ind1, Individual ind2, boolean twoOffspring, Input[] inputs, Constant[] constants, int maxLength, int maxDepth)
	{
		int cutdownLength = 1;
		int len = trainingInputs.length;
		int dim = trainingInputs[0].length;
		if(len >= samplingFactor)
			cutdownLength = len / samplingFactor;
		double[][] cutdownInput = new double[cutdownLength][dim];
		double[] cutdownTarget = new double[cutdownLength];
		int[] indices = Utils.getRandomIndices(len, cutdownLength);
		for(int i=0;i<cutdownLength;i++)
		{
			for(int j=0;j<dim;j++)
				cutdownInput[i][j] = trainingInputs[indices[i]][j];
			cutdownTarget[i] = targetOutputs[indices[i]];
		}
		Individual[] brood = new Individual[broodSize*2];
		double[] fitnesses = new double[broodSize*2];
		double[][] tempTransmissionInfo = new double[broodSize*2][];
		for(int i=0;i<broodSize;i++)
		{
			Individual[] offSpr = simpleCross(ind1, ind2, true, inputs, constants, maxLength, maxDepth);
			//if a valid cross is not found, use the parents as offspring
			if(offSpr == null)
			{
				brood[i*2] = ind1;
				brood[i*2+1] = ind2;
				try
				{
					switch(fitnessFunction)
					{
						case MSE:
							fitnesses[i*2] = ind1.getMSE(cutdownInput,cutdownTarget);
							fitnesses[i*2+1] = ind2.getMSE(cutdownInput,cutdownTarget);
							break;
						case RMSE:
							fitnesses[i*2] = ind1.getRMSE(cutdownInput,cutdownTarget);
							fitnesses[i*2+1] = ind2.getRMSE(cutdownInput,cutdownTarget);
							break;
						case ABSOLUTE_ERROR:
							fitnesses[i*2] = ind1.getAbsoluteError(cutdownInput,cutdownTarget);
							fitnesses[i*2+1] = ind2.getAbsoluteError(cutdownInput,cutdownTarget);
							break;
						case NUMBER_CORRECT_BINARY:
							fitnesses[i*2] = ind1.getNumberCorrectBinary(cutdownInput,cutdownTarget);
							fitnesses[i*2+1] = ind2.getNumberCorrectBinary(cutdownInput,cutdownTarget);
							break;
					}
					tempTransmissionInfo[i*2] = thisTransmissionInfo;
					double[] tti = new double[6];
					tti[0] = thisTransmissionInfo[1];
					tti[1] = thisTransmissionInfo[0];
					tti[2] = thisTransmissionInfo[3];
					tti[3] = thisTransmissionInfo[2];
					tti[4] = thisTransmissionInfo[5];
					tti[5] = thisTransmissionInfo[4];
					tempTransmissionInfo[i*2+1] = tti;
				}
				catch(InvalidInputSizeException iise)
				{
					System.err.println("Invalid inputs. Brood recombination aborted.");
					useBrood = false;
					Individual[] offspr = simpleCross(ind1, ind2, twoOffspring, inputs, constants, maxLength, maxDepth);
					if(offspr!=null) //don't include crossovers that have failed
					{
						if(offspr[0] == null)
						{
							double temp = thisTransmissionInfo[0];
							thisTransmissionInfo[0] = thisTransmissionInfo[1];
							thisTransmissionInfo[1] = temp;
							temp = thisTransmissionInfo[2];
							thisTransmissionInfo[2] = thisTransmissionInfo[3];
							thisTransmissionInfo[3] = temp;
							temp = thisTransmissionInfo[4];
							thisTransmissionInfo[4] = thisTransmissionInfo[5];
							thisTransmissionInfo[5] = temp;
						}
						if(collectTransmissionInfo)
							transmissionInfo.add(thisTransmissionInfo.clone());
						return offspr;
					}
				}
			}
			else
			{
				brood[i*2] = offSpr[0];
				brood[i*2+1] = offSpr[1];
				try
				{
					switch(fitnessFunction)
					{
						case MSE:
							if(offSpr[0] == null)
								fitnesses[i*2] = Double.MAX_VALUE;
							else
								fitnesses[i*2] = offSpr[0].getMSE(cutdownInput,cutdownTarget);
							if(offSpr[1] == null)
								fitnesses[i*2+1] = Double.MAX_VALUE;
							else
								fitnesses[i*2+1] = offSpr[1].getMSE(cutdownInput,cutdownTarget);
							break;
						case RMSE:
							if(offSpr[0] == null)
								fitnesses[i*2] = Double.MAX_VALUE;
							else
								fitnesses[i*2] = offSpr[0].getRMSE(cutdownInput,cutdownTarget);
							if(offSpr[1] == null)
								fitnesses[i*2+1] = Double.MAX_VALUE;
							else
								fitnesses[i*2+1] = offSpr[1].getRMSE(cutdownInput,cutdownTarget);
							break;
						case ABSOLUTE_ERROR:
							if(offSpr[0] == null)
								fitnesses[i*2] = Double.MAX_VALUE;
							else
								fitnesses[i*2] = offSpr[0].getAbsoluteError(cutdownInput,cutdownTarget);
							if(offSpr[1] == null)
								fitnesses[i*2+1] = Double.MAX_VALUE;
							else
								fitnesses[i*2+1] = offSpr[1].getAbsoluteError(cutdownInput,cutdownTarget);
							break;
						case NUMBER_CORRECT_BINARY:
							if(offSpr[0] == null)
								fitnesses[i*2] = -Double.MAX_VALUE;
							else
								fitnesses[i*2] = offSpr[0].getNumberCorrectBinary(cutdownInput,cutdownTarget);
							if(offSpr[1] == null)
								fitnesses[i*2+1] = -Double.MAX_VALUE;
							else
								fitnesses[i*2+1] = offSpr[1].getNumberCorrectBinary(cutdownInput,cutdownTarget);
							break;
					}
					tempTransmissionInfo[i*2] = thisTransmissionInfo;
					double[] tti = new double[6];
					tti[0] = thisTransmissionInfo[1];
					tti[1] = thisTransmissionInfo[0];
					tti[2] = thisTransmissionInfo[3];
					tti[3] = thisTransmissionInfo[2];
					tti[4] = thisTransmissionInfo[5];
					tti[5] = thisTransmissionInfo[4];
					tempTransmissionInfo[i*2+1] = tti;
				}
				catch(InvalidInputSizeException iise)
				{
					System.err.println("Invalid inputs. Brood recombination aborted.");
					useBrood = false;
					Individual[] offspr = simpleCross(ind1, ind2, twoOffspring, inputs, constants, maxLength, maxDepth);
					if(offspr!=null) //don't include crossovers that have failed
					{
						if(offspr[0] == null)
						{
							double temp = thisTransmissionInfo[0];
							thisTransmissionInfo[0] = thisTransmissionInfo[1];
							thisTransmissionInfo[1] = temp;
							temp = thisTransmissionInfo[2];
							thisTransmissionInfo[2] = thisTransmissionInfo[3];
							thisTransmissionInfo[3] = temp;
							temp = thisTransmissionInfo[4];
							thisTransmissionInfo[4] = thisTransmissionInfo[5];
							thisTransmissionInfo[5] = temp;
						}
						if(collectTransmissionInfo)
							transmissionInfo.add(thisTransmissionInfo.clone());
					}
					return offspr;
				}
			}
		}
		Object[] orderedObj = Population.ShellSortIndividuals(brood, fitnesses, fitnessFunction);
		Individual[] orderedInds = (Individual[])(orderedObj[0]);
		Individual[] returnInds = null;
		int[] broodIndices = (int[])(orderedObj[2]);
		if(twoOffspring)
		{
			returnInds = new Individual[2];
			returnInds[0] = orderedInds[0];
			returnInds[1] = orderedInds[1];
		}
		else
		{
			returnInds = new Individual[1];
			returnInds[0] = orderedInds[0];
		}
		if(collectTransmissionInfo)
		{
			double[] ti = tempTransmissionInfo[broodIndices[0]].clone();
			transmissionInfo.add(ti);
		}
		return returnInds;
	}


	private Individual[] simpleCross(Individual ind1, Individual ind2,boolean twoOffspring, Input[] inputs, Constant[] constants, int maxLength, int maxDepth)
	{
		if(ind1.getLength()==1||ind2.getLength()==1)
			return null;
		for(int t=0;t<noOfTries;t++)
		{
			int crossoverPoint1 = -1;
			int crossoverPoint2 = -1;
			if(Double.isNaN(functionProportion))
			{
				if(useExpectedDepth)
				{
					crossoverPoint1 = ind1.getModificationPointFromPoissonDistribution(expectedDepth, true)-1;
					if(crossoverPoint1<0)
						crossoverPoint1 = 0;
					crossoverPoint2 = ind2.getModificationPointFromPoissonDistribution(expectedDepth, true)-1;
					if(crossoverPoint2<0)
						crossoverPoint2 = 0;
				}
				else
				{
					crossoverPoint1 = (int)(rand.nextDouble()*(ind1.getLength()-1));
					crossoverPoint2=(int)(rand.nextDouble()*(ind2.getLength()-1));
				}
			}
			else
			{
				Vector<Component> nodes = ind1.getNodes();
				Vector<Integer> functions = new Vector<Integer>();
				Vector<Integer> terminals = new Vector<Integer>();
				for (int i=0;i<nodes.size();i++)
				{
					if(nodes.get(i) instanceof Function)
						functions.add(i);
					else
						terminals.add(i);
				}
				double selector1 = rand.nextDouble();
				double selector2 = rand.nextDouble();
				if(selector1 < functionProportion)
				{
					if(useExpectedDepth)
						crossoverPoint1 = ind1.getModificationPointFromPoissonDistribution(expectedDepth, functions, true);
					else
						crossoverPoint1 = functions.get((int)(selector2 * functions.size()));
				}
				else
				{
					if(useExpectedDepth)
						crossoverPoint1 = ind1.getModificationPointFromPoissonDistribution(expectedDepth, terminals, true);
					else
						crossoverPoint1 = terminals.get((int)(selector2 * terminals.size()));
				}

				nodes = ind2.getNodes();
				functions = new Vector<Integer>();
				terminals = new Vector<Integer>();
				for (int i=0;i<nodes.size();i++)
				{
					if(nodes.get(i) instanceof Function)
						functions.add(i);
					else
						terminals.add(i);
				}
				selector1 = rand.nextDouble();
				selector2 = rand.nextDouble();
				if(selector1 < functionProportion)
				{
					if(useExpectedDepth)
						crossoverPoint2 = ind2.getModificationPointFromPoissonDistribution(expectedDepth, functions, true);
					else
						crossoverPoint2 = functions.get((int)(selector2 * functions.size()));
				}
				else
				{
					if(useExpectedDepth)
						crossoverPoint2 = ind2.getModificationPointFromPoissonDistribution(expectedDepth, terminals, true);
					else
						crossoverPoint2 = terminals.get((int)(selector2 * terminals.size()));
				}
			}

			Individual[] offspring = cross(ind1,ind2,crossoverPoint1,crossoverPoint2,twoOffspring, inputs, constants);
			if(offspring[0].getLength() <= maxLength && offspring[0].getDepth() <= maxDepth)
			{
				if(offspring.length==2 && offspring[1].getLength()>maxLength || offspring[1].getDepth()>maxDepth)
					offspring[1] = null;
				return offspring;
			}
			else
			{
				offspring[0] = null;
				if(offspring.length == 2 && offspring[1].getLength() <= maxLength && offspring[1].getDepth() <= maxDepth)
					return offspring;
			}
		}
		return null;
	}

	/**halfBroodSize is the number of <i>pairs</i> of offspring created during brood recombination.
	For example, a half brood size of 4 will result in 8 (4 pairs) offspring,
	the best of which will pass through to the next generation.*/
	public void setBroodSize(int halfBroodSize)
	{
		setBroodSize(halfBroodSize,halfBroodSize*2);
	}

	/**halfBroodSize is the number of <i>pairs</i> of offspring created during brood recombination.
	For example, a half brood size of 4 will result in 8 (4 pairs) offspring,
	the best of which will bass through to the next generation. The dataSamplingFactor is the
	rate at which data will be sampled when assessing the fitness of brood members. For example a sampling factor of 10
	will result in 1 in 10 of the data being sampled.*/
	public void setBroodSize(int halfBroodSize, int dataSamplingFactor)
	{
		if(halfBroodSize<=0)
			System.err.println("halfBroodSize must be a positive integer.");
		else if(dataSamplingFactor<=0)
			System.err.println("dataSamplingFactor must be a positive integer.");
		else
		{
			samplingFactor = dataSamplingFactor;
			broodSize = halfBroodSize;
		}
	}

	/**halfBroodSize is the number of <i>pairs</i> of offspring created during brood recombination.
	For example, a half brood size of 4 will result in 8 (4 pairs) offspring,
	the best of which will pass through to the next generation.*/
	public int getHalfBroodSize()
	{
		return broodSize;
	}

	/**The dataSamplingFactor is a variable related to brood recombination. It is the
	rate at which data will be sampled when assessing the fitness of brood members. For example a sampling factor of 10
	will result in 1 in 10 of the data being sampled.*/
	public int getDataSamplingFactor()
	{
		return samplingFactor;
	}

	//not used
	/*public void setPerformanceFunctions(Population.PerformanceFunction[] perfFuncs)
	{
		this.perfFuncs = perfFuncs;
	}*/

	/**In order to use brood recombination the data, i.e. inputs and target outputs, must be set,
	allowing the identification of successful crossover. This method must therefore be called by a Population
	before an attempt is made at crossover.*/
	public void setData(double[][] inputs, double[] targets)
	{
		trainingInputs = inputs;
		targetOutputs = targets;
	}

	/**Sets whether brood recombination should be used.*/
	public void setBroodUse(boolean useBrood)
	{
		this.useBrood = useBrood;
	}

	/**Checks whether brood recombination is being used.*/
	public boolean getBroodUse()
	{
		return useBrood;
	}

	/**This method is only important when using brood recombination.*/
	public void setFitnessFunction(AbstractPopulation.FitnessFunction fitnessFunction)
	{
		this.fitnessFunction = fitnessFunction;
	}
}