package geneticWedge.gp;

import java.util.Arrays;

/**DemicPopulation is an extension of Population that allows only local breeding, and tournaments are perfomed within a local 'deme'.
The concept of a 'location' therefore exists for an Individual within a DemicPopulation.
<p>
When using a multi-objective fitnessFunction, a DemicPopulation may not behave in the expected way. The population size
of a DemicPopulation is always maintained, so setting keepDominatedIndividuals to false has no effect. Instead, the
number of Individuals <i>in the whole population</i> by which an Individual is dominated is used as the fitness function.
<p>
In a similar way, the DIVERSITY_DISTANCE measure finds the average distance of an Individual to the whole population,
not just to the Individuals within its deme. On the other hand, performanceFunctions that do not depend on other members
of the population do not have to consider Individuals outside the Individual's deme.*/
public class DemicPopulation extends Population
{
	int width = 1;
	int length = 1;
	int demeSize = 1;

	/**DemicPopulations have restricted breeding between nearby Individuals. Each Individual therefore has a notional
	geographical location. A DemicPopulation is arranged in a torus, with Individuals at the 'edges' of the population
	able to breed with Individuals on the opposite 'edge'. The total population size is equal to the <i>width</i> multiplied by the <i>length</i>.
	DemicPopulation is designed for use with RANDOM or TOURNAMENT selection. These use local tournaments to determine
	which Individuals to remove and which will parent the replacement Individual. Other selection methods use a roulette wheel
	to select the Individual to be removed and then perform local tournaments to select the parents of the replacement offspring.

	@param width the width of the population
	@param length the length of the population
	@param demeSize the height and width of a deme used to select breeding parents.*/
	public DemicPopulation(int width, int length, int demeSize, PopulationDescriptor desc)
	{
		super(width*length, desc);
		this.width = width;
		this.length = length;
		this.demeSize = demeSize;
	}

	/**If this method is called, the selection method will be deduced from the value
	of steadyState. If true, the selection method will be set to TOURNAMENT. If false
	the selection method will be set to FITNESS_PROPORTIONAL.*/
	public boolean evolve(boolean steadyState, double[][]inputs, double[]targetOutputs, double[][]validationInputs, double[]validationTargetOutputs, StoppingCriterion stopCriterion)
	{
		if(steadyState)
			return evolve(steadyState, SelectionMethod.TOURNAMENT, inputs, targetOutputs, validationInputs, validationTargetOutputs, stopCriterion);
		else
			return evolve(steadyState, SelectionMethod.FITNESS_PROPORTIONAL, inputs, targetOutputs, validationInputs, validationTargetOutputs, stopCriterion);
	}


	/**DemicPopulation is designed to be used with tournament selection in steady state mode. It is possible to run this
	method with alternative inputs, but this may not give the desired output.
	<p>
	EXAMPLE
	<p>
	If this method is called in steadyState mode with a selectMethod that uses roulette wheel selection
	(FITNESS_PROPORTIONAL, RANK or FUSS), elitist selection will not occur, even if an elite size has been set.
	In order to acieve elitist selection with these methods, generational evolution must be applied.
	However, as Individuals are added into the next generation there is no guarantee that offspring bred from
	nearby parents will be near to each other, if using generational  mode.*/
	public boolean evolve(boolean steadyState, SelectionMethod selectMethod, double[][]inputs, double[]targetOutputs, double[][]validationInputs, double[]validationTargetOutputs, StoppingCriterion stopCriterion)
	{
		boolean distanceSet = false;
		if(performanceFunctions!=null)
		{
			for(int i=0;i<performanceFunctions.length;i++)
			{
				switch(performanceFunctions[i])
				{
					case DIVERSITY_DISTANCE:
						setAllDistances();
						break;
					case FUNCTION_COMPLEXITY:
						setAllComplexities();
						break;
					case VARIABLE_MAX:
					case VARIABLE_MIN:
						setAllNumberOfInputsUsed();
						break;
				}
			}
			if(isMultiObjective)
				setDominatedMatrix();
		}
		if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
			setAllDistances();

		try
		{
			if(crossover.getBroodUse())
				crossover.setData(inputs,targetOutputs);
			setFitnessValues(inputs,targetOutputs);
			if(stopCriterion instanceof EvaluativeStoppingCriterion)
			{
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
						((EvaluativeStoppingCriterion)stopCriterion).setMaximise(false);
						break;
					case NUMBER_CORRECT_BINARY:
					case AUROC:
						((EvaluativeStoppingCriterion)stopCriterion).setMaximise(true);
						break;
				}

				if(((EvaluativeStoppingCriterion)stopCriterion).usesValidation())
					setValidationFitnessValues(validationInputs,validationTargetOutputs);
			}

			System.err.println("generation "+generationNumber+". Mean Fitness= "+averageTrainFitness+". Best fitness="+bestTrainFitness);
			//System.err.println("Mean val Fitness= "+averageValidationFitness+". Best val fitness="+bestValidationFitness);
			System.err.println("average length="+getAverageIndividualLength()+", diversity="+getDiversity());
			if(stopCriterion.stopIfNoDiversity()&&getDiversity()==0.0)
				return true;
			while(generationNumber<maxGenerations && !stopCriterion.stopReached(this))
			{
				evolveOneGeneration(steadyState,selectMethod, inputs,targetOutputs);
				setSummaryFitnessValues();
				if(stopCriterion instanceof EvaluativeStoppingCriterion && ((EvaluativeStoppingCriterion)stopCriterion).usesValidation())
					setValidationFitnessValues(validationInputs,validationTargetOutputs);
				if(stopCriterion.stopIfNoDiversity()&&getDiversity()==0.0)
				{
					for(int i=0;i<popSize;i++)
						System.err.println(individuals[i].toString());
					return true;
				}
			}
			setValidationFitnessValues(validationInputs,validationTargetOutputs);
		}
		catch(InvalidInputSizeException iise)
		{
			iise.printStackTrace();
			return false;
		}
		return true;
	}

	protected void evolveOneGeneration(boolean steadyState, SelectionMethod selectMethod, double[][]inputs, double[]targetOutputs)
	{
		runtime.gc();
		try
		{
			if(steadyState)
			{
				if(selectMethod.equals(SelectionMethod.TRUNCATION))
				{
					populationEvolved=0;
					while(!truncationEvolve(true,inputs,targetOutputs)){}
				}
				else
				{
					for(int i=0;i<popSize;i++)
					{
						if(!steadyStateEvolve(selectMethod, inputs,targetOutputs))
							i--;//redo evolution if it has failed
					}
				}
			}
			else
			{
				if(selectMethod.equals(SelectionMethod.TRUNCATION))
				{
					populationEvolved = 0;
					newPopulation = new Individual[popSize];

					//create elite
					if(eliteSize>0)
					{
						Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals()[0]);
						for(int i=0;i<eliteSize;i++)
							newPopulation[i] = orderedIndividuals[i];
						populationEvolved = eliteSize;
					}

					while(!truncationEvolve(false,inputs,targetOutputs)){}
				}
				else
					generationalEvolve(selectMethod, inputs,targetOutputs);

				boolean distanceSet = false;
				//update performance values
				if(performanceFunctions!=null)
				{
					for(int i=0;i<performanceFunctions.length;i++)
					{
						switch(performanceFunctions[i])
						{
							case DIVERSITY_DISTANCE:
								setAllDistances();
								distanceSet = true;
								break;
							case FUNCTION_COMPLEXITY:
								setAllComplexities();
								break;
							case VARIABLE_MAX:
							case VARIABLE_MIN:
								setAllNumberOfInputsUsed();
								break;
						}
					}
				}
				if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
					setAllDistances();
			}

			//setFitnessValues(inputs,targetOutputs);//better than adjusting after each new offspring - more accurate for large errors
		}
		catch(InvalidInputSizeException iise)
		{
			System.err.println("InvalidInputSizeException. Unable to evolve population");
			iise.printStackTrace();
			return;
		}
		catch(InvalidFractionException ife)
		{
			System.err.println("InvalidFractionException. Unable to evolve population");
			ife.printStackTrace();
			return;
		}
		generationNumber++;

		System.err.println("generation "+generationNumber+". Mean Fitness= "+averageTrainFitness+". Best Fitness="+bestTrainFitness);
		//System.err.println("Mean val Fitness= "+averageValidationFitness+". Best val fitness="+bestValidationFitness);
		System.err.println("average length="+getAverageIndividualLength()+", diversity="+getDiversity());
	}

	private boolean truncationEvolve(boolean steadyState, double[][] inputs,double[] targetOutputs)
	{
		//boolean finished = false;
		Individual[] parents = new Individual[mu];
		int[] parentIndices = new int[mu];
		int firstIndex = (int)(rand.nextDouble()*popSize);
		parentIndices[0] = firstIndex;
		parents[0]=individuals[firstIndex];
		for(int i=1;i<mu;i++)
		{
			int nextIndex = getSecondParentIndex(firstIndex);
			parentIndices[i] = nextIndex;
			parents[i]=individuals[nextIndex];
		}
		Individual[] offspring = null;
		double[] fitnesses;
		double[] savedFitnesses = null;
		if(parentsInTruncationPool)
		{
			offspring = new Individual[mu+lambda];
			fitnesses = new double[mu+lambda];
			if(isMultiObjective)
				savedFitnesses = new double[mu+lambda];
		}
		else
		{
			offspring = new Individual[lambda];
			fitnesses = new double[lambda];
			if(isMultiObjective)
				savedFitnesses = new double[lambda];
		}

		Individual[] tempOffspr = getMultipleOffspring(parents, lambda);
		double[][] performances = null;
		for(int i=0;i<lambda;i++)
		{
			offspring[i] = tempOffspr[i];
			try
			{
				fitnesses[i] = getFitness(offspring[i],inputs,targetOutputs);
			}
			catch(InvalidInputSizeException iise)
			{
				iise.printStackTrace();
				return false;
			}
			if(isMultiObjective)
				savedFitnesses[i] = fitnesses[i];

			/*rather than calling adjustFitnessForPerformance(), performance values are evaluated explicitly,
			so that they are available to adjust values for functionComplexities, etc.*/
			if(performanceFunctions!=null)
			{
				int len = performanceFunctions.length;
				performances = new double[lambda][len];
				if(fitnessFunction.equals(FitnessFunction.AUROC))
					fitnesses[i] = 2.0 * fitnesses[i] - 1.0;//normalise to range [0,1]
				else
					fitnesses[i] = 1.0 - 1.0 / (fitnesses[i]+1.0);//normalise to range [0,1]
				for(int j=0;j<len;j++)
				{
					//additionalFitnesses are normalised to range [0,1], with 1 as the best value.
					double additionalFitness = 0.0;
					switch(performanceFunctions[j])
					{
						case PROGRAM_LENGTH:
							performances[i][j] = offspring[i].getLength();
							additionalFitness = 1.0 / performances[i][j];
							break;
						case DIVERSITY_DISTANCE:
							performances[i][j] = getAverageSquaredDistance(offspring[i]);
							additionalFitness = performances[i][j];
							break;
						case FUNCTION_COMPLEXITY:
							performances[i][j] = getComplexity(offspring[i]) + 1.0;
							additionalFitness = 1.0 / performances[i][j];
							break;
						case VARIABLE_MAX:
							performances[i][j] = (double)(getNumberOfInputsUsed(offspring[i]));
							additionalFitness = performances[i][j] / popDescriptor.getInputs().length;
							break;
						case VARIABLE_MIN:
							performances[i][j] = (double)(getNumberOfInputsUsed(offspring[i]));
							additionalFitness = 1.0 - performances[i][j] / popDescriptor.getInputs().length;
							break;
					}
					if(!isMultiObjective)
					{
						switch(fitnessFunction)
						{
							case MSE:
							case RMSE:
							case ABSOLUTE_ERROR:
							case AOREC:
								fitnesses[i] += 1.0 - additionalFitness;//These FitnessFunctions aim to minimise value
								break;
							case NUMBER_CORRECT_BINARY:
							case AUROC:
								fitnesses[i] += additionalFitness;//These FitnessFunctions aim to maximise value
								break;
						}
					}
				}
				if(isMultiObjective)
				{
					try
					{
						//fitnesses[i] = getDominatedCount(offspring[i], inputs, targetOutputs);
						fitnesses[i] = getDominatedCount(fitnesses[i], performances[i], inputs, targetOutputs);
						if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
							fitnesses[i] = startingPopSize - fitnesses[i]; //alter so that least dominated has a HIGH value
					}
					catch(InvalidInputSizeException iise)
					{
						System.err.println("Exception while getting dominated count.");
						System.exit(1);
					}
				}
			}
		}

		if(parentsInTruncationPool)
		{
			for(int i=0;i<mu;i++)
			{
				offspring[lambda+i] = parents[i];
				fitnesses[lambda+i] = allFitnesses[parentIndices[i]];
				if(performanceFunctions != null)
				{
					if(isMultiObjective)
					{
						savedFitnesses[lambda+i] = fitnesses[lambda+i];
						fitnesses[lambda+i] = dominatedCount[parentIndices[i]];
						if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
							fitnesses[lambda+i] = startingPopSize - fitnesses[lambda+i]; //alter so that least dominated has a HIGH value
					}
					else
						adjustFitnessForPerformance(parentIndices[i], fitnesses[lambda+i]);
				}
			}
		}
		Object[] sortedObj = ShellSortIndividuals(offspring, fitnesses, fitnessFunction);
		int[] sortedIndices = (int[])(sortedObj[2]);

		if(steadyState)
		{
			int removalPoolSize = lambda;
			if(parentsInTruncationPool)
				removalPoolSize+=mu;
			int[] removalIndices=new int[removalPoolSize];
			Individual[] removalInds = new Individual[removalPoolSize];
			fitnesses = new double[removalPoolSize];
			//int index = (int)(rand.nextDouble() * popSize);
			int index = getSecondParentIndex(firstIndex);//select Individuals to be removed from the same locale as the parents
			removalIndices[0] = index;
			removalInds[0] = individuals[index];
			fitnesses[0] = allFitnesses[index];
			for(int i=1;i<removalPoolSize;i++)
			{
				//int nextIndex = getSecondParentIndex(index);
				int nextIndex = getSecondParentIndex(firstIndex);
				removalIndices[i] = nextIndex;
				removalInds[i] = individuals[nextIndex];
				fitnesses[i] = allFitnesses[nextIndex];
				if(performanceFunctions != null)
				{
					if(isMultiObjective)
					{
						fitnesses[i] = dominatedCount[index];
						if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
							fitnesses[i] = startingPopSize - fitnesses[i];
					}
					else
						adjustFitnessForPerformance(i, fitnesses[i]);
				}
			}
			Object[] sortedRemovals = ShellSortIndividuals(removalInds, fitnesses, fitnessFunction);
			int[] remIndices = (int[])(sortedRemovals[2]);

			int[] replacementIndices = new int[mu];
			for(int i=0;i<mu;i++)
			{
				replacementIndices[i] = removalIndices[remIndices[removalPoolSize-1-i]];
				individuals[removalIndices[remIndices[removalPoolSize-1-i]]] = ((Individual[])sortedObj[0])[i];
				allFitnesses[removalIndices[remIndices[removalPoolSize-1-i]]] = ((double[])sortedObj[1])[i];
			}
			setPerformanceValues(replacementIndices, performances);

			populationEvolved += mu;
			if(populationEvolved>=popSize)
			{
				populationEvolved = 0;
				return true;
			}
			else
				return false;
		}
		else //it doesn't really make sense having generational evolution with a DemicPopulation
		{
			for(int i=0;i<mu;i++)
			{
				newPopulation[populationEvolved] = ((Individual[])sortedObj[0])[i];
				populationEvolved++;
				if(populationEvolved == popSize)
				{
					individuals = newPopulation;
					populationEvolved = 0;
					return true;
				}
			}
			return false;
		}
		//return false;
	}

	private boolean steadyStateEvolve(SelectionMethod selectMethod, double[][] inputs,double[] targetOutputs) throws InvalidFractionException, InvalidInputSizeException
	{
		//select worst individual - this is done BEFORE selecting parents in contrast to a normal Population
		int worstIndex = -1;
		switch(selectMethod)
		{
			case TOURNAMENT:
			case RANDOM:
				worstIndex = getWorstIndexFromTournament(inputs,targetOutputs);
				break;
			case FUSS:
				/*FUSS uses the method in Population, i.e. searching across the whole population for an individual to remove.
				Demes are not easily applied to FUSS.*/
				worstIndex = getCrowdedIndexUsingFUSS();
				break;
			case FITNESS_PROPORTIONAL:
			case RANK:
				worstIndex = getIndexFromRouletteWheel(selectMethod, false);
				break;
		}


		Individual parent1=individuals[getBestIndexFromTournament(worstIndex,inputs,targetOutputs)];

		//create offspring
		Individual offspring;
		double selector=rand.nextDouble();

		if(selector<mutationProportion)
		{
			offspring=mut.mutate(parent1,maximumLength,maximumDepth,popDescriptor);
			if(offspring==null)
				return false;
		}
		else if(selector<mutationProportion+reproductionProportion)
		{
			offspring=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
		}
		else
		{
			Individual parent2 = individuals[getBestIndexFromTournament(worstIndex,inputs,targetOutputs)];
			Individual[] twoOffspring = crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
			if(twoOffspring==null)
			{
				//System.err.println("Unable to form offspring of appropriate size. Crossover will not occur.");
				return false;
			}
			else if(twoOffspring[0].getLength()>maximumLength || twoOffspring[0].getDepth()>maximumDepth)
				offspring=twoOffspring[1];
			else
				offspring=twoOffspring[0];
			//Individual offspring=Crossover.cross(parent1,parent2,false,popDescriptor.getInputs(),popDescriptor.getConstants())[0];//false indicates just one offspring
		}

		//get new fitness
		double newFitness=0.0;
		switch(fitnessFunction)
		{
			case MSE:
				newFitness = offspring.getMSE(inputs,targetOutputs);
				break;
			case RMSE:
				newFitness = offspring.getRMSE(inputs,targetOutputs);
				break;
			case ABSOLUTE_ERROR:
				newFitness = offspring.getAbsoluteError(inputs,targetOutputs);
				break;
			case NUMBER_CORRECT_BINARY:
				newFitness = offspring.getNumberCorrectBinary(inputs,targetOutputs);
				break;
			case AUROC:
				newFitness = offspring.getAUROC(inputs,targetOutputs,aurocSampleSize);
				break;
			case AOREC:
				newFitness = offspring.getAOREC(inputs,targetOutputs,aurocSampleSize,recFitnessFunction);
				break;
		}

		//replace worst individual by offspring
		individuals[worstIndex]=offspring;
		allFitnesses[worstIndex]=newFitness;

		//averageTrainFitness+=(newFitness-oldFitness)/popSize;
		switch(fitnessFunction)
		{
			case MSE:
			case RMSE:
			case ABSOLUTE_ERROR:
			case AOREC:
				if(newFitness<bestTrainFitness)
					bestTrainFitness=newFitness;
				break;
			case NUMBER_CORRECT_BINARY:
			case AUROC:
				if(newFitness>bestTrainFitness)
					bestTrainFitness=newFitness;
				break;
		}

		boolean distanceSet = false;
		if(performanceFunctions != null)
		{
			for(int i=0;i<performanceFunctions.length;i++)
			{
				switch(performanceFunctions[i])
				{
					case DIVERSITY_DISTANCE:
						for(int j=0;j<popSize;j++)
						{
							if(j<worstIndex)
								allSquaredDistances[j][worstIndex] = Math.pow(getDistance(worstIndex,j),2.0);
							else
								allSquaredDistances[worstIndex][j] = Math.pow(getDistance(worstIndex,j),2.0);
						}
						distanceSet = true;
						break;
					case FUNCTION_COMPLEXITY:
						functionComplexities[worstIndex] = getComplexity(offspring);
						break;
					case VARIABLE_MAX:
					case VARIABLE_MIN:
						noInputsUsed[worstIndex] = getNumberOfInputsUsed(offspring);
						break;
				}
			}
		}
		if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
		{
			for(int j=0;j<popSize;j++)
			{
				if(j<worstIndex)
					allSquaredDistances[j][worstIndex] = Math.pow(getDistance(worstIndex,j),2.0);
				else
					allSquaredDistances[worstIndex][j] = Math.pow(getDistance(worstIndex,j),2.0);
			}
		}

		return true;
	}


	//Not suitable for TRUNCATION selection
	private void generationalEvolve(SelectionMethod selectMethod, double[][] inputs,double[] targetOutputs) throws InvalidFractionException, InvalidInputSizeException
	{
		double[] rouletteVals=null;
		switch(selectMethod)
		{
			case FITNESS_PROPORTIONAL:
			case RANK:
			case FUSS:
				rouletteVals=getRouletteWheelValues(selectMethod, true);
				break;
			//for other methods do nothing
		}
		int len=popSize;
		Individual[] newInds=new Individual[len];
		if(eliteSize>0)
		{
			double[] fitnesses = new double[popSize];
			for(int i=0;i<popSize;i++)
			{
				fitnesses[i] = allFitnesses[i];
				adjustFitnessForPerformance(i,fitnesses[i]);
			}
			Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals(individuals, fitnesses, fitnessFunction)[0]);
			//Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals()[0]);
			for(int i=0;i<eliteSize;i++)
				newInds[i]=orderedIndividuals[i];
		}
		for(int i=eliteSize;i<len;i++)
		{
			Individual parent1=null;
			int demeCentre = (int)(rand.nextDouble() * popSize);
			switch(selectMethod)
			{
				case FITNESS_PROPORTIONAL:
				case RANK:
				case FUSS:
					parent1 = getIndividualFromRouletteWheel(rouletteVals);
					break;
				case TOURNAMENT:
				case RANDOM:
					parent1 = individuals[getBestIndexFromTournament(demeCentre,inputs, targetOutputs)];
					break;
			}
			double selector=rand.nextDouble();

			if(selector<mutationProportion)
			{
				newInds[i]=mut.mutate(parent1,maximumLength,maximumDepth,popDescriptor);
				//if mutation fails, use parent in next generation
				if(newInds[i]==null)
					newInds[i]=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			}
			else if(selector<mutationProportion+reproductionProportion)
			{
				newInds[i]=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			}
			else
			{
				Individual parent2 = null;
				switch(selectMethod)
				{
					case FITNESS_PROPORTIONAL:
					case RANK:
					case FUSS:
						parent2=getIndividualFromRouletteWheel(rouletteVals);
						break;
					case TOURNAMENT:
					case RANDOM:
						parent2 = individuals[getBestIndexFromTournament(demeCentre,inputs, targetOutputs)];
						break;
				}
				Individual[] twoOffspring=crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
				//if crossover fails, use one parent in next generation
				if(twoOffspring==null)
				{
					selector=rand.nextDouble();
					if(selector<0.5)
						newInds[i]=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
					else
						newInds[i]=new Individual(parent2.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
				}
				else if(twoOffspring[0].getLength()>maximumLength || twoOffspring[0].getDepth()>maximumDepth)
					newInds[i]=twoOffspring[1];
				else
					newInds[i]=twoOffspring[0];
			}
		}
		individuals = newInds;
	}

	/**If <i>addition</i> is true, roulette wheel values are set for the addition of an Individual
	to the breeding pool. Otherwise, roulette wheel values will be set for deletion from the population.*/
	private int getIndexFromRouletteWheel(SelectionMethod selectMethod, boolean addition)
	{
		int demePopSize = demeSize*demeSize;
		double[] rouletteWheelValues=new double[demePopSize];
		int demeCentre = (int)(rand.nextDouble() * popSize);

		int demex = demeCentre % width;
		int demey = demeCentre / width;
		double totalFitness=0.0;
		double fitness;
		switch(selectMethod)
		{
			case RANDOM:
				for(int i=0;i<demePopSize;i++)
					rouletteWheelValues[i] = ((double)(i+1))/demePopSize;
				break;

			case FITNESS_PROPORTIONAL:
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
						for(int i=0;i<demeSize;i++)
							for(int j=0;j<demeSize;j++)
							{
								int indexx = (width + demex - demeSize/2 +i) % width;
								int indexy = (length + demey - demeSize/2 +j) % length;
								int index = (indexy * width + indexx)/popSize;
								fitness = allFitnesses[index];
								adjustFitnessForPerformance(index, fitness);
								double val = 1.0 / (fitness + 1.0);
								if(!addition)
									val = 1.0-val;
								rouletteWheelValues[i * demeSize + j] = val + totalFitness;
								totalFitness+=val;
							}
						break;
					case NUMBER_CORRECT_BINARY:
						for(int i=0;i<demeSize;i++)
							for(int j=0;j<demeSize;j++)
							{
								int indexx = (width + demex - demeSize/2 +i) % width;
								int indexy = (length + demey - demeSize/2 +j) % length;
								int index = (indexy * width + indexx)/popSize;
								fitness = allFitnesses[index];
								adjustFitnessForPerformance(index, fitness);
								double val;
								if(performanceFunctions != null)
								{
									if(isMultiObjective)
										val = popSize - dominatedCount[i];
									else
									{
										adjustFitnessForPerformance(i , fitness);
										if(addition)
											val = fitness;
										else
											val = 1 + performanceFunctions.length - fitness;
									}
								}
								else
								{
									if(addition)
										val = fitness +1;
									else
										val = bestTrainFitness - fitness;
								}
								rouletteWheelValues[i * demeSize + j] = val + totalFitness;
								totalFitness+=val;
							}
						break;
					case AUROC:
						for(int i=0;i<demeSize;i++)
							for(int j=0;j<demeSize;j++)
							{
								int indexx = (width + demex - demeSize/2 +i) % width;
								int indexy = (length + demey - demeSize/2 +j) % length;
								int index = (indexy * width + indexx)/popSize;
								fitness = allFitnesses[index];
								adjustFitnessForPerformance(index, fitness);
								double val;
								if(performanceFunctions != null)
								{
									if(isMultiObjective)
										val = popSize - dominatedCount[i];
									else
									{
										adjustFitnessForPerformance(i , fitness);
										if(addition)
											val = fitness;
										else
											val = 1 + performanceFunctions.length - fitness;
									}
								}
								else
								{
									if(addition)
										val = fitness * 2.0 - 1.0;
									else
										val = (1.0 - fitness) * 2.0;
								}
								rouletteWheelValues[i * demeSize + j] = val + totalFitness;
								totalFitness+=val;
							}
						break;
				}
				break;

			case FUSS:
				Individual[] demeInds = new Individual[demePopSize];
				double[] demeFitnesses = new double[demePopSize];
				for(int i=0;i<demeSize;i++)
					for(int j=0;j<demeSize;j++)
					{
						int indexx = (width + demex - demeSize/2 +i) % width;
						int indexy = (length + demey - demeSize/2 +j) % length;
						int index = (indexy * width + indexx)/popSize;
						demeInds[i*demeSize+j] = individuals[index];
						demeFitnesses[i*demeSize+j] = allFitnesses[index];
						adjustFitnessForPerformance(index, demeFitnesses[i*demeSize+j]);
					}
				Object[] orderedObj = ShellSortIndividuals(demeInds, demeFitnesses, fitnessFunction);
				double[] orderedFitnesses = (double[])(orderedObj[1]);
				int[] orderedIndices = (int[])(orderedObj[2]);
				if(performanceFunctions==null || isMultiObjective)//otherwise, normalisation will have been performed by adjustFitnessForPerformance
				{
					switch(fitnessFunction)
					{
						case MSE:
						case RMSE:
						case ABSOLUTE_ERROR:
						case AOREC:
							for(int i=0;i<demePopSize;i++)
								orderedFitnesses[i] = 1.0/(orderedFitnesses[i]+1.0);//fit to range [0,1]. This avoids infinite values and gives more even spread of values
							break;
							//take no action for other fitnessFunctions
					}

				}
				double maxDistance = -1.0;
				if(orderedFitnesses[orderedFitnesses.length-1]>orderedFitnesses[0])
				{
					double averageWidth = (orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[0]) / (popSize-1);
					double val = averageWidth + orderedFitnesses[1] - orderedFitnesses[0];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[0]] = val;
					//totalFitness += val;
					for(int i=1;i<popSize-1;i++)
					{
						val = orderedFitnesses[i+1] - orderedFitnesses[i-1];
						if(val>maxDistance)
							maxDistance = val;
						rouletteWheelValues[orderedIndices[i]] = val;
						//totalFitness += val;
					}
					val = averageWidth + orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[orderedFitnesses.length-2];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[popSize-1]] = val;
					//totalFitness += val;
				}
				else
				{
					double averageWidth = (orderedFitnesses[0] - orderedFitnesses[orderedFitnesses.length-1]) / (popSize-1);
					double val = averageWidth + orderedFitnesses[0] - orderedFitnesses[1];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[0]] = val;
					//totalFitness += val;
					for(int i=1;i<popSize-1;i++)
					{
						val = orderedFitnesses[i-1] - orderedFitnesses[i+1];
						if(val>maxDistance)
							maxDistance = val;
						rouletteWheelValues[orderedIndices[i]] = val;
						//totalFitness += val;
					}
					val = averageWidth + orderedFitnesses[orderedFitnesses.length-2] - orderedFitnesses[orderedFitnesses.length-1];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[popSize-1]] = val;
					//totalFitness += val;
				}
				//totalFitness=0.0;
				for(int i=0;i<demePopSize;i++)
				{
					if(addition)
						totalFitness += rouletteWheelValues[i];
					else
						totalFitness += maxDistance - rouletteWheelValues[i];
					rouletteWheelValues[i] = totalFitness;
				}
				break;

			case RANK:
				Individual[] demInds = new Individual[demePopSize];
				double[] demFitnesses = new double[demePopSize];
				for(int i=0;i<demeSize;i++)
					for(int j=0;j<demeSize;j++)
					{
						int indexx = (width + demex - demeSize/2 +i) % width;
						int indexy = (length + demey - demeSize/2 +j) % length;
						int index = (indexy * width + indexx)/popSize;
						demInds[i*demeSize+j] = individuals[index];
						demFitnesses[i*demeSize+j] = allFitnesses[index];
						adjustFitnessForPerformance(index, demFitnesses[i*demeSize+j]);
					}
				Object[] orderedOb = ShellSortIndividuals(demInds, demFitnesses, fitnessFunction);
				int[] orderedI = (int[])(orderedOb[2]);
				for(int i=0;i<demePopSize;i++)
				{
					if(addition)
						rouletteWheelValues[orderedI[i]] = rankGradient * (demePopSize - i -1) +1;
					else
						rouletteWheelValues[orderedI[i]] = rankGradient * i + 1;
				}
				for(int i=0;i<demePopSize;i++)
				{
					totalFitness += rouletteWheelValues[i];
					rouletteWheelValues[i] = totalFitness;
				}
				break;

			case TOURNAMENT:
			case TRUNCATION:
				System.err.println("Attempt to obtain roulette wheel values for tournament or truncation selection. This is an invalid operation.");
				return -1;
		}
		for(int i=0;i<demePopSize;i++)
			rouletteWheelValues[i] /= totalFitness;

		double selector = 1.0 - rand.nextDouble();
		int selIndex = Arrays.binarySearch(rouletteWheelValues,selector);
		if(selIndex<0)
			selIndex=-(selIndex+1);
		int indexx = (width + demex - demeSize/2 + selIndex % width) % width;
		int indexy = (length + demey - demeSize/2 + selIndex / width) % length;
		int index = (indexy * width + indexx)/popSize;
		return index;
	}

	//This method performs tournament selection within a local deme
	private int getBestIndexFromTournament(int demeCentre, double[][] inputs, double[] targetOutputs) throws InvalidInputSizeException
	{
		if(popSize<tournamentSize)
		{
			System.err.println("Population size is less than tournament size.No evolution will occur.");
			return -1;
		}

		int bestIndex = getSecondParentIndex(demeCentre);
		double bestScore = allFitnesses[bestIndex];
		if(performanceFunctions != null)
		{
			if(isMultiObjective)
			{
				bestScore = dominatedCount[bestIndex];
				bestScore -= 1.0 / getShareValue(bestIndex);//adjust to give preference to unshared Individuals
				if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
					bestScore = popSize - bestScore;
			}
			else
			{
				adjustFitnessForPerformance(bestIndex , bestScore);
				bestScore /= getShareValue(bestIndex);
			}
		}
		else
			bestScore /= getShareValue(bestIndex);

		for(int i=1;i<tournamentSize;i++)
		{
			int index = getSecondParentIndex(demeCentre);
			double score=allFitnesses[index];
			if(performanceFunctions != null)
			{
				if(isMultiObjective)
				{
					score = dominatedCount[index];
					score -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
					if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
						score = popSize - score;
				}
				else
				{
					adjustFitnessForPerformance(index,score);
					score /= getShareValue(index);
				}
			}
			else
				score /= getShareValue(index);

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
					if(score<bestScore)
					{
						bestScore=score;
						bestIndex=index;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
					if(score>bestScore)
					{
						bestScore=score;
						bestIndex=index;
					}
					break;
			}
		}
		return bestIndex;
	}

	//Unlike the corresponding method in Population, this method uses a <i>local</i> tournament within a randomly selected deme.
	private int getWorstIndexFromTournament(double[][] inputs, double[] targetOutputs) throws InvalidInputSizeException
	{
		if(popSize<tournamentSize)
		{
			System.err.println("Population size is less than tournament size.No evolution will occur.");
			return -1;
		}
		int firstIndex = (int)(rand.nextDouble()*popSize);
		int worstIndex = firstIndex;
		double worstScore=allFitnesses[worstIndex];
		if(performanceFunctions != null)
		{
			if(isMultiObjective)
			{
				worstScore = dominatedCount[worstIndex];
				worstScore -= 1.0 / getShareValue(worstIndex);//adjust to give preference to unshared Individuals
				if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
					worstScore = -worstScore;
			}
			else
			{
				adjustFitnessForPerformance(worstIndex , worstScore);
				//worstScore /= getShareValue(worstIndex);
			}
		}
		//else
		//	worstScore /= getShareValue(worstIndex);

		for(int i=1;i<tournamentSize;i++)
		{
			int index=getSecondParentIndex(firstIndex);
			double score=allFitnesses[index];
			if(performanceFunctions != null)
			{
				if(isMultiObjective)
				{
					//make sure that best Individual is kept, i.e. elitist selection
					if(score == bestTrainFitness)
						score = dominatedCount[index] - 1;
					else
						score = dominatedCount[index];
					score -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
					if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC))
						score = -score;
				}
				else
				{
					adjustFitnessForPerformance(index , score);
					//score /= getShareValue(index);
				}
			}
			//else
			//	score /= getShareValue(index);

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
					if(score>worstScore)
					{
						worstScore=score;
						worstIndex=index;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
					if(score<worstScore)
					{
						worstScore=score;
						worstIndex=index;
					}
					break;
			}
		}

		return worstIndex;
	}

	private int getSecondParentIndex(int firstParentIndex)
	{
		int selector = (int)(rand.nextDouble()*demeSize*demeSize);
		int parentx = firstParentIndex % width;
		int parenty = firstParentIndex / width;
		int selectorx = selector % demeSize;
		int selectory = selector / demeSize;

		//get second parent position
		parentx = (width+parentx+selectorx-demeSize/2) % width;
		parenty = (length+parenty+selectory-demeSize/2) % length;

		return parenty * width + parentx;
	}

	/**Returns the width and length of this population. The overall size of this DemicPopulation is <em>width * length</em>.*/
	public int[] getDimensions()
	{
		return new int[]{width, length};
	}
}