package geneticWedge.gp;

/**The use of a FitnessShare object encourages diversity within a population. It 'shares' the fitness of Individuals
that are close together, either genotypically (similar tree structure) or phenotypically (similar fitness), making
them less fit. This aims to prevents premature convergence to a small niche.*/
public class FitnessShare
{
	/**FitnessShareFunction determines the shape of the fitness sharing within a niche.
	No fitness sharing will occur outside a niche.
	Sensitivity to the distance (the rate at which the sharing value falls away with distance)
	increases in the order COUNT, QUADRATIC, TRIANGULAR, SQRT.*/
	public enum FitnessShareFunction
	{
		/**COUNT simply divides the fitness by the number of Individuals within a fixed size niche.*/
		COUNT,
		/**With QUADRATIC sharing, the sharing coefficient falls off as the square of the distance to other Individuals within the niche.*/
		QUADRATIC,
		/**With TRIANGULAR sharing, the sharing coefficient falls off linearly with the distance to other Individuals within the niche.*/
		TRIANGULAR,
		/**With SQRT sharing, the sharing coefficient falls off as the square root of the distance to other Individuals within the niche.*/
		SQRT
	};

	/**ShareType determines the function over which sharing is performed*/
	public enum ShareType
	{
		/**GENE_DISTANCE sharing shares across similar genotypes*/
		GENE_DISTANCE,
		/**OBJECTIVE_FUNCTION sharing shares across Individuals with a similar Fitness (phenotype).*/
		OBJECTIVE_FUNCTION
	};

	private FitnessShareFunction shareFunction = FitnessShareFunction.TRIANGULAR;
	private double nicheSize;

	/**Creates a FitnessShare with a niche size and a FitnessShareFunction.
	@param shareFunction This determines the shape of the sharing function.*/
	public FitnessShare(double nicheSize, FitnessShareFunction shareFunction)
	{
		this.nicheSize = Math.abs(nicheSize);
		this.shareFunction = shareFunction;
	}

	/**Returns the share values (degree of sharing) for the data provided.
	@param data data should be either fitnesses or average squared genotypic distances, depending whether the ShareType is OBJECTIVE_FUNCTION or GENE_DISTANCE*/
	public double[] getShareValuesFromSingleObjective(double[] data)
	{
		int len = data.length;
		double[] share = new double[len];
		for(int i=0;i<len;i++)
		{
			share[i] = 1.0;//1.0 allows for the target point, which is definitely within the niche
			for(int j=i+1;j<len;j++)
			{
				double distance = Math.abs(data[i]-data[j]);
				double individualShare = getIndividualShare(distance);
				share[i] += individualShare;
				share[j] += individualShare;
			}
		}
		return share;
	}

	/**Returns the share value (degree of sharing) for the particular datum identified by index.
	@param data data should be either fitnesses or average squared genotypic distances, depending whether the ShareType is OBJECTIVE_FUNCTION or GENE_DISTANCE
	@param index The index of the target datum within the array <em>data</em>.*/
	public double getShareValuesFromSingleObjective(double[] data, int index)
	{
		int len = data.length;
		double share = 1.0;
		for(int i=0;i<len;i++)
		{
			if(i != index)
			{
				double distance = Math.abs(data[i]-data[index]);
				share += getIndividualShare(distance);
			}
		}
		return share;
	}

	/**Sets the niche size (radius). Only other Individuals within the niche size will affect sharing.*/
	public void setNicheSize(double nicheSize)
	{
		this.nicheSize = nicheSize;
	}

	/**Returns the niche size (radius). Only other Individuals within the niche size will affect sharing.*/
	public double getNicheSize()
	{
		return nicheSize;
	}

	/**This method takes weightings for a multi-objective function as the second input. The weights are applied
	by multiplying the components of the distances along each axis. Dimensions that have a small data range should
	therefore be assigned a large weighting and vice versa.
	<p>
	Weightings may be adjusted dynamically by calling this method. For example, a Population may adjust the weightings
	to the current ranges of the objectives.*/
	public double[] getShareValuesFromMultipleObjectives(double[][] data, double[] weighting)
	{
		int len = data.length;
		double[] share = new double[len];
		for(int i=0;i<len;i++)
		{
			share[i] = 1.0;//1.0 allows for the target point, which is definitely within the niche
			for(int j=i+1;j<len;j++)
			{
				double distance = 0.0;
				if(weighting == null || weighting.length!=len)
					distance = getDistance(data[i],data[j]);
				else
					distance = getDistance(data[i],data[j],weighting);
				double individualShare = getIndividualShare(distance);
				share[i] += individualShare;
				share[j] += individualShare;
			}
		}
		return share;
	}

	//no longer used
	/*public double[] getShareValuesFromDistances(double[][] distances)
	{
		int len = distances.length;
		double[] share = new double[len];
		for(int i=0;i<len;i++)
		{
			share[i] = 1.0;//1.0 allows for the target point, which is definitely within the niche
			for(int j=i+1;j<len;j++)
			{
				double individualShare = getIndividualShare(distances[i][j]);
				share[i] += individualShare;
				share[j] += individualShare;
			}
		}
		return share;
	}*/

	/**Returns the share values (degree of sharing) for the squared distances provided.
	@param squaredDistances The distances between all pairs of Individuals. Only the values above the diagonal
	(2nd dimension greater than 1st dimension) are used.*/
	public double[] getShareValuesFromSquaredDistances(double[][] squaredDistances)
	{
		int len = squaredDistances.length;
		double[] share = new double[len];
		for(int i=0;i<len;i++)
		{
			share[i] = 1.0;//1.0 allows for the target point, which is definitely within the niche
			for(int j=i+1;j<len;j++)
			{
				double individualShare = getIndividualShare(Math.sqrt(squaredDistances[i][j]));
				share[i] += individualShare;
				share[j] += individualShare;
			}
		}
		return share;
	}

	/**Returns the share value (degree of sharing) for the particular datum identified by index.
	@param squaredDistances The distances between all pairs of Individuals
	@param index The index of the target datum within the array <em>data</em>.*/
	public double getShareValuesFromSquaredDistances(double[][] squaredDistances, int index)
	{
		int len = squaredDistances.length;
		double share = 1.0;
		for(int i=0;i<len;i++)
		{
			if(i>index)
				share += getIndividualShare(Math.sqrt(squaredDistances[index][i]));
			else if (i<index)
				share += getIndividualShare(Math.sqrt(squaredDistances[i][index]));
		}
		return share;
	}

	private double getDistance(double[] datum1, double[]datum2)
	{
		double squaredDistance = 0.0;
		for(int i=0;i<datum1.length;i++)
			squaredDistance += Math.pow(datum1[i] - datum2[i] , 2.0);
		return Math.sqrt(squaredDistance);
	}

	//this method does not check the lengths of the input arrays. It is assumed that this has been done by the calling method.
	private double getDistance(double[] datum1, double[]datum2, double[]weighting)
	{
		double squaredDistance = 0.0;
		for(int i=0;i<datum1.length;i++)
			squaredDistance += Math.pow(weighting[i] * (datum1[i] - datum2[i]) , 2.0);
		return Math.sqrt(squaredDistance);
	}

	private double getIndividualShare(double dist)
	{
		switch(shareFunction)
		{
			case COUNT:
				if(dist < nicheSize)
					return 1.0;
				else
					return 0.0;
			case TRIANGULAR:
				if(dist < nicheSize)
					return (nicheSize - dist) / nicheSize;
				else
					return 0.0;
			case QUADRATIC:
				if(dist < nicheSize)
					return 1.0 - Math.pow(dist / nicheSize , 2.0);
				else
					return 0.0;
			case SQRT:
				if(dist < nicheSize)
					return 1.0 - Math.sqrt(dist / nicheSize);
				else
					return 0.0;
			default:
				System.err.println("Unknown Sharing Function");
				return 0.0;
		}
	}

	/**Returns a String representation of this FitnessShare object, identifying its niche size and share function (niche shape)*/
	public String toString()
	{
		return "niches size: " + nicheSize + "share function: "+shareFunction;
	}
}