package geneticWedge.gp;

import java.util.Vector;
import java.util.Random;
import java.util.Arrays;
import java.awt.geom.*;

import geneticWedge.gp.function.*;
import geneticWedge.gp.dataAnalysis.*;

/*import cern.colt.matrix.*;
import cern.colt.matrix.impl.*;
import cern.colt.matrix.linalg.*;*/

/**The Individual class represents a single tree (program). A population is usually made up of a large number of Individuals.*/
public class Individual extends Object
{
	//functions and terminals are used when constructing new Individuals
	private Function[] functions;
	private Terminal[] terminals;

	/*backupLevels create binary branching. When constructing a tree, a backup point is created for TwoInputFunctions
	but not for OneInputFunctions. A different mechanism will be required if N-input functions are introduced.*/
	private Vector<Integer> backupLevels=new Vector<Integer>();

	private Vector<Component> nodes=new Vector<Component>();
	private int maxInitialDepth=10;//used when growing trees, should not be applied to crossover or mutation
	private int treeDepth=0;//the depth of the lowest leaf

	private Random rand=new Random();
	private double growFunctionProportion=0.7;

	private double savedThreshold = Double.NaN;
	private double savedAUROC = Double.NaN;
	private double savedAOREC = Double.NaN;
	private double savedCombinedError = Double.NaN;
	private double[] savedCentroids = null;
	private double[][] savedRegressionCoeffs = null;
	private int noClusters = 3;

	private int[] classificationData = null;

	private double[] poissonRoulette = null;
	private Vector<Vector<Integer>> depthInfo = null;

	/**Constructs an Individual according to parameters held by the IndividualDescriptor.
	These parameters are: maxTreeDepth, minTreeDepth, GrowMethod, functions, constants, inputs, growFunctionProportion*/
	public Individual(IndividualDescriptor descriptor) throws InvalidFractionException
	{
		//this(descriptor.getGrowMethod(),descriptor.getMaxTreeDepth(),descriptor.getNoInputs(),descriptor.getNoConstants(),descriptor.getGrowFunctionProportion());
		this(descriptor.getGrowMethod(),descriptor.getMaxTreeDepth(),descriptor.getMinTreeDepth(),descriptor.getFunctions(),descriptor.getInputs(),descriptor.getConstants(),descriptor.getGrowFunctionProportion());
	}

	/**Constructs an Individual from the parameters provided.*/
	public Individual(IndividualDescriptor.IndividualGrowMethod growMethod, int treeDepth, int minTreeDepth, Function[] functions, Input[] inputs, Constant[] constants, double growFunctionProportion) throws InvalidFractionException
	{
		//this(functions,inputs,constants);
		setComponents(functions,inputs,constants);
		setGrowFunctionProportion(growFunctionProportion);
		switch(growMethod)
		{
			case GROW:
				growIndividual(treeDepth, minTreeDepth);
				//System.out.println("depth="+this.treeDepth+", nodes="+toString());
				break;

			case FULL:
				fullIndividual(treeDepth);
				//System.out.println("depth="+this.treeDepth+", nodes="+toString());
				break;
		}
	}

	/**Constructs an Individual from the parameters provided, using the GROW method.*/
	public Individual(int treeDepth, Function[] functions, Input[] inputs, Constant[] constants, double growFunctionProportion) throws InvalidFractionException
	{
		//this(IndividualDescriptor.IndividualGrowMethod.GROW, treeDepth, noOfInputs, noOfConstants, growFunctionProportion); //assume if a growFunctionProportion is provided that the GROW method should be used
		this(IndividualDescriptor.IndividualGrowMethod.GROW, treeDepth, 0, functions, inputs, constants, growFunctionProportion); //assume if a growFunctionProportion is provided that the GROW method should be used
	}

	/**Conctructs an Individual from the parameters provided.
	@param minDepth This sets the minimum depth of the tree created. It only applies if the GROW method is used; if the FULL method is applied minDepth = treeDepth*/
	public Individual(IndividualDescriptor.IndividualGrowMethod growMethod, int treeDepth, int minDepth, Function[] functions, Input[] inputs, Constant[] constants)
	{
		//this(functions,inputs, constants);
		setComponents(functions,inputs, constants);
		switch(growMethod)
		{
			case GROW:
				growIndividual(treeDepth, minDepth);
				break;

			case FULL:
				fullIndividual(treeDepth);
				break;
		}
	}

	//creates an Individual containing no nodes
	/*public Individual(Function[] functions, Input[] inputs, Constant[] constants)
	{
		int noOfInputs=inputs.length;
		int noConstants=constants.length;
		terminals=new Terminal[noOfInputs+noConstants];
		for(int i=0;i<noOfInputs;i++)
			terminals[i]=inputs[i];
		for(int i=0;i<noConstants;i++)
			terminals[noOfInputs+i]=constants[i];
		this.functions=functions;
	}

	//creates an Individual containing no nodes
	public Individual(Function[] functions, Terminal[] terminals)
	{
		this.terminals = terminals;
		this.functions = functions;
	}*/

	public static Individual getIndividualFromString(String string, Function[] functions, Input[] inputs, Constant[] constants)
	{
		//System.err.println(string);
		String[]compStr = string.split(",");
		Vector<Component> vecTree = new Vector<Component>();
		Vector<Constant> additionalConstants = new Vector<Constant>();
nextComp:
		for(int j=0;j<compStr.length;j++)
		{
			boolean found = false;
			for(int k=0;k<functions.length;k++)
			{
				String str=compStr[j].trim();

				//deal with legacy code!
				if(str.equals("*")||str.equals("×"))
					str="\u00D7";
				else if(str.equals("/")||str.equals("÷"))
					str="\u00F7";

				if(str.equals(functions[k].toString()))
				{
					vecTree.add(functions[k]);
					found = true;
					continue nextComp;
				}
			}
			for(int k=0;k<inputs.length;k++)
			{
				if(compStr[j].trim().equals(inputs[k].toString()))
				{
					vecTree.add(inputs[k]);
					found = true;
					continue nextComp;
				}
			}
			for(int k=0;k<constants.length;k++)
			{
				if(compStr[j].trim().equals(constants[k].toString()))
				{
					vecTree.add(constants[k]);
					found = true;
					continue nextComp;
				}
			}
			//System.err.println("not found:"+compStr[j]+". Adding new constant.");
			Constant newConst = new Constant(Double.parseDouble(compStr[j]));
			additionalConstants.add(newConst);
			vecTree.add(newConst);
		}
		if(additionalConstants.size()>0)
		{
			Constant[] tempConst = new Constant[constants.length+vecTree.size()];
			for(int j=0;j<constants.length;j++)
				tempConst[j] = constants[j];
			for(int j=0;j<additionalConstants.size();j++)
				tempConst[j+constants.length] = additionalConstants.get(j);
			constants = tempConst;
		}
		return new Individual(vecTree, inputs, constants);
	}

	private void setComponents(Function[] functions, Input[] inputs, Constant[] constants)
	{
		int noOfInputs=inputs.length;
		int noConstants=constants.length;
		terminals=new Terminal[noOfInputs+noConstants];
		for(int i=0;i<noOfInputs;i++)
			terminals[i]=inputs[i];
		for(int i=0;i<noConstants;i++)
			terminals[noOfInputs+i]=constants[i];
		this.functions=functions;
	}

	private void setComponents(Function[] functions, Terminal[] terminals)
	{
		this.terminals = terminals;
		this.functions = functions;
	}

	private void setComponents(Input[] inputs, Constant[] constants)
	{
		int noOfInputs=inputs.length;
		int noConstants=constants.length;
		terminals=new Terminal[noOfInputs+noConstants];
		for(int i=0;i<noOfInputs;i++)
			terminals[i]=inputs[i];
		for(int i=0;i<noConstants;i++)
			terminals[noOfInputs+i]=constants[i];
	}

	/**Creates a clone of this Individual, containing the same components in the same tree structure*/
	public Individual clone()
	{
		Individual clone = new Individual((Vector<Component>)nodes.clone(), terminals);
		clone.setClassificationData(classificationData);
		clone.setSavedValues(savedThreshold, savedAUROC, savedAOREC);
		return clone;
	}

	/**Constructs a new Individual containing the Components in newNodes in the specified order.
	Nodes are added from root to leaf and left to right.
	@param terminals The terminals should be provided in order to allow the inputs to be changed for different data.*/
	public Individual(Vector<Component> newNodes, Terminal[] terminals)
	{
		//this(new Function[]{},terminals);
		this.terminals = terminals;
		int currentLevel=0;
		nodes=new Vector<Component>();
		for(int i=0;i<newNodes.size();i++)
		{
			Component node=newNodes.get(i);
			currentLevel++;
			if(currentLevel>treeDepth)
				treeDepth=currentLevel;
			//if(!(node instanceof OneInputFunction))//If it is a one input function we do not have to return to this point (no branching)
			if(node instanceof TwoInputFunction)//If it is a one input function or a terminal we do not have to return to this point (no branching)
				backupLevels.add(currentLevel);
			nodes.add(node);
			if(i<newNodes.size()-1 && node instanceof Terminal)
			{
				currentLevel=backupLevels.get(backupLevels.size()-1);
				backupLevels.remove(backupLevels.size()-1);
			}
		}
	}

	/**Constructs a new Individual containing the Components in newNodes in the specified order.
	Nodes are added from root to leaf and left to right.
	@param inputs The inputs should be provided in order to allow the inputs to be changed for different data.*/
	public Individual(Vector<Component> newNodes, Input[] inputs, Constant[] constants)
	{
		//this(new Function[]{},inputs, constants);
		setComponents(inputs,constants);
		int currentLevel=0;
		nodes=new Vector<Component>();
		for(int i=0;i<newNodes.size();i++)
		{
			Component node=newNodes.get(i);
			currentLevel++;
			if(currentLevel>treeDepth)
				treeDepth=currentLevel;
			//if(!(node instanceof OneInputFunction))//If it is a one input function we do not have to return to this point (no branching)
			if(node instanceof TwoInputFunction)//If it is a one input function or a terminal we do not have to return to this point (no branching)
				backupLevels.add(currentLevel);
			nodes.add(node);
			if(i<newNodes.size()-1 && node instanceof Terminal)
			{
				currentLevel=backupLevels.get(backupLevels.size()-1);
				backupLevels.remove(backupLevels.size()-1);
			}
		}
	}

	/**Returns the Root Mean Square Error for the provided dataset. An exception will be thrown if any data
	contain the wrong number of inputs*/
	public double getRMSE(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return Math.sqrt(getMSE(inputs,targetOutputs));
	}

	/**Returns the Mean Square Error for the provided dataset. An exception will be thrown if any data
	contain the wrong number of inputs*/
	public double getMSE(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double mse=0.0;
		for(int i=0;i<len;i++)
			mse+=getSquaredError(inputs[i],targetOutputs[i]);
		mse/=len;
		return mse;
	}

	/**Returns the Absolute Error for the provided dataset. An exception will be thrown if any data
	contain the wrong number of inputs*/
	public double getAbsoluteError(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double err=0.0;
		for(int i=0;i<len;i++)
			err+=getAbsoluteError(inputs[i],targetOutputs[i]);
		err/=len;
		return err;
	}

	/**Returns the number of correctly classified outputs from the provided dataset. A classification is considered correct if
	the predicted output has the same sign as the target output. An exception will be thrown if any data
	contain the wrong number of inputs*/
	public double getNumberCorrectBinary(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getNumberCorrectBinary(inputs, targetOutputs, 0.0);
	}

	/**Returns the number of correctly classified outputs from the provided dataset. A classification is considered correct if
	the predicted output exceeds the threshold and the target is positive or if the predicted output is less than or equal
	to the threshold and the target output is non-positive. An exception will be thrown if any data
	contain the wrong number of inputs*/
	public double getNumberCorrectBinary(double[][] inputs,double[] targetOutputs, double threshold) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double correct=0.0;
		for(int i=0;i<len;i++)
		{
			//if((getOutput(inputs[i])-threshold)*targetOutputs[i]>0.0)
			double out = getOutput(inputs[i])-threshold;
			if((out>0.0 && targetOutputs[i]>0.0)||out<=0.0 && targetOutputs[i]<=0.0)
				correct++;
		}
		return correct;
	}

	/**The first time this method is called, the cluster centroids will be fixed. It MUST therefore be called
	using the training data before being called with test data.*/
	/*
	public double getClusteredRegressionMSE(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		Algebra alg = new Algebra();
		int dataLen = inputs.length;
		double[][] transformedInputs = new double[dataLen][1];
		for(int i=0;i<dataLen;i++)
			transformedInputs[i][0] = getOutput(inputs[i]);

		int dimensionality = inputs[0].length;

		double mse = 0.0;
		if(savedCentroids == null)
		{
			Vector<Integer>[] clusters = KMeans.clusterData(transformedInputs,noClusters);
			double[][] temp = KMeans.getLastCentroids();
			savedCentroids = new double[noClusters];
			savedRegressionCoeffs = new double[noClusters][dimensionality+1];
			for(int i=0;i<noClusters;i++)
			{
				savedCentroids[i] = temp[i][0];

				int clusterSize = clusters[i].size();

				if(clusterSize==1) //for single member cluster, just predict output value
				{
					for(int j=0;j<dimensionality;j++)
						savedRegressionCoeffs[i][j] = 0.0;
					savedRegressionCoeffs[i][dimensionality] = targetOutputs[clusters[i].get(0)];
				}
				else
				{
					double[][] inputData = new double[clusterSize][dimensionality+1];
					double[][] outputData = new double[clusterSize][1];
					for(int j=0;j<clusterSize;j++)
					{
						for(int k=0;k<dimensionality;k++)
							inputData[j][k] = inputs[clusters[i].get(j)][k];
						inputData[j][dimensionality] = 1.0;
						outputData[j][0] = targetOutputs[clusters[i].get(j)];
					}
					//NOTE: the sample size must be at least as large as the dimensionality. Otherwise there are an
					//infinite number of solutions and an Exception will be thrown. In this situation, set the output
					//to the average of the output values within the cluster
					if(inputData.length>=dimensionality+1)
					{
						DenseDoubleMatrix2D X = new DenseDoubleMatrix2D(inputData);
						DenseDoubleMatrix2D Y = new DenseDoubleMatrix2D(outputData);
						DoubleMatrix2D mat = alg.solve(X,Y);
						for(int j=0;j<dimensionality+1;j++)
							savedRegressionCoeffs[i][j] = mat.get(j,0);
					}
					else
					{
						for(int j=0;j<dimensionality;j++)
							savedRegressionCoeffs[i][j] = 0.0;
						double averageOut = 0.0;
						for(int j=0;j<clusterSize;j++)
							averageOut += outputData[j][0];
						savedRegressionCoeffs[i][dimensionality] = averageOut/clusterSize;
					}

					for(int j=0;j<clusterSize;j++)
					{
						double out = 0.0;
						for(int k=0;k<dimensionality;k++)
							out += savedRegressionCoeffs[i][k] * inputs[clusters[i].get(j)][k];
						out += savedRegressionCoeffs[i][dimensionality];
						mse += Math.pow(out - targetOutputs[clusters[i].get(j)], 2.0);
					}
				}
			}
		}
		else
		{
			Vector<Integer>[] clusters = new Vector[noClusters];
			for(int i=0;i<noClusters;i++)
				clusters[i] = new Vector<Integer>();
			for(int i=0;i<transformedInputs.length;i++)
			{
				int index = -1;
				double minDistance = Double.MAX_VALUE;
				for(int j=0;j<noClusters;j++)
				{
					double dist = Math.pow(transformedInputs[i][0]-savedCentroids[j],2.0);

					if(dist<minDistance)
					{
						minDistance = dist;
						index = j;
					}
				}
				clusters[index].add(i);
			}
			for(int i=0;i<noClusters;i++)
			{
				int clusterSize = clusters[i].size();
				for(int j=0;j<clusterSize;j++)
				{
					double out = 0.0;
					for(int k=0;k<dimensionality;k++)
						out += savedRegressionCoeffs[i][k] * inputs[clusters[i].get(j)][k];
					out += inputs[clusters[i].get(j)][dimensionality];
					mse += Math.pow(out - targetOutputs[clusters[i].get(j)], 2.0);
				}
			}
		}
		return mse/dataLen;
	}*/

	/**This returns an array containing 4 values - True Positives, False Positives, True Negatives and False Negatives.*/
	public int[] getAllClassificationData(double[][] inputs, double[] targetOutputs, double threshold, boolean useCachedValues, boolean saveValues) throws InvalidInputSizeException
	{
		int[] tempClassData;
		if(!useCachedValues || classificationData == null)
		{
			int len=inputs.length;
			if(len!=targetOutputs.length)
				throw new InvalidInputSizeException();
			int TP, FP, TN, FN ;
			TP = FP = TN = FN = 0;
			for(int i=0;i<len;i++)
			{
				double output = getOutput(inputs[i])-threshold;
				if(targetOutputs[i]>0.0)
				{
					if(output>0.0)
						TP++;
					else
						FN++;
				}
				else
				{
					if(output>0.0)
						FP++;
					else
						TN++;
				}
			}
			tempClassData = new int[]{TP, FP, TN, FN};
			if(saveValues)
				//classificationData = tempClassData;
				classificationData = new int[]{TP, FP, TN, FN};
		}
		else
			tempClassData = classificationData;
		return tempClassData;
	}

	/**Returns the sensitivity during a classification task. Sensitivity is defined as TP/(TP+FN).*/
	public double getSensitivity(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getSensitivity(inputs, targetOutputs, 0.0);
	}

	/**Returns the sensitivity during a classification task. Sensitivity is defined as TP/(TP+FN).
	@param threshold The threshold used during classification. Outputs which exceed the threshold are considered to be 'true'.*/
	public double getSensitivity(double[][] inputs,double[] targetOutputs, double threshold) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double correct=0.0;
		double total = 0.0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]>0.0)
			{
				total++;
				if((getOutput(inputs[i])-threshold)>0.0)
					correct++;
			}
		}
		if(total>0.0)
			return correct/total;
		else
			return 0.0;
	}

	/**Returns the specificity during a classification task. Specificity is defined as TN/(TN+FP).*/
	public double getSpecificity(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getSpecificity(inputs, targetOutputs, 0.0);
	}

	/**Returns the specificity during a classification task. Specificity is defined as TN/(TN+FP).
	@param threshold The threshold used during classification. Outputs which exceed the threshold are considered to be 'true'.*/
	public double getSpecificity(double[][] inputs,double[] targetOutputs, double threshold) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double correct=0.0;
		double total = 0.0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]<=0.0)
			{
				total++;
				if((getOutput(inputs[i])-threshold)<=0.0)
					correct++;
			}
		}
		if(total>0.0)
			return correct/total;
		else
			return 0.0;
	}

	/**Returns the Positive Predictive Value (PPV) during a classification task. PPV is defined as TP/(TP+FP).
	It may be viewed as the confidence level of a positive prediction.*/
	public double getPositivePredictiveValue(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getPositivePredictiveValue(inputs, targetOutputs, 0.0);
	}

	/**Returns the Positive Predictive Value (PPV) during a classification task. PPV is defined as TP/(TP+FP).
	It may be viewed as the confidence level of a positive prediction.
	@param threshold The threshold used during classification. Outputs which exceed the threshold are considered to be 'true'.*/
	public double getPositivePredictiveValue(double[][] inputs,double[] targetOutputs, double threshold) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double correct=0.0;
		double total = 0.0;
		for(int i=0;i<len;i++)
		{
			double output = getOutput(inputs[i])-threshold;
			if(output>0.0)
			{
				total++;
				if(targetOutputs[i]>0.0)
					correct++;
			}
		}
		if(total>0.0)
			return correct/total;
		else
			return 0.0;
	}

	/**Returns the Negative Predictive Value (NPV) during a classification task. NPV is defined as TN/(TN+FN).
	It may be viewed as the confidence level of a negative prediction.*/
	public double getNegativePredictiveValue(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getNegativePredictiveValue(inputs, targetOutputs, 0.0);
	}

	/**Returns the Negative Predictive Value (NPV) during a classification task. NPV is defined as TN/(TN+FN).
	It may be viewed as the confidence level of a positive prediction.
	@param threshold The threshold used during classification. Outputs which exceed the threshold are considered to be 'true'.*/
	public double getNegativePredictiveValue(double[][] inputs,double[] targetOutputs, double threshold) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double correct=0.0;
		double total = 0.0;
		for(int i=0;i<len;i++)
		{
			double output = getOutput(inputs[i])-threshold;
			if(output<=0.0)
			{
				total++;
				if(targetOutputs[i]<=0.0)
					correct++;
			}
		}
		if(total>0.0)
			return correct/total;
		else
			return 0.0;
	}

	/**Returns the Area under the ROC curve (AUROC) during a classification task.
	It may be viewed as the confidence level of a positive prediction.
	@param sampleSize The number of data to be sampled when estimating the AUROC. 100 points is usually enough to give a reasonable estimate.*/
	public double getAUROC(double[][] inputs,double[] targetOutputs,int sampleSize) throws InvalidInputSizeException
	{
		return getAUROC(inputs,targetOutputs,sampleSize, false, false);
	}

	/**Returns the Area under the ROC curve (AUROC) during a classification task.
	The AUROC is a measure of the discriminatory power of a classifier.
	@param sampleSize The number of data to be sampled when estimating the AUROC. 100 points is usually enough to give a reasonable estimate.
	@param useSavedValue If true the AUROC will not be recalculated from the data provided. Instead the last saved value will be returned.
	@param saveValues This flag indicates that the AUROC value should be saved by this Individual. It may later be returned by calling this method with useSavedValue = true.*/
	public double getAUROC(double[][] inputs,double[] targetOutputs,int sampleSize, boolean useSavedValue, boolean saveValues) throws InvalidInputSizeException
	{
		double tempAUROC;
		if(!useSavedValue || Double.isNaN(savedAUROC))
		{
			int len = inputs.length;
			if(sampleSize >= len)//if sample size is too large, sample ALL points
				return getAUROC(inputs,targetOutputs, useSavedValue,saveValues);
			int[] indices = Utils.getRandomIndices(len, sampleSize);
			double[][] sampleInputs = new double[sampleSize][];
			double[] sampleTargetOutputs = new double[sampleSize];
			for(int i=0;i<sampleSize;i++)
			{
				sampleInputs[i] = inputs[indices[i]];
				sampleTargetOutputs[i] = targetOutputs[indices[i]];
			}
			Point2D.Double[] points = getAllROCCoordinates(sampleInputs, sampleTargetOutputs);//CHECK THIS
			//Point2D.Double[] points = getROCCoordinates(inputs, targetOutputs, sampleSize);
			if(points == null)
				return Double.NaN;
			if(saveValues)
				savedThreshold = getBestROCThreshold(points, sampleInputs);//cache value to be used later
			ShellSortPoints(points);
			double auroc = points[0].y * points[0].x;
			for(int i=1;i<sampleSize;i++)
				auroc+=(points[i].y+points[i-1].y)*(points[i].x-points[i-1].x);
			auroc+=(1.0+points[inputs.length-1].y)*(1.0-points[inputs.length-1].x);//080108 add last point at (1,1)
			tempAUROC = auroc / 2.0;
			if(saveValues)
				savedAUROC = tempAUROC;
		}
		else
			tempAUROC = savedAUROC;
		return tempAUROC;
	}

	/**Calculates and returns the Area under the ROC curve (AUROC) during a classification task, using all of the provided data.
	The AUROC is a measure of the discriminatory power of a classifier.*/
	public double getAUROC(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		return getAUROC(inputs, targetOutputs, false, false);
	}

	/**Calculates and returns the Area under the ROC curve (AUROC) during a classification task, using all of the provided data.
	The AUROC is a measure of the discriminatory power of a classifier.
	@param useSavedValue If true the AUROC will not be recalculated from the data provided. Instead the last saved value will be returned.
	@param saveValues This flag indicates that the AUROC value should be saved by this Individual. It may later be returned by calling this method with useSavedValue = true.*/
	public double getAUROC(double[][] inputs,double[] targetOutputs, boolean useSavedValue, boolean saveValues) throws InvalidInputSizeException
	{
		double tempAUROC;
		if(!useSavedValue || Double.isNaN(savedAUROC))
		{
			Point2D.Double[] points = null;
			points = getAllROCCoordinates(inputs, targetOutputs);
			if(points == null)
				return Double.NaN;
			if(saveValues)
				savedThreshold = getBestROCThreshold(points, inputs);//cache value to be used later
			ShellSortPoints(points);
			double auroc = points[0].y * points[0].x;
			for(int i=1;i<inputs.length;i++)
				auroc+=(points[i].y+points[i-1].y)*(points[i].x-points[i-1].x);
			auroc+=(1.0+points[inputs.length-1].y)*(1.0-points[inputs.length-1].x);//080108 add last point at (1,1)
			tempAUROC = auroc / 2.0;
			if(saveValues)
				savedAUROC = tempAUROC;
		}
		else
			tempAUROC = savedAUROC;
		return tempAUROC;
	}

	/**This is the sqrt of the sum of the squares of the false negative fraction and the false positive fraction during classification.
	This is the same as the distance between the 'best' point on a ROC curve and (0,1), i.e.the point used to calculate the 'best' thresdhold.*/
	public double getCombinedClassificationError(double[][] inputs,double[] targetOutputs, int sampleSize, boolean useSavedValue, boolean useSavedThresholdValue, boolean saveValues) throws InvalidInputSizeException
	{
		if(sampleSize>inputs.length||sampleSize<1)
			return getCombinedClassificationError(inputs, targetOutputs, useSavedValue, useSavedThresholdValue, saveValues);
		else
		{
			int[] indices = Utils.getRandomIndices(inputs.length, sampleSize);
			double[][] cutdownInputs = new double[sampleSize][];
			double[] cutdownTargets = new double[sampleSize];
			for(int i=0;i<sampleSize;i++)
			{
				cutdownInputs[i] = inputs[indices[i]];
				cutdownTargets[i] = targetOutputs[indices[i]];
			}
			return getCombinedClassificationError(cutdownInputs, cutdownTargets, useSavedValue, useSavedThresholdValue, saveValues);
		}
	}

	/**This is the sqrt of the sum of the squares of the false negative fraction and the false positive fraction during classification.
	This is the same as the distance between the 'best' point on a ROC curve and (0,1), i.e.the point used to calculate the 'best' thresdhold.*/
	public double getCombinedClassificationError(double[][] inputs,double[] targetOutputs, boolean useSavedValue, boolean useSavedThresholdValue, boolean saveValues) throws InvalidInputSizeException
	{
		double threshold;
		double combinedError;
		if(!useSavedValue || Double.isNaN(savedCombinedError))
		{
			Point2D.Double[] points = null;
			points = getAllROCCoordinates(inputs, targetOutputs);
			if(points == null)
				return Double.NaN;
			int bestIndex = getBestROCPoint(points);
			if(!useSavedThresholdValue || Double.isNaN(savedThreshold))
				threshold = getOutput(inputs[bestIndex]);
			else
				threshold = savedThreshold;
			combinedError = Math.sqrt(Math.pow(points[bestIndex].x,2) + Math.pow(1.0 - points[bestIndex].y,2));
			if(saveValues)//cache values to be used later
			{
				savedThreshold = threshold;
				savedCombinedError = combinedError;
			}
		}
		else
			combinedError = savedCombinedError;
		return combinedError;
	}


	/**This method returns the threshold value that results in the point on the ROC closest to (0,1).*/
	public double getBestROCThreshold(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		Point2D.Double[] points = getAllROCCoordinates(inputs,targetOutputs);
		savedThreshold = getBestROCThreshold(points,inputs);
		return savedThreshold;
	}

	/**This method returns the threshold value that results in the point on the ROC closest to (0,1).*/
	public double getBestROCThreshold(double[] outputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		Point2D.Double[] points = getAllROCCoordinates(outputs,targetOutputs);
		int bestIndex = getBestROCPoint(points);
		return outputs[bestIndex];
	}

	/**If <i>useCachedValue</i> is true, <i>inputs</i> and <i>targetOutputs</i> will only be used if a cached value is not available.*/
	public double getBestROCThreshold(double[][] inputs,double[] targetOutputs, boolean useCachedValue) throws InvalidInputSizeException
	{
		if(useCachedValue && !Double.isNaN(savedThreshold))
			return savedThreshold;
		else
		{
			savedThreshold = getBestROCThreshold(inputs, targetOutputs);
			return savedThreshold;
		}
	}

	private double getBestROCThreshold(Point2D.Double[] points, double[][] inputs) throws InvalidInputSizeException
	{
		if(points == null)
			return Double.NaN;
		int bestIndex = getBestROCPoint(points);
		return getOutput(inputs[bestIndex]);
	}

	private int getBestROCPoint(Point2D.Double[] points)
	{
		if(points == null)
			return -1;
		int bestIndex = -1;
		double bestDistance = Double.MAX_VALUE;
		int len = points.length;
		for(int i=0;i<len;i++)
		{
			double sqDistance = Math.pow(points[i].x,2)+Math.pow(1.0-points[i].y,2);//square of distance to (0,1)
			if(sqDistance<bestDistance)
			{
				bestIndex = i;
				bestDistance = sqDistance;
			}
		}
		return bestIndex;
	}

	/**This method is used in calculating AUROC values. It is a public method to enable other classes
	to obtain ROC coordinates to be used, for example, in visualisation.*/
	public Point2D.Double[] getAllROCCoordinates(double[][] inputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double[] thresholds = new double[len];
		for(int i=0;i<len;i++)
			thresholds[i] = getOutput(inputs[i]);

		double output;
		int[] truePositives = new int[len];
		int[] falsePositives = new int[len];
		double targetPositives = 0.0;
		double targetNegatives = 0.0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]>0.0)
				targetPositives++;
			else
				targetNegatives++;
			output = thresholds[i];
			for(int j=0;j<len;j++)
			{
				//if(output>=thresholds[j])
				if(output>thresholds[j])//181006
				{
					if(targetOutputs[i]>0.0)
						truePositives[j]++;
					else
						falsePositives[j]++;
				}
			}
		}
		if(targetPositives == 0 || targetNegatives == 0)
		{
			System.err.println("getAllROCCords returning null."+ (targetPositives == 0));
			int test = 0;
			for(int i=0;i<len;i++)
				if(targetOutputs[i] == 1)
					test++;
			return null;
		}
		Point2D.Double[] points = new Point2D.Double[len];
		for(int i=0;i<len;i++)
			points[i] = new Point2D.Double(falsePositives[i] / targetNegatives, truePositives[i] / targetPositives);
		return points;
	}

	/**This method is used in calculating AUROC values. It is a public method to enable other classes
	to obtain ROC coordinates to be used, for example, in visualisation.*/
	public Point2D.Double[] getAllROCCoordinates(double[] outputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		int len=outputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double[] thresholds = new double[len];
		for(int i=0;i<len;i++)
			thresholds[i] = outputs[i];

		double output;
		int[] truePositives = new int[len];
		int[] falsePositives = new int[len];
		double targetPositives = 0.0;
		double targetNegatives = 0.0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]>0.0)
				targetPositives++;
			else
				targetNegatives++;
			output = thresholds[i];
			for(int j=0;j<len;j++)
			{
				//if(output>=thresholds[j])
				if(output>thresholds[j])//181006
				{
					if(targetOutputs[i]>0.0)
						truePositives[j]++;
					else
						falsePositives[j]++;
				}
			}
		}
		if(targetPositives == 0 || targetNegatives == 0)
		{
			System.err.println("getAllROCCords returning null."+ (targetPositives == 0));
			int test = 0;
			for(int i=0;i<len;i++)
				if(targetOutputs[i] == 1)
					test++;
			return null;
		}
		Point2D.Double[] points = new Point2D.Double[len];
		for(int i=0;i<len;i++)
			points[i] = new Point2D.Double(falsePositives[i] / targetNegatives, truePositives[i] / targetPositives);
		return points;
	}

	/**This method is used in calculateing AUROC values. It is a public method to enable other classes
	to obtain ROC coordinates to be used, for example, in visualisation. A random selection of <i>sampleSize</i>
	points are returned.*/
	public Point2D.Double[] getROCCoordinates(double[][] inputs,double[] targetOutputs,int sampleSize) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		int[] indices = Utils.getRandomIndices(len, sampleSize);
		double[] thresholds = new double[sampleSize];
		for(int i=0;i<sampleSize;i++)
			thresholds[i] = getOutput(inputs[indices[i]]);

		double output;
		int[] truePositives = new int[sampleSize];
		int[] falsePositives = new int[sampleSize];
		double targetPositives = 0;
		double targetNegatives = 0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]>0.0)
				targetPositives++;
			else
				targetNegatives++;
			output = getOutput(inputs[i]);
			for(int j=0;j<sampleSize;j++)
			{
				if(output>=thresholds[j])
				{
					if(targetOutputs[i]>0.0)
						truePositives[j]++;
					else
						falsePositives[j]++;
				}
			}
		}
		Point2D.Double[] points = new Point2D.Double[sampleSize];
		for(int i=0;i<sampleSize;i++)
			points[i] = new Point2D.Double(falsePositives[i] / targetNegatives, truePositives[i] / targetPositives);
		return points;
	}

	/**Returns the Area over the REC curve (AOREC) during a regression task.
	The AOREC is analogous to the AUROC used to select good classifiers.
	@param sampleSize The number of data to be sampled when estimating the AOREC. 100 points is usually enough to give a reasonable estimate.*/
	public double getAOREC(double[][] inputs,double[] targetOutputs,int sampleSize,AbstractPopulation.FitnessFunction fitnessFunction) throws InvalidInputSizeException
	{
		return getAOREC(inputs,targetOutputs,sampleSize,fitnessFunction, false);
	}

	/**Returns the Area over the REC curve (AOREC) during a regression task.
	The AOREC is analogous to the AUROC used to select good classifiers.
	@param sampleSize The number of data to be sampled when estimating the AOREC. 100 points is usually enough to give a reasonable estimate.
	@param useSavedValue If true, this method will return the last calculated AOREC value rather than recalculating it.*/
	public double getAOREC(double[][] inputs,double[] targetOutputs,int sampleSize,AbstractPopulation.FitnessFunction fitnessFunction, boolean useSavedValue) throws InvalidInputSizeException
	{
		if(!useSavedValue || Double.isNaN(savedAOREC))
		{
			if(sampleSize >= inputs.length)//if sample size is too large, sample ALL points
				return getAOREC(inputs,targetOutputs,fitnessFunction);
			Point2D.Double[] points = getRECCoordinates(inputs, targetOutputs, sampleSize, fitnessFunction);
			double cutoff = getMaxNullResidual(targetOutputs,fitnessFunction);
			if(points[0].x>cutoff)
				return cutoff;

			//calculate area under curve, then subtract from total area of curve
			double aurec = points[0].y * points[0].x;
			for(int i=1;i<points.length;i++)
			{
				if(points[i].x>=cutoff)
				{
					aurec+=(cutoff-points[i-1].x)*(points[i-1].y+(points[i].y - points[i-1].y)*(cutoff-points[i-1].x)/(points[i].x - points[i-1].x));
					break;
				}
				aurec+=(points[i].y+points[i-1].y)*(points[i].x-points[i-1].x);
			}
			savedAOREC = points[points.length-1].x - aurec / 2.0;
		}
		return savedAOREC;
	}

	/**Returns the Area over the REC curve (AOREC) during a regression task.
	The AOREC is analogous to the AUROC used to select good classifiers.*/
	public double getAOREC(double[][] inputs,double[] targetOutputs,AbstractPopulation.FitnessFunction fitnessFunction) throws InvalidInputSizeException
	{
		return getAOREC(inputs,targetOutputs,fitnessFunction, false);
	}

	/**Returns the Area over the REC curve (AOREC) during a regression task.
	The AOREC is analogous to the AUROC used to select good classifiers.
	@param useSavedValue If true, this method will return the last calculated AOREC value rather than recalculating it.*/
	public double getAOREC(double[][] inputs,double[] targetOutputs,AbstractPopulation.FitnessFunction fitnessFunction, boolean useSavedValue) throws InvalidInputSizeException
	{
		if(!useSavedValue || Double.isNaN(savedAOREC))
		{
			Point2D.Double[] points = getAllRECCoordinates(inputs, targetOutputs, fitnessFunction);
			double cutoff = getMaxNullResidual(targetOutputs,fitnessFunction);
			if(points[0].x>cutoff)
				return cutoff;

			//calculate area under curve, then subtract from total area of curve
			double aurec = points[0].y * points[0].x;
			for(int i=1;i<points.length;i++)
			{
				if(points[i].x>=cutoff)
				{
					aurec+=(cutoff-points[i-1].x)*(points[i-1].y+(points[i].y - points[i-1].y)*(cutoff-points[i-1].x)/(points[i].x - points[i-1].x));
					break;
				}
				aurec+=(points[i].y+points[i-1].y)*(points[i].x-points[i-1].x);
			}
			savedAOREC = points[points.length-1].x - aurec / 2.0;
		}
		return savedAOREC;
	}

	private double getMaxNullResidual(double[]outputs,AbstractPopulation.FitnessFunction fitnessFunction)
	{
		double maxVal = -Double.MAX_VALUE;
		double minVal = Double.MAX_VALUE;
		double average = 0.0;
		for(int i=0;i<outputs.length;i++)
		{
			average+=outputs[i];
			if(outputs[i]>maxVal)
				maxVal = outputs[i];
			if(outputs[i]<minVal)
				minVal = outputs[i];
		}
		average/=outputs.length;
		double difference = maxVal - average;
		if(average - minVal > difference)
			difference = average - minVal;
		switch(fitnessFunction)
		{
			case MSE:
				return Math.pow(difference,2.0);
			case RMSE:
			case ABSOLUTE_ERROR:
				return difference;
			default:
				return Double.NaN;
		}
	}

	/**This method ia used in calculateing AOREC values. It is a public method to enable other classes
	to obtain REC coordinates to be used, for example, in visualisation.*/
	public Point2D.Double[] getAllRECCoordinates(double[][] inputs,double[] targetOutputs,AbstractPopulation.FitnessFunction fitnessFunction) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double[] residuals = new double[len];
		for(int i=0;i<len;i++)
		{
			switch(fitnessFunction)
			{
				case MSE:
					residuals[i] = getSquaredError(inputs[i],targetOutputs[i]);
					break;
				case RMSE:
					residuals[i] = Math.sqrt(getSquaredError(inputs[i],targetOutputs[i]));
					break;
				case ABSOLUTE_ERROR:
					residuals[i] = getAbsoluteError(inputs[i],targetOutputs[i]);
					break;
				default:
					System.err.println("Attempt to apply getAllRECCoordinates to "+fitnessFunction+" fitness function.");
					return null;
			}
		}
		Arrays.sort(residuals);
		Point2D.Double[] points = new Point2D.Double[len];
		int count = 0;
		double lastResid = 0.0;
		for(int i=0;i<len;i++)
		{
			points[count] = new Point2D.Double(residuals[i],(i+1.0)/len);
			if(residuals[i]>lastResid)
				count++;//only count on if the residual is not duplicated
			lastResid = residuals[i];
		}
		if(count<len)
		{
			Point2D.Double[] uniquePoints = new Point2D.Double[count];
			for(int i=0;i<count;i++)
				uniquePoints[i] = points[i];
			return uniquePoints;
		}
		else
			return points;
	}

	/**This method ia used in calculateing AOREC values. It is a public method to enable other classes
	to obtain REC coordinates to be used, for example, in visualisation. A random selection of <i>sampleSize</i>
	points are returned.*/
	public Point2D.Double[] getRECCoordinates(double[][] inputs,double[] targetOutputs,int sampleSize,AbstractPopulation.FitnessFunction fitnessFunction) throws InvalidInputSizeException
	{
		int len=inputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		int[] indices = Utils.getRandomIndices(len, sampleSize);
		double[] residuals = new double[sampleSize];
		for(int i=0;i<sampleSize;i++)
		{
			switch(fitnessFunction)
			{
				case MSE:
					residuals[i] = getSquaredError(inputs[indices[i]],targetOutputs[indices[i]]);
					break;
				case RMSE:
					residuals[i] = Math.sqrt(getSquaredError(inputs[indices[i]],targetOutputs[indices[i]]));
					break;
				case ABSOLUTE_ERROR:
					residuals[i] = getAbsoluteError(inputs[indices[i]],targetOutputs[indices[i]]);
					break;
				default:
					System.err.println("Attempt to apply getRECCoordinates to "+fitnessFunction+" fitness function.");
					return null;
			}
		}
		Arrays.sort(residuals);
		Point2D.Double[] points = new Point2D.Double[sampleSize];
		int count = 0;
		double lastResid = 0.0;
		for(int i=0;i<sampleSize;i++)
		{
			points[count] = new Point2D.Double(residuals[i],(i+1.0)/sampleSize);
			if(residuals[i]>lastResid)
				count++;//only count on if the residual is not duplicated
			lastResid = residuals[i];
		}
		if(count<sampleSize)
		{
			Point2D.Double[] uniquePoints = new Point2D.Double[count];
			for(int i=0;i<count;i++)
				uniquePoints[i] = points[i];
			return uniquePoints;
		}
		else
			return points;
	}

	/**Returns the square of the error for the single data-point provided.*/
	public double getSquaredError(double[] inputs,double targetOutput) throws InvalidInputSizeException
	{
		return Math.pow(getOutput(inputs)-targetOutput,2);
	}

	/**Returns the absolute error for the single data-point provided.*/
	public double getAbsoluteError(double[] inputs,double targetOutput) throws InvalidInputSizeException
	{
		return Math.abs(getOutput(inputs)-targetOutput);
	}

	/**Returns the threshold adjusted error, i.e. <em>output-threshold</em> for the single input provided.*/
	public double getThresholdAdjustedOutput(double[] input) throws InvalidInputSizeException
	{
		if(Double.isNaN(savedThreshold))
		{
			System.err.println("Threshold not set. Individual returning unadjusted output.");
			return getOutput(input);
		}
		else
			return (getOutput(input) - savedThreshold);
	}

	/**Returns the output of this Individual, given the provided input vector.*/
	public double getOutput(double[] input) throws InvalidInputSizeException
	{
		for(int i=0;i<input.length;i++)
		{
			try
			{
				((Input)terminals[i]).setValue(input[i]);
			}
			catch(Exception e)
			{
				throw new InvalidInputSizeException();
			}
		}

		if(nodes.get(0) instanceof Terminal)
			return ((Terminal)nodes.get(0)).getValue();
		//right-to-left and leaf-to-root - this should be faster and require less memory than going in the opposite direction
		else
		{
			int nodeCount=nodes.size()-1;
			Vector<Double> branchValue=new Vector<Double>();
			Component comp = null;

			try
			{
				while(nodeCount>=0)
				{
					comp = nodes.get(nodeCount);

					if (comp instanceof Terminal)
					{
						branchValue.add(((Terminal)comp).getValue());
					}
					else
					{
						if(comp instanceof TwoInputFunction)
						{
							branchValue.set(branchValue.size()-2,((Function)comp).getOutput(new double[]{branchValue.get(branchValue.size()-1),branchValue.get(branchValue.size()-2)}));
							branchValue.remove(branchValue.size()-1);
						}
						else
							branchValue.set(branchValue.size()-1,((Function)comp).getOutput(new double[]{branchValue.get(branchValue.size()-1)}));
					}
					nodeCount--;
				}
			}
			catch(Exception e)
			{
				e.printStackTrace();
				System.err.println("failed ind:"+toString());
			}

			return branchValue.get(0);
		}
	}

	/**Returns the function represented by this Individual, with all terms expanded out and collected. NOTE: This function is
	intended to be used with Individuals containing only arithmetic functions*/
	public String getExpandedStringFunction()
	{
		int noInputs=0;
		for(int i=0;i<terminals.length;i++)
		{
			if(terminals[i] instanceof Input)
				noInputs++;
		}

		if(nodes.get(0) instanceof Terminal)
			return ((Terminal)nodes.get(0)).toString();
		//right-to-left and leaf-to-root - this should be faster and require less memory than going in the opposite direction
		else
		{
			int nodeCount=nodes.size()-1;
			Vector<Vector<double[]>> branchValue=new Vector<Vector<double[]>>();
			Component comp = null;

			try
			{
				while(nodeCount>=0)
				{
					comp = nodes.get(nodeCount);

					if (comp instanceof Terminal)
					{
						double[] coeffs = new double[noInputs+1];
						if(comp instanceof Constant)
							coeffs[0]=((Constant)comp).getValue();
						else
						{
							int inputNo=0;
							for(int i=0;i<terminals.length;i++)
							{
								if(comp.equals(terminals[i]))
								{
									coeffs[0]=1.0;
									coeffs[inputNo+1]=1.0;
									break;
								}
								if(terminals[i] instanceof Input)
									inputNo++;
							}
						}
						Vector<double[]> vec =new Vector<double[]>();
						vec.add(coeffs);
						branchValue.add(vec);
					}
					else
					{
						if(comp instanceof AdditionFunction)
						{
							Vector<double[]> vec1 = branchValue.get(branchValue.size()-1);
							Vector<double[]> vec2 = branchValue.get(branchValue.size()-2);
							Vector<double[]> newVec = new Vector<double[]>(vec1);
							for(int i=0;i<vec2.size();i++)
							{
								double[] coeffs=vec2.get(i);
								boolean found=false;
nextNewVec:
								for(int j=0;j<newVec.size();j++)
								{
									double[] newCoeffs=newVec.get(i);
									for (int k=1;k<coeffs.length;k++)
									{
										if(coeffs[k]!=newCoeffs[k])
											continue nextNewVec;
									}
									found=true;
									newCoeffs[0]+=coeffs[0];
									break;
								}
								if(!found)
								newVec.add(coeffs);
							}
							branchValue.set(branchValue.size()-2,newVec);
							branchValue.remove(branchValue.size()-1);
						}
						else if(comp instanceof SubtractionFunction)
						{
							Vector<double[]> vec1 = branchValue.get(branchValue.size()-1);
							Vector<double[]> vec2 = branchValue.get(branchValue.size()-2);
							Vector<double[]> newVec = new Vector<double[]>(vec1);
							for(int i=0;i<vec2.size();i++)
							{
								double[] coeffs=vec2.get(i);
								coeffs[0]=-coeffs[0];//negative coefficient for subtraction
								boolean found=false;
nextNewVec:
								for(int j=0;j<newVec.size();j++)
								{
									double[] newCoeffs=newVec.get(i);
									for(int k=1;k<coeffs.length;k++)
									{
										if(coeffs[k]!=newCoeffs[k])
											continue nextNewVec;
									}
									found=true;
									newCoeffs[0]+=coeffs[0];
									break;
								}
								if(!found)
									newVec.add(coeffs);
							}
							branchValue.set(branchValue.size()-2,newVec);
							branchValue.remove(branchValue.size()-1);
						}
						else if(comp instanceof MultiplicationFunction)
						{
							Vector<double[]> vec1 = branchValue.get(branchValue.size()-1);
							Vector<double[]> vec2 = branchValue.get(branchValue.size()-2);
							Vector<double[]> newVec = new Vector<double[]>();
							for(int i=0;i<vec1.size();i++)
							{
								double[] coeffs1=vec1.get(i);
								for(int j=0;j<vec2.size();j++)
								{
									double[] coeffs2=vec2.get(i);
									double[] newCoeffs = new double[noInputs+1];
									newCoeffs[0]=coeffs1[0]*coeffs2[0];
									for(int k=1;k<noInputs+1;k++)
									{
										newCoeffs[k]=coeffs1[k]+coeffs2[k];
									}
									boolean found=false;
nextNewVec3:
									for(int k=0;k<newVec.size();k++)
									{
										double[] testCoeffs = newVec.get(k);
										for(int l=1;l<testCoeffs.length;l++)
										{
											if(testCoeffs[l]!=newCoeffs[l])
												continue nextNewVec3;
										}
										testCoeffs[0]+=newCoeffs[0];
										found=true;
									}
									if(!found)
										newVec.add(newCoeffs);
								}
							}
							branchValue.set(branchValue.size()-2,newVec);
							branchValue.remove(branchValue.size()-1);
						}
						else if(comp instanceof DivisionFunction)
						{
							Vector<double[]> vec1 = branchValue.get(branchValue.size()-1);
							Vector<double[]> vec2 = branchValue.get(branchValue.size()-2);
							Vector<double[]> newVec = new Vector<double[]>();
							if (vec2.size()==1)
							{
								double[] coeffs2=vec1.get(0);
								for(int i=0;i<vec1.size();i++)
								{
									double[] coeffs1=vec1.get(i);
									double[] newCoeffs=new double[coeffs1.length];
									for(int j=0;j<coeffs2.length;j++)
									{
										if(j<coeffs1.length)
											newCoeffs[j]=coeffs1[j]-coeffs2[j];
										else
											newCoeffs[j]=-coeffs2[j];
									}
									newVec.add(newCoeffs);
								}
								branchValue.set(branchValue.size()-2,newVec);
								branchValue.remove(branchValue.size()-1);
							}
							else
							{

							}
						}
						else
						{
							System.err.println("WARNING: attempt to use getExpandedStringFunction with Individual containing complex operators. Returning empty string.");
							return "";
						}
					}
					nodeCount--;
				}
			}
			catch(Exception e)
			{
				e.printStackTrace();
				System.err.println("failed ind:"+toString());
			}

			Vector<double[]> vec = branchValue.get(0);
			return "";//TO FINISH
		}
	}

	/**lambda should be a number between 0.0 and 1.0, i.e. the expected tree depth as a <em>fraction</em>
	of the Individual's tree depth.
	If reverseDistribution is false, the distribution will tail off for large values. If true,
	the distribution will tail off towards zero.*/
	public int getModificationPointFromPoissonDistribution(double lambda, boolean reverseDistribution)
	{
		if(poissonRoulette == null)
		{
			double[] poisson = DataAcquirer.getPoissonDistribution(lambda*treeDepth, 1, treeDepth);
			poissonRoulette = new double[treeDepth];
			double lastVal = 0.0;
			for(int i=0;i<treeDepth;i++)
			{
				if(reverseDistribution)
					poissonRoulette[i]=lastVal+poisson[treeDepth-1-i];
				else
					poissonRoulette[i]=lastVal+poisson[i];
				lastVal = poissonRoulette[i];
			}
		}
		double selector = 1.0 - rand.nextDouble();
		int index = Arrays.binarySearch(poissonRoulette,selector);
		if(index<0)
			index = -(index+1);
		int depth = index+1;
		return selectNodeFromDepth(depth);
	}

	/**lambda should be a number between 0.0 and 1.0, i.e. the expected tree depth as a <em>fraction</em>
	of the Individual's tree depth.
	If reverseDistribution is false, the distribution will tail off for large values. If true,
	the distribution will tail off towards zero.*/
	public int getModificationPointFromPoissonDistribution(double lambda, Vector<Integer> selectedIndices, boolean reverseDistribution)
	{
		if(poissonRoulette == null)
		{
			double[] poisson = DataAcquirer.getPoissonDistribution(lambda*treeDepth, 1, treeDepth);
			poissonRoulette = new double[treeDepth];
			double lastVal = 0.0;
			for(int i=0;i<treeDepth;i++)
			{
				if(reverseDistribution)
					poissonRoulette[i]=lastVal+poisson[treeDepth-1-i];
				else
					poissonRoulette[i]=lastVal+poisson[i];
				lastVal = poissonRoulette[i];
			}
		}
		double selector = 1.0 - rand.nextDouble();
		int index = Arrays.binarySearch(poissonRoulette,selector);
		if(index<0)
			index = -(index+1);
		int depth = index+1;
		return selectNodeFromDepth(selectedIndices, depth);
	}

	private int selectNodeFromDepth(int depth)
	{
		if(depthInfo == null)
		{
			depthInfo = new Vector<Vector<Integer>>();
			for(int i=0;i<treeDepth;i++)
				depthInfo.add(new Vector<Integer>());
			int lev = 0;
			for(int i=0;i<nodes.size();i++)
			{
				Component node=nodes.get(i);
				lev++;
				depthInfo.get(lev-1).add(i);
				if(node instanceof TwoInputFunction)//If it is a one input function or a terminal we do not have to return to this point (no branching)
					backupLevels.add(lev);
				if(i<nodes.size()-1 && node instanceof Terminal)
				{
					lev=backupLevels.get(backupLevels.size()-1);
					backupLevels.remove(backupLevels.size()-1);
				}
			}
		}
		Vector<Integer> nodeList = depthInfo.get(depth-1);
		int selector = (int)(rand.nextDouble()*nodeList.size());
		int ret = nodeList.get(selector);
		if(ret<0||ret>=nodes.size())
			System.err.println("("+ret+","+nodes.size()+")");
		return ret;
	}

	private int selectNodeFromDepth(Vector<Integer> indices, int depth)
	{
		if(depthInfo == null)
		{
			depthInfo = new Vector<Vector<Integer>>();
			for(int i=0;i<treeDepth;i++)
				depthInfo.add(new Vector<Integer>());
				int lev = 0;
			for(int i=0;i<nodes.size();i++)
			{
				Component node=nodes.get(i);
				lev++;
				depthInfo.get(lev-1).add(i);
				if(node instanceof TwoInputFunction)//If it is a one input function or a terminal we do not have to return to this point (no branching)
					backupLevels.add(lev);
				if(i<nodes.size()-1 && node instanceof Terminal)
				{
					lev=backupLevels.get(backupLevels.size()-1);
					backupLevels.remove(backupLevels.size()-1);
				}
			}
		}
		Vector<Integer> nodeList = depthInfo.get(depth-1);
		Vector<Integer> nodesAtDepth = new Vector<Integer>();
		for(int i=0;i<indices.size();i++)
		{
			int target = indices.get(i);
			if(nodeList.contains(target))
				nodesAtDepth.add(target);
		}
		if(nodesAtDepth.size()>0)
		{
			int selector = (int)(rand.nextDouble()*nodesAtDepth.size());
			int ret = nodesAtDepth.get(selector);
			if(ret<0||ret>=nodes.size())
				System.err.println("("+ret+","+nodes.size()+")");
			return ret;
		}
		else //none of the nodes in the Vector provided are at the correct depth, therefore randomly select (ignoring depth)
		{
			int selector = (int)(rand.nextDouble()*indices.size());
			int ret = indices.get(selector);
			if(ret<0||ret>=nodes.size())
				System.err.println("("+ret+","+nodes.size()+")");
			return ret;
		}
	}

	/**Returns the length (number of nodes) in this Individual.*/
	public int getLength()
	{
		return nodes.size();
	}

	/**Returns the depth of this Individual.*/
	public int getDepth()
	{
		return treeDepth;
	}

	/**Sets the probability of a node selected during growth being a Function (rather than a Terminal).
	This indirectly controls the average tree size.*/
	public void setGrowFunctionProportion(double proportion) throws InvalidFractionException
	{
		if(proportion<=0.0||proportion>=1.0)
			throw new InvalidFractionException();
		growFunctionProportion=proportion;
	}

	/**Returns the probability of a node selected during growth being a Function (rather than a Terminal).
	This parameter indirectly controls the average tree size.*/
	public double getGrowFunctionProportion()
	{
		return growFunctionProportion;
	}

	private void growIndividual(int treeDepth, int minTreeDepth)
	{
		if(minTreeDepth<1)
		{
			System.err.println("Attempt to grow Individual with minimum tree depth below 1 in Individual.");
			minTreeDepth = 1;
		}
		maxInitialDepth = treeDepth;//should be at least 2
		while(this.treeDepth < minTreeDepth)
		{
			try
			{
				growTree(false);
			}
			catch(StackOverflowError soe)
			{
				System.err.println("StackOverflowError. TreeDepth=" + maxInitialDepth + ". Reducing tree depth.");
				maxInitialDepth--;
				treeDepth = 0;
			}
		}
	}

	private void fullIndividual(int treeDepth)
	{
		maxInitialDepth = treeDepth;//should be at least 2
		this.treeDepth = treeDepth;
		try
		{
			growTree(true);
		}
		catch(StackOverflowError soe)
		{
			System.err.println("StackOverflowError. Reducing tree depth.");
			fullIndividual(treeDepth - 1);
		}
	}

	private void growTree(boolean isFull) throws StackOverflowError
	{
		nodes.clear();
		backupLevels.clear();
		treeDepth = 0;
		addTreeToEnd(1,isFull);
	}

	/**this recursive method can cause problems with stack size. If StackOverflowError
	occurs, call the program with increased stack size, e.g. java -Xss128M ClassName*/
	private void addTreeToEnd(int currentLevel,boolean isFull)
	{
		if(currentLevel<=0)
			return;
		else if(currentLevel>treeDepth)
			treeDepth=currentLevel;

		Component node=selectNode(currentLevel,isFull);
		nodes.add(node);

		//lefthand branch
		while(!isTerminal(node))
		{
			currentLevel++;
			if(currentLevel>treeDepth)
				treeDepth=currentLevel;
			if(!(node instanceof OneInputFunction))//If it is a one input function we do not have to return to this point (no branching)
				backupLevels.add(currentLevel);
			node=selectNode(currentLevel,isFull);
			nodes.add(node);
		}
		int backupSize=backupLevels.size();
		if(backupSize>0)
		{
			int backup=backupLevels.lastElement();
			backupLevels.remove(backupSize-1);
			addTreeToEnd(backup,isFull);
		}
	}

	private Component selectNode(int treeLevel,boolean isFull)
	{
		if(treeLevel==maxInitialDepth)
			return selectTerminal();
		else if(isFull)
			return selectFunction();
		else
		{
			double selector=0.0;
			try
			{
				selector=rand.nextDouble();
			}
			catch(Exception e)
			{
				e.printStackTrace();
			}
			if(selector<growFunctionProportion)
				return selectFunction();
			else
				return selectTerminal();
		}
	}

	private Function selectFunction()
	{
		double selector=rand.nextDouble();
		return functions[(int)(selector*functions.length)];
	}

	private Terminal selectTerminal()
	{
		double selector=rand.nextDouble();
		return terminals[(int)(selector*terminals.length)];
	}

	private static boolean isTerminal(Component node)
	{
		if(node instanceof Terminal)
			return true;
		else
			return false;
	}

	/**Returns the Components within this Individual in the correct order, working from root to leaf and left to right.*/
	public Vector<Component> getNodes()
	{
		return nodes;
	}

	/**Returns a string representation of this Individual. All components are separated by commas
	and components are listed from root to leaf and left to right.*/
	public String toString()
	{
		String stringRepresentation=nodes.get(0).toString();
		for(int i=1;i<nodes.size();i++)
			stringRepresentation+=","+nodes.get(i).toString();
		return stringRepresentation;
	}

	/**Returns a string representation of this Individual in the format of a mathematical function. Brackets are used to
	make operator priority clear (even when they are not required).*/
	public String toStringFunction()
	{
		String str = "";

		int nodeCount=nodes.size()-1;
		Vector<String> branchValue=new Vector<String>();

		while(nodeCount>=0)
		{
			Component comp=nodes.get(nodeCount);
			if (comp instanceof Terminal)
				branchValue.add(comp.toString());

			else
			{
				//this assumes that TwoInputFunctions are infix operators and OneInputFunctions are prefix operators
				if(comp instanceof TwoInputFunction)
				{
					if(nodeCount == 0)
						branchValue.set(branchValue.size()-2, branchValue.get(branchValue.size()-1) + " " + comp.toString() + " " + branchValue.get(branchValue.size()-2));
					else
						branchValue.set(branchValue.size()-2, "(" + branchValue.get(branchValue.size()-1) + " " + comp.toString() + " " + branchValue.get(branchValue.size()-2) + ")");
					branchValue.remove(branchValue.size()-1);
				}
				else
					branchValue.set(branchValue.size()-1, comp.toString() + "(" + branchValue.get(branchValue.size()-1) + ")");
			}
			nodeCount--;
		}

		return branchValue.get(0);
	}

	/**This method will only return true if all of the nodes in the comparison
	Individual refer to the same Component objects as the reference Individual.
	All Individuals in a Population should therefore be provided with
	references to the same Component objects.*/
	public boolean equals(Individual comparison)
	{
		int len=getLength();
		if(len!=comparison.getLength())
			return false;
		Vector<Component> comparisonNodes=comparison.getNodes();
		for(int i=0;i<len;i++)
		{
			if(!nodes.get(i).equals(comparisonNodes.get(i)))
				return false;
		}
		return true;
	}

	void setClassificationData(int[] data)
	{
		if(data==null || data.length == 4)
			classificationData = data;
	}

	void setSavedValues(double threshold, double AUROC, double AOREC)
	{
		savedThreshold = threshold;
		savedAUROC = AUROC;
		savedAOREC = AOREC;
	}

	//sorts by x-coordinate and then y-coordinate
	private static void ShellSortPoints(Point2D.Double[] points)
	{
		int len = points.length;
		int[] originalIndices = new int[len];
		for(int i=0;i<len;i++)
			originalIndices[i] = i;
		int h = 1;
		//find the largest h value possible
		while ((h * 3 + 1) < len)
		{
			h = 3 * h + 1;
		}

		while( h > 0 )
		{
			//for each set of elements (there are h sets)
			for (int i = h - 1; i < len; i++)
			{
				//pick the last element in the set
				Point2D.Double point = new Point2D.Double(points[i].x,points[i].y);
				int index = originalIndices[i];
				int j = i;
				 /*compare ind to the one before it in the set
				   if they are out of order continue this loop, moving
				   elements "back" to make room for ind to be inserted.*/
				for( j = i; (j >= h) && ((points[j-h].x > point.x)||(points[j-h].x == point.x && points[j-h].y > point.y)); j -= h)
				{
					points[j] = points[j-h];
					originalIndices[j] = originalIndices[j-h];
				}
				//insert ind into the correct place
				points[j] = point;
				originalIndices[j] = index;
			}
			//all sets h-sorted, now decrease set size
			h = h / 3;
		}
	}
}