package geneticWedge.gp;

import geneticWedge.gp.function.*;

/**IndividualDescriptor holds the information required to create an Individual*/
public class IndividualDescriptor
{
	/**Determines whether an Individual is grown by the FULL or GROW method*/
	public static enum IndividualGrowMethod
	{
		/**With the GROW method Terminals and Functions are assigned randomly, leading to irregular trees.*/
		GROW,
		/**With the FULL method all branches are grown to the maximum depth, leading to regular trees.*/
		FULL
	};

	private IndividualGrowMethod growMethod;
	private int maxDepth;
	private int minDepth = 0;
	private Function[] functions;
	private Input[] inputs;
	private Constant[] constants;
	private double functionProportion=0.7;

	/**Creates an IndividualDescriptor with all of the information needed to create a new Individual. */
	public IndividualDescriptor(IndividualGrowMethod growMethod, int maxTreeDepth, Function[] functions, Input[]inputs, Constant[] constants)
	{
		this.growMethod=growMethod;
		maxDepth=maxTreeDepth;
		this.functions=functions;
		this.inputs=inputs;
		this.constants=constants;
	}

	//public IndividualDescriptor(int maxTreeDepth, int noOfInputs, int noOfConstants, double growFunctionProportion)
	/*public IndividualDescriptor(int maxTreeDepth, Function[] functions, Input[]inputs, Constant[] constants, double growFunctionProportion)
	{
		this.growMethod=growMethod;
		maxDepth=maxTreeDepth;
		this.functions=functions;
		this.inputs=inputs;
		this.constants=constants;
		functionProportion=growFunctionProportion;
	}*/

	/**Returns the method that will be used to grow Individuals, i.e. GROW or FULL*/
	public IndividualGrowMethod getGrowMethod()
	{
		return growMethod;
	}

	/**Sets the maximum depth of trees grown using this descriptor.*/
	public void setMaxTreeDepth(int newMaxDepth)
	{
		if(maxDepth<1)
		{
			System.err.println("Attempt to set maximum tree depth below 1 in Individual.");
			maxDepth = 1;
		}
		else
			maxDepth = newMaxDepth;
	}

	/**Returns the maximum depth of trees grown using this descriptor.*/
	public int getMaxTreeDepth()
	{
		return maxDepth;
	}

	/**Sets the minimum depth of trees grown using this descriptor.*/
	public void setMinTreeDepth(int newMinDepth)
	{
		if(newMinDepth <= maxDepth)
		{
			if(newMinDepth<1)
			{
				System.err.println("Attempt to set minimum tree depth below 1 in Individual.");
				minDepth = 1;
			}
			else
				minDepth = newMinDepth;
		}
		else
			System.err.println("IndividualDescriptor: Attempt to set minDepth less than maxDepth");
	}

	/**Returns the minimum depth of trees grown using this descriptor.*/
	public int getMinTreeDepth()
	{
		return minDepth;
	}

	/**Returns the array of Functions used by this descriptor*/
	public Function[] getFunctions()
	{
		return functions;
	}

	/**Returns the array of Inputs used by this descriptor*/
	public Input[] getInputs()
	{
		return inputs;
	}

	/**Returns the array of Constants used by this descriptor*/
	public Constant[] getConstants()
	{
		return constants;
	}

	/**Returns the growFunctionProportion used by this descriptor. This is the probability that a Function (rather than a Terminal) will
	be selected while growing Individuals with the GROW method. The growFunctionProportion indirectly affects average tree size.*/
	public double getGrowFunctionProportion()
	{
		return functionProportion;
	}
}