package geneticWedge.gp;

import java.util.Vector;
import java.util.Random;

import geneticWedge.gp.function.*;

/**Mutation mutates an Individual using a selection of mutation operators.*/
public class Mutation
{
	/**Mutation operators should be chosen with care. Most of the operators are neutral with respect to
	program length, but SHRINK has a bias towards small programs and HOIST has a heavy bias towards
	small programs. EXPAND can have a large bias towards larger programs,
	if setMaximumExpandDepth() is called with a large value*/
	public static enum MutationOperator
	{
		/**This is the 'standard' mutation operator. It replaces a subtree with a randomly generated subtree of similar size.*/
		SUBTREE_REPLACEMENT,
		/**POINT_REPLCEMENT replaces a single component with a randomly generated component with the same arity*/
		POINT_REPLACEMENT,
		/**HOIST creates a new Individual from a randomly chosen subtree of the parent.*/
		HOIST,
		/**SHRINK replaces a randomly chosen subtree with a randomly chosen Terminal. This operator is useful for keeping down the size of trees.*/
		SHRINK,
		/**EXPAND replaces a randomly chosen Terminal with a randomly chosen subtree. This operator should be used with care as it may cause bloat.*/
		EXPAND,
		/**SWAP chooses a non-commutative binary node (such as SUBTRACT) and swaps the subtrees below this node*/
		SWAP,
		/**CONSTANT changes the value of a Constant node by a random amount (generated with a Gaussian distribution).*/
		CONSTANT
	};

	private static Random rand=new Random();
	private MutationOperator[] operators;
	//noOfTries should have public get and set methods.
	//Keeping noOfTries low reduces the chance of bloat - large subtrees are more likely to fail
	private int noOfTries=5;
	private double constantMutationStep=0.1;
	private int maximumExpandDepth=4;//used in EXPAND
	private int maximumReplacementDepth = 15;

	private boolean collectTransmissionInfo = false;
	private Vector<double[]> transmissionInfo=null;

	private boolean useExpectedDepth = false;
	private double expectedDepth = 0.5;

	/**The constructor takes a list of mutation operations to be used. When mutation occurs one of these operators will be chosen at random.
	It is possible to have the same operator more than once, thereby increasing the frequency with which it occurs.*/
	public Mutation(MutationOperator[] ops)
	{
		operators=ops;
	}

	/**This sets the 'expected' depth at which mutation will take place. A Poisson distribution is then created with this
	expected value in order to select the depth at which mutation takes place. If this method is not called, or it is called
	with depth outside the range 0.0-1.0 (exclusive), the mutation site will be selected randomly.*/
	public void setExpectedMutationDepth(double depth)
	{
		if(depth<=0.0||depth>=1.0)
		{
			useExpectedDepth = false;
			return;
		}
		useExpectedDepth = true;
		expectedDepth = depth;
	}

	/**If set to true, collectTransmissionInfo saves information concerning the relationship between parent and offspring
	during mutation operations as well as the mutation operation that occurred.*/
	public void collectTransmissionInfo(boolean collect)
	{
		collectTransmissionInfo = collect;
		if(collect)
			transmissionInfo=new Vector<double[]>();
	}

	/**Returns a vector containing information about the relationship between parent and offspring during mutation operations,
	provided collectTransmissionInfo has been set to true.*/
	public Vector<double[]> getTransmissionInfo()
	{
		return transmissionInfo;
	}

	/**Perform a random mutation on an Individual. The mutation operator is selected randomly from the array of mutation operators provided in the constructor.
	@param maxLength The maximum length (number of nodes) of the offspring
	@param maxDepth The maximum depth of the offspring
	@param popDescriptor This is needed in order to provide the available Components and growFunctionProportion
	(determining the proportion of components in a new subtree that are selected as Functions)*/
	public Individual mutate(Individual ind, int maxLength, int maxDepth, PopulationDescriptor popDescriptor) throws InvalidFractionException
	{
		Function[] functions = popDescriptor.getFunctions();
		Input[] inputs = popDescriptor.getInputs();
		Constant[] constants = popDescriptor.getConstants();
		double growFunctionProportion = popDescriptor.getGrowFunctionProportion();

		Individual offspring=null;

		int parentLength=ind.getLength();
		if(parentLength==1)
		{
			offspring = new Individual(IndividualDescriptor.IndividualGrowMethod.FULL,1,0,functions,inputs,constants,growFunctionProportion);
			if(collectTransmissionInfo)
				transmissionInfo.add(new double[]{1,1,1,0,0});//treat as point mutation
			return offspring;
		}
		int parentDepth=ind.getDepth();
		Vector<Component> nodes=ind.getNodes();

		int index = (int)(rand.nextDouble()*operators.length);
		MutationOperator op = operators[index];
		Vector<Component> offspringNodes=new Vector<Component>(0);

		int mutationPoint;
		int end;
		int len;
		int[] depths;
		double lenChange;
		int selector;
		Vector<Integer> funcIndex = new Vector<Integer>();
		switch(op)
		{
			case SUBTREE_REPLACEMENT:
				if(useExpectedDepth)
				{
					mutationPoint = ind.getModificationPointFromPoissonDistribution(expectedDepth, true)-1;
					if(mutationPoint<0)
						mutationPoint = 0;
				}
				else
					mutationPoint=(int)(rand.nextDouble()*(parentLength-1));
				end=Utils.getEndOfFragment(nodes,mutationPoint);
				double sizeMultiplier = rand.nextDouble()+0.5;//length between 0.5 and 1.5 times parent length
				len = (int)(Math.round((end-mutationPoint)*sizeMultiplier));
				int maxFragmentLength=maxLength-mutationPoint+end-parentLength;
				if(len>maxFragmentLength)
					len=maxLength+end-mutationPoint-parentLength;
				if (len<1)
					len=1;
				depths = Utils.getFragmentDepths(nodes,mutationPoint,end);
				//don't use depth limit as high as 1.5 times fragment depth - can lead to very large trees
				//int depth=(int)(Math.round((depths[1]-depths[0])*(rand.nextDouble()+0.5)));//depth between 0.5 and 1.5 times parent depth
				int depth = depths[1]-depths[0];//removed subtree depth
				if(sizeMultiplier > 1.0)
					depth++;
				if(depth+depths[0]>maxDepth)
					depth=maxDepth-depths[0];
				if(depth<2)
					depth=2;
				Individual subtree=null;
				int count=0;
				do
				{
					count++;
					if(count==noOfTries+1)
					{
						if(collectTransmissionInfo)
							transmissionInfo.add(new double[]{-1,parentDepth,0,0,0});
						return new Individual(nodes,inputs,constants);//after x tries, return copy of original Individual
					}
					subtree = new Individual(IndividualDescriptor.IndividualGrowMethod.GROW,depth,1,functions,inputs,constants,growFunctionProportion);
				}
				while (subtree.getLength()<len || subtree.getLength()>maxFragmentLength);

				Vector<Component> subtreeNodes=subtree.getNodes();
				for(int i=0;i<=mutationPoint;i++)
					offspringNodes.add(nodes.get(i));
				for(int i=0;i<subtreeNodes.size();i++)
					offspringNodes.add(subtreeNodes.get(i));//mutated fragment
				for(int i=end+1;i<nodes.size();i++)
					offspringNodes.add(nodes.get(i));

				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
					transmissionInfo.add(new double[]{0,parentDepth,depths[0],depth,lenChange});
				break;

			case POINT_REPLACEMENT:
				if(useExpectedDepth)
					mutationPoint = ind.getModificationPointFromPoissonDistribution(expectedDepth, true);
				else
					mutationPoint=(int)(rand.nextDouble()*parentLength);
				Vector<Component> sameTypeComps=new Vector<Component>();

				//if n-input Functions are introduced this will have to be modified
				if(nodes.get(mutationPoint) instanceof Terminal)
				{
					for(int j=0;j<inputs.length;j++)
						sameTypeComps.add(inputs[j]);
					for(int j=0;j<constants.length;j++)
						sameTypeComps.add(constants[j]);
				}
				else if(nodes.get(mutationPoint) instanceof OneInputFunction)
				{
					for(int j=0;j<functions.length;j++)
					{
						if(functions[j] instanceof OneInputFunction)
							sameTypeComps.add(functions[j]);
					}
				}
				else if(nodes.get(mutationPoint) instanceof TwoInputFunction)
				{
					for(int j=0;j<functions.length;j++)
					{
						if(functions[j] instanceof TwoInputFunction)
							sameTypeComps.add(functions[j]);
					}
				}

				int replacementIndex=(int)(rand.nextDouble()*sameTypeComps.size());
				for(int i=0;i<parentLength;i++)
				{
					if(i==mutationPoint)
						offspringNodes.add(sameTypeComps.get(replacementIndex));
					else
						offspringNodes.add(nodes.get(i));
				}
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
				{
					end=Utils.getEndOfFragment(nodes,mutationPoint-1);
					depths = Utils.getFragmentDepths(nodes,mutationPoint-1,end);
					transmissionInfo.add(new double[]{1,parentDepth,depths[0],depths[1]-depths[0],lenChange});
				}
				break;

			case HOIST:
				if(useExpectedDepth)
				{
					mutationPoint = ind.getModificationPointFromPoissonDistribution(expectedDepth, false) - 1;
					if(mutationPoint<0)
						mutationPoint = 0;
				}
				else
					mutationPoint=(int)(rand.nextDouble()*(parentLength-1));
				end=Utils.getEndOfFragment(nodes,mutationPoint);

				for(int i = mutationPoint+1;i <= end;i++)
					offspringNodes.add(nodes.get(i));
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
				{
					end=Utils.getEndOfFragment(nodes,mutationPoint);
					depths = Utils.getFragmentDepths(nodes,mutationPoint,end);
					transmissionInfo.add(new double[]{2,parentDepth,depths[0],depths[1]-depths[0],lenChange});
				}
				break;

			case SHRINK:
				for (int i=0;i<nodes.size();i++)
				{
					if(nodes.get(i) instanceof Function)
						funcIndex.add(i);
				}
				if(useExpectedDepth)
				{
					mutationPoint = ind.getModificationPointFromPoissonDistribution(expectedDepth, funcIndex, true) - 1;
					if(mutationPoint<0)
						mutationPoint = 0;
				}
				else
				{
					selector = (int)(rand.nextDouble()*funcIndex.size());
					mutationPoint = funcIndex.get(selector)-1;
				}
				end = Utils.getEndOfFragment(nodes,mutationPoint);

				Vector<Terminal> terms=new Vector<Terminal>();
				for(int j=0;j<inputs.length;j++)
					terms.add(inputs[j]);
				for(int j=0;j<constants.length;j++)
					terms.add(constants[j]);
				selector = (int)(rand.nextDouble()*terms.size());
				Terminal newTerminal = terms.get(selector);

				for(int i=0;i<=mutationPoint;i++)
					offspringNodes.add(nodes.get(i));
				offspringNodes.add(newTerminal);
				for(int i=end+1;i<nodes.size();i++)
					offspringNodes.add(nodes.get(i));
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
				{
					depths = Utils.getFragmentDepths(nodes,mutationPoint,end);
					transmissionInfo.add(new double[]{3,parentDepth,depths[0],depths[1]-depths[0],lenChange});
				}
				break;

			case EXPAND:
				Vector<Integer> t=new Vector<Integer>();
				len = nodes.size();
				for(int i=0;i<nodes.size();i++)
				{
					if(nodes.get(i) instanceof Terminal)
						t.add(i);
				}

				Individual sub = null;
				int c=0;
				int dep;
				do
				{
					c++;
					if(c==noOfTries+1)
					{
						if(collectTransmissionInfo)
							transmissionInfo.add(new double[]{-1,parentDepth,0,0,0});
						return new Individual(nodes,inputs,constants);//after x tries, return copy of original Individual
					}
					if(useExpectedDepth)
						mutationPoint = ind.getModificationPointFromPoissonDistribution(expectedDepth, t, true);
					else
					{
						selector = (int)(rand.nextDouble() * t.size());
						mutationPoint = t.get(selector);
					}
					dep = Utils.getFragmentDepths(nodes,mutationPoint-1,0)[0];
					int maxFragmentDepth = maximumExpandDepth;
					if(maxFragmentDepth > maxDepth - dep) //cut off new subtree if the offspring will be too deep
						maxFragmentDepth = maxDepth - dep;
					if(maxFragmentDepth<=1)//must be at least 2
						continue;
					sub = new Individual(IndividualDescriptor.IndividualGrowMethod.GROW,maxFragmentDepth,1,functions,inputs,constants,growFunctionProportion);
				}
				while (sub==null || sub.getLength() + len - 1 > maxLength);

				Vector<Component> subNodes=sub.getNodes();
				for(int i=0;i<mutationPoint;i++)
					offspringNodes.add(nodes.get(i));
				for(int i=0;i<subNodes.size();i++)
					offspringNodes.add(subNodes.get(i));//mutated fragment
				for(int i=mutationPoint+1;i<nodes.size();i++)
					offspringNodes.add(nodes.get(i));
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
					transmissionInfo.add(new double[]{4,parentDepth,dep,0,lenChange});
				break;

			case SWAP:
				for (int i=0;i<nodes.size();i++)
				{
					if(nodes.get(i) instanceof TwoInputFunction)
						funcIndex.add(i);
				}
				if(funcIndex.size() == 0)
				{
					if(collectTransmissionInfo)
						transmissionInfo.add(new double[]{-1,parentDepth,0,0});
					return new Individual(nodes, inputs, constants);//we cannot perform swap if there are no non-commutative TwoInputFunctions within the Individual
				}
				Vector<TwoInputFunction> twoInputs=new Vector<TwoInputFunction>();
				for(int j=0;j<functions.length;j++)
				{
					Function f = functions[j];
					if(f instanceof TwoInputFunction && !((TwoInputFunction)f).isCommutative())
						twoInputs.add((TwoInputFunction)f);
				}
				if(twoInputs.size() == 0)
				{
					if(collectTransmissionInfo)
						transmissionInfo.add(new double[]{-1,parentDepth,0,0,0});
					return new Individual(nodes, inputs, constants);//we cannot perform swap if there are no non-commutative TwoInputFunctions
				}
				int leftBranch;
				if(useExpectedDepth)
					leftBranch = ind.getModificationPointFromPoissonDistribution(expectedDepth, funcIndex, true);
				else
				{
					selector = (int)(rand.nextDouble()*funcIndex.size());
					leftBranch = funcIndex.get(selector);
				}
				int leftEnd = Utils.getEndOfFragment(nodes,leftBranch);
				int rightEnd = Utils.getEndOfFragment(nodes,leftEnd);

				for(int i=0;i<=leftBranch;i++)
					offspringNodes.add(nodes.get(i));
				for(int i=leftEnd+1;i<=rightEnd;i++)
					offspringNodes.add(nodes.get(i));
				for(int i=leftBranch+1;i<=leftEnd;i++)
					offspringNodes.add(nodes.get(i));
				for(int i=rightEnd+1;i<nodes.size();i++)
					offspringNodes.add(nodes.get(i));
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
				{
					end=Utils.getEndOfFragment(nodes,leftBranch-1);
					depths = Utils.getFragmentDepths(nodes,leftBranch-1,end);
					transmissionInfo.add(new double[]{5,parentDepth,depths[0],depths[1]-depths[0],lenChange});
				}
				break;

			case CONSTANT:
				Vector<Constant> foundConstants=new Vector<Constant>();
				Vector<Vector<Integer>> constantIndices=new Vector<Vector<Integer>>();
				for (int i=0;i<nodes.size();i++)
				{
					Component node = nodes.get(i);
					if(node instanceof Constant)
					{
						int constIndex = foundConstants.indexOf(node);
						if(constIndex == -1)
						{
							foundConstants.add((Constant)node);
							Vector<Integer> newConst = new Vector<Integer>();
							newConst.add(i);
							constantIndices.add(newConst);
						}
						else
						{
							constantIndices.get(constIndex).add(i);
						}
					}
				}
				depth = Integer.MAX_VALUE;
				if(foundConstants.size()==0)
				{
					if(collectTransmissionInfo)
						transmissionInfo.add(new double[]{-1,parentDepth,0,0,0});
					return new Individual(nodes, inputs, constants);//we cannot perform CONSTANT mutation if there are no Constants
				}
				else
				{
					selector = (int)(rand.nextDouble()*foundConstants.size());
					double constantVal = foundConstants.get(selector).getValue();
					constantVal += rand.nextGaussian() * constantMutationStep;
					Constant newConstant =new Constant(constantVal);
					Vector<Integer> changeIndices = constantIndices.get(selector);
					for(int i=0;i<nodes.size();i++)
					{
						if(changeIndices.contains(i))
						{
							offspringNodes.add(newConstant);
							if(collectTransmissionInfo)
							{
								int d = Utils.getDepth(nodes,i);
								if(d<depth)
									depth = d;
							}
						}
						else
							offspringNodes.add(nodes.get(i));
					}
				}
				lenChange = offspringNodes.size()-parentLength;

				if(collectTransmissionInfo)
					transmissionInfo.add(new double[]{6,parentDepth,depth,0,lenChange});
				break;
		}
		return new Individual(offspringNodes, inputs, constants);
	}

	/**The stepSize determines how far constant values move during CONSTANT mutation.
	The step size is equal to the standard deviation of the 'step' made.*/
	public void setConstantMutationStepSize(double stepSize)
	{
		constantMutationStep = stepSize;
	}

	/**The stepSize determines how far constant values move during CONSTANT mutation.
	The step size is equal to the standard deviation of the 'step' made.*/
	public double getConstantMutationStepSize()
	{
		return constantMutationStep;
	}

	/**When the EXPAND mutation takes place, a randomly chosen terminal is replaced by a subtree,
	grown with the GROW method. The parameter maxDepth sets the maximum depth of the tree that is grown.
	If maxDepth is set to a high value (greater than about 20), it may be necessary to increase the stack size,
	e.g. -Xss128M, to avoid stack overflow.*/
	public void setMaximumExpandDepth(int maxDepth)
	{
		if(maxDepth>0)
			maximumExpandDepth = maxDepth;
		else
			System.err.println("Attempt to set non-positive maximumExpandDepth.");
	}

	/**maximumExpandDepth is the maximum allowable depth of the subtree grown during an EXPAND mutation.
	It has a default value of 4.*/
	public int getMaximumExpandDepth()
	{
		return maximumExpandDepth;
	}

	/**When the SUBTREE_REPLACEMENT mutation takes place, a randomly chosen subtree is replaced by a random subtree,
	of similar size grown with the GROW method. The parameter maxDepth sets the maximum depth of the tree that is grown.
	If maxDepth is set to a high value (greater than about 20), it may be necessary to increase the stack size,
	e.g. -Xss128M, to avoid stack overflow.*/
	public void setMaximumReplacementDepth(int maxDepth)
	{
		if(maxDepth>0)
			maximumReplacementDepth = maxDepth;
		else
			System.err.println("Attempt to set non-positive maximumReplacmentDepth.");
	}

	/**maximumReplacementDepth is the maximum allowable depth of the subtree grown during a SUBTREE mutation.
	It has a default value of 4.*/
	public int getMaximumReplacementDepth()
	{
		return maximumReplacementDepth;
	}

	/**Returns the array of mutationOperators used to perform mutation. This array should have been provided to the constructor.*/
	public MutationOperator[] getMutationOperators()
	{
		return operators;
	}
}