package geneticWedge.gp;

import java.util.Random;
import java.util.Arrays;
import java.util.Vector;
import java.util.Iterator;
import java.io.*;

import geneticWedge.gp.function.*;
import geneticWedge.gp.dataAnalysis.DataAcquirer;

/**Population is the standard Population class. A Population contains a number of Individuals and holds information about
the population as a whole such as average fitness. A Population may be evolved by calling the evolve() method.*/
public class Population extends AbstractPopulation
{
	//Most class variables and methods have 'friendly' rather than private access so that they can be used by subclasses within the same package
	static Runtime runtime=Runtime.getRuntime();
	boolean collectTransmissionInfo = false;
	boolean collectOperatorInfo = false;
	Vector<Vector<double[]>> transmissionInfo = null;
	Vector<double[]> currentTransmissionVector = null;
	Vector<Vector<String>> operatorInfo = null;
	Vector<String> currentOperatorVector = null;
	int transmissionWritingFrequency = 100;
	String transmissionInfoFile = null;
	BufferedWriter transmissionWriter = null;
	double[] cumulativeCutDepthCorrelation = null;
	Vector<double[]> allCutDepthCorrelations = null;
	int[] cutDepthCorrCount = null;

	Individual[] individuals;
	Individual[] newPopulation = null;
	Individual[]savedPopulation = null;
	double[] savedFitnesses = null;
	int eliteSize=0;
	double rankGradient = 1.0;
	boolean parentsInTruncationPool=false;
	//truncation selection is also known as (mu,lambda) selection, hence the Greek names!
	int mu = 4;
	int lambda = 8;
	FitnessShare fitnessShare = null;
	FitnessShare.ShareType shareType = FitnessShare.ShareType.GENE_DISTANCE;
	double[] shareWeightings = null;

	/**PerformanceFunctions are used to create multi-objective evaluation functions. They may be used in addition to
	a normal FitnessFunction and are applied either in a multi-objective (Pareto-type) manner or in a single objective manner,
	using a weighted sum of the FitnessFunction and the PerformanceFunctions.*/
	public static enum PerformanceFunction
	{
		/**Aims to minimise program length (number of nodes)*/
		PROGRAM_LENGTH,
		/**Aims to minimise the average genotypic distance of an Individual from all other members of the population*/
		DIVERSITY_DISTANCE,
		/**Aims to minimise the complexity of functions used by an Individual. Simple arithmetic functions are considered to be of low complexity,
		whereas higher order functions such as pow and exp are considered to be of high complexity.*/
		FUNCTION_COMPLEXITY,
		/**Aims to maximise the number of different variables (Inputs) used*/
		VARIABLE_MAX,
		/**Aims to minimise the number of different variables (Inputs) used*/
		VARIABLE_MIN
	};
	PerformanceFunction[] performanceFunctions = null;
	boolean isMultiObjective = true;
	boolean keepDominatedIndividuals = true;
	double[] singleObjectiveRatio = null;
	boolean[][] dominatedBy;
	int[] dominatedCount;
	Random rand=new Random();

	double[][] allSquaredDistances;
	int[] allLengths;
	int[] noInputsUsed;
	int[] functionComplexities;

	int maximumLength=250;
	int maximumDepth=17;

	double mutationProportion=0.0;
	double reproductionProportion=0.0;

	boolean allowCombinedOp = false;

	Mutation mut = new Mutation(new Mutation.MutationOperator[]{Mutation.MutationOperator.SUBTREE_REPLACEMENT});
	Crossover crossover = new Crossover();

	int populationEvolved = 0;//used with TruncationSelection to check when a whole generation has evolved


	private boolean useBrood = false;
	protected boolean printInfo = true;
	private int bestValidationIndex = -1;
	protected int evolutionNumber = 0;

	/**The Population constructor takes the population size and a PopulationDescriptor as variables.*/
	public Population(int populationSize,PopulationDescriptor descriptor)
	{
		int selPressure = descriptor.getInitialSelectionPressure();
		if(selPressure>1)
			populationSize*=selPressure;
		startingPopSize = populationSize;
		popSize = populationSize;

		popDescriptor = descriptor;
		individuals=new Individual[populationSize];
		try
		{
			IndividualDescriptor indDescriptor;
			int minDepth=popDescriptor.getMinTreeDepth();
			int maxDepth=popDescriptor.getMaxTreeDepth();
			mutationProportion=popDescriptor.getMutationProportion();
			reproductionProportion=popDescriptor.getReproductionProportion();
			allowCombinedOp=popDescriptor.getAllowCombinedOp();
			int tempEliteSize=popDescriptor.getEliteSize();
			if(tempEliteSize>0&&tempEliteSize<populationSize)
				eliteSize=tempEliteSize;


			int[] indices = Utils.getRandomIndices(populationSize);
			switch(popDescriptor.getGrowMethod())
			{
				case RAMPED_HALF_AND_HALF:
					Input[] in=popDescriptor.getInputs();
					indDescriptor=new IndividualDescriptor(IndividualDescriptor.IndividualGrowMethod.GROW,minDepth,popDescriptor.getFunctions(),popDescriptor.getInputs(),popDescriptor.getConstants());
					for(int i=0;i<populationSize/2;i++)
					{
						double fraction=0.5;//in case of very small population (2 Individuals) use average depth
						int denom = populationSize/2-1;
						if(denom>0)
							fraction = (double)i/denom;
						int treeDepth=(int)(Math.round(minDepth+(maxDepth-minDepth)*fraction));
						indDescriptor.setMaxTreeDepth(treeDepth);
						indDescriptor.setMinTreeDepth(treeDepth);
						individuals[indices[i]]=new Individual(indDescriptor);
					}
					indDescriptor=new IndividualDescriptor(IndividualDescriptor.IndividualGrowMethod.FULL,minDepth,popDescriptor.getFunctions(),popDescriptor.getInputs(),popDescriptor.getConstants());
					for(int i=populationSize/2;i<populationSize;i++)
					{
						double fraction = 0.5;
						int denom = populationSize - populationSize/2 -1;
						if(denom>0)
							fraction = ((double)i-populationSize/2)/denom;
						int treeDepth=(int)(Math.round(minDepth+(maxDepth-minDepth)*fraction));
						indDescriptor.setMaxTreeDepth(treeDepth);
						individuals[indices[i]]=new Individual(indDescriptor);
					}
					break;

				case GROW:
					minDepth=popDescriptor.getMinTreeDepth();
					maxDepth=popDescriptor.getMaxTreeDepth();
					indDescriptor=new IndividualDescriptor(IndividualDescriptor.IndividualGrowMethod.GROW,minDepth,popDescriptor.getFunctions(),popDescriptor.getInputs(),popDescriptor.getConstants());
					for(int i=0;i<populationSize;i++)
					{
						double fraction = 0.5;
						int denom = populationSize-1;
						if(denom>0)
							fraction = (double)i/denom;
						int treeDepth=(int)(Math.round(minDepth+(maxDepth-minDepth)*fraction));
						indDescriptor.setMaxTreeDepth(treeDepth);
						indDescriptor.setMinTreeDepth(treeDepth);
						individuals[indices[i]]=new Individual(indDescriptor);
					}
					break;

				case FULL:
					indDescriptor=new IndividualDescriptor(IndividualDescriptor.IndividualGrowMethod.FULL,minDepth,popDescriptor.getFunctions(),popDescriptor.getInputs(),popDescriptor.getConstants());
					for(int i=0;i<populationSize;i++)
					{
						double fraction = 0.5;
						int denom = populationSize-1;
						if(denom>0)
							fraction = (double)i/denom;
						int treeDepth=(int)(Math.round(minDepth+(maxDepth-minDepth)*fraction));
						indDescriptor.setMaxTreeDepth(treeDepth);
						individuals[indices[i]]=new Individual(indDescriptor);
					}
					break;
			}
		}
		catch(InvalidFractionException ife)
		{
			System.err.println(ife.getMessage());
		}
	}

	/**If printInfo is set to true, Population will print information concerning it's current status to the 'out' stream during evolution.*/
	public void setPrintStatus(boolean status)
	{
		printInfo = status;
	}

	/***/
	public Individual[] getIndividuals()
	{
		return individuals.clone();
	}

	/**If this method is called, the selection method will be deduced from the value
	of steadyState. If true, the selection method will be set to TOURNAMENT. If false
	the selection method will be set to FITNESS_PROPORTIONAL.*/
	public boolean evolve(boolean steadyState, double[][]inputs, double[]targetOutputs, double[][]validationInputs, double[]validationTargetOutputs, StoppingCriterion stopCriterion)
	{
		if(steadyState)
			return evolve(steadyState, SelectionMethod.TOURNAMENT, inputs, targetOutputs, validationInputs, validationTargetOutputs, stopCriterion);
		else
			return evolve(steadyState, SelectionMethod.FITNESS_PROPORTIONAL, inputs, targetOutputs, validationInputs, validationTargetOutputs, stopCriterion);
	}


	/**If this method is called in steadyState mode with a selectMethod that uses roulette wheel selection
	(FITNESS_PROPORTIONAL, RANK or FUSS), elitist selection will not occur, even if an elite size has been set.
	In order to acieve elitist selection with these methods, generational evolution must be applied.*/
	public boolean evolve(boolean steadyState, SelectionMethod selectMethod, double[][]inputs, double[]targetOutputs, double[][]validationInputs, double[]validationTargetOutputs, StoppingCriterion stopCriterion)
	{
		evolutionNumber = 0;
		if(printInfo)
			System.out.println("Population evolving.");
		try
		{
			if(crossover.getBroodUse())
			{
				crossover.setData(inputs,targetOutputs);
				crossover.setFitnessFunction(fitnessFunction);
			}
			if(printInfo)
				System.out.println("Setting fitness vals");
			setFitnessValues(inputs,targetOutputs);
			if(printInfo)
				System.out.println("Finished setting fitness vals");

			int selPressure = popDescriptor.getInitialSelectionPressure();
			if(selPressure>1)
			{
				setFitnessValues(inputs,targetOutputs);
				Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals()[0]);

				popSize /=selPressure;
				startingPopSize /=selPressure;
				individuals = new Individual[popSize];
				for(int i=0;i<popSize;i++)
					individuals[i] = orderedIndividuals[i];
			}
			if(stopCriterion instanceof EvaluativeStoppingCriterion)
			{
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
					case COMBINED_CLASSIFICATION_ERROR:
					//case CLUSTERED_REGRESSION:
						((EvaluativeStoppingCriterion)stopCriterion).setMaximise(false);
						break;
					case NUMBER_CORRECT_BINARY:
					case AUROC:
					case SENSITIVITY:
					case SPECIFICITY:
					case PPV:
					case NPV:
						((EvaluativeStoppingCriterion)stopCriterion).setMaximise(true);
						break;
				}
				if(printInfo)
					System.out.println("Setting validation fitness vals");
				if(((EvaluativeStoppingCriterion)stopCriterion).usesValidation())
					setValidationFitnessValues(validationInputs,validationTargetOutputs);
				if(printInfo)
					System.out.println("Finished setting validation fitness vals");
				if(stopCriterion instanceof StagnationStoppingCriterion)
					savePopulation();
			}
			runtime.gc();
			boolean distanceSet = false;
			if(performanceFunctions!=null)
			{
				if(printInfo)
					System.out.println("Setting performance functions");
				for(int i=0;i<performanceFunctions.length;i++)
				{
					switch(performanceFunctions[i])
					{
						case DIVERSITY_DISTANCE:
							setAllDistances();
							distanceSet = true;
							break;
						case FUNCTION_COMPLEXITY:
							setAllComplexities();
							break;
						case VARIABLE_MAX:
						case VARIABLE_MIN:
							setAllNumberOfInputsUsed();
							break;
					}
				}
				if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
				{
					setAllDistances();
					distanceSet = true;
				}
				if(printInfo)
					System.out.println("Finished setting performance functions");
				if(isMultiObjective)
				{
					if(printInfo)
						System.out.println("Removing dominated and duplicated individuals");
					if(!keepDominatedIndividuals)
						removeDuplicatedIndividuals();
					setDominatedMatrix();
					if(!keepDominatedIndividuals)
						removeDominatedIndividuals();
					if(printInfo)
						System.out.println("Finished removing dominated and duplicated individuals");
				}
			}
			if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
				setAllDistances();
			if(printInfo)
			{
				System.out.println("Generation "+generationNumber+", Mean Fitness= "+averageTrainFitness+". Best fitness="+bestTrainFitness);
				System.out.println("Mean val Fitness= "+averageValidationFitness+". Best val fitness="+bestValidationFitness);
				System.out.println("average length="+getAverageIndividualLength());
			}
			if(stopCriterion.stopIfNoDiversity()&&hasNoDiversity())
				return true;
			if(collectTransmissionInfo)
			{
				transmissionInfo = new Vector<Vector<double[]>>();
				if(transmissionInfoFile != null)
				{
					try
					{
						transmissionWriter = new BufferedWriter(new FileWriter(transmissionInfoFile));
					}
					catch(IOException ioe)
					{
						System.err.println("Exception creating transmission writer.");
						transmissionInfoFile = null;
						transmissionWriter = null;
					}
				}
				cumulativeCutDepthCorrelation = new double[8];
				cutDepthCorrCount = new int[8];
				allCutDepthCorrelations = new Vector<double[]>();
			}
			while(generationNumber<maxGenerations && !stopCriterion.stopReached(this))
			{
				if(collectTransmissionInfo)
				{
					currentTransmissionVector = new Vector<double[]>();
					transmissionInfo.add(currentTransmissionVector);
				}

				if(stopCriterion instanceof StagnationStoppingCriterion && ((StagnationStoppingCriterion)stopCriterion).lastGenerationWasImproved())
					savePopulation();

				evolveOneGeneration(steadyState,selectMethod, inputs,targetOutputs);

				if(stopCriterion instanceof EvaluativeStoppingCriterion && ((EvaluativeStoppingCriterion)stopCriterion).usesValidation())
					setValidationFitnessValues(validationInputs,validationTargetOutputs);
				if(printInfo)
				{
					System.out.println("generation "+generationNumber+". Mean Fitness= "+averageTrainFitness+". Best Fitness="+bestTrainFitness);
					System.out.println("Mean val Fitness= "+averageValidationFitness+". Best val fitness="+bestValidationFitness);
					System.out.println("average length="+getAverageIndividualLength());
					System.out.println("Best ind= "+individuals[bestIndex].toStringFunction());
					System.out.println("pop size = "+popSize);
				}

				if(stopCriterion.stopIfNoDiversity()&&hasNoDiversity())
				{
					if(transmissionWriter != null)
					{
						try
						{
							transmissionWriter.flush();
							transmissionWriter.close();
						}
						catch(IOException ioe)
						{
							ioe.printStackTrace();
						}
					}
					return true;
				}
			}
			if(stopCriterion instanceof StagnationStoppingCriterion)
			{
				individuals = savedPopulation;
				allFitnesses = savedFitnesses;
				setSummaryFitnessValues();
			}
			setValidationFitnessValues(validationInputs,validationTargetOutputs);
		}
		catch(InvalidInputSizeException iise)
		{
			iise.printStackTrace();
			if(transmissionWriter != null)
			{
				try
				{
					transmissionWriter.flush();
					transmissionWriter.close();
				}
				catch(IOException ioe)
				{
					ioe.printStackTrace();
				}
			}
			return false;
		}
		if(transmissionWriter != null)
		{
			try
			{
				transmissionWriter.flush();
				transmissionWriter.close();
			}
			catch(IOException ioe)
			{
				ioe.printStackTrace();
			}
		}
		if(printInfo)
		{
			System.out.println("FINAL POPULATION. Mean Fitness= "+averageTrainFitness+". Best Fitness="+bestTrainFitness);
			System.out.println("Best ind= "+individuals[bestIndex].toStringFunction());
			if(!Double.isNaN(averageValidationFitness))
			{
				System.out.println("Mean val Fitness= "+averageValidationFitness+". Best val fitness="+bestValidationFitness);
				System.out.println("Best val ind= "+individuals[bestValidationIndex].toStringFunction());
			}
			System.out.println("average length="+getAverageIndividualLength());
			System.out.println("pop size = "+popSize);
		}
		return true;
	}

	protected void evolveOneGeneration(boolean steadyState, SelectionMethod selectMethod, double[][]inputs, double[]targetOutputs)
	{
		runtime.gc();
		try
		{
			if(steadyState)
			{
				if(selectMethod.equals(SelectionMethod.TRUNCATION))
				{
					populationEvolved=0;
					while(!truncationEvolve(true,inputs,targetOutputs)){}
				}
				else
				{
					for(int i=0;i<startingPopSize;i++)
					{
						if(!steadyStateEvolve(selectMethod, inputs,targetOutputs))
							i--;//redo evolution if it has failed
					}
				}
			}
			else
			{
				if(selectMethod.equals(SelectionMethod.TRUNCATION))
				{
					populationEvolved = 0;
					newPopulation = new Individual[startingPopSize];

					//create elite
					if(eliteSize>0)
					{
						int cloneSize = eliteSize;
						if(popSize < eliteSize)
							cloneSize = popSize;

						Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals()[0]);
						for(int i=0;i<cloneSize;i++)
						{
							newPopulation[i] = orderedIndividuals[i];
						}
						populationEvolved = cloneSize;
					}

					while(!truncationEvolve(false,inputs,targetOutputs)){}
				}
				else
				{
					generationalEvolve(selectMethod, inputs,targetOutputs);
				}

				setFitnessValues(inputs,targetOutputs);

				boolean distanceSet = false;
				//update performance values
				if(performanceFunctions!=null)
				{
					for(int i=0;i<performanceFunctions.length;i++)
					{
						switch(performanceFunctions[i])
						{
							case DIVERSITY_DISTANCE:
								setAllDistances();
								distanceSet = true;
								break;
							case FUNCTION_COMPLEXITY:
								setAllComplexities();
								break;
							case VARIABLE_MAX:
							case VARIABLE_MIN:
								setAllNumberOfInputsUsed();
								break;
						}
					}
					if(isMultiObjective)
					{
						if(!keepDominatedIndividuals)
							removeDuplicatedIndividuals();

						setDominatedMatrix();
						if(!keepDominatedIndividuals)
							removeDominatedIndividuals();
					}
				}
				if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
					setAllDistances();
			}
		}
		catch(InvalidInputSizeException iise)
		{
			System.err.println("InvalidInputSizeException. Unable to evolve population");
			iise.printStackTrace();
			return;
		}
		catch(InvalidFractionException ife)
		{
			System.err.println("InvalidFractionException. Unable to evolve population");
			ife.printStackTrace();
			return;
		}
		generationNumber++;
		setSummaryFitnessValues();//This should be done every generation, to keep average and best fitnesses up-to-date
	}

	private void writeTransmissionInfo()
	{
		try
		{
			Vector<Vector<double[]>> transmissionInfo = getTransmissionInfo();
			Vector<double[]> mutationInfo = mut.getTransmissionInfo();
			Vector<double[]> crossoverInfo = crossover.getTransmissionInfo();

			int corrDataLength = transmissionInfo.get(0).size();
			for(int i=1;i<transmissionInfo.size();i++)
				corrDataLength += transmissionInfo.get(i).size();


			int noTransDims = transmissionInfo.get(0).get(0).length;
			int noDims = noTransDims + crossoverInfo.get(0).length;//assume that crossover info has more dims than mutation
			double[][] correlationInfo = new double[noDims][corrDataLength];
			int counter = 0;
			int crossCounter = 0;
			int mutCounter = 0;
			int[] operatorCounters = new int[8];
			for(int i=0;i<transmissionInfo.size();i++)
			{
				Vector<double[]> currentVector = transmissionInfo.get(i);
				for(int j=0;j<currentVector.size();j++)
				{
					double[] transData = currentVector.get(j);
					for(int k=0;k<transData.length;k++)
						correlationInfo[k][counter] = transData[k];

					double repType = Math.round(correlationInfo[0][counter]);
					if(repType > 0)//mutation
					{
						double[] mutData = (double[])(mutationInfo.get(mutCounter));
						mutCounter++;
						for(int k=0;k<mutData.length;k++)
							correlationInfo[noTransDims + k][counter] = mutData[k];

						//sort mutation operators
						if(correlationInfo[6][counter] == -1.0)
						{

						}
						else if(correlationInfo[6][counter] == 1.0)
						{
							if(correlationInfo[7][counter]>1.0)//don't count point mutations with parent length of 1
								operatorCounters[1]++;
						}
						else
							operatorCounters[(int)(correlationInfo[6][counter])]++;

					}
					else if(repType < 0)//crossover
					{
						double[] crossData = (double[])(crossoverInfo.get(crossCounter));
						crossCounter++;
						for(int k=0;k<crossData.length;k++)
							correlationInfo[noTransDims + k][counter] = crossData[k];

						operatorCounters[7]++;
					}
					counter++;
				}
			}


			double[] poFitnessRatio1 = new double[corrDataLength];
			double[] poFitnessRatio2 = new double[corrDataLength];
			for(int i=0;i<corrDataLength;i++)
			{
				poFitnessRatio1[i] = Math.log10(correlationInfo[1][i]/correlationInfo[3][i]);
				if(poFitnessRatio1[i]<0.0)
					poFitnessRatio1[i] = -poFitnessRatio1[i];
				if(correlationInfo[0][i] < 0.0)//only crossover has 2 parents
				{
					poFitnessRatio2[i] = Math.log10(correlationInfo[2][i]/correlationInfo[3][i]);
					if(poFitnessRatio2[i]<0.0)
						poFitnessRatio2[i] = -poFitnessRatio2[i];
				}
			}

			double[][][] operatorInfo = new double[8][13][];
			for(int i=0;i<8;i++)
				for(int j=0;j<13;j++)
					operatorInfo[i][j] = new double[operatorCounters[i]];

			int[] counters = new int[8];
			for(int i=0;i<corrDataLength;i++)
			{
				int op = 7;//crossover
				if(correlationInfo[0][i] == 0.0)//reproduction
					continue;
				else if(correlationInfo[0][i]>0.0)//mutation
					op = (int)(correlationInfo[6][i]);
				if(op == -1)
					continue;
				else if(op == 1)//point mutation
				{
					if(correlationInfo[7][i]>1)//ignore if parental depth = 1
					{
						operatorInfo[1][0][counters[1]] = correlationInfo[4][i];
						operatorInfo[1][1][counters[1]] = correlationInfo[7][i];
						operatorInfo[1][2][counters[1]] = correlationInfo[8][i];
						operatorInfo[1][3][counters[1]] = correlationInfo[9][i];
						operatorInfo[1][4][counters[1]] = poFitnessRatio1[i];
						operatorInfo[1][10][counters[op]] = Math.log10(correlationInfo[1][i]);
						operatorInfo[1][12][counters[op]] = Math.log10(correlationInfo[3][i]);
						counters[1]++;
					}
				}
				else if(op==7)//crossover
				{
					operatorInfo[op][0][counters[op]] = correlationInfo[4][i];
					operatorInfo[op][1][counters[op]] = correlationInfo[5][i];
					operatorInfo[op][2][counters[op]] = correlationInfo[6][i];
					operatorInfo[op][3][counters[op]] = correlationInfo[7][i];
					operatorInfo[op][4][counters[op]] = correlationInfo[8][i];
					operatorInfo[op][5][counters[op]] = correlationInfo[9][i];
					operatorInfo[op][6][counters[op]] = correlationInfo[10][i];
					operatorInfo[op][7][counters[op]] = correlationInfo[11][i];
					operatorInfo[op][8][counters[op]] = poFitnessRatio1[i];
					operatorInfo[op][9][counters[op]] = poFitnessRatio2[i];
					operatorInfo[op][10][counters[op]] = Math.log10(correlationInfo[1][i]);
					operatorInfo[op][11][counters[op]] = Math.log10(correlationInfo[2][i]);
					operatorInfo[op][12][counters[op]] = Math.log10(correlationInfo[3][i]);
					counters[op]++;
				}
				else//mutation
				{
					operatorInfo[op][0][counters[op]] = correlationInfo[4][i];
					operatorInfo[op][1][counters[op]] = correlationInfo[7][i];
					operatorInfo[op][2][counters[op]] = correlationInfo[8][i];
					operatorInfo[op][3][counters[op]] = correlationInfo[9][i];
					operatorInfo[op][4][counters[op]] = poFitnessRatio1[i];
					operatorInfo[op][10][counters[op]] = Math.log10(correlationInfo[1][i]);
					operatorInfo[op][12][counters[op]] = Math.log10(correlationInfo[3][i]);
					counters[op]++;
				}
			}
			double corr;
			transmissionWriter.write("Generation, "+generationNumber+", Evolution no, "+evolutionNumber+", best train fitness,"+bestTrainFitness);
			transmissionWriter.newLine();
			transmissionWriter.write("Correlation, SUBTREE, POINT, HOIST, SHRINK, EXPAND, SWAP, CONSTANT, crossover");
			transmissionWriter.newLine();
			transmissionWriter.write("Distance 1");
			for(int i=0;i<7;i++)
			{
				corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][0],operatorInfo[i][4]});
				if(Double.isNaN(corr)||Double.isInfinite(corr))
					transmissionWriter.write(",");
				else
					transmissionWriter.write(","+corr);
			}
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][0],operatorInfo[7][8]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write(",");
			else
				transmissionWriter.write(","+corr);

			transmissionWriter.newLine();
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][1],operatorInfo[7][9]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write("Distance 2,,,,,,,,");
			else
				transmissionWriter.write("Distance 2,,,,,,,,"+corr);

			transmissionWriter.newLine();
			transmissionWriter.write("Parent depth 1");
			for(int i=0;i<7;i++)
			{
				corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][1],operatorInfo[i][4]});
				if(Double.isNaN(corr)||Double.isInfinite(corr))
					transmissionWriter.write(",");
				else
					transmissionWriter.write(","+corr);
			}
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][2],operatorInfo[7][8]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write(",");
			else
				transmissionWriter.write(","+corr);

			transmissionWriter.newLine();
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][3],operatorInfo[7][9]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write("Parent depth 2,,,,,,,,");
			else
				transmissionWriter.write("Parent depth 2,,,,,,,,"+corr);

			transmissionWriter.newLine();
			transmissionWriter.write("Cut depth 1");
			double[] allCorrs = new double[8];
			for(int i=0;i<7;i++)
			{
				corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][2],operatorInfo[i][4]});
				allCorrs[i] = corr;
				if(Double.isNaN(corr)||Double.isInfinite(corr))
					transmissionWriter.write(",");
				else
				{
					cumulativeCutDepthCorrelation[i] += corr;
					cutDepthCorrCount[i]++;
					transmissionWriter.write(","+corr);
				}
			}
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][4],operatorInfo[7][8]});
			allCorrs[7] = corr;
			allCutDepthCorrelations.add(allCorrs);
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write(",");
			else
			{
				cumulativeCutDepthCorrelation[7] += corr;
				cutDepthCorrCount[7]++;
				transmissionWriter.write(","+corr);
			}

			transmissionWriter.newLine();
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][5],operatorInfo[7][9]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write("Cut depth 2,,,,,,,,");
			else
				transmissionWriter.write("Cut depth 2,,,,,,,,"+corr);

			transmissionWriter.newLine();
			transmissionWriter.write("Depth to leaf 1");
			for(int i=0;i<7;i++)
			{
				if(i==4||i==6)
					transmissionWriter.write(",");//expand and constant always have depth to leaf of 0, therefore exclude them
				else
				{
					corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][3],operatorInfo[i][4]});
					if(Double.isNaN(corr)||Double.isInfinite(corr))
						transmissionWriter.write(",");
					else
						transmissionWriter.write(","+corr);
				}
			}
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][6],operatorInfo[7][8]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write(",");
			else
				transmissionWriter.write(","+corr);

			transmissionWriter.newLine();
			corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[7][7],operatorInfo[7][9]});
			if(Double.isNaN(corr)||Double.isInfinite(corr))
				transmissionWriter.write("Depth to leaf 2,,,,,,,,");
			else
				transmissionWriter.write("Depth to leaf 2,,,,,,,,"+corr);
			transmissionWriter.newLine();

			//System.err.println("Writing");

			//NEW CODE 041007 - write basic PO correlation
			transmissionWriter.write("PO 1");
			for(int i=0;i<8;i++)
			{
				corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][10],operatorInfo[i][12]});
				if(Double.isNaN(corr)||Double.isInfinite(corr))
					transmissionWriter.write(",");
				else
					transmissionWriter.write(","+corr);
			}
			transmissionWriter.newLine();
			transmissionWriter.write("PO 2");
			for(int i=0;i<8;i++)
			{
				corr = DataAcquirer.getCorrelation(new double[][]{operatorInfo[i][11],operatorInfo[i][12]});
				if(Double.isNaN(corr)||Double.isInfinite(corr))
					transmissionWriter.write(",");
				else
					transmissionWriter.write(","+corr);
			}
			transmissionWriter.newLine();
			//END NEW CODE

		}
		catch(IOException ioe)
		{
			System.err.println("IOException while trying to write to "+transmissionInfoFile);
			ioe.printStackTrace();
		}

		mut.getTransmissionInfo().clear();
		crossover.getTransmissionInfo().clear();
		transmissionInfo.clear();
	}

	private boolean truncationEvolve(boolean steadyState, double[][] inputs,double[] targetOutputs)
	{
		Individual[] parents = new Individual[mu];
		int[] parentIndices = Utils.getRandomIndices(mu, popSize);//select without replacement
		for(int i=0;i<mu;i++)
			parents[i]=individuals[parentIndices[i]];

		Individual[] offspring = null;
		double[] fitnesses;
		double[] savedFitnesses = null;
		if(parentsInTruncationPool)
		{
			offspring = new Individual[mu+lambda];
			fitnesses = new double[mu+lambda];
			if(isMultiObjective)
				savedFitnesses = new double[mu+lambda];
		}
		else
		{
			offspring = new Individual[lambda];
			fitnesses = new double[lambda];
			if(isMultiObjective)
				savedFitnesses = new double[lambda];
		}

		Individual[] tempOffspr = getMultipleOffspring(parents, lambda);
		double[][] performances = new double[lambda][];
		boolean distanceSet = false;
		for(int i=0;i<lambda;i++)
		{
			offspring[i] = tempOffspr[i];
			try
			{
				fitnesses[i] = getFitness(offspring[i],inputs,targetOutputs);
			}
			catch(InvalidInputSizeException iise)
			{
				iise.printStackTrace();
				return false;
			}
			if(isMultiObjective)
				savedFitnesses[i] = fitnesses[i];
			/*rather than calling adjustFitnessForPerformance(), performance values are evaluated explicitly,
			so that they are available to adjust values for functionComplexities, etc.*/
			if(performanceFunctions!=null)
			{
				int len = performanceFunctions.length;
				performances[i] = getPerformances(offspring[i]);
				switch(fitnessFunction)
				{
					case AUROC:
						fitnesses[i] = 2.0 * fitnesses[i] - 1.0;//normalise to range [0,1]
						break;
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
					case NUMBER_CORRECT_BINARY:
					//case CLUSTERED_REGRESSION:
						fitnesses[i] = 1.0 - 1.0 / (fitnesses[i]+1.0);//normalise to range [0,1]
						break;
				}

				for(int j=0;j<len;j++)
				{
					//additionalFitnesses are normalised to range [0,1], with 1 as the best value.
					double additionalFitness = 0.0;
					switch(performanceFunctions[j])
					{
						case PROGRAM_LENGTH:
							additionalFitness = 1.0 / performances[i][j];
							break;
						case DIVERSITY_DISTANCE:
							additionalFitness = performances[j][i];
							distanceSet = true;
							break;
						case FUNCTION_COMPLEXITY:
							additionalFitness = 1.0 / (1.0 + performances[i][j]);
							break;
						case VARIABLE_MAX:
							additionalFitness = performances[i][j] / popDescriptor.getInputs().length;
							break;
						case VARIABLE_MIN:
							additionalFitness = 1.0 - performances[j][i] / popDescriptor.getInputs().length;
							break;
					}
					if(!isMultiObjective)
					{
						switch(fitnessFunction)
						{
							case MSE:
							case RMSE:
							case ABSOLUTE_ERROR:
							case AOREC:
							case COMBINED_CLASSIFICATION_ERROR:
							//case CLUSTERED_REGRESSION:
								fitnesses[i] += 1.0 - additionalFitness;//These FitnessFunctions aim to minimise value
								break;
							//case NUMBER_CORRECT_BINARY:
							//case AUROC:
							default:
								fitnesses[i] += additionalFitness;//These FitnessFunctions aim to maximise value
								break;
						}
					}
				}
				if(isMultiObjective)
				{
					try
					{
						fitnesses[i] = getDominatedCount(fitnesses[i], performances[i], inputs, targetOutputs);
						if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
							fitnesses[i] = startingPopSize - fitnesses[i]; //alter so that least dominated has a HIGH value
					}
					catch(InvalidInputSizeException iise)
					{
						System.err.println("Exception while getting dominated count.");
						System.exit(1);
					}
				}
			}
		}

		if(parentsInTruncationPool)
		{
			for(int i=0;i<mu;i++)
			{
				offspring[lambda+i] = parents[i];
				fitnesses[lambda+i] = allFitnesses[parentIndices[i]];
				if(performanceFunctions != null)
				{
					if(isMultiObjective)
					{
						savedFitnesses[lambda+i] = fitnesses[lambda+i];
						fitnesses[lambda+i] = dominatedCount[parentIndices[i]];
						if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
							fitnesses[lambda+i] = startingPopSize - fitnesses[lambda+i]; //alter so that least dominated has a HIGH value
					}
					else
						adjustFitnessForPerformance(parentIndices[i], fitnesses[lambda+i]);
				}
			}
		}
		Object[] sortedObj = ShellSortIndividuals(offspring, fitnesses, fitnessFunction);
		int[] sortedIndices = (int[])(sortedObj[2]);

		if(steadyState)
		{
			if(performanceFunctions!=null && isMultiObjective && !keepDominatedIndividuals)
			{
				int max = lambda;
				if(parentsInTruncationPool)
					max += mu;
				int newCount = 0;
				for(int i=0;i<max;i++)
				{
					Individual newInd = ((Individual[])sortedObj[0])[i];
					//double newFitness = savedFitnesses[sortedIndices[i]];
					double newFitness = fitnesses[sortedIndices[i]];//080208

					if(addIndividual(newInd, newFitness))
						newCount++;

					if(newCount == mu)
						break;
				}
				populationEvolved += mu;
				if(populationEvolved>=startingPopSize)
				{
					populationEvolved = 0;
					return true;
				}
				else
					return false;
			}
			else
			{
				int removalPoolSize = lambda;
				if(parentsInTruncationPool)
					removalPoolSize+=mu;
				if(removalPoolSize>popSize)
					removalPoolSize = popSize;//new 140308, required because selection is without replacement
				int[] removalIndices=new int[removalPoolSize];
				Individual[] removalInds = new Individual[removalPoolSize];
				double[] remFitnesses = new double[removalPoolSize];
				removalIndices = Utils.getRandomIndices(popSize,removalPoolSize);//140308, select without replacement
				for(int i=0;i<removalPoolSize;i++)
				{
					removalInds[i] = individuals[removalIndices[i]];
					remFitnesses[i] = allFitnesses[removalIndices[i]];
					if(performanceFunctions != null)
					{
						if(isMultiObjective)
						{
							fitnesses[i] = dominatedCount[removalIndices[i]];
							if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
								remFitnesses[i] = startingPopSize - remFitnesses[i];
						}
						else
							adjustFitnessForPerformance(i, remFitnesses[i]);
					}
				}
				Object[] sortedRemovals = ShellSortIndividuals(removalInds, remFitnesses, fitnessFunction);
				int[] remIndices = (int[])(sortedRemovals[2]);

				int[] replacementIndices = new int[mu];
				for(int i=0;i<mu;i++)
				{
					int newIndex = removalIndices[remIndices[removalPoolSize-1-i]];
					replacementIndices[i] = newIndex;
					Individual newInd = ((Individual[])sortedObj[0])[i];
					individuals[newIndex] = newInd;
					//double newFitness = savedFitnesses[sortedIndices[i]];
					double newFitness = fitnesses[sortedIndices[i]];//080208

					allFitnesses[replacementIndices[i]] = newFitness;
					switch(fitnessFunction)
					{
						case MSE:
						case RMSE:
						case ABSOLUTE_ERROR:
						case AOREC:
						case COMBINED_CLASSIFICATION_ERROR:
						//case CLUSTERED_REGRESSION:
							if(newFitness<bestTrainFitness)
							{
								bestTrainFitness=newFitness;
								bestIndex = replacementIndices[i];
							}
							break;
						case NUMBER_CORRECT_BINARY:
						case AUROC:
						case SENSITIVITY:
						case SPECIFICITY:
						case PPV:
						case NPV:
							if(newFitness>bestTrainFitness)
							{
								bestTrainFitness=newFitness;
								bestIndex = replacementIndices[i];
							}
							break;
					}
					if(performanceFunctions != null && isMultiObjective)
					{
						updateDominatedMatrix(newInd, newIndex, newFitness);
					}
					if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
					{
						for(int j=0;j<popSize;j++)
						{
							//all distances to new Individual must be calculated
							if(j<newIndex)
								allSquaredDistances[j][newIndex] = Math.pow(getDistance(j,newIndex),2.0);
							else if(j>newIndex)
								allSquaredDistances[newIndex][j] = Math.pow(getDistance(j,newIndex),2.0);
						}
					}
				}
				setPerformanceValues(replacementIndices, performances);

				populationEvolved += mu;
				if(populationEvolved>=startingPopSize)
				{
					populationEvolved = 0;
					return true;
				}
				else
					return false;
			}
		}
		else
		{
			for(int i=0;i<mu;i++)
			{
				newPopulation[populationEvolved] = ((Individual[])sortedObj[0])[i];
				populationEvolved++;
				if(populationEvolved == startingPopSize)
				{
					individuals = (Individual[])(newPopulation.clone());
					popSize = startingPopSize;
					populationEvolved = 0;
					return true;
				}
			}
			return false;
		}
		//return false;
	}

	/*final void adjustFitnessForPerformance(int index)
	{

	}*/

	final void setPerformanceValues(int[] indices, double[][] performances)
	{
		int len = indices.length;
		if(performanceFunctions!=null && !isMultiObjective && len == performances.length)
		{
			for(int i=0;i<indices.length;i++)
				for(int j=0;j<performances[i].length;i++)
				{
					switch(performanceFunctions[j])
					{
						case DIVERSITY_DISTANCE:
							for(int k=0;k<popSize;k++)
							{
								//all distances to new Individual must be calculated
								if(k<indices[i])
									allSquaredDistances[k][indices[i]] = Math.pow(getDistance(indices[i],k),2.0);
								else if(k>indices[i])
									allSquaredDistances[indices[i]][k] = Math.pow(getDistance(indices[i],k),2.0);
							}
							break;
						case FUNCTION_COMPLEXITY:
							functionComplexities[indices[i]] = (int)performances[i][j];
							break;
						case VARIABLE_MAX:
						case VARIABLE_MIN:
							noInputsUsed[indices[i]] = (int)performances[i][j];
							break;
					}
				}
		}
	}

	final Individual[] getMultipleOffspring(Individual[] parents, int noOfOffspring)
	{
		Individual[] offspring = new Individual[noOfOffspring];
		for(int i=0;i<noOfOffspring;i++)
		{
			double selector = rand.nextDouble();
			int index = (int)(selector * mu);
			Individual parent1 = parents[index];
			selector = rand.nextDouble();
			if(selector<mutationProportion)
			{
				try
				{
					offspring[i] = mut.mutate(parent1,maximumLength,maximumDepth,popDescriptor);//may be null
					if(offspring[i]==null)
					{
						//System.err.println("Null offspring");
						offspring[i] = new Individual(parent1.getNodes(), popDescriptor.getInputs(), popDescriptor.getConstants());
					}
				}
				catch(InvalidFractionException ife)
				{
					ife.printStackTrace();
					offspring[i] = new Individual(parent1.getNodes(), popDescriptor.getInputs(), popDescriptor.getConstants());
				}
			}
			else if(selector<mutationProportion+reproductionProportion)
				offspring[i] = new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			else
			{
				selector = rand.nextDouble();
				index = (int)(selector * mu);
				Individual parent2 = parents[index];
				Individual[] twoOffspring = crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
				if(twoOffspring==null)
					offspring[i] = new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());//if no cross is possible, use asexual reproduction
				else if(twoOffspring[0] == null)
					offspring[i] = twoOffspring[1];
				else
					offspring[i] = twoOffspring[0];
			}
		}
		return offspring;
	}

	private boolean steadyStateEvolve(SelectionMethod selectMethod, double[][] inputs,double[] targetOutputs) throws InvalidFractionException, InvalidInputSizeException
	{
		Individual parent1=null;
		Individual parent2=null;
		double parentFitness1 = Double.NaN;
		double parentFitness2 = Double.NaN;
		double[] rouletteVals=null;
		int index;
		long startTime=0;
		long endTime=0;
		if(performanceFunctions!=null && isMultiObjective && !keepDominatedIndividuals)
		{
			index = (int)(rand.nextDouble()*popSize);
			parent1 = individuals[index];
			parentFitness1 = allFitnesses[index];
		}
		else
		{
			switch(selectMethod)
			{
				case RANDOM:
					index = (int)(rand.nextDouble()*popSize);
					parent1 = individuals[index];
					parentFitness1 = allFitnesses[index];
					break;
				case TOURNAMENT:
					parent1 = getBestIndividualFromTournament();
					if(collectTransmissionInfo)
						parentFitness1 = getFitness(parent1, inputs,targetOutputs);
					break;
				case FUSS:
					parent1 = getDiverseIndividualUsingFUSS();
					if(collectTransmissionInfo)
						parentFitness1 = getFitness(parent1, inputs,targetOutputs);
					break;
				case FITNESS_PROPORTIONAL:
				case RANK:
					rouletteVals = getRouletteWheelValues(selectMethod, true);
					parent1 = getIndividualFromRouletteWheel(rouletteVals);
					if(collectTransmissionInfo)
						parentFitness1 = getFitness(parent1, inputs,targetOutputs);
					break;
				case TRUNCATION:
					System.err.println("Attempt to apply steadyStateEvolve to Truncation method. This is invalid as steadyStateEvolve is only meant to create one new Individual.");
					break;
			}
		}

		if(parent1==null)
			return false;

		//create offspring
		Individual offspring;
		double selector=rand.nextDouble();
		double reproductionType = 0.0;
		if(allowCombinedOp)
		{
			if(selector<reproductionProportion)
				offspring=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			else
			{
				if(performanceFunctions!=null && isMultiObjective && !keepDominatedIndividuals)
				{
					index = (int)(rand.nextDouble()*popSize);
					parent2 = individuals[index];
					parentFitness2 = allFitnesses[index];
				}
				else
				{
					switch(selectMethod)
					{
						case RANDOM:
							index = (int)(rand.nextDouble()*popSize);
							parent2 = individuals[index];
							if(collectTransmissionInfo)
								parentFitness2 = allFitnesses[index];
							break;
						case TOURNAMENT:
							parent2 = getBestIndividualFromTournament();
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
							break;
						case FUSS:
							parent2 = getDiverseIndividualUsingFUSS();
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
							break;
						case FITNESS_PROPORTIONAL:
						case RANK:
							parent2 = getIndividualFromRouletteWheel(rouletteVals);
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
						break;
					}
				}
				if(parent2 == null)
					return false;
				Individual[] twoOffspring = crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
				if(twoOffspring==null)
				{
					return false;
				}
				else if(twoOffspring[0] == null)
				{
					Individual temp = parent2;
					parent2 = parent1;
					parent1 = temp;
					double tempFit = parentFitness2;
					parentFitness2 = parentFitness1;
					parentFitness1 = tempFit;
					offspring = twoOffspring[1];
				}
				else
					offspring = twoOffspring[0];
				reproductionType = -1.0;
			}
			selector=rand.nextDouble();
			if(selector<mutationProportion)
			{
				offspring=mut.mutate(offspring,maximumLength,maximumDepth,popDescriptor);
				if(offspring==null)
					return false;
				else if (reproductionType == 0)
					reproductionType = 1.0;
				else
					reproductionType=2;//combined crossover & mutation type
			}
		}
		else
		{
			if(selector<mutationProportion)
			{
				offspring=mut.mutate(parent1,maximumLength,maximumDepth,popDescriptor);
				if(offspring==null)
					return false;
				else
					reproductionType = 1.0;
			}
			else if(selector<mutationProportion+reproductionProportion)
				offspring=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			else
			{
				if(performanceFunctions!=null && isMultiObjective && !keepDominatedIndividuals)
				{
					index = (int)(rand.nextDouble()*popSize);
					parent2 = individuals[index];
					parentFitness2 = allFitnesses[index];
				}
				else
				{
					switch(selectMethod)
					{
						case RANDOM:
							index = (int)(rand.nextDouble()*popSize);
							parent2 = individuals[index];
							if(collectTransmissionInfo)
								parentFitness2 = allFitnesses[index];
							break;
						case TOURNAMENT:
							//parent2 = getBestIndividualFromTournament(inputs,targetOutputs);
							parent2 = getBestIndividualFromTournament();
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
							break;
						case FUSS:
							parent2 = getDiverseIndividualUsingFUSS();
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
							break;
						case FITNESS_PROPORTIONAL:
						case RANK:
							parent2 = getIndividualFromRouletteWheel(rouletteVals);
							if(collectTransmissionInfo)
								parentFitness2 = getFitness(parent2, inputs,targetOutputs);
						break;
					}
				}
				if(parent2 == null)
					return false;
				Individual[] twoOffspring = crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
				if(twoOffspring==null)
				{
					return false;
				}
				else if(twoOffspring[0] == null)
				{
					Individual temp = parent2;
					parent2 = parent1;
					parent1 = temp;
					double tempFit = parentFitness2;
					parentFitness2 = parentFitness1;
					parentFitness1 = tempFit;
					offspring = twoOffspring[1];
				}
				else
					offspring = twoOffspring[0];
				reproductionType = -1.0;
			}
		}

		//get new fitness
		double newFitness=0.0;
		switch(fitnessFunction)
		{
			case MSE:
				newFitness = offspring.getMSE(inputs,targetOutputs);
				break;
			case RMSE:
				newFitness = offspring.getRMSE(inputs,targetOutputs);
				break;
			case ABSOLUTE_ERROR:
				newFitness = offspring.getAbsoluteError(inputs,targetOutputs);
				break;
			case NUMBER_CORRECT_BINARY:
				newFitness = offspring.getNumberCorrectBinary(inputs,targetOutputs);
				break;
			case AUROC:
				newFitness = offspring.getAUROC(inputs,targetOutputs,aurocSampleSize,false,true);
				break;
			case AOREC:
				newFitness = offspring.getAOREC(inputs,targetOutputs,aurocSampleSize,recFitnessFunction);
				break;
			case SENSITIVITY:
				newFitness = offspring.getSensitivity(inputs,targetOutputs);
				break;
			case SPECIFICITY:
				newFitness = offspring.getSpecificity(inputs,targetOutputs);
				break;
			case PPV:
				newFitness = offspring.getPositivePredictiveValue(inputs,targetOutputs);
				break;
			case NPV:
				newFitness = offspring.getNegativePredictiveValue(inputs,targetOutputs);
				break;
			case COMBINED_CLASSIFICATION_ERROR:
				newFitness = offspring.getCombinedClassificationError(inputs,targetOutputs,aurocSampleSize, false, false, true);
				break;
			/*case CLUSTERED_REGRESSION:
				newFitness = offspring.getClusteredRegressionMSE(inputs, targetOutputs);
				break;*/
		}

		if(performanceFunctions!=null && isMultiObjective && !keepDominatedIndividuals)
			addIndividual(offspring,newFitness);
		else
		{
			//select worst individual
			int worstIndex = -1;
			switch(selectMethod)
			{
				case RANDOM:
					worstIndex = (int)(rand.nextDouble()*popSize);
					break;
				case TOURNAMENT:
					worstIndex = getWorstIndexFromTournament();
					break;
				case FUSS:
					worstIndex = getCrowdedIndexUsingFUSS();
					break;
				case FITNESS_PROPORTIONAL:
				case RANK:
					rouletteVals = getRouletteWheelValues(selectMethod, false);
					worstIndex = getIndexFromRouletteWheel(rouletteVals);
					break;
			}

			//replace worst individual by offspring
			individuals[worstIndex] = offspring;
			allFitnesses[worstIndex] = newFitness;
			if(performanceFunctions != null && isMultiObjective)
				updateDominatedMatrix(offspring, worstIndex, newFitness);

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(newFitness<bestTrainFitness)
					{
						bestTrainFitness=newFitness;
						bestIndex = worstIndex;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(newFitness>bestTrainFitness)
					{
						bestTrainFitness=newFitness;
						bestIndex = worstIndex;
					}
					break;
			}

			boolean distanceSet = false;
			if(performanceFunctions != null)
			{
				for(int i=0;i<performanceFunctions.length;i++)
				{
					switch(performanceFunctions[i])
					{
						case DIVERSITY_DISTANCE:
							for(int j=0;j<popSize;j++)
							{
								if(j<worstIndex)
									allSquaredDistances[j][worstIndex] = Math.pow(getDistance(worstIndex,j),2.0);
								else
									allSquaredDistances[worstIndex][j] = Math.pow(getDistance(worstIndex,j),2.0);
							}
							distanceSet = true;
							break;
						case FUNCTION_COMPLEXITY:
							functionComplexities[worstIndex] = getComplexity(offspring);
							break;
						case VARIABLE_MAX:
						case VARIABLE_MIN:
							noInputsUsed[worstIndex] = getNumberOfInputsUsed(offspring);
							break;
					}
				}
			}
			if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
			{
				for(int j=0;j<popSize;j++)
				{
					if(j<worstIndex)
						allSquaredDistances[j][worstIndex] = Math.pow(getDistance(worstIndex,j),2.0);
					else
						allSquaredDistances[worstIndex][j] = Math.pow(getDistance(worstIndex,j),2.0);
				}
			}
		}
		if(collectTransmissionInfo)
		{
			double[] transInfo = new double[6];
			transInfo[0] = reproductionType;
			transInfo[1] = parentFitness1;
			transInfo[2] = parentFitness2;
			transInfo[3] = newFitness;
			transInfo[4] = getDistance(parent1, offspring);

			if(parent2 == null)
				transInfo[5] = Double.NaN;
			else
				transInfo[5] = getDistance(parent2, offspring);
			currentTransmissionVector.add(transInfo);
		}

		evolutionNumber++;
		if(transmissionWriter != null && evolutionNumber % transmissionWritingFrequency == 0)
			writeTransmissionInfo();

		return true;
	}


	//Not suitable for TRUNCATION selection
	private void generationalEvolve(SelectionMethod selectMethod, double[][] inputs,double[] targetOutputs) throws InvalidFractionException, InvalidInputSizeException
	{
		double[] rouletteVals=null;
		switch(selectMethod)
		{
			case FITNESS_PROPORTIONAL:
			case RANK:
			case FUSS:
				rouletteVals = getRouletteWheelValues(selectMethod, true);
				break;
			//for other methods do nothing
		}
		int len = startingPopSize;
		Individual[] newInds=new Individual[len];
		int cloneSize = eliteSize;
		if(popSize < eliteSize)
			cloneSize = popSize;
		if(cloneSize>0)
		{

			double[] fitnesses = new double[popSize];
			for(int i=0;i<popSize;i++)
			{
				fitnesses[i] = allFitnesses[i];
				adjustFitnessForPerformance(i,fitnesses[i]);
			}
			Individual[] orderedIndividuals = (Individual[])(ShellSortIndividuals(individuals, fitnesses, fitnessFunction)[0]);
			for(int i=0;i<cloneSize;i++)
				newInds[i]=orderedIndividuals[i];
		}
		for(int i = cloneSize;i<len;i++)
		{
			Individual parent1 = null;
			switch(selectMethod)
			{
				case FITNESS_PROPORTIONAL:
				case RANK:
				case FUSS:
					parent1 = getIndividualFromRouletteWheel(rouletteVals);
					break;
				case TOURNAMENT:
					parent1 = getBestIndividualFromTournament();
					break;
				case RANDOM:
					parent1 = individuals[(int)(rand.nextDouble()*popSize)];
					break;
			}
			double selector = rand.nextDouble();

			if(selector<mutationProportion)
			{
				newInds[i] = mut.mutate(parent1,maximumLength,maximumDepth,popDescriptor);
				//if mutation fails, use parent in next generation
				if(newInds[i] == null)
					newInds[i] = new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			}
			else if(selector<mutationProportion+reproductionProportion)
			{
				newInds[i]=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
			}
			else
			{
				Individual parent2 = null;
				switch(selectMethod)
				{
					case FITNESS_PROPORTIONAL:
					case RANK:
					case FUSS:
						parent2 = getIndividualFromRouletteWheel(rouletteVals);
						break;
					case TOURNAMENT:
						parent2 = getBestIndividualFromTournament();
						break;
					case RANDOM:
						parent2 = individuals[(int)(rand.nextDouble()*popSize)];
						break;
				}
				Individual[] twoOffspring=crossover.cross(parent1,parent2,true,popDescriptor.getInputs(),popDescriptor.getConstants(), maximumLength, maximumDepth);
				//if crossover fails, use one of the parents in the next generation
				if(twoOffspring==null)
				{
					selector=rand.nextDouble();
					if(selector<0.5)
						newInds[i]=new Individual(parent1.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
					else
						newInds[i]=new Individual(parent2.getNodes(),popDescriptor.getInputs(),popDescriptor.getConstants());
				}
				else if (twoOffspring[0] == null)
					newInds[i]=twoOffspring[1];
				else
					newInds[i]=twoOffspring[0];
			}
		}
		individuals = (Individual[])(newInds.clone());
	}

	Individual getIndividualFromRouletteWheel(double[] rouletteVals)
	{
		return individuals[getIndexFromRouletteWheel(rouletteVals)];
	}

	private int getIndexFromRouletteWheel(double[] rouletteVals)
	{
		double selector = 1.0 - rand.nextDouble();
		int index = Arrays.binarySearch(rouletteVals,selector);
		if(index<0)
			index = -(index+1);
		return index;
	}

	/**If <i>addition</i> is true, roulette wheel values are set for the addition of an Individual
	to the breeding pool. Otherwise, roulette wheel values will be set for deletion from the population.*/
	final double[] getRouletteWheelValues(SelectionMethod selectMethod, boolean addition)
	{
		double[] rouletteWheelValues=new double[popSize];
		double totalFitness=0.0;
		double fitness;
		double val;
		double[] shareVals = null;
		switch(selectMethod)
		{
			case RANDOM:
				for(int i=0;i<popSize;i++)
					rouletteWheelValues[i] = ((double)(i+1))/popSize;
				break;

			case FITNESS_PROPORTIONAL:
				if(fitnessShare!=null && addition)//don't use sharing for removal
					shareVals = getAllShareValues();
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
					/*case CLUSTERED_REGRESSION:
						for(int i=0;i<popSize;i++)
						{
							fitness = allFitnesses[i];
							if(performanceFunctions != null)
							{
								if(isMultiObjective)
									fitness = dominatedCount[i];
								else
									adjustFitnessForPerformance(i , fitness);
							}

							val = 1.0/(fitness+1.0);
							if(!addition)
								val = 1.0-val;
							if(shareVals != null && shareVals[i] != 0.0)
							{
								//if(addition)
									val /= shareVals[i];//fitness sharing
								//else
								//	val *= shareVals[i];//fitness sharing - don't apply to removal
							}
							rouletteWheelValues[i] = val + totalFitness;
							totalFitness+=val;
						}
						break;*/
					case NUMBER_CORRECT_BINARY:
						for(int i=0;i<popSize;i++)
						{
							fitness = allFitnesses[i];
							if(performanceFunctions != null)
							{
								if(isMultiObjective)
								{
									if(addition)
										val = popSize - dominatedCount[i];
									else
										val = 1 + dominatedCount[i];
								}
								else
								{
									adjustFitnessForPerformance(i , fitness);
									if(addition)
										val = fitness;
									else
										val = 1 + performanceFunctions.length - fitness;
								}
							}
							else
							{
								if(addition)
									val = fitness +1;
								else
									val = bestTrainFitness - fitness;
							}
							if(shareVals != null && shareVals[i] != 0.0)
								val /= shareVals[i];//fitness sharing
							rouletteWheelValues[i] = val + totalFitness;
							totalFitness+=val;
						}
						break;
					case AUROC:
						for(int i=0;i<popSize;i++)
						{
							fitness = allFitnesses[i];
							//double val;
							if(performanceFunctions != null)
							{
								if(isMultiObjective)
									val = popSize - dominatedCount[i];
								else
								{
									adjustFitnessForPerformance(i , fitness);
									if(addition)
										val = fitness;
									else
										val = 1 + performanceFunctions.length - fitness;
								}
							}
							else
							{
								if(addition)
									val = fitness * 2.0 - 1.0;
								else
									val = (1.0 - fitness) * 2.0;
							}
							if(shareVals != null && shareVals[i] != 0.0)
								val /= shareVals[i];//fitness sharing
							rouletteWheelValues[i] = val + totalFitness;
							totalFitness+=val;
						}
						break;
					case SENSITIVITY:
					case SPECIFICITY:
					case PPV:
					case NPV:
						for(int i=0;i<popSize;i++)
						{
							fitness = allFitnesses[i];
							//double val;
							if(performanceFunctions != null)
							{
								if(isMultiObjective)
									val = popSize - dominatedCount[i];
								else
								{
									adjustFitnessForPerformance(i , fitness);
									if(addition)
										val = fitness;
									else
										val = 1 + performanceFunctions.length - fitness;
								}
							}
							else
							{
								if(addition)
									val = fitness;
								else
									val = 1.0 - fitness;
							}
							if(shareVals != null && shareVals[i] != 0.0)
								val /= shareVals[i];//fitness sharing
							rouletteWheelValues[i] = val + totalFitness;
							totalFitness+=val;
						}
						break;
					case COMBINED_CLASSIFICATION_ERROR:
						for(int i=0;i<popSize;i++)
						{
							fitness = allFitnesses[i];
							if(performanceFunctions != null)
							{
								if(isMultiObjective)
									val = popSize - dominatedCount[i];
								else
								{
									adjustFitnessForPerformance(i , fitness);
									if(addition)
										val = fitness;
									else
										val = 1 + performanceFunctions.length - fitness;
								}
							}
							else
							{
								if(addition)
									val = 1.0 - fitness;
								else
									val = fitness;
							}
							if(shareVals != null && shareVals[i] != 0.0)
								val /= shareVals[i];//fitness sharing
							rouletteWheelValues[i] = val + totalFitness;
							totalFitness+=val;
						}
						break;
				}
				break;

			case FUSS:
				double[] fitnesses = new double[popSize];
				//If performanceFunctions are used, adjustFitnessForPerformance will normalise values
				if(performanceFunctions!=null && !isMultiObjective)
				{
					for(int i=0;i<popSize;i++)
						fitnesses[i] = allFitnesses[i];
				}
				else
				{
					switch(fitnessFunction)
					{
						case MSE:
						case RMSE:
						case ABSOLUTE_ERROR:
						case AOREC:
						//case CLUSTERED_REGRESSION:
							for(int i=0;i<popSize;i++)
								fitnesses[i] = 1.0 - 1.0 / (allFitnesses[i] + 1.0);//fit to range [0,1]. This avoids infinite values and gives more even spread of values
							break;
						/*case NUMBER_CORRECT_BINARY:
							for(int i=0;i<popSize;i++)
								fitnesses[i] = allFitnesses[i] / bestTrainFitness;//fit to range [0,1]*/
					}
				}

				for(int i=0;i<popSize;i++)
					adjustFitnessForPerformance(i, fitnesses[i]);

				Object[] orderedObj = ShellSortIndividuals(individuals , fitnesses, fitnessFunction);
				double[] orderedFitnesses = (double[])(orderedObj[1]);
				int[] orderedIndices = (int[])(orderedObj[2]);

				double maxDistance = -1.0;
				if(orderedFitnesses[orderedFitnesses.length-1]>orderedFitnesses[0])
				{
					double averageWidth = (orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[0]) / (popSize-1);
					val = averageWidth + orderedFitnesses[1] - orderedFitnesses[0];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[0]] = val;
					for(int i=1;i<popSize-1;i++)
					{
						val = orderedFitnesses[i+1] - orderedFitnesses[i-1];
						if(val>maxDistance)
							maxDistance = val;
						rouletteWheelValues[orderedIndices[i]] = val;
					}
					val = averageWidth + orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[orderedFitnesses.length-2];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[popSize-1]] = val;
				}
				else
				{
					double averageWidth = (orderedFitnesses[0] - orderedFitnesses[orderedFitnesses.length-1]) / (popSize-1);
					val = averageWidth + orderedFitnesses[0] - orderedFitnesses[1];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[0]] = val;
					for(int i=1;i<popSize-1;i++)
					{
						val = orderedFitnesses[i-1] - orderedFitnesses[i+1];
						if(val>maxDistance)
							maxDistance = val;
						rouletteWheelValues[orderedIndices[i]] = val;
					}
					val = averageWidth + orderedFitnesses[orderedFitnesses.length-2] - orderedFitnesses[orderedFitnesses.length-1];
					if(val>maxDistance)
						maxDistance = val;
					rouletteWheelValues[orderedIndices[popSize-1]] = val;
				}
				for(int i=0;i<popSize;i++)
				{
					if(addition)
						totalFitness += rouletteWheelValues[i];
					else
						totalFitness += maxDistance - rouletteWheelValues[i];
					rouletteWheelValues[i] = totalFitness;
				}
				break;

			case RANK:
				double[] fits = new double[popSize];
				for(int i=0;i<popSize;i++)
					fits[i] = allFitnesses[i];

				if(fitnessShare!=null && addition)//don't use sharing for removal
					shareVals = getAllShareValues();

				for(int i=0;i<popSize;i++)
				{
					if(performanceFunctions != null)
					{
						if(isMultiObjective)
						{
							fits[i] = dominatedCount[i];
							if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
								fits[i] = popSize + 1 - fits[i];
						}
						else
							adjustFitnessForPerformance(i , fits[i]);
					}
					if(shareVals != null && shareVals[i] != 0.0)
						fits[i] /= shareVals[i];//fitness sharing
				}


				Object[] orderedOb = ShellSortIndividuals(individuals , fits, fitnessFunction);
				int[] orderedI = (int[])(orderedOb[2]);
				for(int i=0;i<popSize;i++)
				{
					if(addition)
						rouletteWheelValues[orderedI[i]] = rankGradient * (popSize - i -1) +1;
					else
						rouletteWheelValues[orderedI[i]] = rankGradient * i + 1;
				}
				for(int i=0;i<popSize;i++)
				{
					totalFitness += rouletteWheelValues[i];
					rouletteWheelValues[i] = totalFitness;
				}
				break;

			case TOURNAMENT:
			case TRUNCATION:
				System.err.println("Attempt to obtain roulette wheel values for tournament or truncation selection. This is an invalid operation.");
				return null;
		}
		for(int i=0;i<popSize;i++)
			rouletteWheelValues[i] /= totalFitness;
		return rouletteWheelValues;
	}

	final Individual getDiverseIndividualUsingFUSS()
	{
		Object[] orderedObj = ShellSortIndividuals();
		Individual[] orderedInds = (Individual[])(orderedObj[0]);
		double[] orderedFitnesses = (double[])(orderedObj[1]);
		switch(fitnessFunction)
		{
			case MSE:
			case RMSE:
			case ABSOLUTE_ERROR:
			case AOREC:
			//case CLUSTERED_REGRESSION:
				for(int i=0;i<popSize;i++)
					orderedFitnesses[i] = 1.0/(orderedFitnesses[i]+1.0);//fit to range [0,1]. This avoids infinite values and gives more even spread of values
				break;
				//take no action for other fitnessFunctions
		}
		double max = orderedFitnesses[0];
		double min = orderedFitnesses[popSize-1];
		boolean reverseArray=false;
		if(max<min)
		{
			double temp = min;
			min=max;
			max=temp;
			for(int i=0;i<popSize/2;i++)
			{
				temp = orderedFitnesses[i];
				orderedFitnesses[i] = orderedFitnesses[popSize-1-i];
				orderedFitnesses[popSize-1-i] = orderedFitnesses[i];
			}
			reverseArray=true;
		}
		double width = (max-min)/(popSize-1);
		max+=width/2.0;
		min-=width/2.0;
		double selector=rand.nextDouble()*(max-min)+min;
		double[] boundaries = new double[popSize];
		for(int i=0;i<popSize-1;i++)
			boundaries[i] = (orderedFitnesses[i]+orderedFitnesses[i+1])/2.0;
		boundaries[popSize-1] = max;
		int index = Arrays.binarySearch(boundaries,selector);
		if(index<0)
			index = -(index+1);
		if(reverseArray)
			index = popSize-1-index;
		return orderedInds[index];
	}

	final int getCrowdedIndexUsingFUSS()
	{
		Object[] orderedObj = ShellSortIndividuals();
		double[] orderedFitnesses = (double[])(orderedObj[1]);
		int[] orderedIndices = (int[])(orderedObj[2]);
		switch(fitnessFunction)
		{
			case MSE:
			case RMSE:
			case ABSOLUTE_ERROR:
			case AOREC:
			//case CLUSTERED_REGRESSION:
				for(int i=0;i<popSize;i++)
					orderedFitnesses[i] = 1.0/(orderedFitnesses[i]+1.0);//fit to range [0,1]. This avoids infinite values and gives more even spread of values
				break;
				//take no action for other fitnessFunctions
		}

		double[] nearestDistance = new double[popSize];

		if(orderedFitnesses[orderedFitnesses.length-1]>orderedFitnesses[0])
		{
			double averageWidth = (orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[0]) / (popSize-1);
			nearestDistance[0] = averageWidth + orderedFitnesses[1] - orderedFitnesses[0];
			for(int i=1;i<popSize-1;i++)
				nearestDistance[i] = orderedFitnesses[i+1] - orderedFitnesses[i-1];
			nearestDistance[popSize-1] = averageWidth + orderedFitnesses[orderedFitnesses.length-1] - orderedFitnesses[orderedFitnesses.length-2];
		}
		else
		{
			double averageWidth = (orderedFitnesses[0] - orderedFitnesses[orderedFitnesses.length-1]) / (popSize-1);
			nearestDistance[0] = averageWidth + orderedFitnesses[0] - orderedFitnesses[1];
			for(int i=1;i<popSize-1;i++)
				nearestDistance[i] = orderedFitnesses[i-1] - orderedFitnesses[i+1];
			nearestDistance[popSize-1] = averageWidth + orderedFitnesses[orderedFitnesses.length-2] - orderedFitnesses[orderedFitnesses.length-1];
		}

		int index=0;
		double smallestDistance = Double.MAX_VALUE;
		for(int i=0;i<popSize;i++)
		{
			if(nearestDistance[i]<smallestDistance)
			{
				index=i;
				smallestDistance = nearestDistance[i];
			}
		}
		return orderedIndices[index];
	}

	private Individual getBestIndividualFromTournament()
	{
		if(popSize<tournamentSize)
		{
			System.err.println("Population size ("+popSize+") is less than tournament size ("+tournamentSize+").No evolution will occur.");
			return null;
		}
		int index=(int)(rand.nextDouble()*popSize);
		Individual bestInd=individuals[index];
		double bestScore = allFitnesses[index]; //use cached value
		if(performanceFunctions != null)
		{
			if(isMultiObjective)
			{
				bestScore = dominatedCount[index];
				bestScore -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
				if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
					bestScore = popSize - bestScore;
			}
			else
			{
				adjustFitnessForPerformance(index , bestScore);
				bestScore /= getShareValue(index);
			}
		}
		else
			bestScore /= getShareValue(index);

		for(int i=1;i<tournamentSize;i++)
		{
			index=(int)(rand.nextDouble()*popSize);
			Individual ind=individuals[index];
			double score = allFitnesses[index];
			if(performanceFunctions != null)
			{
				if(isMultiObjective)
				{
					score = dominatedCount[index];
					score -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
					if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
						score = popSize - score;
				}
				else
				{
					adjustFitnessForPerformance(index , score);
					score /= getShareValue(index);
				}
			}
			else
				score /= getShareValue(index);

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(score<bestScore)
					{
						bestScore=score;
						bestInd=ind;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(score>bestScore)
					{
						bestScore=score;
						bestInd=ind;
					}
					break;
			}
		}
		return bestInd;
	}

	//not used 070906
	private Individual getWorstIndividualFromTournament()
	{
		if(popSize<tournamentSize)
		{
			System.err.println("Population size is less than tournament size.No evolution will occur.");
			return null;
		}
		int[] indices = Utils.getRandomIndices(popSize, tournamentSize);
		int index = indices[0];
		Individual worstInd=individuals[index];
		double worstScore=allFitnesses[index];
		if(performanceFunctions != null)
		{
			if(isMultiObjective)
			{
				//make sure that best Individual is kept, i.e. elitist selection
				if(worstScore == bestTrainFitness)
					worstScore = dominatedCount[index] - 1;
				else
					worstScore = dominatedCount[index];
				worstScore -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
				if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
					worstScore = -worstScore;
			}
			else
			{
				adjustFitnessForPerformance(index , worstScore);
			}
		}

		for(int i=1;i<tournamentSize;i++)
		{
			index = indices[i];
			Individual ind = individuals[index];
			double score = allFitnesses[index];
			if(performanceFunctions != null)
			{
				if(isMultiObjective)
				{
					//make sure that best Individual is kept, i.e. elitist selection
					if(score == bestTrainFitness)
						score = dominatedCount[index] - 1;
					else
						score = dominatedCount[index];
					score -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
					if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
						score = -score;
				}
				else
				{
					adjustFitnessForPerformance(index , score);
				}
			}

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(score>worstScore)
					{
						worstScore=score;
						worstInd=ind;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(score<worstScore)
					{
						worstScore=score;
						worstInd=ind;
					}
					break;
			}
		}
		return worstInd;
	}

	private int getWorstIndexFromTournament()
	{
		if(popSize<tournamentSize)
		{
			System.err.println("Population size is less than tournament size.No evolution will occur.");
			return -1;
		}
		int[] indices = null;
		if(tournamentSize<=popSize)
			indices = Utils.getRandomIndices(popSize, tournamentSize);//new 261006. Select without repetitions if possible
		else
		{
			indices = new int[tournamentSize];
			for(int i=0;i<tournamentSize;i++)
				indices[i] = (int)(rand.nextDouble() * popSize);
		}
		int worstIndex = indices[0];

		double worstScore = allFitnesses[worstIndex];
		if(performanceFunctions != null)
		{
			if(isMultiObjective)
			{
				//make sure that best Individual is kept, i.e. elitist selection
				if(worstScore == bestTrainFitness)
					worstScore = dominatedCount[worstIndex] - 1;
				else
					worstScore = dominatedCount[worstIndex];
				worstScore -= 1.0 / getShareValue(worstIndex);//adjust to give preference to unshared Individuals
				if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
					worstScore = -worstScore;
			}
			else
			{
				adjustFitnessForPerformance(worstIndex , worstScore);
			}
		}

		for(int i=1;i<tournamentSize;i++)
		{
			int index = indices[i];//new 261006. Select without repetitions

			double score = allFitnesses[index];

			if(performanceFunctions != null)
			{
				if(isMultiObjective)
				{
					//make sure that best Individual is kept, i.e. elitist selection
					if(score == bestTrainFitness)
						score = dominatedCount[index] - 1;
					else
						score = dominatedCount[index];
					score -= 1.0 / getShareValue(index);//adjust to give preference to unshared Individuals
					if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
						score = -score;
				}
				else
				{
					adjustFitnessForPerformance(index , score);
				}
			}

			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(score>worstScore)
					{
						worstScore=score;
						worstIndex=index;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(score<worstScore)
					{
						worstScore=score;
						worstIndex=index;
					}

					break;
			}
		}
		return worstIndex;
	}

	/**When a performanceFunction is used, the fitness of an Individual is adjusted using this method.*/
	final void adjustFitnessForPerformance(int i, double fitness)
	{
		if(performanceFunctions!=null && !isMultiObjective)
		{
			if(fitnessFunction.equals(FitnessFunction.AUROC))
				fitness = 2.0 * fitness - 1.0;//normalise to range [0,1], keeping ordering (large-to-small or vice versa) the same
			else if(fitnessFunction.equals(FitnessFunction.MSE)||fitnessFunction.equals(FitnessFunction.RMSE)||fitnessFunction.equals(FitnessFunction.ABSOLUTE_ERROR)||fitnessFunction.equals(FitnessFunction.AOREC)/*||fitnessFunction.equals(FitnessFunction.CLUSTERED_REGRESSION)*/)
				fitness = 1.0 - 1.0 / (fitness + 1.0);//normalise to range [0,1], keeping ordering (large-to-small or vice versa) the same
			//for other fitnessFunctions do nothing

			for(int j=0;j<performanceFunctions.length;j++)
			{
				//additionalFitnesses are normalised to range [0,1], with 1 as the best value.
				double additionalFitness = 0.0;
				switch(performanceFunctions[j])
				{
					case PROGRAM_LENGTH:
						additionalFitness = 1.0 / individuals[i].getLength();
						break;
					case DIVERSITY_DISTANCE:
						additionalFitness = getAverageSquaredDistance(i);
						break;
					case FUNCTION_COMPLEXITY:
						additionalFitness = 1.0 / (functionComplexities[i] + 1.0);
						break;
					case VARIABLE_MAX:
						additionalFitness = ((double)noInputsUsed[i]) / popDescriptor.getInputs().length;
						break;
					case VARIABLE_MIN:
						additionalFitness = 1.0 - ((double)noInputsUsed[i]) / popDescriptor.getInputs().length;
						break;
				}
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
					case COMBINED_CLASSIFICATION_ERROR:
					//case CLUSTERED_REGRESSION:
						fitness += 1.0 - additionalFitness;//These FitnessFunctions aim to minimise value
						break;
					//case NUMBER_CORRECT_BINARY:
					//case AUROC:
					default:
						fitness += additionalFitness;//These FitnessFunctions aim to maximise value
						break;
				}
			}
		}
	}

	final double getShareValue(int index)
	{
		if(fitnessShare == null)
			return 1.0;
		else
		{
			switch(shareType)
			{
				case GENE_DISTANCE:
					return fitnessShare.getShareValuesFromSquaredDistances(allSquaredDistances, index);
				case OBJECTIVE_FUNCTION:
					if(shareWeightings == null || performanceFunctions == null)
						return fitnessShare.getShareValuesFromSingleObjective(allFitnesses,index);
					else //this will be slow if called in steady state mode and may lead to oscillation in generational mode - use with care!!
					{
						double[] allData = new double[1+performanceFunctions.length];
						allData[0] = allFitnesses[index];
						for(int j=0;j<performanceFunctions.length;j++)
						{
							switch(performanceFunctions[j])
							{
								case DIVERSITY_DISTANCE:
									allData[j+1] = getAverageSquaredDistance(index);
									break;
								case FUNCTION_COMPLEXITY:
									allData[j+1] = functionComplexities[index];
									break;
								case VARIABLE_MAX:
								case VARIABLE_MIN:
									allData[j+1] = noInputsUsed[index];
									break;
							}
						}
						return fitnessShare.getShareValuesFromMultipleObjectives(new double[][]{allData},shareWeightings)[0];
					}
				default:
					return 1.0;
			}
		}
	}

	final double[] getAllShareValues()
	{
		double[] shareVals = new double[popSize];
		switch(shareType)
		{
			case GENE_DISTANCE:
				shareVals = fitnessShare.getShareValuesFromSquaredDistances(allSquaredDistances);
				break;
			case OBJECTIVE_FUNCTION:
				if(shareWeightings == null || performanceFunctions == null)
					shareVals = fitnessShare.getShareValuesFromSingleObjective(allFitnesses);
				else //this will be slow if called in steady state mode and may lead to oscillation in generational mode - use with care!!
				{
					double[][] allData = new double[popSize][1+performanceFunctions.length];
					for(int i=0;i<popSize;i++)
					{
						allData[i][0] = allFitnesses[i];
						for(int j=0;j<performanceFunctions.length;j++)
						{
							switch(performanceFunctions[j])
							{
								case DIVERSITY_DISTANCE:
									allData[i][j+1] = getAverageSquaredDistance(i);
									break;
								case FUNCTION_COMPLEXITY:
									allData[i][j+1] = functionComplexities[i];
									break;
								case VARIABLE_MAX:
								case VARIABLE_MIN:
									allData[i][j+1] = noInputsUsed[i];
									break;
							}
						}
					}
					shareVals = fitnessShare.getShareValuesFromMultipleObjectives(allData,shareWeightings);
				}
				break;
			default:
				System.err.println("Unknown ShareType.");
				return null;
		}
		return shareVals;
	}

	/**If set to true, information concerning the relationship between parents and offspring will be saved.*/
	public void collectTransmissionInfo(boolean collect)
	{
		collectTransmissionInfo = collect;
		mut.collectTransmissionInfo(collect);
		crossover.collectTransmissionInfo(collect);
	}

	/**Sets parameters for saving transmission information (if used).
	@param frequency The frequency with which transmission information is saved (in number of evaluations)
	@param transmissionInfoFile A pathname describing the file to which transmission information should be written*/
	public void setTransmissionWritingParameters(int frequency, String transmissionInfoFile)
	{
		if(frequency>0)
			transmissionWritingFrequency = frequency;
		this.transmissionInfoFile = transmissionInfoFile;
	}

	/**Returns a vector containing transmission information concerning the relationships between parents and offspring*/
	public Vector<Vector<double[]>> getTransmissionInfo()
	{
		return transmissionInfo;
	}

	/**This method may be used if transmission information has been collected. It returns the average correlation between
	the depth at which crossover occurs and the log of the ratio of the parent and offspring fitnesses.*/
	public double[] getAverageCutDepthCorrelation()
	{
		if(cumulativeCutDepthCorrelation == null)
			return null;
		for(int i=0;i<cumulativeCutDepthCorrelation.length;i++)
			cumulativeCutDepthCorrelation[i] /= cutDepthCorrCount[i];
		return cumulativeCutDepthCorrelation;
	}

	/**This method may be used if transmission information has been collected. It returns the correlations between
	the depth at which crossover occurs and the log of the ratio of the parent and offspring fitnesses.*/
	public Vector<double[]> getAllCutDepthCorrelations()
	{
		return allCutDepthCorrelations;
	}

	final void setFitnessValues(double[][]inputs,double[]targetOutputs) throws InvalidInputSizeException
	{
		allFitnesses=new double[popSize];
		averageTrainFitness=0.0;
		bestTrainFitness=Double.MAX_VALUE;
		if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
			bestTrainFitness=-Double.MAX_VALUE;
		for(int i=0;i<popSize;i++)
		{
			double val=0.0;
			switch(fitnessFunction)
			{
				case MSE:
					val=individuals[i].getMSE(inputs,targetOutputs);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case RMSE:
					val=individuals[i].getRMSE(inputs,targetOutputs);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case ABSOLUTE_ERROR:
					val=individuals[i].getAbsoluteError(inputs,targetOutputs);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case AOREC:
					val=individuals[i].getAOREC(inputs,targetOutputs,aurocSampleSize,recFitnessFunction);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case NUMBER_CORRECT_BINARY:
					val=individuals[i].getNumberCorrectBinary(inputs,targetOutputs);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case AUROC:
					val=individuals[i].getAUROC(inputs,targetOutputs,aurocSampleSize,true,true);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case SENSITIVITY:
					val=individuals[i].getSensitivity(inputs,targetOutputs);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case SPECIFICITY:
					val=individuals[i].getSpecificity(inputs,targetOutputs);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case PPV:
					val=individuals[i].getPositivePredictiveValue(inputs,targetOutputs);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case NPV:
					val=individuals[i].getNegativePredictiveValue(inputs,targetOutputs);
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case COMBINED_CLASSIFICATION_ERROR:
					val=individuals[i].getCombinedClassificationError(inputs,targetOutputs,aurocSampleSize, true, true, true);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				/*case CLUSTERED_REGRESSION:
					val=individuals[i].getClusteredRegressionMSE(inputs,targetOutputs);
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;*/
			}
			allFitnesses[i]=val;
			averageTrainFitness+=val;
		}
		assert(!Double.isNaN(averageTrainFitness));
		averageTrainFitness/=popSize;
	}

	private void setBestFitness()
	{
		bestTrainFitness=Double.MAX_VALUE;
		if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
			bestTrainFitness=-Double.MAX_VALUE;
		for(int i=0;i<popSize;i++)
		{
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(allFitnesses[i]<bestTrainFitness)
					{
						bestTrainFitness=allFitnesses[i];
						bestIndex=i;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(allFitnesses[i]>bestTrainFitness)
					{
						bestTrainFitness=allFitnesses[i];
						bestIndex=i;
					}
					break;
			}
		}
	}
	/*This updates the bestTrainFitness, bestIndex and averageTrainFitness values. It does not carry out a complete
	re-evaluation of fitness values, as does setFitnessValues.*/
	final void setSummaryFitnessValues()
	{
		averageTrainFitness=0.0;
		bestTrainFitness=Double.MAX_VALUE;
		if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
			bestTrainFitness=-Double.MAX_VALUE;
		for(int i=0;i<popSize;i++)
		{
			double val=allFitnesses[i];
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(val<bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(val>bestTrainFitness)
					{
						bestTrainFitness=val;
						bestIndex=i;
					}
					break;
			}
			averageTrainFitness+=val;
		}
		averageTrainFitness/=popSize;
	}

	final void setValidationFitnessValues(double[][]inputs,double[]targetOutputs) throws InvalidInputSizeException
	{
		if(inputs==null || targetOutputs==null)
			return;
		averageValidationFitness=0.0;
		bestValidationFitness=Double.MAX_VALUE;
		if(fitnessFunction.equals(FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunction.equals(FitnessFunction.AUROC)||fitnessFunction.equals(FitnessFunction.SENSITIVITY)||fitnessFunction.equals(FitnessFunction.SPECIFICITY)||fitnessFunction.equals(FitnessFunction.PPV)||fitnessFunction.equals(FitnessFunction.NPV))
			bestValidationFitness=-Double.MAX_VALUE;
		for(int i=0;i<popSize;i++)
		{
			double val=0.0;
			switch(fitnessFunction)
			{
				case MSE:
					val=individuals[i].getMSE(inputs,targetOutputs);
					if(val<bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case RMSE:
					val=individuals[i].getRMSE(inputs,targetOutputs);
					if(val<bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case ABSOLUTE_ERROR:
					val=individuals[i].getAbsoluteError(inputs,targetOutputs);
					if(val<bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case AOREC:
					val=individuals[i].getAOREC(inputs,targetOutputs,aurocSampleSize, recFitnessFunction);
					if(val<bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case NUMBER_CORRECT_BINARY:
					val=individuals[i].getNumberCorrectBinary(inputs,targetOutputs);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case AUROC:
					val=individuals[i].getAUROC(inputs,targetOutputs,aurocSampleSize,false,false);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case SENSITIVITY:
					val=individuals[i].getSensitivity(inputs,targetOutputs);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case SPECIFICITY:
					val=individuals[i].getSpecificity(inputs,targetOutputs);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case PPV:
					val=individuals[i].getPositivePredictiveValue(inputs,targetOutputs);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case NPV:
					val=individuals[i].getNegativePredictiveValue(inputs,targetOutputs);
					if(val>bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				case COMBINED_CLASSIFICATION_ERROR:
					val=individuals[i].getCombinedClassificationError(inputs,targetOutputs,aurocSampleSize, false, false, false);
					if(val<bestValidationFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex = i;
					}
					break;
				/*case CLUSTERED_REGRESSION:
					val=individuals[i].getClusteredRegressionMSE(inputs,targetOutputs);
					if(val<bestTrainFitness)
					{
						bestValidationFitness=val;
						bestValidationIndex=i;
					}
					break;*/
			}
			averageValidationFitness+=val;
		}
		averageValidationFitness/=popSize;
	}

	/**This method returns an Object array. The first element in the array is a copy of
	this Population's Individuals, sorted into order from better to worse fitness. The
	second element is an array containing the corresponding training fitnesses. The third array
	contains the original indices of the Individuals*/
	public Object[] ShellSortIndividuals()
	{
		return ShellSortIndividuals(individuals, allFitnesses, fitnessFunction);
	}

	/**This method returns an Object array. The first element in the array is a copy of
	this Population's Individuals, sorted into order from better to worse fitness. The
	second element is an array containing the corresponding training fitnesses. The third array
	contains the original indices of the Individuals*/
	public Object[] ShellSortIndividuals(double[] objectiveData, boolean sortAscending)
	{
		return ShellSortIndividuals(individuals, objectiveData, sortAscending);
	}

	/**This method uses the Shell sort algorithm to order an array of Individuals according to theoir fitness.
	The Object array returned contains 3 arrays: the Individuals in the correct order, the corresponding fitnesses
	and the indices of the original array in order of fitness (best to worst). Note: individuals, popSize and allFitnesses
	are LOCAL variables, allowing the sorting of any Individual[].*/
	public static Object[] ShellSortIndividuals(Individual[] individuals, double[]allFitnesses, FitnessFunction fitnessFunction)
	{
		switch(fitnessFunction)
		{
			case MSE:
			case RMSE:
			case ABSOLUTE_ERROR:
			case AOREC:
			case COMBINED_CLASSIFICATION_ERROR:
			//case CLUSTERED_REGRESSION:
				return ShellSortIndividuals(individuals, allFitnesses,true);
			case NUMBER_CORRECT_BINARY:
			case AUROC:
			case SENSITIVITY:
			case SPECIFICITY:
			case PPV:
			case NPV:
				return ShellSortIndividuals(individuals, allFitnesses,false);
			default:
				System.err.println("Unknown FitnessFunction used in ShellSort.");
				return null;
		}
	}

	/**This method uses the Shell sort algorithm to order an array of Individuals according to theoir fitness.
	The Object array returned contains 3 arrays: the Individuals in the correct order, the corresponding fitnesses
	and the indices of the original array in order of fitness. Note: individuals, popSize and allFitnesses
	are LOCAL variables, allowing the sorting of any Individual[].
	@param sortAscending If true, the Individuals will be sorted from low to high fitness, if false
	they will be sorted from high to low. The value chosen for <i>sortAscending</i> may be different for different fitness functions:
	for some fitness functions low fitnesses are preferrred; for others high fitnesses are preferred.*/
	public static Object[] ShellSortIndividuals(Individual[] individuals, double[]allFitnesses, boolean sortAscending)
	{
		int popSize = individuals.length;
		Individual[] sortedInds=(Individual[])(individuals.clone());
		double[] fitnesses = allFitnesses.clone();
		int[] originalIndices = new int[popSize];
		for(int i=0;i<popSize;i++)
			originalIndices[i] = i;
		int h = 1;
		//find the largest h value possible
		while ((h * 3 + 1) < popSize)
		{
			h = 3 * h + 1;
		}

        while( h > 0 )
        {
            //for each set of elements (there are h sets)
            for (int i = h - 1; i < popSize; i++)
            {
                //pick the last element in the set
                double fitness = fitnesses[i];
                Individual ind = sortedInds[i];
                int index = originalIndices[i];
                int j = i;
                 /*compare ind to the one before it in the set
                   if they are out of order continue this loop, moving
                   elements "back" to make room for ind to be inserted.*/
                if(sortAscending)
                {
					for( j = i; (j >= h) && (fitnesses[j-h] > fitness); j -= h)
					{
						fitnesses[j] = fitnesses[j-h];
						sortedInds[j] = sortedInds[j-h];
						originalIndices[j] = originalIndices[j-h];
					}
				}
				else
				{
					for( j = i; (j >= h) && (fitnesses[j-h] < fitness); j -= h)
					{
						fitnesses[j] = fitnesses[j-h];
						sortedInds[j] = sortedInds[j-h];
						originalIndices[j] = originalIndices[j-h];
					}
				}
                //insert ind into the correct place
                fitnesses[j] = fitness;
                sortedInds[j] = ind;
                originalIndices[j] = index;
            }
            //all sets h-sorted, now decrease set size
            h = h / 3;
        }
        return new Object[]{sortedInds, fitnesses, originalIndices};
	}

	/**Returns the fitness of the best Individual on the training data.*/
	public double getBestTrainingFitness()
	{
		return bestTrainFitness;
	}

	/**Returns the fitness of the best Individual on the validation data.*/
	public double getBestValidationFitness()
	{
		return bestValidationFitness;
	}

	/**Returns the average fitness of all Individuals on the training data.*/
	public double getAverageTrainingFitness()
	{
		return averageTrainFitness;
	}

	/**Returns the average fitness of all Individuals on the validation data.*/
	public double getAverageValidationFitness()
	{
		return averageValidationFitness;
	}

	/**Returns the average fitness of all Individual on the data provided. If <i>recalculateValues</i> is false this
	method simply returns the average training fitness.*/
	public double getAverageFitness(double[][]inputs, double[]targetOutputs, boolean recalculateValues) throws InvalidInputSizeException
	{
		if(inputs==null || targetOutputs==null)
			throw new InvalidInputSizeException();
		double averageFitness=0.0;
		for(int i=0;i<popSize;i++)
		{
			double val = 0.0;
			if(recalculateValues)
				val = getFitness(individuals[i], inputs, targetOutputs);
			else
				val = allFitnesses[i];
			averageFitness+=val;
		}
		averageFitness/=popSize;
		return averageFitness;
	}

	final double getFitness(Individual ind, double[][]inputs, double[]targetOutputs) throws InvalidInputSizeException
	{
		double val = Double.NaN;
		switch(fitnessFunction)
		{
			case MSE:
				val=ind.getMSE(inputs,targetOutputs);
				break;
			case RMSE:
				val=ind.getRMSE(inputs,targetOutputs);
				break;
			case ABSOLUTE_ERROR:
				val=ind.getAbsoluteError(inputs,targetOutputs);
				break;
			case NUMBER_CORRECT_BINARY:
				val=ind.getNumberCorrectBinary(inputs,targetOutputs);
				break;
			case AUROC:
				val=ind.getAUROC(inputs,targetOutputs,aurocSampleSize,true,true);
				break;
			case AOREC:
				val=ind.getAOREC(inputs,targetOutputs,aurocSampleSize,recFitnessFunction);
				break;
			case SENSITIVITY:
				val = ind.getSensitivity(inputs,targetOutputs);
				break;
			case SPECIFICITY:
				val = ind.getSpecificity(inputs,targetOutputs);
				break;
			case PPV:
				val = ind.getPositivePredictiveValue(inputs,targetOutputs);
				break;
			case NPV:
				val = ind.getNegativePredictiveValue(inputs,targetOutputs);
				break;
			case COMBINED_CLASSIFICATION_ERROR:
				val=ind.getCombinedClassificationError(inputs,targetOutputs,aurocSampleSize, true, true, true);
				break;
			/*case CLUSTERED_REGRESSION:
				val=ind.getClusteredRegressionMSE(inputs,targetOutputs);
				break;*/
		}
		return val;
	}

	final double[] getPerformances(Individual ind)
	{
		if(performanceFunctions == null)
			return new double[0];
		int len = performanceFunctions.length;
		double[] performances = new double[len];
		for(int i=0;i<len;i++)
		{
			switch(performanceFunctions[i])
			{
				case PROGRAM_LENGTH:
					performances[i] = ind.getLength();
					break;
				case DIVERSITY_DISTANCE:
					performances[i] = getAverageSquaredDistance(ind);
					break;
				case FUNCTION_COMPLEXITY:
					performances[i] = getComplexity(ind);
					break;
				case VARIABLE_MAX:
				case VARIABLE_MIN:
					performances[i] = getNumberOfInputsUsed(ind);
					break;
			}
		}
		return performances;
	}

	/**Setting a low maximumIndividualLength is a way of controlling bloat. However, if a very low value
	is set, there may be too much bias towards small programs, leading to loss of diversity.
	The default value is 250 nodes.*/
	public void setMaximumIndividualLength(int maxLength)
	{
		if(maxLength>0)
			maximumLength=maxLength;
	}

	/**Returns the maximum length of an Individual created by crossover or mutation.
	Setting a low maximumIndividualLength is a way of controlling bloat. However, if a very low value
	is set, there may be too much bias towards small programs, leading to loss of diversity.
	The default value is 250 nodes.*/
	public int getMaximumIndividualLength()
	{
		return maximumLength;
	}

	/**Setting a low maximumIndividualDepth is a way of controlling bloat. However, if a very low value
	is set, there may be too much bias towards small programs, leading to loss of diversity.
	The default value is 17.*/
	public void setMaximumIndividualDepth(int maxDepth)
	{
		if(maxDepth>0)
			maximumDepth=maxDepth;
	}

	/**Retruns the maximum depth of an Individual created by crossover or mutation.
	Setting a low maximumIndividualDepth is a way of controlling bloat. However, if a very low value
	is set, there may be too much bias towards small programs, leading to loss of diversity.
	The default value is 17.*/
	public int getMaximumIndividualDepth()
	{
		return maximumDepth;
	}

	/**Returns the average length (number of nodes) of all Individuals in this Population.*/
	public double getAverageIndividualLength()
	{
		double averageLength=0.0;
		for(int i=0;i<popSize;i++)
			averageLength+=individuals[i].getLength();
		averageLength/=popSize;
		return averageLength;
	}

	/**getDiversity() counts the number of <i>different</i> Individuals
	   in the Population, subtracts 1 and divides by (population size - 1).
	   This results in a diversity of 1.0 if all Individuals are different
	   and a diversity of 0.0 if all Individuals are the same.*/
	public double getDiversity()
	{
		int len=popSize;
		if(len == 1)
			return 0.0;
		boolean[] repeated=new boolean[len];
		int count=len;
		for(int i=0;i<len-1;i++)
		{
			if(repeated[i])
				continue;
			for(int j=i+1;j<len;j++)
			{
				if(repeated[j])
					continue;
				if(individuals[i].equals(individuals[j]))
				{
					repeated[j]=true;
					count--;
				}
			}
		}
		//return (double)count/(double)len;
		return (double)(count-1)/(double)(len-1);
	}

	public boolean hasNoDiversity()
	{
		if(popSize==1)
			return true;
		for (int i=1;i<popSize;i++)
		{
			if(!individuals[i].equals(individuals[i-1]))
				return false;
		}
		return true;
	}

	/**Returns the best Individual on the training data. This method should only be called after evolving the Population.
	Otherwise, the best Individual will not have been calculated.*/
	public Individual getBestIndividual()
	{
		return individuals[bestIndex];
	}

	/**This method will return null if no evolution has occurred (because no validation has been performed) or if a validation
	set was not provided or a validation criterion was not used.*/
	public Individual getBestValidatedIndividual()
	{
		if(bestValidationIndex<0)
			return null;
		else
			return individuals[bestValidationIndex];
	}

	/**Sets the rank gradient to be used during RANK selection. A higher gradient will lead to a greater probability
	of selecting fitter Individuals. The default gradient is 1.0.*/
	public void setRankGradient(double gradient)
	{
		rankGradient = gradient;
	}

	/**Returns the rank gradient to be used during RANK selection. A higher gradient will lead to a greater probability
	of selecting fitter Individuals. The default gradient is 1.0.*/
	public double getRankGradient()
	{
		return rankGradient;
	}

	/**If set to true, brood recombination will be used during crossover*/
	public void setBroodUse(boolean useBrood)
	{
		this.useBrood=useBrood;
	}

	/**If set to true, brood recombination will be used during crossover*/
	public boolean getBroodUse()
	{
		return useBrood;
	}

	/**If set to true, the Population size will be maintained when using a multi-objective evaluation function (PerformanceFunctions).
	Individuals will be selected on the basis of dominance ranking, i.e. a count of the number of
	other Individuals by which a particular Individual is dominated. If false, all non-dominated Individuals will be kept in a Population,
	leading to fluctuations in Population size.*/
	public void setKeepDominated(boolean keepDominated)
	{
		keepDominatedIndividuals = keepDominated;
	}

	/**If set to true, the Population size will be maintained when using a multi-objective evaluation function (PerformanceFunctions).
	Individuals will be selected on the basis of dominance ranking, i.e. a count of the number of
	other Individuals by which a particular Individual is dominated. If false, all non-dominated Individuals will be kept in a Population,
	leading to fluctuations in Population size.*/
	public boolean getKeepDominated()
	{
		return keepDominatedIndividuals;
	}

	/**Sets parameters to be used during TRUNCATION selection.*/
	public void setTruncationParameters(boolean includeParentsInPool, int numberOfParents, int numberOfOffspring)
	{
		if(numberOfParents<=0||numberOfParents>popSize||numberOfOffspring<=0||(!includeParentsInPool && numberOfOffspring<=numberOfParents))
		{
			System.err.println("Invalid truncation parameters.");
			return;
		}
		parentsInTruncationPool = includeParentsInPool;
		mu = numberOfParents;
		lambda = numberOfOffspring;
	}

	/**This method should only be called after evolving the population.
	Otherwise, fitness values will not have been calculated.*/
	public double[] getFitnesses()
	{
		return allFitnesses.clone();
	}

	/**Returns the FitnessFunction used to evaluate Individuals within this Population.*/
	public FitnessFunction getFitnessFunction()
	{
		return fitnessFunction;
	}

	/**Sets the FitnessFunction used to evaluate Individuals within this Population.*/
	public void setFitnessFunction(FitnessFunction func)
	{
		fitnessFunction=func;
	}

	/**Sets the PerformanceFunctions used to evaluate Individuals within this Population. PerformanceFunctions
	allow the incorporation of additional objectives into the objective function. They include program length,
	program diversity and program complexity (number of higher order Functions). If <i>isMultiObjective</i> is true,
	a Pareto approach will be taken in selecting Individuals. If false, a weighted combination of the FitnessFunction
	and the PerformanceFunctions will be used as a single objective (see setSingleObjectiveRatio).*/
	public void setPerformanceFunctions(PerformanceFunction[] perfFuncs, boolean isMultiObjective)
	{
		performanceFunctions = perfFuncs;
		this.isMultiObjective = isMultiObjective;
		int len = perfFuncs.length;
		if(singleObjectiveRatio == null||singleObjectiveRatio.length!=len)
		{
			singleObjectiveRatio = new double[len];
			for(int i=0;i<len;i++)
				singleObjectiveRatio[i] = 1.0;
		}
	}

	/**Sets the weightings of performanceFunctions. All values are relative to a value of 1.0 for the FitnessFunction.
	This ratio will be ignored if <i>isMultiObjective</i> has been set to true by setPerformanceFunction.*/
	public boolean setSingleObjectiveRatio(double[] ratio)
	{
		int len = performanceFunctions.length;
		if(performanceFunctions != null && ratio.length != len)
		{
			System.err.println("Invalid ratio length for single objective performance function.");
			return false;
		}
		singleObjectiveRatio = new double[len];
		for(int i=0;i<len;i++)
			singleObjectiveRatio[i] = ratio[i];
		return true;
	}

	/**Returns an array of the PerformanceFunctions used in selecting Individuals. If no PerformanceFunctions have been
	applied this method returns null.*/
	public PerformanceFunction[] getPerformanceFunctions()
	{
		return performanceFunctions;
	}

	/**If true, and PerformanceFunctions have been applied, multi-objective (Pareto-style) selection will occur.
	In other words, selection based on dominance ranking.*/
	public boolean getMultiObjective()
	{
		return isMultiObjective;
	}

	/**Sets the Crossover object to be used by this Population.*/
	public void setCrossover(Crossover newCross)
	{
		crossover = newCross;
	}

	/**Returns the Crossover object to be used by this Population.*/
	public Crossover getCrossover()
	{
		return crossover;
	}

	/**Sets the Mutation object to be used by this Population.*/
	public void setMutation(Mutation newMut)
	{
		mut=newMut;
	}

	/**Returns the Mutation object to be used by this Population.*/
	public Mutation getMutation()
	{
		return mut;
	}

	final void setAllNumberOfInputsUsed()
	{
		noInputsUsed = new int[popSize];
		for(int i=0;i<popSize;i++)
			noInputsUsed[i] = getNumberOfInputsUsed(individuals[i]);
	}

	final int getNumberOfInputsUsed(Individual ind)
	{
		Vector<Component> nodes = ind.getNodes();
		Vector<Input> inputsFound = new Vector<Input>();
		int noOfInputsUsed = 0;
		for(int j=0;j<nodes.size();j++)
		{
			Component node = nodes.get(j);
			if(node instanceof Input && !inputsFound.contains(node))
			{
				inputsFound.add((Input)node);
				noOfInputsUsed++;
			}
		}
		return noOfInputsUsed;
	}

	//not currently used, 040906
	/*public double getAverageInputsUsed()
	{
		double total = 0.0;
		for(int i=0;i<popSize;i++)
			total+=noInputsUsed[i];
		total /= popSize;
		return total;
	}*/

	final void setAllComplexities()
	{
		functionComplexities = new int[popSize];
		for(int i=0;i<popSize;i++)
			functionComplexities[i] = getComplexity(individuals[i]);
	}

	final int getComplexity(Individual ind)
	{
		Vector<Component> nodes = ind.getNodes();
		int complexity = 0;
		for(int j=0;j<nodes.size();j++)
		{
			Component node = nodes.get(j);
			if(node instanceof Function)
				complexity += ((Function)node).getComplexity();
		}
		return complexity;
	}

	/**Returns the average function complexity of all Individuals within the population.
	Individuals with a high function complexity are those which contain more 'complex' functions such as exp or pow*/
	public double getAverageComplexity()
	{
		double total = 0.0;
		for(int i=0;i<popSize;i++)
			total+=functionComplexities[i];
		total /= popSize;
		return total;
	}

	final void setAllDistances()
	{
		allSquaredDistances = new double[popSize][popSize];
		for(int i=0;i<popSize;i++)
			for(int j=i+1;j<popSize;j++)
				allSquaredDistances[i][j] = Math.pow(getDistance(i,j),2.0);
	}

	/**Returns a 2-D array containing the 'distances' between pairs of Individuals, in the 'upper right' sector
	(above the leading diagonal).*/
	public final double[][] getAllSquaredDistances(boolean recalculate)
	{
		if(allSquaredDistances == null || recalculate)
			setAllDistances();
		return allSquaredDistances.clone();
	}

	/**Returns the distances between selected Individuals in the Population, denoted by their indices,
	and all other members of the Population. The returned array has dimensions indices.length() - by - popSize.*/
	public final double[][] getSquaredDistances(int[] indices, boolean recalculate)
	{
		int len = indices.length;
		double[][] distances = new double[len][popSize];
		if(allSquaredDistances == null || recalculate)
		{
			for(int i=0;i<len;i++)
				for(int j=0;j<popSize;j++)
					distances[i][j] = Math.pow(getDistance(indices[i],j),2.0);
		}
		else
		{
			for(int i=0;i<len;i++)
				for(int j=0;j<popSize;j++)
				{
					if(j>indices[i])
						distances[i][j] = allSquaredDistances[indices[i]][j];
					else
						distances[i][j] = allSquaredDistances[j][indices[i]];
				}
		}
		return distances;
	}

	/**Returns the average squared distance between the Individual denoted by index and all other members of the population.*/
	public double getAverageSquaredDistance(int index)
	{
		if(allSquaredDistances == null)
			setAllDistances();
		double total = 0.0;
		for(int i=0;i<popSize;i++)
		{
			if(i<index)
				total += allSquaredDistances[i][index];
			else if(i>index)
				total += allSquaredDistances[index][i];
		}
		total /= (popSize-1);
		return total;
	}

	/**Obtains the average distance that would occur <i>if</i> the new Individual were to replace the Individual currently at index <i>index</i>.*/
	public double getAverageSquaredDistance(int index, Individual newInd)
	{
		if(allSquaredDistances == null)
			setAllDistances();
		double total = 0.0;
		for(int i=0;i<popSize;i++)
		{
			if(i<index)
				total += allSquaredDistances[i][index];
			else if(i>index)
				total += allSquaredDistances[index][i];
		}
		total += Math.pow(getDistance(newInd,index), 2.0);
		return total / popSize;
	}

	/**Obtains the average distance between <i>ind</i> and the Population's current members*/
	public double getAverageSquaredDistance(Individual ind)
	{
		double total = 0.0;
		for(int i=0;i<popSize;i++)
			total += Math.pow(getDistance(ind, i),2.0);
		return total / popSize;
	}

	/**Fitness sharing aims to prevent Individuals within a small niche from dominating the Population.
	The fitness of an Individual is deivided by a factor, which is determined by the number of other
	Individuals within a niche near to that Individual.
	@param fitnessShare The FitnessSahre object that determines how fitness sharing will operate.
	@param shareType determines whether fitness sharing will occur over the genotype (tree structural similarity) or the phenotype (fitness)*/
	public void setFitnessShare(FitnessShare fitnessShare, FitnessShare.ShareType shareType)
	{
		this.fitnessShare = fitnessShare;
		this.shareType = shareType;
	}

	/**Returns the fitness share object. Returns null if fitness sharing is not implemented.*/
	public FitnessShare getFitnessShare()
	{
		return fitnessShare;
	}

	/**Returns the fitness type. Returns null if fitness sharing is not implemented.*/
	public FitnessShare.ShareType getShareType()
	{
		return shareType;
	}

	/**Sets the share weightings. If sharing and multi-objective performance functions are both implemented, share weightings are used to
	determine the share factors. Performance functions with a high share weighting will be shared to a greater extent,
	leading to greater diversity in this functions.*/
	public void setShareWeightings(double[] weightings, boolean isMultiObjectivelyShared)
	{
		shareWeightings = weightings;
		//isMultiShared = isMultiObjectivelyShared;
	}

	/**Returns the share weightings. If sharing and multi-objective performance functions are both implemented, share weightings are used to
	determine the share factors. Performance functions with a high share weighting will be shared to a greater extent,
	leading to greater diversity in this functions.*/
	public double[] getShareWeightings()
	{
		return shareWeightings;
	}

	//note: this is NOT the same as de Jong's measure of 'distance'
	final double getDistance(Individual ind, int index)
	{
		return getDistance(ind, individuals[index]);
	}

	final double getDistance(Individual ind1, Individual ind2)
	{
		return Utils.getDistance(ind1,ind2);
	}

	//note: this is NOT the same as de Jong's measure of 'distance'
	final double getDistance(int index1, int index2)
	{
		return getDistance(individuals[index1], index2);
	}

	/**Add individual should be called when adding an Individual to a Population with non-null performanceFunctions
	and with keepDominatedIndividuals set to false. It will not add an individual that is dominated by or is a
	duplicate of an existing member of the population. It removes Individuals that are dominated by the new Individual*/
	private boolean addIndividual(Individual newInd , double newFitness)
	{
		double[] performances = getPerformances(newInd);
		boolean[][] dominanceStatus = getDominanceStatus(newFitness, performances);
		for(int i=0;i<dominanceStatus.length;i++)
		{
			if(dominanceStatus[i][0] || newInd.equals(individuals[i]))
				return false;//new Individual is dominated or is a duplicate, therefore don't add
		}

		Individual[] newInds = new Individual[popSize+1];
		double[] newFitnesses = new double[popSize+1];
		for(int i=0;i<popSize;i++)
		{
			newInds[i] = individuals[i];
			newFitnesses[i] = allFitnesses[i];
		}
		newInds[popSize] = newInd;
		individuals = newInds;
		newFitnesses[popSize] = newFitness;
		allFitnesses = newFitnesses;
		popSize++;

		boolean distanceSet = false;
		for(int i=0;i<performanceFunctions.length;i++)
		{
			switch(performanceFunctions[i])
			{
				case DIVERSITY_DISTANCE:
					double[][] tempDistances = new double[popSize][popSize];
					for(int j=0;j<popSize-1;j++)
					{
						for(int k=j+1;k<popSize-1;k++)
							tempDistances[j][k] = allSquaredDistances[j][k];
						tempDistances[j][popSize-1] = Math.pow(getDistance(newInd,j),2.0);
					}
					allSquaredDistances = tempDistances;
					distanceSet = true;
					break;
				case FUNCTION_COMPLEXITY:
					int[] newComplexities = new int[popSize];
					for(int j=0;j<popSize-1;j++)
						newComplexities[j] = functionComplexities[j];
					//newComplexities[popSize-1] = getComplexity(newInd);
					newComplexities[popSize-1] = (int)performances[i];
					functionComplexities = newComplexities;
					break;
				case VARIABLE_MAX:
				case VARIABLE_MIN:
					int[] newUsed = new int[popSize];
					for(int j=0;j<popSize-1;j++)
						newUsed[j] = noInputsUsed[j];
					//newUsed[popSize-1] = getNumberOfInputsUsed(newInd);
					newUsed[popSize-1] = (int)performances[i];
					noInputsUsed = newUsed;
					break;
			}
		}
		if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
		{
			double[][] tempDistances = new double[popSize][popSize];
			for(int j=0;j<popSize-1;j++)
			{
				for(int k=j+1;k<popSize-1;k++)
					tempDistances[j][k] = allSquaredDistances[j][k];
				tempDistances[j][popSize-1] = Math.pow(getDistance(newInd,j),2.0);
			}
			allSquaredDistances = tempDistances;
		}
		if(distanceSet)
			setDominatedMatrix();//distances must be recalculated
		else
			expandDominatedMatrix();
		switch(fitnessFunction)
		{
			case MSE:
			case RMSE:
			case ABSOLUTE_ERROR:
			case AOREC:
			case COMBINED_CLASSIFICATION_ERROR:
			//case CLUSTERED_REGRESSION:
				if(newFitness<bestTrainFitness)
				{
					bestTrainFitness = newFitness;
					bestIndex = popSize-1;
				}
				break;
			case NUMBER_CORRECT_BINARY:
			case AUROC:
			case SENSITIVITY:
			case SPECIFICITY:
			case PPV:
			case NPV:
				if(newFitness>bestTrainFitness)
				{
					bestTrainFitness = newFitness;
					bestIndex = popSize-1;
				}
				break;
		}
		removeDominatedIndividuals();
		return true;
	}

	protected void removeDominatedIndividuals()
	{
		int dominatedIndex = -1;
		int dominatedBy = 0;
		for(int i=0;i<dominatedCount.length;i++)
		{
			if(dominatedCount[i]>dominatedBy)
			{
				dominatedBy = dominatedCount[i];
				dominatedIndex = i;
			}
		}
		while(dominatedIndex!=-1)
		{
			popSize--;
			Individual[] tempIndividuals = new Individual[popSize];
			double[] tempFitnesses = new double[popSize];
			boolean resetBest = false;
			for(int i=0;i<popSize+1;i++)
			{
				if(i>dominatedIndex)
				{
					if(i==bestIndex)
						bestIndex--;
					tempIndividuals[i-1] = individuals[i];
					tempFitnesses[i-1] = allFitnesses[i];
				}
				else if(i<dominatedIndex)
				{
					tempIndividuals[i] = individuals[i];
					tempFitnesses[i] = allFitnesses[i];
				}
				else if(i==bestIndex)
					resetBest = true;//Reset bestIndex. Should only happen if 2 Individuals have the same fitness
			}
			allFitnesses = new double[popSize];
			individuals = new Individual[popSize];
			for(int i=0;i<popSize;i++)
			{
				individuals[i] = tempIndividuals[i];
				allFitnesses[i] = tempFitnesses[i];
			}

			if(resetBest)
				setBestFitness();

			boolean distanceSet = false;
			for(int i=0;i<performanceFunctions.length;i++)
			{
				switch(performanceFunctions[i])
				{
					case DIVERSITY_DISTANCE:
						//setAllDistances();//all distances should be recalculated with the new population - not necessary
						double[][] tempDistances = new double[popSize][popSize];
						for(int j=0;j<popSize+1;j++)
						{
							int index1 = j;
							if(j==dominatedIndex)
								continue;
							else if(j>dominatedIndex)
								index1--;
							for(int k=j+1;k<popSize+1;k++)
							{
								int index2 = k;
								if(k==dominatedIndex)
									continue;
								else if(index2>dominatedIndex)
									index2--;
								tempDistances[index1][index2] = allSquaredDistances[j][k];
							}
						}
						allSquaredDistances = tempDistances;
						distanceSet = true;
						break;
					case FUNCTION_COMPLEXITY:
						int[] newComplexities = new int[popSize];
						for(int j=0;j<popSize+1;j++)
						{
							if(j>dominatedIndex)
								newComplexities[j-1] = functionComplexities[j];
							else if(j<dominatedIndex)
								newComplexities[j] = functionComplexities[j];
						}
						functionComplexities = newComplexities;
						break;
					case VARIABLE_MAX:
					case VARIABLE_MIN:
						int[] newUsed = new int[popSize];
						for(int j=0;j<popSize+1;j++)
						{
							if(j>dominatedIndex)
								newUsed[j-1] = noInputsUsed[j];
							else if(j<dominatedIndex)
								newUsed[j] = noInputsUsed[j];
						}
						noInputsUsed = newUsed;
						break;
				}
			}
			if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
			{
				double[][] tempDistances = new double[popSize][popSize];
				for(int j=0;j<popSize+1;j++)
				{
					int index1 = j;
					if(j==dominatedIndex)
						continue;
					else if(j>dominatedIndex)
						index1--;
					for(int k=j+1;k<popSize+1;k++)
					{
						int index2 = k;
						if(k==dominatedIndex)
							continue;
						else if(index2>dominatedIndex)
							index2--;
						tempDistances[index1][index2] = allSquaredDistances[j][k];
					}
				}
				allSquaredDistances = tempDistances;
			}
			if(distanceSet)//distances must be re-calculated
				setDominatedMatrix();
			else
				shrinkDominatedMatrix(dominatedIndex);

			dominatedIndex = -1;
			dominatedBy = 0;
			for(int i=0;i<dominatedCount.length;i++)
			{
				if(dominatedCount[i]>dominatedBy)
				{
					dominatedBy = dominatedCount[i];
					dominatedIndex = i;
				}
			}
		}
	}

	protected void removeDuplicatedIndividuals()
	{
		Vector<Integer> duplicateIndices = new Vector<Integer>();
		Vector<Individual> newPop = new Vector<Individual>();
		for(int i=0;i<popSize;i++)
		{
			boolean duplicate = false;
			for(int j=0;j<newPop.size();j++)
			{
				if(newPop.get(j).equals(individuals[i]))
				{
					duplicateIndices.add(i);
					duplicate = true;
					break;
				}
			}
			if(!duplicate)
				newPop.add(individuals[i]);
		}
		if(duplicateIndices.size() == 0)
		{
			for(int i=0;i<popSize;i++)
				for(int j=i+1;j<popSize;j++)
				{
					if(individuals[i].equals(individuals[j]))
					{
						System.err.println("ERROR1");
						System.exit(1);
					}
				}
			return;
		}
		individuals = new Individual[0];
		individuals = (Individual[])(newPop.toArray(individuals));
		int newSize = individuals.length;

		double[] newFitnesses = new double[newSize];
		int[] newComplexities = new int[newSize];
		int[] newInputs = new int[newSize];
		double[][] newDistances = new double[newSize][newSize];
		int found = 0;

		boolean distanceSet = false;
		for(int i=0;i<popSize;i++)
		{
			if(!duplicateIndices.contains(i))
			{
				newFitnesses[found] = allFitnesses[i];
				if(performanceFunctions != null)
				{
					for(int j=0;j<performanceFunctions.length;j++)
					{
						switch(performanceFunctions[j])
						{
							case DIVERSITY_DISTANCE:
								int found2 = 0;
								for(int k=i+1;k<popSize;k++)
								{
									if(!duplicateIndices.contains(k))
									{
										newDistances[found][found2] = allSquaredDistances[i][k];
										found2++;
									}
								}
								distanceSet = true;
								break;
							case FUNCTION_COMPLEXITY:
								newComplexities[found] = functionComplexities[i];
								break;
							case VARIABLE_MAX:
							case VARIABLE_MIN:
								newInputs[found] = noInputsUsed[i];
								break;
						}
					}
				}
				if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
				{
					int found2 = 0;
					for(int k=i+1;k<popSize;k++)
					{
						if(!duplicateIndices.contains(k))
						{
							newDistances[found][found2] = allSquaredDistances[i][k];
							found2++;
						}
					}
				}
				found++;
			}
		}
		allFitnesses = newFitnesses;

		if(performanceFunctions != null)
		{
			for(int j=0;j<performanceFunctions.length;j++)
			{
				switch(performanceFunctions[j])
				{
					case DIVERSITY_DISTANCE:
						allSquaredDistances = newDistances;
						break;
					case FUNCTION_COMPLEXITY:
						functionComplexities = newComplexities;
						break;
					case VARIABLE_MAX:
					case VARIABLE_MIN:
						noInputsUsed = newInputs;
						break;
				}
			}
		}
		if(fitnessShare != null && shareType.equals(FitnessShare.ShareType.GENE_DISTANCE) && !distanceSet)
			allSquaredDistances = newDistances;

		popSize = newSize;

		for(int i=0;i<popSize;i++)
			for(int j=i+1;j<popSize;j++)
			{
				if(individuals[i].equals(individuals[j]))
				{
					System.err.println("ERROR2");
					System.exit(1);
				}
			}
	}

	final void setDominatedMatrix()
	{
		dominatedBy = new boolean[popSize][popSize];
		dominatedCount = new int[popSize];
		double[] squaredDistances = new double[popSize];
		for(int i=0;i<popSize;i++)
			squaredDistances[i] = getAverageSquaredDistance(i);
		for(int i=0;i<popSize;i++)
		{
			for(int j=i+1;j<popSize;j++)
			{
				boolean iBetterThanj = false;
				boolean jBetterThani = false;
				switch(fitnessFunction)
				{
					case MSE:
					case RMSE:
					case ABSOLUTE_ERROR:
					case AOREC:
					case COMBINED_CLASSIFICATION_ERROR:
					//case CLUSTERED_REGRESSION:
						if(allFitnesses[i]<allFitnesses[j])
							iBetterThanj = true;
						else if(allFitnesses[i]>allFitnesses[j])
							jBetterThani = true;
						break;
					case NUMBER_CORRECT_BINARY:
					case AUROC:
					case SENSITIVITY:
					case SPECIFICITY:
					case PPV:
					case NPV:
						if(allFitnesses[i]>allFitnesses[j])
							iBetterThanj = true;
						else if(allFitnesses[i]<allFitnesses[j])
							jBetterThani = true;
						break;
				}
				if(performanceFunctions!=null)
				{
					for(int k=0;k<performanceFunctions.length;k++)
					{
						switch(performanceFunctions[k])
						{
							case PROGRAM_LENGTH:
								int leni = individuals[i].getLength();
								int lenj = individuals[j].getLength();
								if(leni < lenj)
									iBetterThanj = true;
								else if(leni > lenj)
									jBetterThani = true;
								break;
							case DIVERSITY_DISTANCE:
								double squaredDistancei = squaredDistances[i];
								double squaredDistancej = squaredDistances[j];
								if(squaredDistancei > squaredDistancej)
									iBetterThanj = true;
								else if(squaredDistancei < squaredDistancej)
									jBetterThani = true;
								break;
							case FUNCTION_COMPLEXITY:
								if(functionComplexities[i] < functionComplexities[j])
									iBetterThanj = true;
								else if(functionComplexities[i] > functionComplexities[j])
									jBetterThani = true;
								break;
							case VARIABLE_MAX:
								if(noInputsUsed[i] > noInputsUsed[j])
									iBetterThanj = true;
								else if(noInputsUsed[i] < noInputsUsed[j])
									jBetterThani = true;
								break;
							case VARIABLE_MIN:
								if(noInputsUsed[i] < noInputsUsed[j])
									iBetterThanj = true;
								else if(noInputsUsed[i] > noInputsUsed[j])
									jBetterThani = true;
								break;
						}
					}
				}
				if(iBetterThanj && ! jBetterThani)
				{
					dominatedBy[j][i] =true;
					dominatedCount[j]++;
				}
				else if(jBetterThani && ! iBetterThanj)
				{
					dominatedBy[i][j] =true;
					dominatedCount[i]++;
				}
			}
		}
	}

	final void expandDominatedMatrix()
	{
		if(dominatedBy.length!=popSize-1)
		{
			System.err.println("ERROR. expandDominatedMatrix should only be called when the population size has increased by 1.");
			return;
		}
		boolean[][] tempDominatedBy = new boolean[popSize][popSize];
		int[] tempDominatedCount = new int[popSize];
		for(int i=0;i<popSize-1;i++)
		{
			tempDominatedCount[i] = dominatedCount[i];
			for(int j=0;j<popSize-1;j++)
				tempDominatedBy[i][j] = dominatedBy[i][j];
		}

		int j = popSize-1;
		for(int i=0;i<popSize-1;i++)
		{
			boolean iBetterThanj = false;
			boolean jBetterThani = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(allFitnesses[i]<allFitnesses[j])
						iBetterThanj = true;
					else if(allFitnesses[i]>allFitnesses[j])
						jBetterThani = true;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(allFitnesses[i]>allFitnesses[j])
						iBetterThanj = true;
					else if(allFitnesses[i]<allFitnesses[j])
						jBetterThani = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int k=0;k<performanceFunctions.length;k++)
				{
					switch(performanceFunctions[k])
					{
						case PROGRAM_LENGTH:
							int leni = individuals[i].getLength();
							int lenj = individuals[j].getLength();
							if(leni < lenj)
								iBetterThanj = true;
							else if(leni > lenj)
								jBetterThani = true;
							break;
						case DIVERSITY_DISTANCE:
							double squaredDistancei = getAverageSquaredDistance(i);
							double squaredDistancej = getAverageSquaredDistance(j);
							if(squaredDistancei > squaredDistancej)
								iBetterThanj = true;
							else if(squaredDistancei < squaredDistancej)
								jBetterThani = true;
							break;
						case FUNCTION_COMPLEXITY:
							if(functionComplexities[i] < functionComplexities[j])
								iBetterThanj = true;
							else if(functionComplexities[i] > functionComplexities[j])
								jBetterThani = true;
							break;
						case VARIABLE_MAX:
							if(noInputsUsed[i] > noInputsUsed[j])
								iBetterThanj = true;
							else if(noInputsUsed[i] < noInputsUsed[j])
								jBetterThani = true;
							break;
						case VARIABLE_MIN:
							if(noInputsUsed[i] < noInputsUsed[j])
								iBetterThanj = true;
							else if(noInputsUsed[i] > noInputsUsed[j])
								jBetterThani = true;
							break;
					}
				}
			}
			if(iBetterThanj && ! jBetterThani)
			{
				tempDominatedBy[j][i] =true;
				tempDominatedCount[j]++;
			}
			else if(jBetterThani && ! iBetterThanj)
			{
				tempDominatedBy[i][j] =true;
				tempDominatedCount[i]++;
			}
		}
		dominatedBy = tempDominatedBy;
		dominatedCount = tempDominatedCount;
	}

	final void shrinkDominatedMatrix(int removalIndex)
	{
		if(dominatedBy.length != popSize+1)
		{
			System.err.println("ERROR. shrinkDominatedMatrix should only be called when the population size has decreased by 1.");
			return;
		}
		boolean[][] tempDominatedBy = new boolean[popSize][popSize];
		int[] tempDominatedCount = new int[popSize];
		int count1 = 0;
		for(int i=0;i<popSize+1;i++)
		{
			if(i == removalIndex)
				continue;
			else
				count1++;
			tempDominatedCount[count1-1] = dominatedCount[i];

			if(dominatedBy[i][removalIndex])
				tempDominatedCount[count1-1]--;

			int count2 = 0;
			for(int j=0;j<popSize+1;j++)
			{
				if(j == removalIndex)
					continue;
				else
					count2++;
				tempDominatedBy[count1-1][count2-1] = dominatedBy[i][j];
			}
		}
		dominatedBy = tempDominatedBy;
		dominatedCount = tempDominatedCount;
	}

	//copies one row and column of the dominatedBy matrix into another row, adjusting dominatedCount appropriately.
	//not used 270906
	/*final void repeatDominatedMatrix(int from, int to)
	{
		if(dominatedBy[from][to])
		{
			dominatedBy[from][to] = false;
			dominatedCount[from]--;
		}
		dominatedCount[to] = dominatedCount[from];
		for(int i=0;i<popSize;i++)
		{
			if(dominatedBy[i][from] && !dominatedBy[i][to])
				dominatedCount[i]++;
			else if(!dominatedBy[i][from] && dominatedBy[i][to])
				dominatedCount[i]--;
			dominatedBy[i][to] = dominatedBy[i][from];
			dominatedBy[to][i] = dominatedBy[from][i];
		}
	}*/

	/*this uses the same algorithm as getDominanceStatus. However, this method is used when replacing an individual, i.e.
	keepDominatedIndividuals is true, whereas getDominanceStatus is used when adding an individual, i.e. keepDominatedIndividuals is false.*/
	private void updateDominatedMatrix(Individual newInd, int index, double fitness)
	{
		dominatedCount[index] = 0;
		for(int i=0;i<popSize;i++)
		{
			if(i==index)
				continue;
			boolean iBetterThanNew = false;
			boolean newBetterThani = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(fitness<allFitnesses[i])
						newBetterThani = true;
					else if(fitness>allFitnesses[i])
						iBetterThanNew = true;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(fitness>allFitnesses[i])
						newBetterThani = true;
					else if(fitness<allFitnesses[i])
						iBetterThanNew = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int k=0;k<performanceFunctions.length;k++)
				{
					switch(performanceFunctions[k])
					{
						case PROGRAM_LENGTH:
							int lenNew = newInd.getLength();
							int leni = individuals[i].getLength();
							if(leni < lenNew)
								iBetterThanNew = true;
							else if(leni > lenNew)
								newBetterThani = true;
								break;
						case DIVERSITY_DISTANCE:
							double squaredDistancei = getAverageSquaredDistance(i);
							double squaredDistanceNew = getAverageSquaredDistance(newInd);
							if(squaredDistancei > squaredDistanceNew)
								iBetterThanNew = true;
							else if(squaredDistancei < squaredDistanceNew)
								newBetterThani = true;
							break;
						case FUNCTION_COMPLEXITY:
							int newComplexity = getComplexity(newInd);
							if(functionComplexities[i] < newComplexity)
								iBetterThanNew = true;
							else if(functionComplexities[i] > newComplexity)
								newBetterThani = true;
							break;
						case VARIABLE_MAX:
							int noInputs = getNumberOfInputsUsed(newInd);
							if(noInputsUsed[i] > noInputs)
								iBetterThanNew = true;
							else if(noInputsUsed[i] < noInputs)
								newBetterThani = true;
							break;
						case VARIABLE_MIN:
							int noIns = getNumberOfInputsUsed(newInd);
							if(noInputsUsed[i] < noIns)
								iBetterThanNew = true;
							else if(noInputsUsed[i] > noIns)
								newBetterThani = true;
							break;
					}
				}
			}
			if(iBetterThanNew && ! newBetterThani)
			{
				if(dominatedBy[i][index])
				{
					dominatedBy[i][index] = false;
					dominatedCount[i]--;
				}
				dominatedBy[index][i] = true;
				dominatedCount[index]++;
			}
			else if(newBetterThani && ! iBetterThanNew)
			{
				if(!dominatedBy[i][index])
				{
					dominatedBy[i][index] = true;
					dominatedCount[i]++;
				}
				dominatedBy[index][i] = false;
			}
			else
			{
				if(dominatedBy[i][index])
				{
					dominatedBy[i][index] = false;
					dominatedCount[i]--;
				}
				dominatedBy[index][i] = false;
			}
		}
	}

	/*this is quicker than the method below, but is less precise in comparing diversity_distances, using
	the current values rather than the values that would occur if the new Individual were added.*/
	private boolean[][] getDominanceStatus(double fitness, double[] performances)
	{
		boolean[][] status = new boolean[popSize][2];
		for(int i=0;i<popSize;i++)
		{
			boolean iBetterThanNew = false;
			boolean newBetterThani = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(fitness<allFitnesses[i])
						newBetterThani = true;
					else if(fitness>allFitnesses[i])
						iBetterThanNew = true;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(fitness>allFitnesses[i])
						newBetterThani = true;
					else if(fitness<allFitnesses[i])
						iBetterThanNew = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int k=0;k<performanceFunctions.length;k++)
				{
					switch(performanceFunctions[k])
					{
						case PROGRAM_LENGTH:
							int lenNew = (int)performances[k];
							int leni = individuals[i].getLength();
							if(leni < lenNew)
								iBetterThanNew = true;
							else if(leni > lenNew)
								newBetterThani = true;
							break;
						case DIVERSITY_DISTANCE:
							double squaredDistancei = getAverageSquaredDistance(i);//not as accurate as the method below, which allows for the presence of the new Individual
							//double squaredDistancei = getAverageSquaredDistance(i,newInd);
							double squaredDistanceNew = performances[k];
							if(squaredDistancei > squaredDistanceNew)
								iBetterThanNew = true;
							else if(squaredDistancei < squaredDistanceNew)
								newBetterThani = true;
							break;
						case FUNCTION_COMPLEXITY:
							int newComplexity = (int)performances[k];
							if(functionComplexities[i] < newComplexity)
								iBetterThanNew = true;
							else if(functionComplexities[i] > newComplexity)
								newBetterThani = true;
							break;
						case VARIABLE_MAX:
							int noInputs = (int)performances[k];
							if(noInputsUsed[i] > noInputs)
								iBetterThanNew = true;
							else if(noInputsUsed[i] < noInputs)
								newBetterThani = true;
							break;
						case VARIABLE_MIN:
							int noIns = (int)performances[k];
							if(noInputsUsed[i] < noIns)
								iBetterThanNew = true;
							else if(noInputsUsed[i] > noIns)
								newBetterThani = true;
							break;
					}
				}
			}
			if(iBetterThanNew && ! newBetterThani)
				status[i][0] = true;
			else if(newBetterThani && ! iBetterThanNew)
				status[i][1] = true;
		}
		return status;
	}

	//Quicker to use known values for fitness and performance as in method above, although this involves approximation in the diversity_distance measure
	private boolean[][] getDominanceStatus(Individual newInd, double fitness)
	{
		boolean[][] status = new boolean[popSize][2];
		for(int i=0;i<popSize;i++)
		{
			boolean iBetterThanNew = false;
			boolean newBetterThani = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(fitness<allFitnesses[i])
						newBetterThani = true;
					else if(fitness>allFitnesses[i])
						iBetterThanNew = true;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(fitness>allFitnesses[i])
						newBetterThani = true;
					else if(fitness<allFitnesses[i])
						iBetterThanNew = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int k=0;k<performanceFunctions.length;k++)
				{
					switch(performanceFunctions[k])
					{
						case PROGRAM_LENGTH:
							int lenNew = newInd.getLength();
							int leni = individuals[i].getLength();
							if(leni < lenNew)
								iBetterThanNew = true;
							else if(leni > lenNew)
								newBetterThani = true;
							break;
						case DIVERSITY_DISTANCE:
							double squaredDistancei = getAverageSquaredDistance(i,newInd);
							double squaredDistanceNew = getAverageSquaredDistance(newInd);
							if(squaredDistancei > squaredDistanceNew)
								iBetterThanNew = true;
							else if(squaredDistancei < squaredDistanceNew)
								newBetterThani = true;
							break;
						case FUNCTION_COMPLEXITY:
							int newComplexity = getComplexity(newInd);
							if(functionComplexities[i] < newComplexity)
								iBetterThanNew = true;
							else if(functionComplexities[i] > newComplexity)
								newBetterThani = true;
							break;
						case VARIABLE_MAX:
							int noInputs = getNumberOfInputsUsed(newInd);
							if(noInputsUsed[i] > noInputs)
								iBetterThanNew = true;
							else if(noInputsUsed[i] < noInputs)
								newBetterThani = true;
							break;
						case VARIABLE_MIN:
							int noIns = getNumberOfInputsUsed(newInd);
							if(noInputsUsed[i] < noIns)
								iBetterThanNew = true;
							else if(noInputsUsed[i] > noIns)
								newBetterThani = true;
							break;
					}
				}
			}
			if(iBetterThanNew && ! newBetterThani)
				status[i][0] = true;
			else if(newBetterThani && ! iBetterThanNew)
				status[i][1] = true;
		}
		return status;
	}

	final int getDominatedCount(double fitness, double[] performances, double[][] inputs, double[] targetOutputs) throws InvalidInputSizeException
	{
		int count = 0;
		outer:
		for(int i=0;i<popSize;i++)
		{
			boolean worse = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
			}
			for(int j=0;j<performances.length;j++)
			{
				switch(performanceFunctions[j])
				{
					case PROGRAM_LENGTH:
						int len = (int)performances[j];
						if(len < individuals[i].getLength())
							continue outer;
						else if(len > individuals[i].getLength())
							worse = true;
						break;
					case DIVERSITY_DISTANCE:
						double squaredDistance = performances[j];
						if(squaredDistance > getAverageSquaredDistance(i))
							continue outer;
						else if(squaredDistance < getAverageSquaredDistance(i))
							worse = true;
						break;
					case FUNCTION_COMPLEXITY:
						int complexity = (int)performances[j];
						if(complexity < functionComplexities[i])
							continue outer;
						else if(complexity > functionComplexities[i])
							worse = true;
						break;
					case VARIABLE_MAX:
						int noInputs = (int)performances[j];
						if(noInputs > noInputsUsed[i])
							continue outer;
						else if(noInputs < noInputsUsed[i])
							worse = true;
						break;
					case VARIABLE_MIN:
						int numInputs = (int)performances[j];
						if(numInputs < noInputsUsed[i])
							continue outer;
						else if(numInputs > noInputsUsed[i])
							worse = true;
						break;
				}
			}
			//The test Individual has been 'beaten' in at least 1 dimension and is at least equalled in every dimension (no 'continue'), therefore it is dominated
			if(worse)
				count++;
		}
		return count;
	}

	//no longer used. More efficient to use known values for fitness and performance as in method above
	/*final int getDominatedCount(Individual testInd, double[][] inputs, double[] targetOutputs) throws InvalidInputSizeException
	{
		int count = 0;
		double fitness;
		outer:
		for(int i=0;i<popSize;i++)
		{
			boolean worse = false;
			switch(fitnessFunction)
			{
				case MSE:
					fitness = testInd.getMSE(inputs, targetOutputs);
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
				case RMSE:
					fitness = testInd.getRMSE(inputs, targetOutputs);
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
				case ABSOLUTE_ERROR:
					fitness = testInd.getAbsoluteError(inputs, targetOutputs);
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
				case AOREC:
					fitness = testInd.getAOREC(inputs, targetOutputs,aurocSampleSize,recFitnessFunction);
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
				case NUMBER_CORRECT_BINARY:
					fitness = testInd.getNumberCorrectBinary(inputs, targetOutputs);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case AUROC:
					fitness = testInd.getAUROC(inputs, targetOutputs,aurocSampleSize,true,true);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case SENSITIVITY:
					fitness = testInd.getSensitivity(inputs, targetOutputs);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case SPECIFICITY:
					fitness = testInd.getSpecificity(inputs, targetOutputs);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case PPV:
					fitness = testInd.getPositivePredictiveValue(inputs, targetOutputs);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case NPV:
					fitness = testInd.getNegativePredictiveValue(inputs, targetOutputs);
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
				case COMBINED_CLASSIFICATION_ERROR:
					fitness = testInd.getCombinedClassificationError(inputs, targetOutputs,aurocSampleSize, true, true, true);
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int j=0;j<performanceFunctions.length;j++)
				{
					switch(performanceFunctions[j])
					{
						case PROGRAM_LENGTH:
							int len = testInd.getLength();
							if(len < individuals[i].getLength())
								continue outer;
							else if(len > individuals[i].getLength())
								worse = true;
							break;
						case DIVERSITY_DISTANCE:
							double squaredDistance = getAverageSquaredDistance(testInd);
							if(squaredDistance > getAverageSquaredDistance(i))
								continue outer;
							else if(squaredDistance < getAverageSquaredDistance(i))
								worse = true;
							break;
						case FUNCTION_COMPLEXITY:
							int complexity = getComplexity(testInd);
							if(complexity < functionComplexities[i])
								continue outer;
							else if(complexity > functionComplexities[i])
								worse = true;
							break;
						case VARIABLE_MAX:
							int noInputs = getNumberOfInputsUsed(testInd);
							if(noInputs > noInputsUsed[i])
								continue outer;
							else if(noInputs < noInputsUsed[i])
								worse = true;
							break;
						case VARIABLE_MIN:
							int numInputs = getNumberOfInputsUsed(testInd);
							if(numInputs < noInputsUsed[i])
								continue outer;
							else if(numInputs > noInputsUsed[i])
								worse = true;
							break;
					}
				}
				//The test Individual has been 'beaten' in at least 1 dimension and is at least equalled in every dimension (no 'continue'), therefore it is dominated
				if(worse)
					count++;
			}
			else if(worse)
				count++;
		}
		return count;
	}*/

	private boolean isDominated(Individual testInd, double fitness)
	{
		outer:
		for(int i=0;i<popSize;i++)
		{
			boolean worse = false;
			switch(fitnessFunction)
			{
				case MSE:
				case RMSE:
				case ABSOLUTE_ERROR:
				case AOREC:
				case COMBINED_CLASSIFICATION_ERROR:
				//case CLUSTERED_REGRESSION:
					if(fitness < allFitnesses[i])
						continue;
					else if(fitness > allFitnesses[i])
						worse = true;;
					break;
				case NUMBER_CORRECT_BINARY:
				case AUROC:
				case SENSITIVITY:
				case SPECIFICITY:
				case PPV:
				case NPV:
					if(fitness > allFitnesses[i])
						continue;
					else if(fitness < allFitnesses[i])
						worse = true;
					break;
			}
			if(performanceFunctions!=null)
			{
				for(int j=0;j<performanceFunctions.length;j++)
				{
					switch(performanceFunctions[j])
					{
						case PROGRAM_LENGTH:
							int len = testInd.getLength();
							if(len < individuals[i].getLength())
								continue outer;
							else if(len > individuals[i].getLength())
								worse = true;
							break;
						case DIVERSITY_DISTANCE:
							double squaredDistance = getAverageSquaredDistance(testInd);
							if(squaredDistance > getAverageSquaredDistance(i))
								continue outer;
							else if(squaredDistance < getAverageSquaredDistance(i))
								worse = true;
							break;
						case FUNCTION_COMPLEXITY:
							int complexity = getComplexity(testInd);
							if(complexity < functionComplexities[i])
								continue outer;
							else if(complexity > functionComplexities[i])
								worse = true;
							break;
						case VARIABLE_MAX:
							int noInputs = getNumberOfInputsUsed(testInd);
							if(noInputs > noInputsUsed[i])
								continue outer;
							else if(noInputs < noInputsUsed[i])
								worse = true;
							break;
						case VARIABLE_MIN:
							int numInputs = getNumberOfInputsUsed(testInd);
							if(numInputs < noInputsUsed[i])
								continue outer;
							else if(numInputs > noInputsUsed[i])
								worse = true;
							break;
					}
				}
				//The test Individual has been 'beaten' in at least 1 dimension and is at least equalled in every dimension (no 'continue'), therefore it is dominated
				if(worse)
					return true;
			}
			else if(worse)
				return true;
		}
		return false;
	}

	protected void savePopulation()
	{
		savedPopulation = new Individual[popSize];
		savedFitnesses = new double[popSize];
		for(int i=0;i<popSize;i++)
		{
			savedPopulation[i] = individuals[i].clone();
			savedFitnesses[i] = allFitnesses[i];
		}
	}
}