package geneticWedge.gp;

import geneticWedge.gp.function.*;
import java.util.Vector;

/**PopulationDescriptor holds the information required to create a Population object. Properties include minimum
and maximum tree depths, grow method (GROW, FULL or RAMPED_HALF_AND_HALF), Function and Terminal lists, etc.*/
public class PopulationDescriptor
{
	/**PopulationGrowMethod determines the method used to 'grow' the Individuals within a Population*/
	public static enum PopulationGrowMethod
	{
		/**The GROW method selects Terminals and Functions randomly. This creates uneven trees containing subtrees of different sizes.*/
		GROW,
		/**If the FULL method is used to create trees, Terminals are only used when the maximum tree depth is
		reached. All subtrees therefore have the same depth.*/
		FULL,
		/**The RAMPED_HALF_AND_HALF method creates a mixture of trees by combining the GROW and FULL methods
		and by using varying tree depths between the maximum and minimum tree depths.*/
		RAMPED_HALF_AND_HALF
	};

	private PopulationGrowMethod growMethod;
	//maxDepth and minDepth describe the starting population.
	//In order to set the maximum depth and length of a Population's Individuals during evolution,
	//call setMaximumIndividualDepth, setMaximumIndividualLength on Population
	private int maxDepth;
	private int minDepth;
	//private int noOfInputs;
	//private int noOfConstants;
	private Function[] functions;
	private Input[] inputs;
	private Constant[] constants;
	private Vector<Double> mutatedConstants = new Vector<Double>();
	private double functionProportion=0.7;
	private double mutateProportion=0.01;
	private double reproduceProportion=0.1;
	private int eliteSize=0;
	private int initialSelectionPressure = 1;

	private boolean allowCombinedOp=false;

	/**Creates a PopulationDescriptor object with a growMethod, minimum and maximum tree depths and arrays of Functions,
	Constants and Inputs to be used in constructing Individuals.*/
	public PopulationDescriptor(PopulationGrowMethod growMethod, int minTreeDepth, int maxTreeDepth, Function[] functions, Input[] inputs, Constant[] constants) throws InvalidRangeException
	{
		if(maxTreeDepth<minTreeDepth||minTreeDepth<1)
			throw new InvalidRangeException();
		this.growMethod=growMethod;
		minDepth=minTreeDepth;
		maxDepth=maxTreeDepth;
		//this.noOfInputs=noOfInputs;
		//this.noOfConstants=noOfConstants;
		this.functions=functions;
		this.inputs=inputs;
		this.constants=constants;
	}

	/**Creates a PopulationDescriptor object with a growMethod, minimum and maximum tree depths, arrays of Functions,
	Constants and Inputs to be used in constructing Individuals and a growFunctionProportion. <i>growFunctionProportion</i>
	is the probability of selecting a Function (rather than a Terminal) when creating an Individual using the GROW method.*/
	public PopulationDescriptor(PopulationGrowMethod growMethod, int minTreeDepth, int maxTreeDepth, Function[] functions, Input[] inputs, Constant[] constants, double growFunctionProportion) throws InvalidRangeException
	{
		this(growMethod,minTreeDepth,maxTreeDepth,functions,inputs,constants);
		functionProportion=growFunctionProportion;
	}

	public void setInitialisationParameters(int initialSelectionPressure)
	{
		this.initialSelectionPressure = initialSelectionPressure;
	}

	public int getInitialSelectionPressure()
	{
		return initialSelectionPressure;
	}

	/**Returns the PopulationGrowMethod used when constructing a Population.*/
	public PopulationGrowMethod getGrowMethod()
	{
		return growMethod;
	}

	/**Returns the minimum tree depth used when constructing a Population.*/
	public int getMinTreeDepth()
	{
		return minDepth;
	}

	/**Returns the maximum tree depth used when constructing a Population.*/
	public int getMaxTreeDepth()
	{
		return maxDepth;
	}

	/**Returns the array of Functions used when constructing a Population.*/
	public Function[] getFunctions()
	{
		return functions;
	}

	/**Returns the array of Inputs used when constructing a Population.*/
	public Input[] getInputs()
	{
		return inputs;
	}

	/**Returns the array of Constants used when constructing a Population.*/
	public Constant[] getConstants()
	{
		return constants;
	}

	/**Returns the growFunctionProportion used when constructing a Population. <i>growFunctionProportion</i>
	is the probability of selecting a Function (rather than a Terminal) when creating an Individual using the GROW method.*/
	public double getGrowFunctionProportion()
	{
		return functionProportion;
	}

	/**Sets the probabilies of using reproduction (cloning) and mutation operators during evolution. When crossover and reproduction
	are not used crossover will occur.*/
	public void setMutationAndReproductionProportions(double mutProportion, double repProportion) throws IllegalMutationReproductionRatesException
	{
		if(mutProportion<0.0||repProportion<0.0||mutProportion+repProportion>1.0)
			throw new IllegalMutationReproductionRatesException();
		mutateProportion=mutProportion;
		reproduceProportion=repProportion;
		allowCombinedOp=false;
	}

	/**Sets the probabilies of using reproduction (cloning) and mutation operators during evolution. When reproduction
	is not used crossover will occur. If <i>allowCombinedOp</i> is true, mutation may occur <i>in addition to</i> reproduction or crossover.
	The occurrence of muation is assessed separately after crossover/reproduction. If false, only one operator
	aill be applied at a time, so the total probability of mutation, crossover or reproduction is 1.0.*/
	public void setMutationAndReproductionProportions(double mutProportion, double repProportion, boolean allowCombinedOp) throws IllegalMutationReproductionRatesException
	{
		if(allowCombinedOp)
		{
			if(mutProportion<0.0||repProportion<0.0||mutProportion>1.0||repProportion>1.0)
				throw new IllegalMutationReproductionRatesException();
			mutateProportion=mutProportion;
			reproduceProportion=repProportion;
		}
		else
			setMutationAndReproductionProportions(mutProportion, repProportion);
		this.allowCombinedOp=allowCombinedOp;
	}

	/**Sets the probability of mutation occurring. IllegalMutationReproductionRatesException will be thrown if an
	attempt is made to set a mutation rate outside the range [0,1]. If allowCombinedOp is false, this Exception
	will also be thrown if the combined mutation and reproduction rates exceed 1.0, since these operations cannot occur together.*/
	public void setMutationProportion(double proportion) throws IllegalMutationReproductionRatesException
	{
		if(proportion<0.0||proportion>1.0||!allowCombinedOp && proportion+reproduceProportion>1.0)
			throw new IllegalMutationReproductionRatesException();
		mutateProportion=proportion;
	}

	/**Sets the probability of mutation occurring. IllegalMutationReproductionRatesException will be thrown if an
	attempt is made to set a mutation rate outside the range [0,1]. If allowCombinedOp is false, this Exception
	will also be thrown if the combined mutation and reproduction rates exceed 1.0, since these operations cannot occur together.*/
	public void setMutationProportion(double proportion, boolean allowCombinedOp) throws IllegalMutationReproductionRatesException
	{
		if(allowCombinedOp)
		{
			if(proportion<0.0||proportion>1.0)
				throw new IllegalMutationReproductionRatesException();
			mutateProportion=proportion;
		}
		else
			setMutationProportion(proportion);
		this.allowCombinedOp=allowCombinedOp;
	}

	/**Sets the probability of reproduction (cloning) occurring. IllegalMutationReproductionRatesException will be thrown if an
	attempt is made to set a reproduction rate outside the range [0,1]. If allowCombinedOp is false, this Exception
	will also be thrown if the combined mutation and reproduction rates exceed 1.0, since these operations cannot occur together.*/
	public void setReproductionProportion(double proportion) throws IllegalMutationReproductionRatesException
	{
		if(proportion<0.0||proportion>1.0||!allowCombinedOp &&proportion+mutateProportion>1.0)
			throw new IllegalMutationReproductionRatesException();
		reproduceProportion=proportion;
	}

	/**Sets the probability of reproduction (cloning) occurring. IllegalMutationReproductionRatesException will be thrown if an
	attempt is made to set a reproduction rate outside the range [0,1]. If allowCombinedOp is false, this Exception
	will also be thrown if the combined mutation and reproduction rates exceed 1.0, since these operations cannot occur together.*/
	public void setReproductionProportion(double proportion, boolean allowCombinedOp) throws IllegalMutationReproductionRatesException
	{
		if(allowCombinedOp)
		{
			if(proportion<0.0||proportion>1.0)
				throw new IllegalMutationReproductionRatesException();
			reproduceProportion=proportion;
		}
		else
			setReproductionProportion(proportion);
		this.allowCombinedOp=allowCombinedOp;
	}

	/**Returns the probability of mutation occurring during the creation of offspring.*/
	public double getMutationProportion()
	{
		return mutateProportion;
	}

	/**Returns the probability of reproduction (cloning) occurring during the creation of offspring.*/
	public double getReproductionProportion()
	{
		return reproduceProportion;
	}

	/**Returns the size of the elite. Elite Individuals will be transferred to the next generation without
	having to undergo selection. Having an elite prevents the loss of the best Individual(s) through stochastic sampling.*/
	public int getEliteSize()
	{
		return eliteSize;
	}

	/**Sets the size of the elite. Elite Individuals will be transferred to the next generation without
	having to undergo selection.*/
	public void setEliteSize(int eliteSize)
	{
		if(eliteSize>0)
			this.eliteSize=eliteSize;
	}

	/**When CONSTANT_VALUE mutation is included as a mutation operation, new Constant objects may be created.
	These should be added to PopulationDescriptor so that they can be used during future mutation operations.*/
	public void addMutatedConstant(double value)
	{
		mutatedConstants.add(value);
	}

	/**When CONSTANT_VALUE mutation is included as a mutation operation, new Constant objects may be created.
	These should be added to PopulationDescriptor so that they can be used during future mutation operations.
	This method returns the values of the Constant objects created via CONSTANT_VALUE mutation.*/
	public Vector<Double> getMutatedConstants()
	{
		return mutatedConstants;
	}

	/**When CONSTANT_VALUE mutation is included as a mutation operation, new Constant objects may be created.
	These are added to PopulationDescriptor so that they can be used during future mutation operations.
	This method removes these mutated Constants. They will not then be available during future mutation operations.*/
	public void resetMutatedConstants()
	{
		mutatedConstants.clear();
	}

	/**Returns allowCombinedOp. If true, mutation may occur <i>in addition to<i> crossover or reproduction.
	If false, only one operation may occur at a time, so the combined probability of crossover, reproduction or mutation is 1.0.*/
	public boolean getAllowCombinedOp()
	{
		return allowCombinedOp;
	}
}