package geneticWedge.gp;

import java.util.Vector;

/**StagnationStoppingCriterion stops evolution if no improvement in some measure of Fitness has occurred for some time.
The 'patience' of the stopping method (the allowed number of generations without improvement) may be set, as may
the value against which to assess improvement: training or validation error; best or averaged value.
Note: if both useValidationData and useTrainingData are true, the average of the training and validation fitnesses will be used.
If both useValidationData and useTrainingData are false, the training fitness will be used.*/
public class StagnationStoppingCriterion extends EvaluativeStoppingCriterion
{
	private int patience=0;
	private boolean averageValues=true;
	//private Vector<Double> oldValues=new Vector<Double>();
	private int badCount=0;
	private double lastGoodValue=Double.NaN;

	/**StagnationStoppingCriterion constructor.
	@param patienceParameter The number of generations that are allowed without any improvement. If set to 0,
	evolution will stop as soon as there is no improvement. Otherwise, the Population will be restored to its
	state at the optimum value after evolution has been stopped.
	@param averageValues If true, improvement in average fitness will be used as a stopping criterion, otherwise improvement in best value will be used.
	@param useValidationData If true, improvement will be assessed against fitnesses on the validation data, otherwise against fitnesses on the training data.*/
	public StagnationStoppingCriterion(int patienceParameter, boolean averageValues, boolean useValidationData)
	{
		if(patienceParameter>=0)
			patience=patienceParameter;
		this.averageValues=averageValues;
		this.useValidationData=useValidationData;
		this.useTrainingData = !useValidationData;
	}

	/**StagnationStoppingCriterion constructor.
	@param patienceParameter The number of generations that are allowed without any improvement. If set to 0,
	evolution will stop as soon as there is no improvement. Otherwise, the Population will be restored to its
	state at the optimum value after evolution has been stopped.
	@param averageValues If true, improvement in average fitness will be used as a stopping criterion, otherwise improvement in best value will be used.
	@param useValidationData If true, improvement will be assessed against fitnesses on the validation data.
	Note: if both useValidationData and useTrainingData are true, the average of the training and validation fitnesses will be used.
	@param useTrainingData If true, improvement will be assessed against fitnesses on the training data.
	Note: if both useValidationData and useTrainingData are true, the average of the training and validation fitnesses will be used.*/
	public StagnationStoppingCriterion(int patienceParameter, boolean averageValues, boolean useTrainingData, boolean useValidationData)
	{
		if(patienceParameter>=0)
			patience=patienceParameter;
		this.averageValues=averageValues;
		this.useTrainingData=useTrainingData;
		this.useValidationData=useValidationData;
	}

	/**Returns true if the stopping condition has been reached (no improvement in evaluation function for a fixed number of generations).*/
	public boolean stopReached(AbstractPopulation pop)
	{
		double val=0.0;
		if(averageValues)
		{
			if(useValidationData)
			{
				val=pop.getAverageValidationFitness();
				if(useTrainingData)
				{
					val += pop.getAverageTrainingFitness();
					val /=2.0;
				}
			}
			else
				val=pop.getAverageTrainingFitness();
		}
		else
		{
			if(useValidationData)
			{
				val=pop.getBestValidationFitness();
				if(useTrainingData)
				{
					val += pop.getBestTrainingFitness();
					val /=2.0;
				}
			}
			else
				val=pop.getBestTrainingFitness();
		}
		//System.err.println("val="+val);
		//first value
		if(Double.isNaN(lastGoodValue))
		{
			lastGoodValue=val;
			return false;
		}

		if(maximiseValue)
		{
			if(val>lastGoodValue)
			{
				lastGoodValue=val;
				badCount=0;
				return false;
			}
			else
			{
				badCount++;
				if(badCount>=patience)
				{
					lastGoodValue=Double.NaN;
					badCount=0;
					return true;
				}
				else
					return false;
			}

		}
		else
		{
			if(val<lastGoodValue)
			{
				lastGoodValue=val;
				badCount=0;
				return false;
			}
			else
			{
				badCount++;
				if(badCount>=patience)
				{
					lastGoodValue=Double.NaN;
					badCount=0;
					return true;
				}
				else
					return false;
			}
		}
	}

	/**Returns a String representation of this StagnationStoppingCriterion. Useful for reporting results.*/
	public String toString()
	{
		String str="StagnationStoppingCriterion. ";
		if(averageValues)
			str+="Average ";
		else
			str+="Best ";

		if(useValidationData)
			str+="validation value";
		else
			str+="training value";

		if(maximiseValue)
			str+=" shows no increase for more than "+patience+" generations.";
		else
			str+=" shows no decrease for more than "+patience+" generations.";

		return str;
	}

	/**Returns true if there was an improvment in evaluation function during creation of the last generation.*/
	public boolean lastGenerationWasImproved()
	{
		return (badCount==0);
	}
}