package geneticWedge.gp;

import java.util.*;
import java.awt.geom.*;
import geneticWedge.gp.function.*;

/**Utils contains a number of static methods of use to various members of the gp package.*/
public final class Utils
{
	/**Converts String representations of class membership into numeric values. This is needed because input values
	are numeric. When using a GP for classification class membership tags must therefore be converted into double values.
	+1 is used to indictae membership, -1 to indicate non-membership.*/
	public static double[][] stringToNumericTargets(String[] stringTargets, String[] targetClasses)
	{
		int dataLen = stringTargets.length;
		int noOfClasses = targetClasses.length;
		double[][] numericTargets = new double[dataLen][noOfClasses];
		for(int i=0;i<dataLen;i++)
		{
			for(int j=0;j<noOfClasses;j++)
			{
				if(stringTargets[i].trim().equals(targetClasses[j]))
					numericTargets[i][j] = 1.0;
				else
					numericTargets[i][j] = -1.0;
			}
		}
		return numericTargets;
	}

	/**This method calculates the index of the last node within a subtree. It is essential when performing mutation
	and crossover operations beacuse it allows the removal/transfer of a subtree intact, using its start and end indices.*/
	public static int getEndOfFragment(Vector<Component> nodes, int startIndex)
	{

		int noBranches=1;
		int endIndex=startIndex;
		do
		{
			endIndex++;
			if(nodes.size()==endIndex)
			{
				endIndex--;
				noBranches--;
			}
			else if(nodes.get(endIndex) instanceof Terminal)
				noBranches--;
			else if(nodes.get(endIndex) instanceof TwoInputFunction)
				noBranches++;
		}
		while(noBranches>0);
		return endIndex;
	}

	/**This method calulates the minimum and maximum depths occurring between 2 nodes (those represented by startIndex and endIndex).
	If startIndex and endIndex are the first and last nodes within a subtree, the difference between the 2 depths
	is the depth of the subtree.*/
	public static int[] getFragmentDepths(Vector<Component> nodes, int startIndex, int finishIndex)
	{
		int minDepth=0;
		int maxDepth=0;
		int depth=0;
		Vector<Integer> backupDepths=new Vector<Integer>();
		for(int i=0;i<startIndex+1;i++)
		{
			depth++;
			if(nodes.get(i) instanceof Terminal)
			{
				int noBackups=backupDepths.size();
				if(noBackups==0)
					break;
				int lastBranch=noBackups-1;
				depth=backupDepths.get(lastBranch);
				backupDepths.remove(lastBranch);
			}
			else if(nodes.get(i) instanceof TwoInputFunction)
				backupDepths.add(depth);
		}
		maxDepth = minDepth = depth;
		backupDepths=new Vector<Integer>();
		for(int i=startIndex+1;i<finishIndex+1;i++)
		{
			depth++;
			if(depth>maxDepth)
				maxDepth=depth;
			if(nodes.get(i) instanceof Terminal)
			{
				int noBackups=backupDepths.size();
				if(noBackups==0)
					return new int[]{minDepth,maxDepth};
				int lastBranch=noBackups-1;
				depth=backupDepths.get(lastBranch);
				backupDepths.remove(lastBranch);
			}
			else if(nodes.get(i) instanceof TwoInputFunction)
				backupDepths.add(depth);
		}
		return new int[]{minDepth,maxDepth};
	}

	/**This method returns the depth of the node represented by <i>index</i>*/
	public static int getDepth(Vector<Component> nodes, int index)
	{
		int depth=0;
		Vector<Integer> backupDepths=new Vector<Integer>();
		for(int i=0;i<index+1;i++)
		{
			depth++;
			if(nodes.get(i) instanceof Terminal)
			{
				int noBackups=backupDepths.size();
				if(noBackups==0)
					break;
				int lastBranch=noBackups-1;
				depth=backupDepths.get(lastBranch);
				backupDepths.remove(lastBranch);
			}
			else if(nodes.get(i) instanceof TwoInputFunction)
				backupDepths.add(depth);
		}
		return depth;
	}

	/**Returns an array containing <i>numberOfIndices</i>. These are selected from the numbers 1 to <i>length</i>
	without repetition. Therefore, if length < numberOfIndices <i>null</i> will be returned.*/
	public static int[] getRandomIndices(int length, int numberOfIndices)
	{
		if(numberOfIndices>length||numberOfIndices<0)
		{
			System.err.println("Utils: Invalid number of random indices requested");
			return null;
		}
		Random rand=new Random();
		int[]indices = new int [numberOfIndices];
		Vector<Integer> vectorIndices=new Vector<Integer>();
		for(int i=0;i<length;i++)
			vectorIndices.add(i);
		int index;
		for(int i=0;i<numberOfIndices;i++)
		{
			index=(int)(rand.nextDouble()*(length-i));
			indices[i]=vectorIndices.get(index);
			vectorIndices.remove(index);
		}
		return indices;
	}

	/**Returns the integers between 1 and <i>length</i> (inclusive) in random order. This is useful for randomly permuting data.*/
	public static int[] getRandomIndices(int length)
	{
		Random rand=new Random();
		int[]indices = new int [length];

		/*Vector<Integer> vectorIndices=new Vector<Integer>();
		for(int i=0;i<length;i++)
			vectorIndices.add(i);
		int index;
		for(int i=0;i<length;i++)
		{
			index=(int)(rand.nextDouble()*(length-i));
			indices[i]=vectorIndices.get(index);
			vectorIndices.remove(index);
		}
		return indices;*/
		//new method 101108. Faster, but there is a small chance of duplicate values
		double[] vals = new double[length];
		double[] originalVals = new double[length];
		for(int i=0;i<length;i++)
		{
			vals[i]=rand.nextDouble();
			originalVals[i]=vals[i];
		}
		Arrays.sort(vals);
		for(int i=0;i<length;i++)
		{
			indices[i]=Arrays.binarySearch(vals,originalVals[i]);
			if(i>0 && vals[i]==vals[i-1])
			{
				System.err.println("ERROR. Duplicate values when creating indices!");
				return getRandomIndices(length);//recursive method - could be dangerous!
			}
		}
		return indices;
	}

	/**Sorts <i>values</i> using the Shell sort algorithm. If sortAscending is true, the values will be
	in order from smallest to largest. The returned array contains the indices of <i>values</i> in the requested order.*/
	public static int[] ShellSortValues(double[]values, boolean sortAscending)
	{
		int len = values.length;
		double[] sortedVals=(double[])(values.clone());
		int[] originalIndices = new int[len];
		for(int i=0;i<len;i++)
			originalIndices[i] = i;
		int h = 1;
		//find the largest h value possible
		while ((h * 3 + 1) < len)
		{
			h = 3 * h + 1;
		}

		while( h > 0 )
		{
			//for each set of elements (there are h sets)
			for (int i = h - 1; i < len; i++)
			{
				//pick the last element in the set
				double val = sortedVals[i];
				int index = originalIndices[i];
				int j = i;
				 /*compare value to the one before it in the set
				   if they are out of order continue this loop, moving
				   elements "back" to make room for ind to be inserted.*/
				for( j = i; (j >= h) && ((!sortAscending && sortedVals[j-h] < val) || (sortAscending && sortedVals[j-h] > val)); j -= h)
				{
					sortedVals[j] = sortedVals[j-h];
					originalIndices[j] = originalIndices[j-h];
				}
				//insert val into the correct place
				sortedVals[j] = val;
				originalIndices[j] = index;
			}
			//all sets h-sorted, now decrease set size
			h = h / 3;
		}
		return originalIndices;
	}

	/**Sorts <i>values</i> using the Shell sort algorithm, in alphabetical order.
	The returned array contains the indices of <i>values</i> in the requested order.*/
	public static int[] ShellSortStrings(String[]values)
	{
		int len = values.length;
		String[] sortedVals=(String[])(values.clone());
		int[] originalIndices = new int[len];
		for(int i=0;i<len;i++)
			originalIndices[i] = i;
		int h = 1;
		//find the largest h value possible
		while ((h * 3 + 1) < len)
		{
			h = 3 * h + 1;
		}

		while( h > 0 )
		{
			//for each set of elements (there are h sets)
			for (int i = h - 1; i < len; i++)
			{
				//pick the last element in the set
				String val = sortedVals[i];
				int index = originalIndices[i];
				int j = i;
				 /*compare value to the one before it in the set
				   if they are out of order continue this loop, moving
				   elements "back" to make room for ind to be inserted.*/
				for( j = i; (j >= h) &&  sortedVals[j-h].compareTo(val)>0; j -= h)
				{
					sortedVals[j] = sortedVals[j-h];
					originalIndices[j] = originalIndices[j-h];
				}
				//insert val into the correct place
				sortedVals[j] = val;
				originalIndices[j] = index;
			}
			//all sets h-sorted, now decrease set size
			h = h / 3;
		}
		return originalIndices;
	}

	/**Sorts <i>values</i> using the Shell sort algorithm, returning them in ascending order.
	The returned array contains the indices of <i>values</i> in order.*/
	public static int[] ShellSortValues(double[]values)
	{
		return ShellSortValues(values, true);
	}

	/**This method takes a vector of indices that are assumed to be ordered.
	It performs roulette wheel rank selection, with the early indices favoured if <i>favourFirst</i> is true. It is designed
	to be used for selecting several indices, for example when performing multiple-tree crossover.*/
	public static int[] getIndicesProbabilistically(Vector<Integer> initialIndices, boolean favourFirst)
	{
		Random rand = new Random();
		Vector<Integer> indices = (Vector<Integer>)initialIndices.clone();
		int len = indices.size();
		int[] newIndices = new int[len];
		double total = len*(len+1)/2;
		double[] rouletteVals = new double[len];
		double lastVal = 0.0;
		if(favourFirst)
		{
			for(int i=0;i<len;i++)
			{
				rouletteVals[len-i] = lastVal + (len-i) + 1;
				lastVal = rouletteVals[len-i];
			}
		}
		else
		{
			for(int i=0;i<len;i++)
			{
				rouletteVals[i] = lastVal + i + 1;
				lastVal = rouletteVals[i];
			}
		}
		for(int i=0;i<len;i++)
		{
			double selector = rand.nextDouble() * total;
			int index = Arrays.binarySearch(rouletteVals , selector);
			if(index <0)
				index = -(index+1);
			newIndices[i] = indices.get(index);
			indices.remove(index);
			if(index ==0)
				total -= rouletteVals[index];
			else
				total -= (rouletteVals[index] - rouletteVals[index-1]);
			rouletteVals = removeRouletteValue(rouletteVals, index);
		}
		return newIndices;
	}

	/**This method takes a vector of indices that are assumed to be ordered.
	It performs roulette wheel rank selection, with the early indices favoured if <i>favourFirst</i> is true. It is designed
	to be used for selecting several indices, for example when performing multiple-tree crossover.*/
	public static int[] getIndicesProbabilistically(int[] initialIndices, boolean favourFirst)
	{
		Random rand = new Random();
		Vector<Integer> indices = new Vector<Integer>();
		int len = initialIndices.length;
		for(int i=0;i<len;i++)
			indices.add(initialIndices[i]);
		int[] newIndices = new int[len];
		double total = len*(len+1)/2;
		double[] rouletteVals = new double[len];
		double lastVal = 0.0;
		if(favourFirst)
		{
			for(int i=0;i<len;i++)
			{
				rouletteVals[i] = lastVal + len - i;
				lastVal = rouletteVals[i];
			}
		}
		else
		{
			for(int i=0;i<len;i++)
			{
				rouletteVals[i] = lastVal + i + 1;
				lastVal = rouletteVals[i];
			}
		}
		for(int i=0;i<len;i++)
		{
			double selector = rand.nextDouble() * total;
			int index = Arrays.binarySearch(rouletteVals , selector);
			if(index <0)
				index = -(index+1);
			newIndices[i] = indices.get(index);
			indices.remove(index);
			if(index ==0)
				total -= rouletteVals[index];
			else
				total -= (rouletteVals[index] - rouletteVals[index-1]);
			rouletteVals = removeRouletteValue(rouletteVals, index);
		}
		return newIndices;
	}

	private static double[] removeRouletteValue(double[] array, int index)
	{
		int len = array.length - 1;
		double[] newArray = new double[len];
		double removedValue = 0.0;
		if(index == 0)
			removedValue = array[0];
		else
			removedValue = array[index] - array[index-1];
		for(int i=0;i<len+1;i++)
		{
			if(i<index)
				newArray[i] = array[i];
			else if(i>index)
				newArray[i-1] = array[i]-removedValue;
		}
		return newArray;
	}

	/**Returns the unique classes avaialable from a list of class memberships, sorted into ascending (alphabetical) order.*/
	public static String[] getOutputClasses(String[] targetOutputs)
	{
		String [] outputClasses;
		Vector<String> vectorClasses = new Vector<String>();
		for(int i=0;i<targetOutputs.length;i++)
		{
			String str = targetOutputs[i].trim();
			if(!vectorClasses.contains(str))
				vectorClasses.add(str);
		}
		int len = vectorClasses.size();
		outputClasses = new String[len];
		outputClasses = vectorClasses.toArray(outputClasses);
		Arrays.sort(outputClasses);
		return outputClasses;
	}

	/**This method returns the genotypic distance between 2 Individuals. The trees of the 2 Individuals are overlain
	and identical nodes are identified starting from the root. However, once a dissimilar node has been identified,
	the remaining subtree is considered to be different. The distance between the Individuals is then calculated as
	(max-noMatches)/max, where max is the number of nodes in the larger Individual and noMatches is the number of matching nodes.*/
	public static double getDistance(Individual ind1, Individual ind2)
	{
		Vector<Component> nodes1 = (Vector<Component>)(ind1.getNodes().clone());
		Vector<Component> nodes2 = (Vector<Component>)(ind2.getNodes().clone());
		return getDistance(nodes1,nodes2);
	}

	/**This method returns the genotypic distance between 2 component lists. The trees of the 2 Individuals are overlain
	and identical nodes are identified starting from the root. However, once a dissimilar node has been identified,
	the remaining subtree is considered to be different. The distance between the Individuals is then calculated as
	(max-noMatches)/max, where max is the number of nodes in the larger Individual and noMatches is the number of matching nodes.*/
	public static double getDistance(Vector<Component> nodes1, Vector<Component> nodes2)
	{
		int size1 = nodes1.size();
		int size2 = nodes2.size();
		int diff = size2 - size1;
		int min = size1;
		int max = size2;
		if(diff<0)
		{
			diff = -diff;
			min = size2;
			max = size1;
			Vector<Component> temp = new Vector<Component>(nodes1);
			nodes1 = new Vector<Component>(nodes2);
			nodes2 = new Vector<Component>(temp);
		}
		int noMatches = 0;
		if(nodes1.get(0).equals(nodes2.get(0)))
			noMatches++;
		else
			return 1.0;

		int secondIndex = 1;
		for(int i=1;i<min;i++)
		{
			if(nodes1.get(i).equals(nodes2.get(secondIndex)))
				noMatches++;
			else
			{
				i = Utils.getEndOfFragment(nodes1, i-1);
				secondIndex = Utils.getEndOfFragment(nodes2, secondIndex-1);
			}
			secondIndex++;
			if(secondIndex==max)
				break;
		}
		return ((double)(max - noMatches)) / max;//This formula used, otherwise trees with very different sizes may be classed as similar
		//return ((double)(min - noMatches)) / min; //favours smaller trees. Similar to deJong
	}


	//public static Hashtable<Integer,Double>[] getNumberedInputs(Hashtable<String,Integer>[] inputs, String[] names)
	public static Hashtable<Integer,Double>[] getNumberedInputs(Hashtable<String,Double>[] inputs, String[] names)
	{
		int datalen = inputs.length;
		Hashtable<Integer,Double>[] numberedInputs = (Hashtable<Integer,Double>[])new Hashtable[datalen];
		for(int i=0;i<datalen;i++)
		{
			numberedInputs[i] = new Hashtable<Integer,Double>();
			Enumeration<String> keys = inputs[i].keys();
			Enumeration<Double> elements = inputs[i].elements();
			while(keys.hasMoreElements())
			{
				String name = keys.nextElement();
				double valObj = elements.nextElement();
				numberedInputs[i].put(new Integer(Arrays.binarySearch(names,name)),valObj);
			}
		}
		return numberedInputs;
	}

	//sorts by x-coordinate and then y-coordinate
	private static void ShellSortPoints(Point2D.Double[] points)
	{
		int len = points.length;
		int[] originalIndices = new int[len];
		for(int i=0;i<len;i++)
			originalIndices[i] = i;
		int h = 1;
		//find the largest h value possible
		while ((h * 3 + 1) < len)
		{
			h = 3 * h + 1;
		}

		while( h > 0 )
		{
			//for each set of elements (there are h sets)
			for (int i = h - 1; i < len; i++)
			{
				//pick the last element in the set
				Point2D.Double point = new Point2D.Double(points[i].x,points[i].y);
				int index = originalIndices[i];
				int j = i;
				 /*compare ind to the one before it in the set
				   if they are out of order continue this loop, moving
				   elements "back" to make room for ind to be inserted.*/
				for( j = i; (j >= h) && ((points[j-h].x > point.x)||(points[j-h].x == point.x && points[j-h].y > point.y)); j -= h)
				{
					points[j] = points[j-h];
					originalIndices[j] = originalIndices[j-h];
				}
				//insert ind into the correct place
				points[j] = point;
				originalIndices[j] = index;
			}
			//all sets h-sorted, now decrease set size
			h = h / 3;
		}
	}

	/**Calculates and returns the Area under the ROC curve (AUROC) during a classification task, using all of the provided data.
	The AUROC is a measure of the discriminatory power of a classifier.*/
	public static double getAUROC(double[] outputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		double tempAUROC;
		Point2D.Double[] points = null;
		points = getAllROCCoordinates(outputs, targetOutputs);
		if(points == null)
			return Double.NaN;
		ShellSortPoints(points);
		double auroc = points[0].y * points[0].x;
		for(int i=1;i<outputs.length;i++)
			auroc+=(points[i].y+points[i-1].y)*(points[i].x-points[i-1].x);
		auroc+=(1.0+points[outputs.length-1].y)*(1.0-points[outputs.length-1].x);//080108 add last point at (1,1)
		return auroc / 2.0;
	}

	/**This method is used in calculating AUROC values. It is a public method to enable other classes
	to obtain ROC coordinates to be used, for example, in visualisation.*/
	public static Point2D.Double[] getAllROCCoordinates(double[] outputs,double[] targetOutputs) throws InvalidInputSizeException
	{
		int len=outputs.length;
		if(len!=targetOutputs.length)
			throw new InvalidInputSizeException();
		double output;
		int[] truePositives = new int[len];
		int[] falsePositives = new int[len];
		double targetPositives = 0.0;
		double targetNegatives = 0.0;
		for(int i=0;i<len;i++)
		{
			if(targetOutputs[i]>0.0)
				targetPositives++;
			else
				targetNegatives++;
			output = outputs[i];
			for(int j=0;j<len;j++)
			{
				if(output>outputs[j])
				{
					if(targetOutputs[i]>0.0)
						truePositives[j]++;
					else
						falsePositives[j]++;
				}
			}
		}
		if(targetPositives == 0 || targetNegatives == 0)
		{
			System.err.println("getAllROCCords returning null."+ (targetPositives == 0));
			int test = 0;
			for(int i=0;i<len;i++)
				if(targetOutputs[i] == 1)
					test++;
			return null;
		}
		Point2D.Double[] points = new Point2D.Double[len];
		for(int i=0;i<len;i++)
			points[i] = new Point2D.Double(falsePositives[i] / targetNegatives, truePositives[i] / targetPositives);
		return points;
	}
}