package geneticWedge.gp.dataAnalysis;

import geneticWedge.gp.*;
import geneticWedge.gp.function.*;
import java.util.*;
import java.io.*;

/**DataAcquirer allows the performance of various statistical operations on a dataset. Quantities that
may be calculated include mean, median, variance, standard deviation, Pearson's R. Information may also
be extracted from a results file or from a Population object, including average tree length, number of times a Function or Input is used, etc.*/
public class DataAcquirer
{
	private static long [] factorials =
	{
	   1L /* 0 */,
	   1L /* 1 */,
	   2L /* 2 */,
	   6L /* 3 */,
	   24L /* 4 */,
	   120L /* 5 */,
	   720L /* 6 */ ,
	   5040L /* 7 */ ,
	   40320L /* 8 */ ,
	   362880L /* 9 */ ,
	   3628800L /* 10 */ ,
	   39916800L /* 11 */ ,
	   479001600L /* 12 */ ,
	   6227020800L /* 13 */ ,
	   87178291200L /* 14 */ ,
	   1307674368000L /* 15 */ ,
	   20922789888000L /* 16 */ ,
	   355687428096000L /* 17 */ ,
	   6402373705728000L /* 18 */ ,
	   121645100408832000L /* 19 */ ,
	   2432902008176640000L /* 20 */
	};

	/**ScalarData refer to scalar properties of an Individual.
	They may be contrasted with VectorData which are vector quantities
	referring to the balance of components within Individuals*/
	public enum ScalarData
	{
		/**PROGRAM_LENGTH is the number of nodes within a tree.*/
		PROGRAM_LENGTH,
		/**PROGRAM_DEPTH is the depth of a tree.*/
		PROGRAM_DEPTH,
		/**PROGRAM_FITNESS is the fitness of an Individual, as determined by the FitnessFunction of the Population.*/
		PROGRAM_FITNESS,
		/**PROGRAM_DIVERSITY_DISTANCE is the average genotypic distance of an Individual from all other members of the Population.*/
		PROGRAM_DIVERSITY_DISTANCE
	};

	/**VectorData refer to the balance of components within each individual. They should be used within the method
	getVectorData(). To check which components are referred to, by obtaining an array of components in the same order as the
	numeric array obtained from getVectorData(), call getFunctions(), getInputs(), getConstants() or getComponents().*/
	public enum VectorData
	{
		/**The number of times each Function is used by a population of Individuals.*/
		FUNCTION_USE,
		/**The number of times each Input is used by a population of Individuals.*/
		INPUT_USE,
		/**The number of times each Constant is used by a population of Individuals.*/
		CONSTANT_USE,
		/**The number of times each Component is used by a population of Individuals.*/
		COMPONENT_USE
	};

	private Population pop = null;

	/**Constructor takes a Population object as an argument, allowing the extraction of information from a particular Population.*/
	public DataAcquirer(Population population)
	{
		pop = population;
	}

	/**Returns the minimum and maximum values of a dataset.*/
	public static double[] getMinMax(double[] data)
	{
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for(int i=0;i<data.length;i++)
		{
			if(data[i]>max)
				max = data[i];
			if(data[i]<min)
				min = data[i];
		}
		return new double[]{min, max};
	}

	/**This method partitions the data into equal sized partitions. The lower end of each partition is
	open (does not contain the boundary value), but the upper end is closed (contains the boundary value).
	@return an Object array containing a double array of the boundary values and an int array of the
	population count of each category. Note that the second array will have one less member than the first array.*/
	public static Object[] partitionData(double[] data, double width)
	{
		width = Math.abs(width);
		double[] minMax = getMinMax(data);

		double firstDivisionDouble = minMax[0]/width;
		int firstDivision = 0;
		if(firstDivisionDouble<0.0)
			firstDivision = (int)(Math.floor(firstDivisionDouble));
		else
			firstDivision = (int)(Math.ceil(firstDivisionDouble));
		if(firstDivision == firstDivisionDouble)
			firstDivision--;
		double minimum = firstDivision * width;

		double lastDivisionDouble = minMax[1]/width;
		int lastDivision = 0;
		if(lastDivisionDouble<0.0)
			lastDivision = (int)(Math.floor(lastDivisionDouble));
		else
			lastDivision = (int)(Math.ceil(lastDivisionDouble));
		double maximum = lastDivision * width;

		double[] boundaries = new double[lastDivision - firstDivision +1];
		for(int i = firstDivision;i<=lastDivision;i++)
			boundaries[i - firstDivision] = i * width;
		int[] count = new int[lastDivision - firstDivision];
		for(int i=0;i<data.length;i++)
		{
			int index = Arrays.binarySearch(boundaries, data[i]);
			if(index<0)
				index = -index-2;
			else
				index--;
			count[index]++;
		}
		return new Object[]{boundaries, count};
	}

	/**Returns the mean value of a set of data.*/
	public static double getMean(double[] data)
	{
		double total = 0.0;
		int len = data.length;
		for(int i=0;i<len;i++)
			total += data[i];
		return total/len;
	}

	/**Returns the median value of a set of data.*/
	public static double getMedian(double[] data)
	{
		int len = data.length;
		if(len == 1)
			return data[0];
		double[] sortedData = data.clone();
		Arrays.sort((double[])(sortedData));
		if(sortedData.length % 2 == 0)
			return (sortedData[len/2]+sortedData[len/2-1])/2.0;
		else
			return sortedData[len/2];
	}

	/**Returns the variance of a set of data.*/
	public static double getVariance(double[] data)
	{
		int len = data.length;
		double totalSquare = 0.0;
		double total = 0.0;
		for(int i=0;i<len;i++)
		{
			totalSquare += Math.pow(data[i],2.0);
			total += data[i];
		}
		double mean = total / len;
		return totalSquare / len - Math.pow(mean,2.0);
	}

	/**Returns the standard deviation of a set of data.*/
	public static double getStandardDeviation(double[] data)
	{
		return Math.sqrt(getVariance(data));
	}

	/**Returns the Input use within a population, as represented in a results file. Each key within the Hashtable
	is an Input name. The corresponding element is the average number of times that Input is used per Individual.*/
	public static Hashtable<String,Double> getInputUseFromFile(String filename) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("INPUTS"));
		String[] inputNames = line.substring(7).split(",");

		int noNames = inputNames.length;
		for(int i=0;i<noNames;i++)
		{
			inputNames[i] = inputNames[i].trim();
		}
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();
		int count = 0;
		double[] inputCount = new double[noNames];
		while(line!=null && !line.equals(""))
		{
			if(!(line.startsWith("Number of")||line.startsWith("Best threshold")))
			{
				count++;
				String[] items = line.split("[,() ]");
				for(int i=0;i<items.length;i++)
				{
					for(int j=0;j<noNames;j++)
					{
						if(items[i].trim().equals(inputNames[j]))
						{
							inputCount[j]++;
							break;
						}
					}
				}
			}
			line = reader.readLine();
		}
		//System.err.println("count="+count);
		Hashtable<String,Double> hash = new Hashtable<String,Double>();
		for(int i=0;i<noNames;i++)
			hash.put(inputNames[i],inputCount[i] / count);
		return hash;
	}

	/**Returns the Input use within a population, as represented in a results file. Each key within the Hashtable
	is an Input name. The corresponding element is the fraction of all inputs that are the target Input.*/
	public static Hashtable<String,Double> getFractionalInputUseFromFile(String filename) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("INPUTS"));
		String[] inputNames = line.substring(7).split(",");

		int noNames = inputNames.length;
		for(int i=0;i<noNames;i++)
		{
			inputNames[i] = inputNames[i].trim();
		}
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();
		int count = 0;
		int nodeCount = 0;
		double[] inputCount = new double[noNames];
		while(line!=null && !line.equals(""))
		{
			if(!(line.startsWith("Number of")||line.startsWith("Best threshold")))
			{
				count++;
				String[] items = line.split("[,() ]");
				for(int i=0;i<items.length;i++)
				{
					for(int j=0;j<noNames;j++)
					{
						if(items[i].trim().equals(inputNames[j]))
						{
							inputCount[j]++;
							nodeCount++;
							break;
						}
					}
				}
			}
			line = reader.readLine();
		}
		Hashtable<String,Double> hash = new Hashtable<String,Double>();
		for(int i=0;i<noNames;i++)
			hash.put(inputNames[i],inputCount[i] / nodeCount);
		return hash;
	}

	/**Returns the terminal subtree use within a population, as represented in a results file. Each key within the Hashtable
	is a subtree name. The corresponding element is the average number of times that subtree is used per Individual.
	At the moment this method is severely limited: only binary (3-element) subtrees containing a +, -, / or * function
	are identified.*/
	public static Hashtable<String,Integer> getTerminalSubtreesFromFile(String filename) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("INPUTS"));
		String[] inputNames = line.substring(7).split(",");
		int noNames = inputNames.length;
		for(int i=0;i<noNames;i++)
		{
			inputNames[i] = inputNames[i].trim();
		}
		Hashtable<String,Integer> hash = new Hashtable<String,Integer>();
		for(int i=0;i<noNames;i++)
			hash.put(inputNames[i],0);

		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();

		int startIndex = -1;
		int endIndex = -1;
		String substr = "";
		String[] elements;
		String subtreeStr;
		while(line!=null && !line.equals(""))
		{
			startIndex=line.indexOf("(");
			while(startIndex>=0)
			{
				subtreeStr = "";
				endIndex = line.indexOf(")",startIndex);
				startIndex = line.lastIndexOf("(",endIndex)+1;
				substr = line.substring(startIndex,endIndex);
				elements = substr.split(" ");

				if(elements.length == 3)
				{
					if(elements[1].equals("*"))
					{
						boolean firstElementFound = false;
						for(int i=0;i<noNames;i++)
						{
							if(elements[0].equals(inputNames[i]))
							{
								boolean secondElementFound = false;
								for(int j=0;j<noNames;j++)
								{
									if(elements[2].equals(inputNames[j]))
									{
										if(i<=j)
											subtreeStr = substr;
										else
											subtreeStr = inputNames[j]+" * "+inputNames[i];
										secondElementFound = true;
										break;
									}
								}
								if(!secondElementFound)
									subtreeStr = elements[0];
								firstElementFound = true;
								break;
							}
						}
						if(!firstElementFound)
						{
							for(int i=0;i<noNames;i++)
							{
								if(elements[2].equals(inputNames[i]))
								{
									subtreeStr = elements[2];
									break;
								}
							}
						}
						if(subtreeStr!="")
							addToHash(hash,subtreeStr);
					}
					else if(elements[1].equals("/"))
					{
						boolean firstElementFound = false;
						for(int i=0;i<noNames;i++)
						{
							if(elements[0].equals(inputNames[i]))
							{
								boolean secondElementFound = false;
								for(int j=0;j<noNames;j++)
								{
									if(elements[2].equals(inputNames[j]))
									{
										if(i!=j)
											subtreeStr = substr;
											secondElementFound = true;
										break;
									}
								}
								if(!secondElementFound)
									subtreeStr = elements[0];
								firstElementFound = true;
								break;
							}
						}
						if(!firstElementFound)
						{
							for(int i=0;i<noNames;i++)
							{
								if(elements[2].equals(inputNames[i]))
								{
									subtreeStr = "1 / "+elements[2];
									break;
								}
							}
						}
						if(subtreeStr!="")
							addToHash(hash,subtreeStr);
					}
					else if(elements[1].equals("+")||elements[1].equals("-"))
					{
						boolean firstElementFound = false;
						boolean secondElementFound = false;
						for(int i=0;i<noNames;i++)
						{
							if(elements[0].equals(inputNames[i]))
							{
								subtreeStr = elements[0];
								addToHash(hash,subtreeStr);
								firstElementFound = true;
							}
							if(elements[2].equals(inputNames[i]))
							{
								if(!elements[2].equals(elements[0]))
								{
									subtreeStr = elements[2];
									addToHash(hash,subtreeStr);
								}
								secondElementFound = true;
							}
							if(firstElementFound && secondElementFound)
								break;
						}
					}
				}

				startIndex = line.indexOf("(",endIndex);
			}

			line = reader.readLine();
		}
		return hash;
	}

	private static void addToHash(Hashtable<String,Integer> hash, String entry)
	{
		if(hash.containsKey(entry))
			hash.put(entry,hash.get(entry)+1);
		else
			hash.put(entry,1);
	}

	/**Returns the Function use within a population, as represented in a results file. Each key within the Hashtable
	is a Function name. The corresponding element is the average number of times that Function is used per Individual.*/
	public static Hashtable<String,Double> getFunctionUseFromFile(String filename) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("FUNCTIONS"));
		String[] functionNames = line.substring(10).split(",");
		int noNames = functionNames.length;
		for(int i=0;i<noNames;i++)
		{
			functionNames[i] = functionNames[i].trim();
		}
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();
		int count = 0;
		double[] functionCount = new double[noNames];
		while(line!=null && !line.equals(""))
		{
			count++;
			String[] items = line.split("[,()]");
			for(int i=0;i<items.length;i++)
			{
				for(int j=0;j<noNames;j++)
				{
					if(items[i].trim().equals(functionNames[j]))
					{
						functionCount[j]++;
						break;
					}
				}
			}
			line = reader.readLine();
		}
		Hashtable<String,Double> hash = new Hashtable<String,Double>();
		for(int i=0;i<noNames;i++)
			hash.put(functionNames[i],functionCount[i] / count);
		return hash;
	}

	/**This method returns the lengths of all Individuals within a Population, as represented within a results file.
	An example of operator string is "+*-/". It is required so that the splits between nodes can be identified*/
	public static int[] getProgramLengthsFromFile(String filename, String operatorString) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();
		Vector<Integer> vecLengths = new Vector<Integer>();
		while(line!=null && !line.equals(""))
		{
			int cutoff = line.indexOf(", training fitness");
			if(cutoff == -1)
			{
				line = reader.readLine();
				continue;
			}
			String[] items = line.substring(0,cutoff).split("[,()"+operatorString+"]");
			int len = 0;
			for(int i=0;i<items.length;i++)
			{
				if(!(items[i].trim().equals("")))
					len++;
			}
			items = line.substring(0,cutoff).split("["+operatorString+"]");
			len+=items.length-1;
			vecLengths.add(len);
			line = reader.readLine();
		}
		int[] lengths = new int[vecLengths.size()];
		for(int i=0;i<vecLengths.size();i++)
			lengths[i] = vecLengths.get(i).intValue();
		return lengths;
	}

	/**This method returns the length of the best Individual within a Population, as represented within a results file.
	An example of operator string is "+*-/". It is required so that the splits between nodes can be identified*/
	public static int getBestProgramLengthFromFile(String filename, String operatorString, boolean maximiseFitness) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
		String line="";
		do
		{
			line = reader.readLine();
		}
		while(!line.startsWith("BEST"));
		line = reader.readLine();
		Vector<Integer> vecLengths = new Vector<Integer>();
		String[] bestLine = null;
		double bestFitness = Double.MAX_VALUE;
		if(maximiseFitness)
			bestFitness = -Double.MAX_VALUE;
		while(line!=null && !line.equals(""))
		{
			int cutoff = line.indexOf(", training fitness");
			if(cutoff == -1)
			{
				line = reader.readLine();
				continue;
			}
			String[] splitLine = line.split(", training fitness");
			double valFitness = Double.parseDouble(splitLine[1].split(",")[1].split("=")[1]);
			if(maximiseFitness && valFitness > bestFitness)
			{
				bestFitness = valFitness;
				bestLine = splitLine;
			}
			else if(!maximiseFitness && valFitness < bestFitness)
			{
				bestFitness = valFitness;
				bestLine = splitLine;
			}
			line = reader.readLine();
		}
		String[] items = bestLine[0].split("[,()"+operatorString+"]");
		int len = 0;
		for(int i=0;i<items.length;i++)
		{
			if(!(items[i].trim().equals("")))
				len++;
		}
		items = bestLine[0].split("["+operatorString+"]");
		len+=items.length-1;
		return len;
	}

	/**This method returns the tree diversity and the fitness diversity for a population. Both values are in the range [0,1].*/
	public static double[] getDiversitiesFromFile(String filename) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(filename));
		String line = reader.readLine();
		while(!line.startsWith("BEST"))
		{
			line = reader.readLine();
		}
		int popSize = Integer.parseInt((line.split(" "))[1]);
		String[] allLines = new String[popSize];
		String[] allFitnesses = new String[popSize];

		if(popSize == 1)
			return new double[]{0.0,0.0};
		boolean[]treeDuplicated = new boolean[popSize];
		boolean[]fitnessDuplicated = new boolean[popSize];
		int treeDiversity=popSize;
		int fitnessDiversity=popSize;
		for(int i=0;i<popSize;i++)
		{
			allLines[i] = reader.readLine();
			try
			{
				allFitnesses[i]= allLines[i].substring(allLines[i].indexOf("training fitness"));
			}
			catch(NullPointerException npe)//in case of error in file
			{
				System.err.println("PROBLEM WITH FILE: "+filename);
				return null;
			}
		}
		for(int i=0;i<popSize-1;i++)
		{
			for(int j=i+1;j<popSize;j++)
			{
				if(!treeDuplicated[i] && !treeDuplicated[j] && allLines[i].equals(allLines[j]))
				{
					treeDuplicated[j]=true;
					treeDiversity--;
				}
				if(!fitnessDuplicated[i] && !fitnessDuplicated[j] && allFitnesses[i].equals(allFitnesses[j]))
				{
					fitnessDuplicated[j]=true;
					fitnessDiversity--;
				}
			}
		}
		return new double[]{(treeDiversity-1.0)/(popSize-1.0), (fitnessDiversity-1.0)/(popSize-1.0)};
	}

	/**Returns the correlation coefficient (R^2) for 2 variables.*/
	public static double getCorrelation(double[][] data)
	{
		boolean transposeMatrix = false;
		if(data[0].length!=2)
		{
			if(data.length!=2)
			{
				System.err.println("ERROR: getCorrelation requires an input array with 2 as one of its dimensions.");
				return Double.NaN;
			}
			else
				transposeMatrix = true;
		}
		int len = data.length;
		if(transposeMatrix)
			len = data[0].length;
		double x = 0.0;
		double y = 0.0;
		double xy =0.0;
		double xSquared = 0.0;
		double ySquared = 0.0;
		int invalidDataCount = 0;
		for(int i=0;i<len;i++)
		{
			if(transposeMatrix)
			{
				if(Double.isInfinite(data[0][i])||Double.isInfinite(data[1][i])||Double.isNaN(data[0][i])||Double.isNaN(data[1][i]))
				{
					invalidDataCount++;
					continue;
				}
				x += data[0][i];
				y += data[1][i];
				xy += data[0][i] * data[1][i];
				xSquared += Math.pow(data[0][i],2.0);
				ySquared += Math.pow(data[1][i],2.0);
			}
			else
			{
				if(Double.isInfinite(data[i][0])||Double.isInfinite(data[i][1])||Double.isNaN(data[i][0])||Double.isNaN(data[i][1]))
				{
					invalidDataCount++;
					continue;
				}
				x += data[i][0];
				y += data[i][1];
				xy += data[i][0] * data[i][1];
				xSquared += Math.pow(data[i][0],2.0);
				ySquared += Math.pow(data[i][1],2.0);
			}
		}
		len -= invalidDataCount;
		double numerator = len * xy - x * y;
		double denominator = Math.sqrt(len * xSquared - Math.pow(x, 2.0)) * Math.sqrt(len * ySquared - Math.pow(y, 2.0));
		return numerator / denominator;
	}

	/**Returns the required ScalarData for all Individuals within the Population*/
	public double[] getScalarData(ScalarData dataType)
	{
		int popSize = pop.getPopulationSize();
		double[] data = new double[popSize];
		Individual[] inds = pop.getIndividuals();
		switch(dataType)
		{
			case PROGRAM_LENGTH:
				for(int i=0;i<popSize;i++)
					data[i] = inds[i].getLength();
				break;
			case PROGRAM_DEPTH:
				for(int i=0;i<popSize;i++)
					data[i] = inds[i].getDepth();
				break;
			case PROGRAM_FITNESS:
				double[] fitnesses = pop.getFitnesses();
				for(int i=0;i<popSize;i++)
					data[i] = fitnesses[i];
				break;
			case PROGRAM_DIVERSITY_DISTANCE:
				for(int i=0;i<popSize;i++)
					data[i] = pop.getAverageSquaredDistance(i);
				break;
		}
		return data;
	}

	/**Returns the required ScalarData for all Individuals within the Population.
	<i>orderByIndex</i> determines how the individuals will be sorted before obtaining the ScalarData. If set to less than
	0 or greater than or equal to dataType length the individuals will be left in their original order.*/
	public double[][] getScalarData(ScalarData[] dataType, int orderByIndex)
	{
		int popSize = pop.getPopulationSize();
		int dim = dataType.length;
		double[][] data = new double[popSize][dim];
		Individual[] inds = null;
		boolean ordered = false;
		boolean orderAscending = true;
		if(orderByIndex<0 || orderByIndex>=dim)
			inds = pop.getIndividuals();
		else
		{
			double[] orderingData = new double[popSize];
			switch(dataType[orderByIndex])
			{
				case PROGRAM_LENGTH:
					for(int i=0;i<popSize;i++)
						orderingData[i] = inds[i].getLength();
					break;
				case PROGRAM_DEPTH:
					for(int i=0;i<popSize;i++)
						orderingData[i] = inds[i].getDepth();
					break;
				case PROGRAM_FITNESS:
					double[] fitnesses = pop.getFitnesses();
					for(int i=0;i<popSize;i++)
						orderingData[i] = fitnesses[i];
					Population.FitnessFunction fitnessFunc = pop.getFitnessFunction();
					if(fitnessFunc.equals(Population.FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunc.equals(Population.FitnessFunction.AUROC))
						orderAscending = false;
					break;
				case PROGRAM_DIVERSITY_DISTANCE:
					for(int i=0;i<popSize;i++)
						orderingData[i] = pop.getAverageSquaredDistance(i);
					orderAscending = false;
				break;
			}
			Object[] sortedObj = pop.ShellSortIndividuals(orderingData, orderAscending);
			inds = (Individual[])(sortedObj[0]);
			double[] orderedData = (double[])(sortedObj[1]);
			for(int i=0;i<popSize;i++)
				data[i][orderByIndex] = orderedData[i];
			ordered = true;
		}

		for(int j=0;j<dim;j++)
		{
			if(j == orderByIndex)//this data has already been set
				continue;
			switch(dataType[j])
			{
				case PROGRAM_LENGTH:
					for(int i=0;i<popSize;i++)
						data[i][j] = inds[i].getLength();
					break;
				case PROGRAM_DEPTH:
					for(int i=0;i<popSize;i++)
						data[i][j] = inds[i].getDepth();
					break;
				case PROGRAM_FITNESS:
					double[] fitnesses = pop.getFitnesses();
					for(int i=0;i<popSize;i++)
						data[i][j] = fitnesses[i];
					break;
				case PROGRAM_DIVERSITY_DISTANCE:
					for(int i=0;i<popSize;i++)
						data[i][j] = pop.getAverageSquaredDistance(i);
					break;
			}
		}
		return data;
	}

	/**Returns the Functions available to members of the Population*/
	public Function[] getFunctions()
	{
		PopulationDescriptor desc = pop.getPopulationDescriptor();
		return desc.getFunctions();
	}

	/**Returns the Inputs available to members of the Population*/
	public Input[] getInputs()
	{
		PopulationDescriptor desc = pop.getPopulationDescriptor();
		return desc.getInputs();
	}

	/**Returns the Constants available to members of the Population*/
	public Constant[] getConstants()
	{
		PopulationDescriptor desc = pop.getPopulationDescriptor();
		Constant[] constants = desc.getConstants();
		int constLen = constants.length;
		Vector<Double> mutatedConstants = desc.getMutatedConstants();
		int mutatedConstLen = mutatedConstants.size();
		Constant[] allConstants = new Constant[constLen + mutatedConstLen];
		for(int i=0;i<constLen;i++)
			allConstants[i] = constants[i];
		for(int i=0;i<mutatedConstLen;i++)
			allConstants[constLen + i] = new Constant(mutatedConstants.get(i));
		return allConstants;
	}

	/**Returns the Components available to members of the Population*/
	public Component[] getComponents()
	{
		PopulationDescriptor desc = pop.getPopulationDescriptor();
		Function[] funcs = desc.getFunctions();
		int funcLen = funcs.length;
		Input[] inputs = desc.getInputs();
		int inputLen = inputs.length;
		Constant[] constants = desc.getConstants();
		int constLen = constants.length;
		Vector<Double> mutatedConstants = desc.getMutatedConstants();
		int mutatedConstLen = mutatedConstants.size();
		Component[] allComps = new Component[funcLen + inputLen + constLen + mutatedConstLen];
		for(int i=0;i<funcLen;i++)
			allComps[i] = funcs[i];
		for(int i=0;i<inputLen;i++)
			allComps[funcLen + i] = inputs[i];
		for(int i=0;i<constLen;i++)
			allComps[funcLen + inputLen + i] = constants[i];
		for(int i=0;i<mutatedConstLen;i++)
			allComps[funcLen + inputLen + constLen + i] = new Constant(mutatedConstants.get(i));
		return allComps;
	}

	/** Returns the requested VectorData for all Individuals within the Population.
	The ScalarData <i>sortBy</i> determines the order in which this VectorData will be returned.
	If sortBy is null, the data will be returned in the same order as the individuals in the population.*/
	public double[][] getVectorData(VectorData dataType, ScalarData sortBy)
	{
		Individual[] inds = pop.getIndividuals();
		if(sortBy == null)
			return getVectorData(inds, dataType);
		int popSize = inds.length;
		double[] orderingData = new double[popSize];
		boolean orderAscending = true;
		switch(sortBy)
		{
			case PROGRAM_LENGTH:
				for(int i=0;i<popSize;i++)
					orderingData[i] = inds[i].getLength();
				break;
			case PROGRAM_DEPTH:
				for(int i=0;i<popSize;i++)
					orderingData[i] = inds[i].getDepth();
				break;
			case PROGRAM_FITNESS:
				double[] fitnesses = pop.getFitnesses();
				for(int i=0;i<popSize;i++)
					orderingData[i] = fitnesses[i];
				Population.FitnessFunction fitnessFunc = pop.getFitnessFunction();
				if(fitnessFunc.equals(Population.FitnessFunction.NUMBER_CORRECT_BINARY)||fitnessFunc.equals(Population.FitnessFunction.AUROC))
					orderAscending = false;
				break;
			case PROGRAM_DIVERSITY_DISTANCE:
				for(int i=0;i<popSize;i++)
					orderingData[i] = pop.getAverageSquaredDistance(i);
				orderAscending = false;
			break;
		}
		Object[] sortedObj = pop.ShellSortIndividuals(orderingData, orderAscending);
		inds = (Individual[])(sortedObj[0]);
		return getVectorData(inds, dataType);
	}

	/**This method returns data for the Individuals in the same order as in the population.
	@see #getVectorData(VectorData, ScalarData)*/
	public double[][] getVectorData(VectorData dataType)
	{
		return getVectorData(pop.getIndividuals(), dataType);
	}

	/**Returns a Poisson distribution as a set of probabilities of values between <i>min</i> and <i>max</i>.
	@param lambda The expected value*/
	public static double[] getPoissonDistribution(double lambda, int min, int max)
	{
		if(min>max)
			return null;
		else if(min==max)
			return new double[]{1.0};
		else
		{
			double[] distrib = new double[max-min+1];
			if(min<=0||max>20)
			{
				for(int i=min;i<=max;i++)
					distrib[i] = 1.0 / (max-min+1);//factorial greater than 20 cannot be held in a long. Therefore use a flat distribution
			}
			else
			{
				double total = 0.0;
				for(int i=min;i<=max;i++)
				{
					distrib[i-min] = Math.exp(-lambda) * Math.pow(lambda,i)/factorials[i];
					total += distrib[i-min];
				}
				for(int i=min;i<=max;i++)
				{
					distrib[i-min] /= total;
				}
			}
			return distrib;
		}
	}

	//Obtains Vector data for all members of the array inds
	private double[][] getVectorData(Individual[] inds, VectorData dataType)
	{
		int popSize = inds.length;
		double[][] data = null;
		PopulationDescriptor desc = pop.getPopulationDescriptor();
		int len;
		Component[] comps = null;
		switch(dataType)
		{
			case FUNCTION_USE:
				comps = desc.getFunctions();
				len = comps.length;
				data = new double[popSize][len];
				for(int i=0;i<popSize;i++)
				{
					Vector<Component> nodes = inds[i].getNodes();
					for(int j=0;j<nodes.size();j++)
					{
						Component node = nodes.get(j);////
						for(int k=0;k<len;k++)
						{
							if(node.equals(comps[k]))
							{
								data[i][k]++;
								break;
							}
						}
					}
				}
				break;
			case INPUT_USE:
				comps = desc.getInputs();
				len = comps.length;
				data = new double[popSize][len];
				for(int i=0;i<popSize;i++)
				{
					Vector<Component> nodes = inds[i].getNodes();
					for(int j=0;j<nodes.size();j++)
					{
						Component node = nodes.get(j);
						for(int k=0;k<len;k++)
						{
							if(node.equals(comps[k]))
							{
								data[i][k]++;
								break;
							}
						}
					}
				}
				break;
			case CONSTANT_USE:
				comps = desc.getConstants();
				Vector<Double> mutatedConstants = desc.getMutatedConstants();
				int originalLen = comps.length;
				int mutatedLen = mutatedConstants.size();
				data = new double[popSize][originalLen + mutatedLen];
				for(int i=0;i<popSize;i++)
				{
					Vector<Component> nodes = inds[i].getNodes();

					nextNode:
					for(int j=0;j<nodes.size();j++)
					{
						Component node = nodes.get(j);
						for(int k=0;k<originalLen;k++)
						{
							if(node.equals(comps[k]))
							{
								data[i][k]++;
								continue nextNode;
							}
						}
						for(int k=0;k<mutatedLen;k++)
						{
							if(node.equals(mutatedConstants.get(k)))
							{
								data[i][originalLen + k]++;
								continue nextNode;
							}
						}
					}
				}
				break;
			case COMPONENT_USE:
				Function[] functions = desc.getFunctions();
				int funcLen = functions.length;
				Input[] inputs = desc.getInputs();
				int inputLen = inputs.length;

				Constant[] constants = desc.getConstants();
				Vector<Double> mutatedConsts = desc.getMutatedConstants();
				int originalConstLen = constants.length;
				int mutatedConstLen = mutatedConsts.size();
				data = new double[popSize][funcLen + inputLen + originalConstLen + mutatedConstLen];
				for(int i=0;i<popSize;i++)
				{
					Vector<Component> nodes = inds[i].getNodes();

					nextNode:
					for(int j=0;j<nodes.size();j++)
					{
						Component node = nodes.get(j);
						for(int k=0;k<funcLen;k++)
						{
							if(node.equals(functions[k]))
							{
								data[i][k]++;
								continue nextNode;
							}
						}
						for(int k=0;k<inputLen;k++)
						{
							if(node.equals(inputs[k]))
							{
								data[i][funcLen + k]++;
								continue nextNode;
							}
						}
						for(int k=0;k<originalConstLen;k++)
						{
							if(node.equals(constants[k]))
							{
								data[i][funcLen + inputLen + k]++;
								continue nextNode;
							}
						}
						for(int k=0;k<mutatedConstLen;k++)
						{
							if(node.equals(mutatedConsts.get(k)))
							{
								data[i][funcLen + inputLen + originalConstLen + k]++;
								continue nextNode;
							}
						}
					}
				}
				break;
		}
		return data;
	}
}