package geneticWedge.gp.io;

import java.io.*;
import java.util.Vector;
import geneticWedge.gp.*;

/**This class contains static methods for writing data to csv files*/
public class CSVWriter
{
	/**Writes data from a NumericData object to a csv file. The data will be 'normalised', i.e. converted using a linear transformation to
	the range [minNormalisedValue, maxNormalisedValue]. Variable names and datum labels will be automatically inserted if they are 'known' to the NumericData object.
	@param randomiseOrder If true, the order of the data will be randomly permuted.*/
	public static void writeFile(String filename, NumericData numericData, boolean randomiseOrder, double minNormalisedValue, double maxNormalisedValue) throws IOException
	{
		String[] names = numericData.getNames();
		String[] labels = numericData.getLabels();
		double[][] data = numericData.normaliseData(minNormalisedValue, maxNormalisedValue);
		writeFile(filename, names, labels, data, randomiseOrder);
	}

	/**Writes data from a NumericData object to a csv file. Variable names and datum labels will be automatically inserted if they are 'known' to the NumericData object.
	@param randomiseOrder If true, the order of the data will be randomly permuted.*/
	public static void writeFile(String filename, NumericData numericData, boolean randomiseOrder) throws IOException
	{
		String[] names = numericData.getNames();
		String[] labels = numericData.getLabels();
		double[][] data = numericData.getData();
		writeFile(filename, names, labels, data, randomiseOrder);
	}

	/**Writes data to a csv file. Variable names and datum labels will be inserted appropriately.
	@param randomiseOrder If true, the order of the data will be randomly permuted.*/
	public static void writeFile(String filename, String[] names, String[] labels, double[][]data, boolean randomiseOrder) throws IOException
	{
		int len = data.length;
		int[] randomIndices = Utils.getRandomIndices(len);
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
		String line = "";
		if(names!=null)
		{
			for(int i=0;i<names.length;i++)
				line+=","+names[i];
			writer.write(line);
			writer.newLine();
		}
		for(int i=0;i<len;i++)
		{
			line = "";
			if(labels != null)
			{
				if(randomiseOrder)
					line=labels[randomIndices[i]]+",";
				else
					line=labels[i]+",";
			}
			if(randomiseOrder)
				line+=data[randomIndices[i]][0];
			else
				line+=data[i][0];
			for(int j=1;j<data[i].length;j++)
			{
				if(randomiseOrder)
					line+=","+data[randomIndices[i]][j];
				else
					line+=","+data[i][j];
			}
			writer.write(line);
			writer.newLine();
		}
		writer.flush();
		writer.close();
	}

	/**Writes data to a csv file, normalising it to the range [minNormalisedValue,maxNormalisedValue].
	Variable names and datum labels will be inserted appropriately.
	@param randomiseOrder If true, the order of the data will be randomly permuted.*/
	public static void writeFile(String filename, String[] names, String[] labels, NumericData numericData , boolean randomiseOrder, double minNormalisedValue, double maxNormalisedValue) throws IOException
	{
		double[][]data = numericData.normaliseData(minNormalisedValue, maxNormalisedValue);
		int len = data.length;
		int[] randomIndices = Utils.getRandomIndices(len);
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
		String line = "";
		if(names!=null)
		{
			for(int i=0;i<names.length;i++)
				line+=","+names[i];
			writer.write(line);
			writer.newLine();
		}
		for(int i=0;i<len;i++)
		{
			line = "";
			if(labels != null)
			{
				if(randomiseOrder)
					line=labels[randomIndices[i]]+",";
				else
					line=labels[i]+",";
			}
			if(randomiseOrder)
				line+=data[randomIndices[i]][0];
			else
				line+=data[i][0];
			for(int j=1;j<data[i].length;j++)
			{
				if(randomiseOrder)
					line+=","+data[randomIndices[i]][j];
				else
					line+=","+data[i][j];
			}
			writer.write(line);
			writer.newLine();
		}
		writer.flush();
		writer.close();
	}
}