package geneticWedge.gp.io;

/**This class holds numeric data as well as variable names and data labels. It can perform operations on the data
that it holds, such as normalisation*/
public class NumericData
{
	private String[] names;
	private double[][] data;
	private String[] labels;
	private double[][] normFactors;
	private int dimensionality;
	private int dataLength;

	/**constructs a NumericData object with data, variable names and data labels.*/
	public NumericData(String[] names, double[][]data, String[] labels)
	{
		this.names=names;
		this.data=data;
		this.labels = labels;
		dataLength=data.length;
		dimensionality=data[0].length;
	}

	/**Returns the data held by this NumericData object.*/
	public double[][] getData()
	{
		return data;
	}

	/**Returns selected data held by this NumericData object.*/
	public double[][] getData(int[]variableIndices)
	{
		int dim = variableIndices.length;
		double[][] cutdownData = new double[dataLength][dim];
		for(int i=0;i<dataLength;i++)
			for(int j=0;j<dim;j++)
				cutdownData[i][j] = data[i][variableIndices[j]];
		return cutdownData;
	}

	/**Returns the length of the data.*/
	public int getDataLength()
	{
		return dataLength;
	}

	/**Returns the dimensionality of the data.*/
	public int getDimensionality()
	{
		return dimensionality;
	}

	/**Returns the variable names held by this NumericData object. If no names have been set this method will return null.*/
	public String[] getNames()
	{
		return names;
	}

	/**Returns the data 'labels' held by this NumericData object.
	If no labels have been set this method will return null.*/
	public String[] getLabels()
	{
		return labels;
	}

	/**Returns the data held by this object, normalised to the range [-1,1]*/
	public double[][] normaliseData()
	{
		return normaliseData(-1.0,1.0);
	}

	/**Returns the data held by this object, normalised to the range [minVal,maxVal]*/
	public double[][] normaliseData(double minVal, double maxVal)
	{
		double[][] normData=new double[dataLength][dimensionality];
		normFactors=new double[dimensionality][2];
		for(int i=0;i<dimensionality;i++)
		{
			double max=-Double.MAX_VALUE;
			double min=Double.MAX_VALUE;
			for(int j=0;j<dataLength;j++)
			{
				if(data[j][i]>max)
					max=data[j][i];
				if(data[j][i]<min)
					min=data[j][i];
			}
			double multiplicationFactor=(maxVal-minVal)/(max-min);
			double subtractionFactor=((max+min)*multiplicationFactor-(maxVal+minVal))/2.0;
			normFactors[i]=new double[]{subtractionFactor,multiplicationFactor};
		}
		for(int i=0;i<dimensionality;i++)
			for(int j=0;j<dataLength;j++)
				//normData[j][i]=(data[j][i]-normFactors[i][0])*normFactors[i][1];
				normData[j][i]=data[j][i]*normFactors[i][1]-normFactors[i][0];

		return normData;
	}

	/**Returns the values of the data within the variables identified by <i>variableIndices</i>, normalised to the range [-1,1]*/
	public double[][] normaliseData(int[] variableIndices)
	{
		return normaliseData(-1.0,1.0, variableIndices);
	}

	/**Returns the values of the data within the variables identified by <i>variableIndices</i>, normalised to the range [minVal,maxVal]*/
	public double[][] normaliseData(double minVal, double maxVal, int[]variableIndices)
	{
		int dimensionality = variableIndices.length;
		double[][] normData=new double[dataLength][dimensionality];
		normFactors=new double[dimensionality][2];
		for(int i=0;i<dimensionality;i++)
		{
			double max=-Double.MAX_VALUE;
			double min=Double.MAX_VALUE;
			for(int j=0;j<dataLength;j++)
			{
				if(data[j][variableIndices[i]]>max)
					max=data[j][variableIndices[i]];
				if(data[j][variableIndices[i]]<min)
					min=data[j][variableIndices[i]];
			}
			double multiplicationFactor=(maxVal-minVal)/(max-min);
			double subtractionFactor=((max+min)*multiplicationFactor-(maxVal+minVal))/2.0;
			normFactors[i]=new double[]{subtractionFactor,multiplicationFactor};
		}
		for(int i=0;i<dimensionality;i++)
			for(int j=0;j<dataLength;j++)
				//normData[j][i]=(data[j][i]-normFactors[i][0])*normFactors[i][1];
				normData[j][i]=data[j][variableIndices[i]]*normFactors[i][1]-normFactors[i][0];

		return normData;
	}

	/**Normalises newData using the most recently used normalisation factors. WARNING: this method will throw an Exception
	if normalisation has not been preformed previously, or if newData does not have the same dimensionality as the data used to create this object.*/
	public double[][] normaliseData(double[][] newData)
	{
		int dataLength = newData.length;
		double[][] normData = new double[dataLength][dimensionality];
		for(int i=0;i<dimensionality;i++)
			for(int j=0;j<dataLength;j++)
				normData[j][i]=data[j][i]*normFactors[i][1]-normFactors[i][0];
		return normData;
	}

	/**This method uses the normalisation constants set as a side effect in normaliseData().
	The input data is assumed to have the same dimensionality as the original data. If this is
	not the case, or if normaliseData has not been called previously, results are unpredictable
	and Exceptions may be thrown.*/
	public double[][] denormaliseData(double[][] normData)
	{
		int dataLength=normData.length;
		int dimensionality = normData[0].length;
		double[][] denormData=new double[dataLength][dimensionality];
		for(int i=0;i<dataLength;i++)
			for(int j=0;j<dimensionality;j++)
			{
				//denormData[i][j]=normData[i][j] / normFactors[j][1] + normFactors[j][0];
				denormData[i][j] = (normData[i][j] + normFactors[j][0]) / normFactors[j][1];
			}
		return denormData;
	}

	/**This method uses the normalisation constants set as a side effect in normaliseData().
	If an incorrect index is provided or if normaliseData has not been called previously
	results are unpredictable and Exceptions may be thrown.*/
	public double[] denormaliseData(double[] normData, int index)
	{
		int dataLength=normData.length;
		double[] denormData=new double[dataLength];
		for(int i=0;i<dataLength;i++)
			denormData[i] = normData[i] / normFactors[i][1] + normFactors[i][0];
		return denormData;
	}
}