package geneticWedge.gp.io;

import geneticWedge.gp.*;
import geneticWedge.gp.function.*;
import geneticWedge.gp.dataAnalysis.DataAcquirer;

import java.io.*;

/**Class for writing standard results files. An evolved Population should be provided as the second argument to writeResults.
The other methods provide useful utilities for recording results in alternative formats. They are particularly useful
when a number of 'runs' have been performed and it is required that the results from different runs are stored in a single file.*/
public class ResultWriter
{
	private boolean writeFunction = false;

	/**Writes results of evolution for the Population provided. Performances on the training and validation data are recorded, as are the training parameters.*/
	public boolean writeResults(String filename, Population pop, StoppingCriterion stopCrit, boolean isSteadyState, int numberOfProgramsToReport, double[][]trainInputs, double[]trainTargets, double[][]valInputs, double[]valTargets)
	{
		return writeResults(filename, pop, stopCrit, isSteadyState, numberOfProgramsToReport, trainInputs, trainTargets, valInputs, valTargets, null, null);
	}

	/**Sets the format in which functions will be displayed.
	@param useFunction if true, this ResultWriter will write GP trees (Individuals) as mathematical functions, if false they will be written as comma-sparated lists*/
	public void setFunctionFormat(boolean useFunction)
	{
		writeFunction = useFunction;
	}

	/**Writes results of evolution for the Population provided. Performances on the training, validation and test dataare recorded, as are the training parameters.*/
	public boolean writeResults(String filename, Population pop, StoppingCriterion stopCrit, boolean isSteadyState, int numberOfProgramsToReport,double[][]trainInputs, double[]trainTargets, double[][]valInputs, double[]valTargets, double[][]testInputs, double[]testTargets)
	{
		try
		{
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
			//BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(filename),"UTF-8"));
			if(pop instanceof DemicPopulation)
				writer.write("DEMIC POPULATION");
			else
				writer.write("STANDARD POPULATION");
			writer.newLine();
			PopulationDescriptor desc=pop.getPopulationDescriptor();
			String line="CONTROL PARAMETERS";
			writer.write(line);
			writer.newLine();
			line="------------------";
			writer.write(line);
			writer.newLine();
			writer.newLine();

			line = "GROW METHOD: " +desc.getGrowMethod().toString();
			writer.write(line);
			writer.newLine();

			line = "FUNCTIONS: ";
			Function[] funcs=desc.getFunctions();
			if(funcs.length==0)
				line+="NONE";
			else
			{
				line+=funcs[0];
				for(int i=1;i<funcs.length;i++)
					line+=", "+funcs[i];
			}
			writer.write(line);
			writer.newLine();

			line = "INPUTS: ";
			Input[] inputs=desc.getInputs();
			if(inputs.length==0)
				line+="NONE";
			else
			{
				line+=inputs[0];
				for(int i=1;i<inputs.length;i++)
					line+=", "+inputs[i];
			}
			writer.write(line);
			writer.newLine();

			line = "CONSTANTS: ";
			Constant[] consts=desc.getConstants();
			if(consts.length==0)
				line+="NONE";
			else
			{
				line+=consts[0];
				for(int i=1;i<consts.length;i++)
					line+=", "+consts[i];
			}
			writer.write(line);
			writer.newLine();

			line="FITNESS MEASURE: "+pop.getFitnessFunction();
			writer.write(line);
			writer.newLine();

			Population.PerformanceFunction[] perfFuncs = pop.getPerformanceFunctions();
			if(perfFuncs!=null&&perfFuncs.length>0)
			{
				line = "PERFORMANCE MEASURES: " + perfFuncs[0];
				for(int i=1;i<perfFuncs.length;i++)
					line+=", "+perfFuncs[i];
				writer.write(line);
				writer.newLine();
				if(pop.getMultiObjective())
					writer.write("Performance function is multi-objective.");
				else
					writer.write("Performance function is single-objective.");
				writer.newLine();
			}

			FitnessShare fitnessShare = pop.getFitnessShare();
			if(fitnessShare != null)
			{
				writer.write("FITNESS SHARE: " + fitnessShare + ", share type: "+pop.getShareType());
				writer.newLine();
			}

			line="SELECTION MECHANISM: ";
			if(isSteadyState)
				line+="Steady state. Tournament size = "+pop.getTournamentSize();
			else
				line+="Generational. Elite size = "+desc.getEliteSize();
			writer.write(line);
			writer.newLine();

			line="MUTATION RATE: "+desc.getMutationProportion();
			writer.write(line);
			writer.newLine();

			Mutation.MutationOperator[] mutationOps = pop.getMutation().getMutationOperators();
			if(mutationOps!=null && mutationOps.length>0)
			{
				line = "MUTATION OPERATORS: "+mutationOps[0];
				for(int i=1;i<mutationOps.length;i++)
					line += ", "+mutationOps[i];
				writer.write(line);
				writer.newLine();
			}

			line="ASEXUAL REPRODUCTION RATE: "+desc.getReproductionProportion();
			writer.write(line);
			writer.newLine();

			if(pop.getCrossover().getBroodUse())
			{
				writer.write("Crossover uses brood recombination.");
				writer.newLine();
			}

			line="STOPPING CRITERION: "+stopCrit;
			writer.write(line);
			writer.newLine();

			int popSize=pop.getPopulationSize();
			line="POPULATION SIZE: "+popSize;
			writer.write(line);
			writer.newLine();

			line="MAX GENERATIONS: "+pop.getMaxGenerations();
			writer.write(line);
			writer.newLine();

			line="MAX PROGRAM LENGTH: "+pop.getMaximumIndividualLength();
			writer.write(line);
			writer.newLine();

			line="MAX PROGRAM DEPTH: "+pop.getMaximumIndividualDepth();
			writer.write(line);
			writer.newLine();

			writer.newLine();
			line="RESULTS";
			writer.write(line);
			writer.newLine();
			line="-------";
			writer.write(line);
			writer.newLine();
			writer.newLine();

			writer.write("Number of generations: "+pop.getGenerationNumber());
			writer.newLine();

			writer.write("Average fitnesses:-");
			writer.newLine();
			writer.write("\tTraining: "+pop.getAverageTrainingFitness());
			writer.newLine();
			writer.write("\tValidation: "+pop.getAverageValidationFitness());
			writer.newLine();
			try
			{
				writer.write("\tTest: "+pop.getAverageFitness(testInputs,testTargets,true));
			}
			catch(InvalidInputSizeException e)
			{
				writer.write("INVALID TEST DATA");
			}

			writer.newLine();

			line = "training variance = ";
			if(trainTargets == null)
				line += " INVALID TRAIN DATA,";
			else
				line += DataAcquirer.getVariance(trainTargets)+",";

			line += "validation variance = ";
			if(valTargets == null)
				line += " INVALID VALIDATION DATA,";
			else
				line += DataAcquirer.getVariance(valTargets)+",";

			line += "test variance = ";
			if(testTargets == null)
				line += " INVALID TEST DATA";
			else
				line += DataAcquirer.getVariance(testTargets);
			writer.write(line);

			writer.newLine();

			if(numberOfProgramsToReport>popSize)
				numberOfProgramsToReport=popSize;
			if(numberOfProgramsToReport>0)
			{
				Object[] indsAndFitnesss = pop.ShellSortIndividuals();
				Individual[] orderedInds = (Individual[])(indsAndFitnesss[0]);
				double[] fitnesses = (double[])(indsAndFitnesss[1]);
				line="BEST "+numberOfProgramsToReport+" PROGRAMS:-";
				writer.write(line);
				writer.newLine();
				for(int i=0;i<numberOfProgramsToReport;i++)
				{
					if(writeFunction)
						line=orderedInds[i].toStringFunction() + ", training fitness = " + fitnesses[i];
					else
						line=orderedInds[i] + ", training fitness = " + fitnesses[i];

					try
					{
						if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
						{
							double fitness=0.0;
							switch(pop.getFitnessFunction())
							{
								case MSE:
									fitness=orderedInds[i].getMSE(valInputs,valTargets);
									break;
								case RMSE:
									fitness=orderedInds[i].getRMSE(valInputs,valTargets);
									break;
								case ABSOLUTE_ERROR:
									fitness=orderedInds[i].getAbsoluteError(valInputs,valTargets);
									break;
								case NUMBER_CORRECT_BINARY:
									fitness=orderedInds[i].getNumberCorrectBinary(valInputs,valTargets);
									break;
								case AUROC:
									fitness=orderedInds[i].getAUROC(valInputs,valTargets);
									break;
								case AOREC:
									fitness=orderedInds[i].getAOREC(valInputs,valTargets,pop.getRECFitnessFunction());
									break;
								case SENSITIVITY:
									fitness=orderedInds[i].getSensitivity(valInputs,valTargets);
									break;
								case SPECIFICITY:
									fitness=orderedInds[i].getSpecificity(valInputs,valTargets);
									break;
								case PPV:
									fitness=orderedInds[i].getPositivePredictiveValue(valInputs,valTargets);
									break;
								case NPV:
									fitness=orderedInds[i].getNegativePredictiveValue(valInputs,valTargets);
									break;
								case COMBINED_CLASSIFICATION_ERROR:
									fitness=orderedInds[i].getCombinedClassificationError(valInputs,valTargets,false, true, false);
									break;
							}
							line+=", validation fitness = "+fitness;
						}

					}
					catch(InvalidInputSizeException e)
					{
						line+=", INVALID VALIDATION DATA";
					}

					try
					{
						if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
						{
							double fitness=0.0;
							switch(pop.getFitnessFunction())
							{
								case MSE:
									fitness=orderedInds[i].getMSE(testInputs,testTargets);
									break;
								case RMSE:
									fitness=orderedInds[i].getRMSE(testInputs,testTargets);
									break;
								case ABSOLUTE_ERROR:
									fitness=orderedInds[i].getAbsoluteError(testInputs,testTargets);
									break;
								case NUMBER_CORRECT_BINARY:
									fitness=orderedInds[i].getNumberCorrectBinary(testInputs,testTargets);
									break;
								case AUROC:
									fitness=orderedInds[i].getAUROC(testInputs,testTargets);
									break;
								case AOREC:
									fitness=orderedInds[i].getAOREC(testInputs,testTargets,pop.getRECFitnessFunction());
									break;
								case SENSITIVITY:
									fitness=orderedInds[i].getSensitivity(valInputs,valTargets);
									break;
								case SPECIFICITY:
									fitness=orderedInds[i].getSpecificity(valInputs,valTargets);
									break;
								case PPV:
									fitness=orderedInds[i].getPositivePredictiveValue(valInputs,valTargets);
									break;
								case NPV:
									fitness=orderedInds[i].getNegativePredictiveValue(valInputs,valTargets);
									break;
							}
							line+=", test fitness = "+fitness;
						}

					}
					catch(InvalidInputSizeException e)
					{
						line+=", INVALID TEST DATA";
					}
					writer.write(line);

					try
					{
						switch(pop.getFitnessFunction())
						{
							case AUROC:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									//double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets, true);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, threshold, true, true);
									line = "Number of true targets correct = "+classifData[0] + " from "+ (classifData[0]+classifData[3]) +".";
									writer.write(line);
									writer.newLine();
									line = "Number of false targets correct = "+classifData[2] + " from "+ (classifData[2]+classifData[1]) +".";
									writer.write(line);

									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, threshold, false, false);
										line = "Number of true validation targets correct = "+valClassifData[0] + " from "+ (valClassifData[0]+valClassifData[3]) +".";
										writer.write(line);
										writer.newLine();
										line = "Number of false validation targets correct = "+valClassifData[2] + " from "+ (valClassifData[2]+valClassifData[1]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, threshold, false, false);
										line = "Number of true test targets correct = "+testClassifData[0] + " from "+ (testClassifData[0]+testClassifData[3]) +".";
										writer.write(line);
										writer.newLine();
										line = "Number of false test targets correct = "+testClassifData[2] + " from "+ (testClassifData[2]+testClassifData[1]) +".";
										writer.write(line);
									}

									/*double noCorrect = orderedInds[i].getNumberCorrectBinary(trainInputs,trainTargets,threshold);
									line = "Training correct classification rate: "+noCorrect*100.0/trainTargets.length + "%";
									writer.write(line);

									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										noCorrect = orderedInds[i].getNumberCorrectBinary(valInputs,valTargets,threshold);
										line = "Validation correct classification rate: "+noCorrect*100.0/valTargets.length + "%";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										noCorrect = orderedInds[i].getNumberCorrectBinary(valInputs,valTargets,threshold);
										line = "Test correct classification rate: "+noCorrect*100.0/testTargets.length + "%";
										writer.write(line);
									}*/
								}
								break;
							case SENSITIVITY:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									/*double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();*/
									//int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets,threshold);
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, 0.0, true, true);
									line = "Number train correct = "+classifData[0] + " from "+ (classifData[0]+classifData[3]) +".";
									writer.write(line);


									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, 0.0, false, false);
										line = "Number validation correct = "+valClassifData[0] + " from "+ (valClassifData[0]+valClassifData[3]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, 0.0, false, false);
										line = "Number test correct = "+testClassifData[0] + " from "+ (testClassifData[0]+testClassifData[3]) +".";
										writer.write(line);
									}
								}
								break;
							case SPECIFICITY:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									/*double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();*/
									//int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets,threshold);
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, 0.0, true, true);
									line = "Number train correct = "+classifData[2] + " from "+ (classifData[2]+classifData[1]) +".";
									writer.write(line);


									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, 0.0, false, false);
										line = "Number validation correct = "+valClassifData[2] + " from "+ (valClassifData[2]+valClassifData[1]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, 0.0, false, false);
										line = "Number test correct = "+testClassifData[2] + " from "+ (testClassifData[2]+testClassifData[1]) +".";
										writer.write(line);
									}
								}
								break;
							case PPV:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									/*double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();*/
									//int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets,threshold);
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, 0.0, true, true);
									line = "Number train correct = "+classifData[0] + " from "+ (classifData[0]+classifData[1]) +".";
									writer.write(line);


									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, 0.0, false, false);
										line = "Number validation correct = "+valClassifData[0] + " from "+ (valClassifData[0]+valClassifData[1]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, 0.0, false, false);
										line = "Number test correct = "+testClassifData[0] + " from "+ (testClassifData[0]+testClassifData[1]) +".";
										writer.write(line);
									}
								}
								break;
							case NPV:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									/*double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();*/
									//int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets,threshold);
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, 0.0, true, true);
									line = "Number train correct = "+classifData[2] + " from "+ (classifData[2]+classifData[3]) +".";
									writer.write(line);


									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, 0.0, false, false);
										line = "Number validation correct = "+valClassifData[2] + " from "+ (valClassifData[2]+valClassifData[3]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, 0.0, false, false);
										line = "Number test correct = "+testClassifData[2] + " from "+ (testClassifData[2]+testClassifData[3]) +".";
										writer.write(line);
									}
								}
								break;
							case COMBINED_CLASSIFICATION_ERROR:
								if(trainInputs!=null && trainTargets!=null)
								{
									writer.newLine();
									//double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets);
									double threshold = orderedInds[i].getBestROCThreshold(trainInputs,trainTargets,true);
									line = "Best threshold from ROC="+threshold;
									writer.write(line);
									writer.newLine();
									int[] classifData = orderedInds[i].getAllClassificationData(trainInputs,trainTargets, threshold, true, true);
									line = "Number of true targets correct = "+classifData[0] + " from "+ (classifData[0]+classifData[3]) +".";
									writer.write(line);
									writer.newLine();
									line = "Number of false targets correct = "+classifData[2] + " from "+ (classifData[2]+classifData[1]) +".";
									writer.write(line);

									if(valInputs!=null && valTargets!=null && valInputs.length==valTargets.length)
									{
										writer.newLine();
										int[] valClassifData = orderedInds[i].getAllClassificationData(valInputs,valTargets, threshold, false, false);
										line = "Number of true validation targets correct = "+valClassifData[0] + " from "+ (valClassifData[0]+valClassifData[3]) +".";
										writer.write(line);
										writer.newLine();
										line = "Number of false validation targets correct = "+valClassifData[2] + " from "+ (valClassifData[2]+valClassifData[1]) +".";
										writer.write(line);
									}
									if(testInputs!=null && testTargets!=null && testInputs.length==testTargets.length)
									{
										writer.newLine();
										int[] testClassifData = orderedInds[i].getAllClassificationData(testInputs,testTargets, threshold, false, false);
										line = "Number of true test targets correct = "+testClassifData[0] + " from "+ (testClassifData[0]+testClassifData[3]) +".";
										writer.write(line);
										writer.newLine();
										line = "Number of false test targets correct = "+testClassifData[2] + " from "+ (testClassifData[2]+testClassifData[1]) +".";
										writer.write(line);
									}
								}
								break;
						}
					}
					catch(InvalidInputSizeException iise)
					{
						writer.write("Exception while obtaining classification accuracies.");
					}
					writer.newLine();
				}
			}
			writer.flush();
			writer.close();
			return true;
		}
		catch(IOException ioe)
		{
			System.err.println("Problem writing results.");
			ioe.printStackTrace();
			return false;
		}
	}

	/**This method writes the outputs resulting from training, validation and test data to file.
	The outputs may be any Object since all Objects have a toString method.*/
	public boolean writeOutputs(String filename, String heading, Object[][] trainOutputs, Object[][] validationOutputs, Object[][] testOutputs)
	{
		try
		{
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
			writer.write(heading);
			writer.newLine();
			if(trainOutputs!=null)
			{
				writer.newLine();
				writer.write("TRAINING DATA");
				writer.newLine();
				for(int i=0;i<trainOutputs.length;i++)
				{
					String line = trainOutputs[i][0].toString();
					for(int j=1;j<trainOutputs[i].length;j++)
						line+=","+trainOutputs[i][j];
					writer.write(line);
					writer.newLine();
				}
			}
			if(validationOutputs!=null)
			{
				writer.newLine();
				writer.write("VALIDATION DATA");
				writer.newLine();
				for(int i=0;i<validationOutputs.length;i++)
				{
					String line = validationOutputs[i][0].toString();
					for(int j=1;j<validationOutputs[i].length;j++)
						line+=","+validationOutputs[i][j];
					writer.write(line);
					writer.newLine();
				}
			}
			if(testOutputs!=null)
			{
				writer.newLine();
				writer.write("TEST DATA");
				writer.newLine();
				for(int i=0;i<testOutputs.length;i++)
				{
					String line = testOutputs[i][0].toString();
					for(int j=1;j<testOutputs[i].length;j++)
						line+=","+testOutputs[i][j];
					writer.write(line);
					writer.newLine();
				}
			}
			writer.flush();
			writer.close();
			return true;
		}
		catch(IOException ioe)
		{
			System.err.println("Problem writing outputs.");
			ioe.printStackTrace();
			return false;
		}
	}

	/**This method writes an array of data to a file. It is useful for saving result summaries.*/
	public boolean writeSummaryInfo(String filename, String heading, double[] summaryData)
	{
		try
		{
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
			writer.write(heading);
			writer.newLine();
			String line = "";
			if(summaryData.length>0)
				line+=summaryData[0];
			for(int i=1;i<summaryData.length;i++)
				line+=","+summaryData[i];
			writer.write(line);
			writer.flush();
			writer.close();
			return true;
		}
		catch(IOException ioe)
		{
			System.err.println("Problem writing summary.");
			ioe.printStackTrace();
			return false;
		}
	}

	/**This method writes a 2-D array of data to a file. It is useful for saving result summaries.*/
	public boolean writeSummaryInfo(String filename, String heading, double[][] summaryData)
	{
		try
		{
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
			writer.write(heading);
			writer.newLine();
			for(int i=0;i<summaryData.length;i++)
			{
				String line = "";
				if(summaryData[i].length>0)
					line+=summaryData[i][0];
				for(int j=1;j<summaryData[i].length;j++)
					line+=","+summaryData[i][j];
				writer.write(line);
				writer.newLine();
			}
			writer.flush();
			writer.close();
			return true;
		}
		catch(IOException ioe)
		{
			System.err.println("Problem writing summary.");
			ioe.printStackTrace();
			return false;
		}
	}

	/**This method writes a 3-D array of data to a file. It is useful for saving result summaries.
	It is assumed that the data is made up of 2-D data for each of a number of different 'runs'.*/
	public boolean writeSummaryInfo(String filename, String heading, double[][][] summaryData)
	{
		try
		{
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(filename)));
			writer.write(heading);
			writer.newLine();
			for(int i=0;i<summaryData.length;i++)
			{
				writer.write("run "+(i+1));
				writer.newLine();
				for(int j=0;j<summaryData[i].length;j++)
				{
					String line = ""+(summaryData[i][j][0]);
					for(int k=1;k<summaryData[i][j].length;k++)
						line+=","+summaryData[i][j][k];
					writer.write(line);
					writer.newLine();
				}
			}
			writer.flush();
			writer.close();
			return true;
		}
		catch(IOException ioe)
		{
			System.err.println("Problem writing summary.");
			ioe.printStackTrace();
			return false;
		}
	}
}