package geneticWedge.gp.io;

import java.io.*;
import java.util.Vector;

import geneticWedge.gp.Utils;

/**This class contains static methods for reading files and returning StringData or NumericData objects.
It also contains utility methods to randomly order data and to remove duplicates from datasets.*/
public class TextReader
{
	/**This method reads data from a file. Separate data are identified via a separator character(s).
	If the first row of the csv file contains variable names,this will be automatically detected by the reader and
	the variable names may be retrieved from the NumericData object by calling getNames().
	@param labelsInFirstColumn If true, the first column will be treated as data labels rather than data.
	The labels may be retrieved from the NumericData object by calling getLabels().*/
	public static NumericData readFile(String filename, String separator, boolean labelsInFirstColumn) throws FileNotFoundException, IOException, NumberFormatException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line=reader.readLine().trim();
		String[] items=line.split(separator);
		String[] names=null;
		Vector<double[]> dataVector=new Vector<double[]>();
		Vector<String> labelVector=null;

		int dimensionality=items.length;
		if(labelsInFirstColumn)
		{
			dimensionality--;
			labelVector=new Vector<String>();
		}
		double[] values=new double[dimensionality];

		try
		{
			for(int i=0;i<dimensionality;i++)
			{
				String item;
				if(labelsInFirstColumn)
				{
					try
					{
						item=items[i+1].trim();
					}
					catch(Exception e)
					{
						System.err.println("EXCEPTION");
						item="1";
					}
				}
				else
					item=items[i].trim();
				if(item.equals(""))
					values[i]=Double.NaN;
				else
					values[i]=Double.parseDouble(item);
			}
			dataVector.add((double[])(values.clone()));
			if(labelsInFirstColumn)
				labelVector.add(items[0].trim());
		}
		catch(NumberFormatException nfe)
		{
			names=new String[dimensionality];
			for(int i=0;i<dimensionality;i++)
			{
				if(labelsInFirstColumn)
					names[i]=items[i+1];
				else
					names[i]=items[i];
			}
		}
		line=reader.readLine();
		while(line!=null)
		{
			items=line.trim().split(separator);
			if(labelsInFirstColumn)
				labelVector.add(items[0].trim());
			String item;
			for(int i=0;i<dimensionality;i++)
			{
				if(labelsInFirstColumn)
				{
					try
					{
						item=items[i+1].trim();
					}
					catch(Exception e)
					{
						System.err.println("EXCEPTION");
						item="1";
					}
				}
				else
					item=items[i].trim();
				if(item.equals(""))
					values[i]=Double.NaN;
				else
					values[i]=Double.parseDouble(item);
			}
			dataVector.add((double[])(values.clone()));
			line=reader.readLine();
		}
		reader.close();

		String[] labels = null;
		if(labelsInFirstColumn)
		{
			labels = new String[labelVector.size()];
			for(int i=0;i<labelVector.size();i++)
				labels[i] = labelVector.get(i);
		}
		double[][] data=new double[dataVector.size()][dimensionality];
		dataVector.toArray(data);
		return new NumericData(names,data,labels);
	}

	/**Returns a StringData object by reading a file named <i>filename</i>.
	@param filename A path to the csv file
	@param separator The character(s) used as a data separator
	@param namesInFirstRow If true, the first column will be treated as variable names rather than data.
	The names will be saved and can be retrieved from the NumericData object by calling getNames()
	@param labelsInFirstColumn If true, the first column will be treated as String labels rather than data.
	The labels will be saved and can be retrieved from the NumericData object by calling getLabels()*/
	public static StringData readStringFile(String filename, String separator, boolean namesInFirstRow, boolean labelsInFirstColumn) throws FileNotFoundException, IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line=reader.readLine().trim();
		String[] items=line.split(separator);
		String[] names=null;
		int dimensionality=items.length;
		Vector<String> labelVector = null;

		if(labelsInFirstColumn)
		{
			dimensionality--;
			labelVector=new Vector<String>();
		}
		String[] values=new String[dimensionality];
		Vector<String[]> dataVector=new Vector<String[]>();

		for(int i=0;i<dimensionality;i++)
		{
			String item;
			if(labelsInFirstColumn)
				item=items[i+1].trim();
			else
				item = items[i].trim();
			values[i]=item;
		}
		if(namesInFirstRow)
			names = (String[])(values.clone());
		else
		{
			if(labelsInFirstColumn)
				labelVector.add(items[0].trim());
			dataVector.add((String[])(values.clone()));
		}

		line=reader.readLine();
		int count = 0;
		while(line!=null)
		{
			items=line.trim().split(separator);
			String item;
			if(labelsInFirstColumn)
				labelVector.add(items[0].trim());
			for(int i=0;i<dimensionality;i++)
			{
				try
				{
					if(labelsInFirstColumn)
						item=items[i+1].trim();
					else
						item = items[i].trim();
					values[i]=item;
				}
				catch(ArrayIndexOutOfBoundsException aaibe)
				{
					if(i==dimensionality-1)
					{
						System.err.println("TextReader: Last element is empty");
						values[i]=null;
					}
				}
			}
			dataVector.add((String[])(values.clone()));
			line=reader.readLine();
			count++;
			if(count%1000 == 0)
				System.err.println("line "+count);
		}
		String[] labels = null;
		if(labelsInFirstColumn)
		{
			labels = new String[labelVector.size()];
			for(int i=0;i<labelVector.size();i++)
				labels[i] = labelVector.get(i);
		}
		String[][] data=new String[dataVector.size()][dimensionality];
		data = dataVector.toArray(data);
		return new StringData(names,data,labels);
	}

	/**Randomises the data in the file <i>filename</i> and writes it to the file <i>newFileName</i>.
	@param namesInFirstRow True if <i>filename<i> contains column headings, false otherwise.*/
	public static void randomiseLines(String filename, String newFileName, boolean namesInFirstRow)
	{
		BufferedReader reader = null;
		BufferedWriter writer = null;
		try
		{
			reader=new BufferedReader(new FileReader(filename));
			writer = new BufferedWriter(new FileWriter(new File(newFileName)));
			Vector<String> lines = new Vector<String>();
			String line = reader.readLine();
			boolean firstLine = true;
			if(namesInFirstRow)
			{
				writer.write(line);
				firstLine=false;
			}
			else
				lines.add(line);
			line = reader.readLine();
			while(line!=null)
			{
				if(!line.trim().equals(""))
					lines.add(line);
				line = reader.readLine();
			}
			int noLines = lines.size();
			int[] indices = Utils.getRandomIndices(noLines);
			for(int i=0;i<noLines;i++)
			{
				if(firstLine)
					firstLine = false;
				else
					writer.newLine();
				writer.write(lines.get(indices[i]));
			}
		}
		catch(IOException ioe)
		{

		}
		finally
		{
			try
			{
				reader.close();
				writer.flush();
				writer.close();
			}
			catch(IOException ioe)
			{
				ioe.printStackTrace();
				System.exit(1);
			}
		}
	}

	/**Removes duplicated data from the dataset in <i>filename<i> and writes the 'clean' dataset to <i>newFileName</i>.
	@param labelsInFirstColumn True if the first column contains data labels, false otherwise
	@param separator The separator used in <i>filename</i>.*/
	public static void removeDuplicates(String filename, String newFileName, boolean labelsInFirstColumn, String separator) throws IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		Vector<String> lines = new Vector<String>();
		Vector<String> names = new Vector<String>();
		String line = reader.readLine();
		while(line!=null)
		{
			if(labelsInFirstColumn)
			{
				names.add(line.split(separator)[0]);
			}
			lines.add(line);
			line = reader.readLine();
		}
		reader.close();
		int noLines = lines.size();
		boolean[] duplicated = new boolean[noLines];
		for(int i=0;i<noLines;i++)
		{
			if(!duplicated[i])
			{
				for(int j=i+1;j<noLines;j++)
				{
					if((labelsInFirstColumn&& names.get(i).equals(names.get(j))) ||lines.get(i).equals(lines.get(j)))
					{
						duplicated[i] = true;
						duplicated[j] = true;
					}
				}
			}
		}
		BufferedWriter summaryWriter = new BufferedWriter(new FileWriter(new File(newFileName)));
		boolean firstLine=true;
		for(int i=0;i<noLines;i++)
		{
			if(!duplicated[i])
			{
				if(firstLine)
					firstLine=false;
				else
					summaryWriter.newLine();
				summaryWriter.write(lines.get(i));
			}
		}
		summaryWriter.flush();
		summaryWriter.close();
	}
}