package geneticWedge.gp.io;

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import javax.swing.tree.*;
import java.util.*;

/**This class can read and extract data from XML files.
The data may be present in the file as elements or as attributes, as selected by the <i>mode</i>.
Note that all data with the same name will be agglomerated. If an element and an attribute have the same
name the mode should therefore be set to ELEMENTS_ONLY or ATTIBUTES_ONLY in order to select the desired data.
Element-names are case-insensitive, e.g. 'Data' and 'data' will be treated as the same element*/
public class XMLTreeHandler extends DefaultHandler
{
	/**IncludeData determines which data will be retrieved from the XML file. This is useful when an element and an attribute have the same name.*/
	public static enum IncludeData
	{
		/**Elements and attributes both retrieved*/
		ELEMENTS_AND_ATTRIBUTES,
		/**Only elements are retrieved*/
		ELEMENTS_ONLY,
		/**Only attributess are retrieved*/
		ATTRIBUTES_ONLY
	};
	private IncludeData mode;

	private DefaultMutableTreeNode current;
	private HashMap namespaceMappings;
	private Locator locator;
	private static Vector<Vector<String>> vectorData = new Vector<Vector<String>>();
	private static Vector<String> elements = new Vector<String>();
	private int currentElement = -1;
	private String currentElementValue = "";
	private Vector<Integer> nodelist=new Vector<Integer>();

	/**Constructor*/
	public XMLTreeHandler(String uri, IncludeData mode)
	{
		current = new DefaultMutableTreeNode("XML Document: "+uri);
		namespaceMappings = new HashMap();
		this.mode = mode;
	}

	/**Receive notification of character data inside an element.*/
	public void characters(char[] ch, int start, int length) throws SAXException
	{
		String s = new String(ch, start, length);
		s = s.trim();
		DefaultMutableTreeNode data=new DefaultMutableTreeNode("Character Data: '"+s+"'");
		current.add(data);
		currentElementValue+=s;
	}

	/**Receive notification of a processing instruction.*/
	public void ProcessingInstructions(String target, String data) throws SAXException
	{
		DefaultMutableTreeNode pi= new DefaultMutableTreeNode("PI (target = '"+target+"', data = '"+data+"')");
		current.add(pi);
	}

	/**Receive notification of the start of a Namespace mapping.*/
	public void startPrefixMapping(String prefix, String uri)
	{
		namespaceMappings.put(uri, prefix);
	}

	/**Receive notification of the end of a Namespace mapping.*/
	public void endPrefixMapping(String prefix)
	{
		for(Iterator i=namespaceMappings.keySet().iterator();i.hasNext();)
		{
			String uri = (String)i.next();
			String thisPrefix = (String)namespaceMappings.get(uri);
			if(prefix.equals(thisPrefix))
			{
				namespaceMappings.remove(uri);
				break;
			}
		}
	}

	/**Receive notification of the start of an element.*/
	public void startElement(String namespaceURI, String localName, String qName, Attributes atts) throws SAXException
	{
		if(!mode.equals(IncludeData.ATTRIBUTES_ONLY))
		{
			String elementName = localName.trim().toLowerCase();
			currentElement = elements.indexOf(elementName);
			if(currentElement==-1)
			{
				elements.add(elementName);
				currentElement = elements.size()-1;
				vectorData.add(new Vector<String>());
			}
			nodelist.add(currentElement);
			currentElementValue = "";
		}
		DefaultMutableTreeNode element = new DefaultMutableTreeNode("Element: "+localName);
		current.add(element);
		current = element;

		//determine namespace
		if(namespaceURI.length()>0)
		{
			String prefix=(String)namespaceMappings.get(namespaceURI);
			if(prefix.equals(""))
				prefix="[NONE]";

			DefaultMutableTreeNode namespace=new DefaultMutableTreeNode("Namespace: prefix = '"+prefix + "', URI = '"+namespaceURI+"'");
			current.add(namespace);
		}

		int attLength = atts.getLength();

		//process attributes
		for(int i=0;i<attLength;i++)
		{
			if(!mode.equals(IncludeData.ELEMENTS_ONLY))
			{
				String attName = atts.getLocalName(i).trim().toLowerCase();
				int currentAtt = elements.indexOf(attName);
				if(currentAtt==-1)
				{
					elements.add(attName);
					currentAtt = elements.size()-1;
					vectorData.add(new Vector<String>());
				}
				vectorData.get(currentAtt).add(atts.getValue(i));
			}

			DefaultMutableTreeNode attribute = new DefaultMutableTreeNode("Attribute (name = '"+atts.getLocalName(i)+"', value= '"+atts.getValue(i)+"')");
			String attURI = atts.getURI(i);
			if(attURI.length()>0)
			{
				String attPrefix=(String)namespaceMappings.get(namespaceURI);
				if(attPrefix.equals(""))
					attPrefix = "[NONE]";

				DefaultMutableTreeNode attNamespace = new DefaultMutableTreeNode("Namespace: prefix = '"+attPrefix+"', URI = '"+attURI+"'");
				attribute.add(attNamespace);
			}
			current.add(attribute);
		}
	}

	/**Receive notification of the end of an element.*/
	public void endElement(String namespaceURI, String localName, String qName) throws SAXException
	{
		current = (DefaultMutableTreeNode)current.getParent();

		if(!mode.equals(IncludeData.ATTRIBUTES_ONLY))
		{
			vectorData.get(currentElement).add(currentElementValue);
			nodelist.remove(nodelist.size()-1);
			if(nodelist.size()>0)
				currentElement = nodelist.lastElement();
			currentElementValue = "";
		}
	}

	/**Receive a Locator object for document events.*/
	public void SetDocumentLocator(Locator loc)
	{
		locator=loc;
	}

	/**Receive notification of a recoverable parser error.*/
	public void error(SAXParseException e) throws SAXException
	{
		warning(e);
	}

	/**Report a fatal XML parsing error.*/
	public void fatalError(SAXParseException e) throws SAXException
	{
		warning(e);
	}

	/**Receive notification of a parser warning.*/
	public void warning(SAXParseException e) throws SAXException
	{
		e.printStackTrace();
		throw new SAXException("Error or warning occurred");
	}

	/**This method returns all data identified by the variable name <i>dataName</i>.
	It ignores case, e.g. 'Data' and 'data' will be treated as the same element.*/
	public String[] getData(String dataName)
	{
		if(dataName==null)
			return null;
		int index = elements.indexOf(dataName.trim().toLowerCase());
		if(index == -1)
			return null;
		else
		{
			String[] returnArray = new String[0];
			return (String[])(vectorData.get(index).toArray(returnArray));
		}
	}

	/**This method returns the variable names retrieved from the XML file.*/
	public String[] getDataNames()
	{
		String[] returnArray = new String[0];
		return elements.toArray(returnArray);
	}
}