package geneticWedge.gp.postProcess;

import geneticWedge.gp.Utils;
import java.io.*;
import java.util.Vector;

/**PostProcessor allows access to various properties <i>after</i> an evolutionary run. Available properties include:
<ul>
<li>R-squared values (Pearson correlation)
<li>Individual fitnesses
<li>The best Individual (as a String)
<li>Best threshold (for classification tasks)
<li>Classification information (TP, FP, TN, FP)
</ul>*/
public class PostProcessor
{
	/**SELECT_MODE sets the method by which the 'best' individual is selected.*/
	public static enum SELECT_MODE{VAL_ONLY, TRAIN_ONLY, TEST_ONLY, BETTER_TRAIN, AVERAGE};

	/**getAllRSquared returns an array containing R-squared values (Pearson coefficients) for training, validation and test data.*/
	public static double[] getAllRSquared(String filename, SELECT_MODE mode) throws IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line = reader.readLine();
		while(!line.contains("training variance"))
		{
			line = reader.readLine();
		}
		String[] split = line.split(",");
		double[] variances = new double[]{Double.NaN,Double.NaN, Double.NaN};
		for (int i=0;i<3;i++)
		{
			try
			{
				variances[i] = Double.parseDouble((split[i].split("="))[1]);
			}
			catch(Exception e)
			{
				System.err.println("Error while getting variance.");
				e.printStackTrace();
			}
		}


		while(!line.contains("training fitness"))
		{
			line = reader.readLine();
		}
		split = line.split("training fitness");
		line = split[split.length-1];

		split = line.split(",");
		if(split.length<1)
		{
			System.err.println("PostProcessor: Result file has unknown format. Returning Nan.");
			return new double[]{Double.NaN,Double.NaN,Double.NaN};
		}
		double bestTrain = Double.parseDouble(split[0].split("=")[1]);
		double bestValidation = Double.NaN;
		double bestTest = Double.NaN;
		double bestAverage = Double.NaN;
		if(split.length>=2)
		{
			bestValidation = Double.parseDouble(split[1].split("=")[1]);
			bestAverage = bestTrain+bestValidation;
			if(split.length>=3)
				bestTest = Double.parseDouble(split[2].split("=")[1]);
		}
		line = reader.readLine();
		while(line!=null)
		{
			//new 120607
			while(line!=null && !line.contains("training fitness"))
			{
				line = reader.readLine();
			}
			if (line == null)
				break;
			split = line.split("training fitness");
			line = split[split.length-1];

			split = line.split(",");

			if(split.length<1)
				return new double[]{bestTrain,bestValidation,bestTest};
			else
			{
				double train = Double.parseDouble(split[0].split("=")[1]);
				double validation = Double.NaN;
				double test = Double.NaN;
				double average = Double.NaN;
				if(split.length>=2)
				{
					validation = Double.parseDouble(split[1].split("=")[1]);
					average = train+validation;
					if(split.length>=3)
				 		test = Double.parseDouble(split[2].split("=")[1]);
				}
				if((mode == SELECT_MODE.VAL_ONLY && validation<bestValidation)
				||(mode == SELECT_MODE.BETTER_TRAIN && validation<bestValidation && train<=validation)
				||(mode == SELECT_MODE.AVERAGE && average<bestAverage)
				||(mode == SELECT_MODE.TRAIN_ONLY && train<bestTrain)
				||(mode == SELECT_MODE.TEST_ONLY && test<bestTest))
				{
					bestTrain = train;
					bestValidation = validation;
					bestTest = test;
					bestAverage = average;
				}
			}
			line = reader.readLine();
		}
		return new double[]{1.0 - bestTrain/variances[0],1.0 - bestValidation/variances[1],1.0 - bestTest/variances[2]};
	}

	/**getAllFitnesses returns the training, validation and test fitnesses for the 'best' individual, selected on validation fitness.
	@param getLowest if true, returns the individual with the lowest validation fitness, otherwise the individual with the highest validation fitness.*/
	public static double[] getAllFitnesses(String filename, boolean getLowest) throws IOException
	{
		return getAllFitnesses(filename,getLowest,SELECT_MODE.VAL_ONLY);
	}

	/**getAllFitnesses returns the training, validation and test fitnesses for the 'best' individual, selected using the criterion <i>mode</i>.
	@param getLowest if true, returns the individual with the lowest fitness, otherwise the individual with the highest fitness.*/
	public static double[] getAllFitnesses(String filename, boolean getLowest, SELECT_MODE mode) throws IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line = reader.readLine();
		while(!line.contains("training fitness"))
		{
			line = reader.readLine();
		}

		//new 120607
		String[] split = line.split("training fitness");
		line = split[split.length-1];

		split = line.split(",");
		if(split.length<1)
		{
			System.err.println("PostProcessor: Result file has unknown format. Returning Nan.");
			return new double[]{Double.NaN,Double.NaN,Double.NaN};
		}

		//all indices reduced by 1, 120607
		double bestTrain = Double.parseDouble(split[0].split("=")[1]);
		double bestValidation = Double.NaN;
		double bestTest = Double.NaN;
		double bestAverage = Double.NaN;
		if(split.length>=2)
		{
			bestValidation = Double.parseDouble(split[1].split("=")[1]);
			bestAverage = (bestTrain+bestValidation)/2.0;
			if(split.length>=3)
				bestTest = Double.parseDouble(split[2].split("=")[1]);
		}
		line = reader.readLine();
		while(line!=null)
		{
			//new 120607
			while(line!=null && !line.contains("training fitness"))
			{
				line = reader.readLine();
			}
			if (line == null)
				break;
			split = line.split("training fitness");
			line = split[split.length-1];

			split = line.split(",");

			if(split.length<2)
				return new double[]{bestTrain,bestValidation,bestTest};
			else
			{
				//all indices reduced by 1, 120607
				double train = Double.parseDouble(split[0].split("=")[1]);
				double validation = Double.NaN;
				double test = Double.NaN;
				double average = Double.NaN;
				if(split.length>=2)
				{
					validation = Double.parseDouble(split[1].split("=")[1]);
					average = (train + validation) / 2.0;
					if(split.length>=3)
					{
				 		test = Double.parseDouble(split[2].split("=")[1]);
					}
				}
				switch (mode)
				{
					case VAL_ONLY:
						if(getLowest && validation<bestValidation)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						else if(!getLowest && validation>bestValidation)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						break;
					case TRAIN_ONLY:
						if(getLowest && train<bestTrain)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						else if(!getLowest && train>bestTrain)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						break;
					case TEST_ONLY:
						if(getLowest && test<bestTest)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						else if(!getLowest && test>bestTest)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						break;
					case BETTER_TRAIN:
						if(getLowest && validation<bestValidation && train<=validation)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						else if(!getLowest && validation>bestValidation && train>=validation)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
						}
						break;

					case AVERAGE:
						if(getLowest && average<bestAverage)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
							bestAverage = average;
						}
						else if(!getLowest && average>bestAverage)
						{
							bestTrain = train;
							bestValidation = validation;
							bestTest = test;
							bestAverage = average;
						}
						break;
				}
			}
			line = reader.readLine();
		}
		System.err.println("best test="+bestTest);
		return new double[]{bestTrain,bestValidation,bestTest};
	}

	/**getBestIndividualAsString returns the 'best' individual as a string, containing all nodes as a comma-separated list.
	@param getLowest if true, a string representation of the individual with the lowest fitness is returend, otherwise a string representation of the individual with the highest fitness.
	@param mode The criterion used to select the 'best' individual*/
	public static String getBestIndividualAsString(String filename, boolean getLowest, SELECT_MODE mode) throws IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line = reader.readLine();
		while(!line.contains("training fitness"))
		{
			line = reader.readLine();
		}

		String bestString = "";
		//new 120607
		String[] split = line.split("training fitness");
		String treeString = split[0].substring(0,split[0].length()-2);
		bestString = new String(treeString);

		line = split[split.length-1];

		split = line.split(",");
		if(split.length<1)
		{
			System.err.println("PostProcessor: Result file has unknown format. Returning Nan.");
			return "";
		}

		//all indices reduced by 1, 120607
		double bestTrain = Double.parseDouble(split[0].split("=")[1]);
		double bestValidation = Double.NaN;
		double bestTest = Double.NaN;
		double bestAverage = Double.NaN;
		if(split.length>=2)
		{
			bestValidation = Double.parseDouble(split[1].split("=")[1]);
			bestAverage = (bestTrain + bestValidation)/2.0;
			if(split.length>=3)
				bestTest = Double.parseDouble(split[2].split("=")[1]);
		}
		line = reader.readLine();
		while(line!=null)
		{
			//new 120607
			while(line!=null && !line.contains("training fitness"))
			{
				line = reader.readLine();
			}
			if (line == null)
				break;
			split = line.split("training fitness");
			treeString = split[0].substring(0,split[0].length()-2);
			line = split[split.length-1];

			split = line.split(",");

			double train = Double.parseDouble(split[0].split("=")[1]);
			double validation = Double.NaN;
			double test = Double.NaN;
			double average = Double.NaN;
			if(split.length>=2)
			{
				validation = Double.parseDouble(split[1].split("=")[1]);
				average = (train + validation) / 2.0;
				if(split.length>=3)
				{
					test = Double.parseDouble(split[2].split("=")[1]);
				}
			}
			switch (mode)
			{
				case VAL_ONLY:
					if(getLowest && validation<bestValidation)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					else if(!getLowest && validation>bestValidation)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					break;
				case TRAIN_ONLY:
					//System.err.println("best train="+bestTrain);
					if(getLowest && train<bestTrain)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					else if(!getLowest && train>bestTrain)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					break;
				case TEST_ONLY:
					if(getLowest && test<bestTest)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					else if(!getLowest && test>bestTest)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					break;
				case BETTER_TRAIN:
					if(getLowest && validation<bestValidation && train<=validation)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					else if(!getLowest && validation>bestValidation && train>=validation)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestString = treeString;
					}
					break;

				case AVERAGE:
					if(getLowest && average<bestAverage)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestAverage = average;
						bestString = treeString;
					}
					else if(!getLowest && average>bestAverage)
					{
						bestTrain = train;
						bestValidation = validation;
						bestTest = test;
						bestAverage = average;
						bestString = treeString;
					}
					break;
			}
			line = reader.readLine();
		}
		return bestString;
	}

	/**getBestValidationAndTestFitness returns the validation and test fitnesses for the 'best' individual, selected on validation fitness.
	@param getLowest if true, returns the individual with the lowest validation fitness, otherwise the individual with the highest validation fitness.*/
	public static double[] getBestValidationAndTestFitness(String filename, boolean getLowest) throws IOException
	{
		BufferedReader reader=new BufferedReader(new FileReader(filename));
		String line = reader.readLine();
		while(!line.contains("training fitness"))
		{
			line = reader.readLine();
		}
		String[] split = line.split(",");
		if(split.length!=4)
		{
			System.err.println("PostProcessor: Result file has unknown format. Returning Nan.");
			return new double[]{Double.NaN,Double.NaN};
		}
		double bestValue = Double.parseDouble(split[3].split("=")[1]);
		double bestValidation = Double.parseDouble(split[2].split("=")[1]);
		line = reader.readLine();
		while(line!=null)
		{
			split = line.split(",");
			if(split.length<4)
				return new double[]{bestValidation,bestValue};
			else
			{
				double validation = Double.parseDouble(split[2].split("=")[1]);
				double test = Double.parseDouble(split[3].split("=")[1]);
				if(getLowest && validation<bestValidation)
				{
					bestValidation = validation;
					bestValue = test;
				}
				else if(!getLowest && validation>bestValidation)
				{
					bestValidation = validation;
					bestValue = test;
				}
			}
			line = reader.readLine();
		}
		return new double[]{bestValidation,bestValue};
	}

	/**
	@param targetNumber The number of predictions that will have the desired output (positive or negative) once the threshold has been set.
	@param targetIsPositive Whether <i>targetNumber</i> refers to the number of positive outputs (true) or negative outputs (false).
	@param data An array containing the raw outputs and the target (binary) outputs.
	@return An array containing 5 elements: the best threshold and the following 4 values given this threshold: TP, FP, TN, FN.
	*/
	public static double[] setThreshold(int targetNumber, boolean targetIsPositive, double[][] data)
	{
		int len = data[0].length;
		if(targetNumber>len)
		{
			System.err.println("PostProcessor: targetNumber must be less than or equal to the number of data points.");
			return null;
		}
		int[] indices = Utils.ShellSortValues(data[0], false);
		double threshold = data[0][indices[targetNumber-1]];

		//find mid-point between positive and negative predictions
		if(targetNumber == len)
			threshold+=(data[0][indices[len-1]]-data[0][indices[len-2]])/2.0;//assume at least 2 points
		else
			threshold+=(data[0][indices[targetNumber]]-data[0][indices[targetNumber-1]])/2.0;

		double TP, FP, TN, FN;
		TP = FP = TN = FN = 0.0;
		for(int i=0;i<len;i++)
		{
			if(data[0][i]>threshold)
			{
				if(data[1][i]>0.0)
					TP++;
				else
					FP++;
			}
			else
			{
				if(data[1][i]>0.0)
					FN++;
				else
					TN++;
			}
		}
		return new double[]{threshold, TP, FP, TN, FN};
	}

	/**setThresholdForPPV attamepts to set a threshold (cut-off) value that will yield the desired PPV (positive predictive value).
	It returns an array containing 5 values: a suggested threshold  value, number of true positives, number of false positives, number of true negatives and number of false negatives.
	@param data should be a 2-by-n array containing data values in the first dimension and target outputs (+/-1) in the second dimension.*/
	public static double[] setThresholdForPPV(double targetPPV, double[][] data)
	{
		int[] indices = Utils.ShellSortValues(data[0], true);
		double TP = 0;
		double FP = 0;
		double TN = 0;
		double FN = 0;
		int len = indices.length;
		for(int i=0;i<len;i++)
		{
			if(data[1][i]>0.0)
				TP++;
			else
				FP++;
		}
		double PPV = TP / (TP + FP);
		if(PPV > targetPPV)
			return new double[]{data[0][indices[0]] - (data[0][indices[1]] - data[0][indices[0]]) / 2.0, TP, FP, 0.0, 0.0};

		for(int count = 0;count<len;count++)
		{
			if(data[1][indices[count]]>0.0)
			{
				TP--;
				FN++;
			}
			else
			{
				FP--;
				TN++;
			}
			PPV = TP / (TP + FP);
			if(PPV > targetPPV)
			{
				if(count == len-1)
					return new double[]{data[0][indices[count]] + (data[0][indices[count]] - data[0][indices[count-1]]) / 2.0, TP, FP, TN, FN};
				else
					return new double[]{(data[0][indices[count]] + data[0][indices[count+1]]) / 2.0, TP, FP, TN, FN};
			}
		}
		System.err.println("PostProcessor error: Unable to set threshold to achieve target PPV of "+targetPPV+". Returning null.");
		return null;
	}

	/**getAllClassificationData returns the numbers of true positives, false positives, true negatives and false negatives, given a data array and a threshold (cut-off).
	@param data should be a 2-by-n array containing data values in the first dimension and target outputs (+/-1) in the second dimension.*/
	public static double[] getAllClassificationData(double[][] data, double threshold)
	{
		int len = data[0].length;

		double TP, FP, TN, FN;
		TP = FP = TN = FN = 0.0;
		for(int i=0;i<len;i++)
		{
			if(data[0][i]>threshold)
			{
				if(data[1][i]>0.0)
					TP++;
				else
					FP++;
			}
			else
			{
				if(data[1][i]>0.0)
					FN++;
				else
					TN++;
			}
		}
		return new double[]{threshold, TP, FP, TN, FN};
	}
}